/*******************************************************************************
*                                                                              *
*   (C) 1997-2004 by Ernst W. Mayer.                                           *
*                                                                              *
*  This program is free software; you can redistribute it and/or modify it     *
*  under the terms of the GNU General Public License as published by the       *
*  Free Software Foundation; either version 2 of the License, or (at your      *
*  option) any later version.                                                  *
*                                                                              *
*  This program is distributed in the hope that it will be useful, but WITHOUT *
*  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or       *
*  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for   *
*  more details.                                                               *
*                                                                              *
*  You should have received a copy of the GNU General Public License along     *
*  with this program; see the file GPL.txt.  If not, you may view one at       *
*  http://www.fsf.org/licenses/licenses.html, or obtain one by writing to the  *
*  Free Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA     *
*  02111-1307, USA.                                                            *
*                                                                              *
*******************************************************************************/

/* USE 8-COLUMN TABS TO PROPERLY VIEW THIS FILE - IF THE STRINGS IN THE FOLLOWING LINES DON'T LINE UP PERFECTLY,
   PLEASE MODIFY YOUR EDITOR'S TAB SETTINGS SO THAT THEY DO:

	0123456789
        0123456789
*/

/*
!   COMPILING THE PROGRAM: The following list gives the best compile options for various Unix
!   platforms on which the code has been run successfully. If your particular hardware/software
!   is not represented, please let me know what kinds of results you get with your compiler
!   (especially which optimization sequence worked best). First gunzip (or gzip -d) the .gz file,
!   then unpack the tar archive using tar xvf *tar, then compile and link as follows:
!
!  *(1A) Compaq C compiler (cc) for TruUnix V5.0+:
!
!   21064/ev4/ev45:         cc -o Mlucas -O4 -assume accuracy_sensitive -unroll 1 -arch ev4     -tune ev4     -Olimit 100000 *.c -lm
!   21164/ev5/ev56/generic: cc -o Mlucas -O4 -assume accuracy_sensitive -unroll 1 -arch generic -tune generic -Olimit 100000 *.c -lm
!   21164/ev6/ev67/ev68:    cc -o Mlucas -O4 -assume accuracy_sensitive -unroll 1 -arch ev6     -tune ev6     -Olimit 100000 *.c -lm
!
!   NOTES:
!
!   - Do NOT use the <-fast> flag for TruUnix V5 and above - the code will not run properly.
!
!   - I've found (on 21064, 21164 and 21264) that using -O5 will increase compile time, and may actually hurt performance of this code.
!     If you want to try it on your system, note that when using -O5, you *must* also use the <-assume accuracy_sensitive> option.
!
!   - Using 'generic' as the argument to the -tune and -arch flags has the advantage that an executable compiled with it will run on any
!     generation of the Alpha architecture and usually performs well across a broad variety of systems, whereas 'ev6' *only* runs on the ev6.
!     But 'ev6' should also give slightly better performance on the ev6. So if for some reason you need just a single binary use 'generic,'
!     otherwise it's better to use the appropriate architecture-specific tunings.
!
!  *(1B) Compaq C compiler (ccc) for Linux: same as for Unix, but use ccc instead of cc, and note that the Linux compiler ignores the -Olimit flag.
!
!
!  *(2) SGI/Irix: compile using (for pre-R10000 CPU):
!
!   cc -o Mlucas -O3 -n32 -OPT:IEEE_arithmetic=1:roundoff=1:fast_bit_intrinsics=ON *.c
!
!   On R10000 and above, compile using:
!
!   cc -o Mlucas -mips4 -O3 -n32 -r10000 -IPA -OPT:Olimit=10000:const_copy_limit=20000:fast_bit_intrinsics=ON:IEEE_arithmetic=1:roundoff=1:fast_nint=ON *.c
!
!  *(3) SPARC/Solaris : Please use the SunPro C compiler - gcc does a poor job at optimizing for the Sparc.
!   Bill Rea, who has done extensive testing on the SPARC, recommends first building an
!   executable that is suitable for run-time profiling using the -xprofile=collect compile flag,
!   then doing lots of profiling (you could just use the built-in self-test feature for this),
!   then recompiling using the above sequence, this time replacing -xprofile=collect with -xprofile=use.
!   The -Bstatic flag binds all the needed run-time library files into the binary, making it easy to migrate
!   the code to other systems with slightly different versions of the OS and run-time libraries.
!
!
!  *(4) IBM RS6000/SP1/SP2/SP3: I don't have access to any such system. Are there any volunteers who want to try a build?
!
!  *(5) Mac/PPC, Mac Project C compiler (based on gcc) or gcc:
!
!   cc (or gcc) -o Mlucas -O3 *.c -lm
!
!
!		      *     *     *     *     *
!
!
!   RUNNING THE PROGRAM: see the README file: http://hogranch.com//mayer/README.html
*/

#include "Mlucas.h"

/* Globals */

/*...program version with patch suffix... */
const char VERSION[] = "2.8x";			/* <--- a suffix of x, y, or z indicates a beta code. */

const char OFILE     [] = "results.txt";
const char CONFIGFILE[] = "mlucas.cfg";		/* Configuration File: contains allowed FFT lengths
						   and allows user to control (at runtime, and in
						   a modifiable way) which of a predefined set
						   of FFT radices gets used for each runlength.
						   This file is re-read after each checkpoint. */
const char RANGEFILE [] = "worktodo.ini";	/* File containing exponents to be tested. New exponents
						   may be appended at will, while the program is running. */
char STATFILE[15];				/* This is set at runtime, based on the exponent being processed. */

#if defined(USE_THREADS)
	const char ntfile[] = "nthreads.ini";	/* Allows us to control the no. of threads dynamically at runtime. */
#endif

int INTERACT;

double ISRT2;
double RND_A, RND_B;	/* Used for fast NINT emulation; set in util.c. */
double AME;		/* Avg. Max. Err., used for setting exponent/FFT-length breakpoints */

int NTHREADS;		/* No. of threads. If multithreading disabled, set = 1. */

uint32 DAT_BITS, PAD_BITS;	/* Array padding parameters */

const uint32 PMAX=154000000;	/* maximum exponent allowed */

int NRADICES, RADIX_SET, RADIX_VEC[10];	/* RADIX_VEC[] stores sequence of complex FFT radices used.	*/

/****************/

int ernstMain(long exponent, int fft_length, int selftest, int iterations, int radixset, int error_checking, char *res64, int scrnFlag, double *runtime)
{
  int resFlag = 0;
/*
!...Code to test primality of Mersenne numbers, using arbitrary-precision (array-integer) arithmetic.
!   Author: Ernst W. Mayer.
!
!   This version (v2.8x) dated ?? ??? 2004.
!
!   Accomplishes the Mersenne-mod squaring via the weighted discrete Fourier transform technique
!   of Crandall and Fagin (Mathematics of Computation 62 (205), pp.305-324, January 1994).
!
!   For each binary exponent p, generates first p-2 terms of the Lucas-Lehmer sequence,
!
!       s(n+1) = s(n)**2 - 2, s(0) = 4
!
!   modulo N = 2^p - 1. If N divides s(p-2) (specifically, the (p-2)th residue == 0 modulo N), N is prime.
!   If p > 1398269 (congratulations Joel and George!) and p != 2976221 (congrats Gordon and George!)
!   and p != 3021377 (congrats Roland, George and Scott!), and p != 6972593 (congrats Nayan, George and Scott!),
!   and p != 13466917 (congrats Michael C., George and Scott!) and p != 20996011 (congrats Michael S., George and Scott!)
!   fame (or at least a fleeting sort of notoriety :-) awaits. If not, maybe you'll have learned something
!   nonetheless. In either case, go have a beer...(assuming you're of age, and are not a recovering alcoholic -
!   wouldn't want to be accused by irate parents of driving the kiddies to the bottle ;-)
!
!   Oh, and make it a good sort. Life's too short to drink crappy brewhahas, as a college buddy of
!   mine used to call them.
!
!   Calls MERS_MOD_SQUARE and a host of small FFT-pass and utility routines.
!
!
!   REVISION CONTROL:
!
!   02/27/97: v2.1 - DWT, power-of-2 FFT with precomputed trigs
!
!   04/28/97: v2.2 - dumb-but-effective unrolled-outer-loop FFT,
!		     fixed some redundancies in indexing, better memory usage in FFT.
!
!   05/15/97: v2.3 - automatic setting of FFT length and bits per digit;
!  -07/27/97       - added a second error check, based on a combination of absolute and
!		     relative difference between two FFT checksums;
!		   - speeded up carry propagation by faster NINT and interleaving with FFT bit reversal;
!		   - implemented slightly cache-friendlier FFT;
!		   - in interactive mode prints out average # of bits/digit;
!		   - in non-interative mode reads file, thus allowing range tests;
!		   - writes to status and restart files every 10,000 iterations,			!* NB: changed default to
!		     (to customize this, simply change the value of parameter INCR below),		!* 2000 iterations as of V2.4
!		     and automatically looks for a restart/range file when started;
!		     if it finds either, starts in non-interactive (range test) mode,
!		     otherwise starts in interactive mode.
!		   - purge integer*8's to improve portability;
!		   - convert base_index from integer*1 to individual bits of default integers;		!* NB: obsolete as of V2.6
!		   - allow for automatic invocation of extended (> 8-byte) precision real for
!		     accuracy-sensitive, once-initialized parameters (e.g. DWT weight factors,
!		     FFT sincos arrays) on systems which support this.
!		   - replace integer multiplies/divides by powers of 2 in critical loops
!		     with bitwise leftward/rightward shifts;
!
!   10/24/97: v2.4 - radix-4 FFT. Uglier but faster...
!   11/05/97: v2.4b- radix-8 FFT. Even uglier, even faster.
!		   - use a single DWT weights array for forward and inverse weights.
!		   - implemented one-pass combined FFT wrapper/squaring routine
!		     (see documentation in routine FFT_SQUARE). 5-10% faster,
!		     and slightly more accurate, to boot.
!   02/19/98: v2.4c- Combined FFT/square routine, specialized to FFT lengths of				!* NB: Long-runlength specialization
!		     128, 256 and 512K (exponents ranging from 1.33-10.2 million.) 7-10% faster.	!* is obsolete as of V2.5C
!   03/27/98; v2.4d- Fixed bug in 64-bit residue printout (affected ~1 of 16 hex digits on average.)
!		   - Added cache-friendlier A(j,j+1,j+2,j+3) --> B(k,k+1,k+n/2,k+n/2+1) indexing	!* NB: obsolete as of V2.5C,
!		     scheme to bit-reversal part of carry propagation loop. For N=256K, this gives	!* which does no bit reversals
!		     roughly a 3% speedup.
!
!   05/15/98: v2.5 - Radix-2,4,8,16, 2-D padded array FFT scheme. Much L1-cache-friendlier, especially at large N.
!                  - code can test exponents up to ~20 million.
!   07/20/98: v2.5b- Switched to a Nussbaumer-type radix-16 complex transform, with multiplies by external-loop
!                    sincos data (r^0-r^15) occurring immediately after array data are loaded from memory
!                    This version of the radix-16 FFT has significantly fewer floating-point (FP) multiplies
!                     and adds than the previous one, but does not pipeline as well
!                    (in the V2.5 FFT, mults and adds are well-mixed throughout the loop body, whereas now
!                    mults dominate early on and adds in the middle), so may give no significant increase in speed.
!                    Still, the reduced FP opcount reduces the level of accumulated roundoff error, so the increase
!                    in accuracy (and hence the range of useable exponents for a given FFT length) is worth it.
!                    Note also that the new version requires significantly larger sincos arrays than the old one -
!                    this extra bandwidth could be alleviated via use of recurrences, but this is not implemented
!                    here for reasons of accuracy (and when I tried it, it wasn't significantly faster than
!                    the non-recurrence version; good hardware seems to have no problems with the extra bandwidth,
!                    probably because the sincos data arrays have separate sections for each pass, i.e. sincos
!                    data access is in terms of sequential data blocks as the FFT proceeds.
!                  - Changed DWT weights array to zero-offset, so (k)th weight is wt(k-1) and its inverse
!                    is wt(n-k)*0.5 (rather than wt(k) and wt(n+2-k)*0.5 in previous version).
!                  - Removed extra calling argument (interact) from call to hex_res64 in main, changed nbits to
!                    intent(out) in typedefs of hex_res64 (thanks to Gord Palameta for pointing these out.)
!   11/19/98: v2.5c- Modularized the radix-2,4,8,16 transforms so they can be combined in essentially any fashion.
!                  - Forward FFT is now a decimation-in-frequency algorithm (DIFA). No more bit reversals!
!                  - Well, that's not completely true. There are no more explicit bit reversals, but the
!                    wrapper/pointwise-square step in FFT_SQUARE still deals with bit-reversed data. The
!                    good news is, I found a nice cache-friendly way of dealing with these. This makes a
!                    significant (10-20% speedup) difference at long runlengths, e.g. N >= 256K.
!                  - In preparation for the parallel modular tranbform (where multiplies will be relatively costly,)
!                    the radix-4 and 8 transforms are now also of the reduced-multiply variety. (Radix 2 is moot.)
!                  - User can specify how many iterations to do for timing test, then e.g. [T(200)-T(100)]/100
!                    gives an accurate per-iteration timing without regard to the data initialization time.
!
!   02/21/99: v2.6x- Added non-power-of-2 runlengths, by way of radix-3,5,7 forward DIF and inverse DIT routines
!                    which are done before and after the power-of-2 radices, respectively. This breaks up each
!                    interval between adjacent powers of 2 into four evenly spaced subintervals.
!                  - Changed from a 2-D array storage scheme with PSMALL*16 columns and an arbitrary number of padding
!                    elements to a 1-D scheme, still with PSMALL*16 padding blocks, each padding block now having length
!                    a power of 2. The 1-D scheme eliminates the integer multiplies needed by the hardware to get
!                    memory addresses in a 2-D scheme, which doesn't make a big difference in an all-floating-point
!                    transform version, but will pay off when we add the parallel modular transform, i.e. when integer
!                    multiplies become precious. Making PAD a power of 2 simplifies the calculation of padded-
!                    array indices from linear unpadded ones (e.g. from loop indices.) See the documentation in
!                    radix16_dif_pass for details. This scheme will also generalize easily to one having
!                    essentially arbitrary (but still power-of-two) data and padding block sizes.
!                  - Went to an on-the-fly wordsize computation, which needs just an integer subtract, AND and shift
!                    to return the wordsize (i.e. whether it's a bigword, having CEILING(p/n) bits, or a smallword,
!                    having FLOOR(p/n) bits, in the Crandall/Fagin mixed-radix "irrational base" scheme.) This
!                    eliminates the n-bit BASE_INDEX array, which needed a shift, an AND and an IBITS (i.e. also
!                    three integer operations) to extract the wordsize. Even if it gives no speed increase, it
!                    looks a lot better. And of course on many machines it will give a small speed increase.
!
!   06/06/99: v2.6a- Do the final-radix inverse DIT FFT pass, the inverse DWT weighting, the carry
!                    propagation step (with fractional roundoff error check), the forward DWT weighting,
!                    and the initial-radix forward DIF FFT pass in a single pass through the data array, rather than five.
!                    Initial trial implementation of this was for eadix-3*2^n only, but allowed David Willmore
!                    to perform the verification of the 38th known Mersenne prime, M6972593 in "just" 14 CPU-days
!                    on a 500MHz Alpha 21264, rather than the estimated 18 days it would have needed for v2.6x.
!                    Not bad for two days' coding and debugging effort...
!   08/10/99:      - After settling into my new digs in Cupertino, finally got round to implementing the above
!                    fused DIT-passN/carry/DIF-pass1 strategy for radices 5, 7, 8 and 16.
!   08/20/99: v2.6b- Added a fused DIF-passN/wrapper-square/DIT-pass1 routine, thus rolling formerly 5 passes
!                    through the big data arrays to effect these steps into just one, ugly monster pass routine.
!                    Not nearly as big a speedup as with the fused carry stuff, but hey, every little bit counts.
!   08/20/99: v2.6c- Fixed bug related to multiexponent runs with different FFT lengths (not all routines
!                    were reinitializing properly);
!                  - Implemented a reduced-sincos-array scheme, where twiddle arrays store only selected powers
!                    of a basic exponential exp(I*x), and partial recurrences (chosen for an optimal balance of
!                    efficiency and accuracy) are used to generate the other twiddles. 5-10% gain at long runlengths.
!                    (The larger gains are on lower-bandwidth/smaller-cache systems.)
!
!   08/20/99: v2.7x- In-place FFT, with DWT weights and sincos data obtained using multiplies of data from two
!                    small tables each. Code needs just 1/4 the memory of v2.6. Better cache utilization means
!                    a 20-40% speedup on Alpha, 10-25% on MIPS.
!                  - Simplified range test behavior so program reads just the first entry in the rangefile,
!                    leaving users free to add or modify any assignments besides the one currently running.
!                  - Stats file and savefiles now named differently for each exponent, e.g. p01234567.stat,
!                    p01234567, q01234567; added savefile residue checksum.
!                  - Results file now named results.txt, assignments file named worktodo.ini, to conform
!                    with Prime95 convention ("It's the look and feel thing, Mister President...")
!   09/30/99: v2.7y- Modified padding strategy to cause less thrashing (esp. for power-of-2 runlengths).
!                  - Put global array padding parameters in a module Mdata.
!                  - Other minor bugfixes.
!
!   11/21/99: v2.7z- Fixed hex_res64 re-init bug. Thanks to Brian Beesley for discovering it,
!                    and George Woltman for pointing out the bit-shifted upper residue bits that result.
!                    Note that the bug didn't affect the actual calculation. just the printing of Res64s.
!                  - Removed leading space in front of 'M' in outputs - the PrimeNet server doesn't like these.
!                  - Updated my and George's e-mail addresses.
!                  - The carry loop was running more slowly than I expected, so took a hard look at it. (The fused final-DIT/
!                    carry/initial-DIF routine typically needs around 20% of the total runtime, so multiply each of the listed
!                    speedups in () by 5 to see how much the change in question speeded up that particular routine.)
!                  - Changed the base(0:1) array from int*4 to real*8, thus eliminating an unneeded type conversion. (~1% speedup)
!                  - Rearranged the data in the WT1 weights table to ensure sequential access during carry propagation. (~3% speedup)
!                  - Converted repeated references to SI(NWT-L) datum in carry loop to a local integer constant. (~1% speedup)
!   03/07/00: v2.7a- Added radix-9 capability.
!                  - Eliminated large nearly-duplicate code sections in the radix*_ditN_cy_dif1 and radix16_wrapper_square
!                    routines (which required a tiny amount of added program logic) in order to improve L1 instruction
!                    cache usage. All computation-intensive modules should now be < 32KB in terms of object code size,
!                    assuming the compiler loop unrolling is not getting out of hand (e.g. use -pipeline -unroll 1 on Alpha.)
!                  - Program now checks for old savefiles every time a new exponent (not just a new run) is started,
!                    allowing users to concatenate exponents and savefiles from other machines in a single directory,
!                    if desired. (Thanks to Siegmar Szlavik for suggesting this.)
!   01/20/01: v2.7b- Added radix-11,"13",15 capability. I put 13 in quotes since it's implemented as a pseudoradix:
!		     rather than implement an actual (ugly & probably inefficient) radix13_ditN_cy_dif1 routine, we
!		     approximate 13*8 by 3*5*7, which is done as a radix-7 pass followed by a radix-15 pass. The former
!		     allows us to make use of the already-existing radix7_ditN_cy_dif1 routine, while the latter's
!		     forward-DIF and inverse-DIT incarnations are easily hacked from the just-written radix15_ditN_cy_dif1
!		     routine.
!                  - Improved pair_square function called by wrapper_square routines - moved multiplies by scaling
!                    factors 2/N into carry loop (where they need only be done infrequently) and improved operation
!                    sequence used for generating the complex terms - thanks to Guillermo Valor for the latter.
!                  - Added radix-32,64 FFT-pass and wrapper/square routines. These allow runlengths up to 2^23 to be
!		     done using as few as 6 passes through the data per squaring.
!                  - Allow various combinations of FFT radices to be tried (manually) while a run is in progress, by
!		     reading an mlucas.cfg configuration file prior to beginning each iteration block. This allows
!		     the user to explore which radix combination is best for their platform and assignment at runtime.
!   01/20/01: v2.7c- Implemented much cache-friendlier FFT data access scheme. Runtimes drop 10-30% on most systems,
!  -??/??/02         especially at larger runlengths and on smaller-cache systems.
!                  - Implemented a radix-13 front-and-back-end routine. All intervals between adjacent power-of-2
!                    runlengths can now be split  into 8 roughly equal-sized pieces.
!                  - Composite front-and-back-end DFT routines now use twiddleless permuted-input algorithm, whenever
!                    possible, i.e. when radix R is the product of 2 coprime radices. Floating-point opcounts for each
!                    set of R complex inputs, for the old and new versions, are as follows:
!
!                     Radix R       v2.7b           v2.7c
!                                 FADD  FMUL      FADD  FMUL
!                     -------     ----  ----      ----  ----
!                        5          34    10        34    10 (no change)
!                        6          40    16        36     8 (10% fewer add, 50% fewer mul)
!                        7          60    36        72    16 (20%  more add, 50% fewer mul)
!                        8          52     4        52     4 (no change)
!                        9          80    40        80    40 (no change)
!                       10          96    32        88    16 ( 8% fewer add, 50% fewer mul)
!                       11         168    40       168    40 (no change)
!                       12         104    32        96    16 ( 8% fewer add, 50% fewer mul)
!                       13       not available     190    62 (for C version; f90 has 192 FADD, 144 FMUL)
!                       14         160    96       160    32 (               67% fewer mul)
!                       15         178    82       162    50 ( 9% fewer add, 39% fewer mul)
!                       16         144    20       144    20 (no change)
!                       18       not available     196    80
!
!                    Here is how the opcounts for the optimized convolutions can be calculated a priori:
!
!                    Odd Prime radix R: Let R = 2*S + 1. Then, S is the length of the subconvolutions
!                    which are key to efficiently performing the length-R DFT in my implementation. These
!                    subconvolutions may be of either the cyclic or acyclic variety, depending on S.  I believe
!                    (but have not formally proved) that whenever S is odd we need only cyclics, whereas when
!                    S is even we need a mixture of cyclic and acyclic subconvolutions (cyclics to calculate the
!                    cosine terms in the DFT matrix, acyclics for the sine terms). If the polynomial x^S -+ 1
!                    (- for cyclic, + for acyclic) factorizes over the rationals, these subconvolutions can be
!                    implemented efficiently using smaller convolutions modulo the factors of the relevant polynomial
!                    and Chinese Remainder Theorem (CRT) reconstruction - see any decent introductory number theory
!                    text for more on this, and the comments in radix13_ditN_cy_dif1.c for some worked examples.
!                    Let A_cc(x) and M_cc(x) denote the number of real adds and multiplies, respectively,
!                    needed to effect a length-x real cyclic convolution (A_ac(x) and M_ac(x) for length-x acyclic).
!                    Here are the opcounts for the short-length convolutions of interest in the present application:
!
!                                            x   A_cc   M_cc   A_ac   M_ac   Comments
!                                           --  -----  -----  -----  -----   --------
!                                            1      0      1
!                                            2      4      2      3      3
!                                            3     11      4
!                                            5     31     10
!                                            6     40      8     41     15   Low-add variants need 34/12 and 35/19, resp.
!                                            8     46     14     ??     ??
!                                            9     74     19
!
!                    To effect a length-R complex DFT then requires
!
!                                4*[  R + A_cc      ] additions, 4* M_cc         multiplies  if S odd;
!                                2*[2*R + A_cc +A_ac] additions, 2*[M_cc + M_ac] multiplies  if S even.
!
!                    Let A_dft(x) and M_dft(x) denote the number of real adds and multiplies, respectively,
!                    needed to effect a length-x complex DFT. Here are the opcounts for odd-prime-length DFTs
!                    of interest in the present application, based on the above short-length-convolution opcounts:
!
!                                            x  A_dft  M_dft   Comments
!                                           --  -----  -----   --------
!                                            3     12      4
!                                            5     34     10
!                                            7     72     16   Direct computation of length-3 subconvos ==> 60 add, 36 mul - may need fewer clock cycles.
!                                           11    168     40   (150,100) for LO_ADD version
!                                           13    190     62   Low-mul variant of subconvos needs 214 add, 46 mul. Selesnick needs just 188/40.
!                                           17    252?    56?  Only if can use cyclic for both sets of terms, which seems dubious. Selesnick needs 274/82.
!                                           19    372     76   Selesnick needs 404/72.
!
!                    Composite radix R = p*q.
!                    Let A_dft(x) and M_dft(x) denote the number of real adds and multiplies, respectively,
!                    needed to effect a length-x complex DFT. If p and q are coprimeThen, we can use Rader's
!                    index permutation trick to eliminate the twiddle multiplies in between the sets of
!                    length-p and length-q DFTs, in which case we need precisely
!
!                                q*A|M_dft(p) + p*A|M_dft(q) additions|multiplies to do a length-p*q DFT.
!
!                    If p and q are not coprime (e.g. R = 4, 8, 9, 16) we need twiddles in between the
!                    radix-p and radix-q passes; whether these twiddles are trivial (or strictly real
!                    rather than complex) depends on the structure of the order-R roots of unity. Powers
!                    of 2 tend to have a greater fraction of trivial twiddles.
!
!                    Here are the opcounts for the various prime-and-composite-length DFTs used in the present application:
!
!                                            x  A_dft  M_dft   Comments
!                                           --  -----  -----   --------
!                                            2      4      0
!                                            3     12      4
!                                            4     16      0   p=q=2 not coprime, so need twiddles, but these = +-I, so are free.
!                                            5     34     10
!                                            6     36      8   p=3, q=2
!                                            7     72     16   		( 60, 36 for LO_ADD version)
!                                            8     52      4   p=4, q=2 not coprime, so need twiddles, which amount to 2 real*complex muls.
!                                            9     80     40   p=q=3 not coprime, so need twiddles; Nussbaumer version needs 88/20.
!                                           10     88     16   p=5, q=2
!                                           11    168     40		(150,100 for LO_ADD version)
!                                           12     96     16   p=3, q=4
!                                           13    190     62
!                                           14    172     32   p=7, q=2	(148, 72 for LO_ADD version)
!                                           15    162     50   p=5, q=3
!                                           16    144     24   p=q=4 not coprime, so need twiddles, which amount to 4 real*(complex+complex)
!                                                              and 4 complex*complex multiplies. Nussbaumer needs 148/20.
!                                           18    196     80   p=9, q=2
!                                           20    216     40   p=5, q=4
!                                           22    380     80   p=11,q=2	(344,200 for LO_ADD version)
!                                           24    252     44   p=3, q=8
!                                           26    432    124   p=13,q=2
!                                           28    400     64   p=7, q=4	(352,144 for LO_ADD version)
!                                           30    384    100   p=15,q=2
!                                           32    376     88   p=8, q=4 not coprime, so need twiddles, which amount to 4 real*(complex+complex)
!                                                              and 16 complex*complex multiplies.
!                                           36    534    160   p=9, q=4
!                                           40    532    120   p=5, q=8
!                                           44    848    160   p=11,q=4	(776,400 for LO_ADD version)
!                                           48    624    136   p=3, q=16
!                                           52    968    248   p=13,q=4
!                                           56    940    156   p=7, q=8	(844,316 for LO_ADD version)
!                                           60    888    200   p=15,q=4
!                                           64    922    216   p=q=8 not coprime, so need twiddles, which amount to 4 real*(complex+complex)
!                                                              and 44 complex*complex multiplies.
!
!                  - Fixed bug which caused error checking to be turned off beyond ERR_ITER even if RO warnings were
!                    issued before this point.
!                  - Fixed bug which attempted to write message to stdout when starting a new exponent after another
!                    one halted due to fatal roundoff error.
!                  - Since twiddles for forward FFT are re-used for many blocks of data before being recalculated,
!                    these are now done with extra accuracy (i.e. one table multiply per twiddle.)
!                  - (C version only) Implemented prefetch macros for selected platforms.
!                  - Numerous small bugfixes and things related to porting the code to x86 and Mac.
!
!   TO DO LIST:
!
!   Performance-related:
!   (0) Continue to improve cache/TLB behavior of code. This should include the following enhancements at a minimum:
!
!	- Implement inverse (DIT) FFT which does twiddle multiplies at END of processing of each block of data,
!	  and which thus allows each set of twiddles to be re-used for many consecutive data blocks, similarly
!	  to the forward FFT. This will allow DIT twiddles to be computed with extra accuracy and no loss of speed.
!	- With above DIT scheme, carry step now only will need weights (and not sincos) data tables, which will
!	  hopefully decrease contention for cache lines.
!	- Extend block-FFT strategy to include more than just the initial FFT pass, thus allowing the blocks
!	  to be smaller than n/radix(pass 1). This should help at very large runlengths and on systems with
!	  small L1 caches.
!
!   (1) Implement hybrid complex floating-point FFT with a modular complex (i.e. Gaussian integer) transform
!       over the Galois field GF(M61^2), where M61 = 2^61 - 1. See Richard Crandall's preprint, "Integer convolution
!       via split-radix fast Galois transform" for details on the fast Galois transform (FGT), freely available at:
!       http://www.perfsci.com/free/techpapers/confgt.ps. On hardware with good 64-bit integer capabilities, this
!       will really rock.
!
!   (2) Implement critical portions of code in assembly language (first for Alpha, possibly also for MIPS and SPARC.)
!       I suggest beginning with the radix-8 forward and inverse transform passes, i.e. subroutines radix8_dif_pass
!	and radix8_dit_pass, which (unlike their radix-16 brothers) can be coded with explicit access to no more
!	than 32 floating-point registers. (Using < 32 integer registers in the hybrid FFT/FGT will be trickier,
!	since we need up to 8-10 registers for array indices.)
!
!   Functionality-related:
!   (*) Add sieve-based factoring module which allows sieving to > 64 bits.
!   (*) Read assignments from a worktodo file in Prime95 format.
!   (*) Need code to allow automatic connection to PrimeNet server, e.g. for Unix systems.
!   (*) Convert savefile format to bytewise platform-independent format?
!   (*) Add P-1 factoring module.
!   (*) An (optional) X-based GUI would be nice...
!
!
!   ACKNOWLEDGEMENTS: special thanks are due to the following people:
!
!    Richard Crandall and Barry Fagin - for the discrete weighted transform.
!
!    Peter Montgomery - for help with theoretical and implementation questions,
!	loan of his fast factoring code, and for testing early versions of the program on his SGI
!	(and uncovering several compiler bugs along the way).
!
!    Luke Welsh - fu"r das Max und Moritz Weissbier Glas.
!
!    George Woltman - for organizing the Great Internet Mersenne Prime Search,
!	and for freely sharing his time and keen computational insights.
!
!    Scott Kurowski - for the GIMPS PrimeNet server.
!
!    Jason Papadopoulos - for his valuable perspectives regarding FFT machine
!    implementations and the various machine architectures.
!
!    Ed Haletky, John Henning and Greg Gaertner (Digital Equipment Corp.) - EH for
!       patiently answering N ( >> 1) questions regarding the Alpha architecture,
!	JH and GG for helping to shape the version of the program submitted for
!       the SPEC98 benchmark search.
!
!    Guillermo Ballester Valor - for many interesting discussions on optimization
!       and several usueful suggestions for improving the code.
!
!
!    ...as well as the numerous people who have been kind enough to try the code out
!    and send timings, compiler options for various architectures and suggestions
!    for improvement.
!
!    For the v2.7c release, special thanks to:
!
!       * Paul Novarese (Compaq computer Corp.; Alpha/TruUnix compile czar)
!       * Brian Beesley (U. Ulster, UK; Alpha/Linux compile czar)
!       * Bill Rea (U. Canterbury, New Zealand; Sparc compile czar)
!       * John Pierce (For hosting the ftp archive on hogranch.com)
!
!   For UPDATES, PATCHES and USAGE INSTRUCTIONS, see
!
!	http://hogranch.com//mayer/README.html
!
!   Send QUESTIONS, COMMENTS or SUGGESTIONS to me at <ewmayer@aol.com>.
!
!	Alright, let's go hunt for some primes!
!
*/

/*...Various general run parameters.. */
	const uint32 ntestmax=1000;		/* maximum number of exponents in the worktodo.ini file */
	const uint32 incr=2000;			/* number of iterations between checkpoints */
	uint32 timing_test_iters;

/*...scalars and fixed-size arrays...	*/
	 int32 i,j,nbits;
	uint32 dum,err_iter,ilo,ihi = 0,isprime,kblocks,n,npad = 0,p,iu,ju;
/*	uint64 res64;	*/
	double sum1,sum2;

/*...Known Mersenne prime exponents.	*/

	const uint32 nknown=41, knowns[41] = {2,3,5,7,13,17,19,31,61,89,107,127,521,607,1279,2203,2281,3217,4253,4423,9689,9941
		,11213,19937,21701,23209,44497,86243,110503,132049,216091,756839,859433,1257787,1398269,2976221,3021377,6972593
		,13466917,20996011,24036583};

/*...What a bunch of characters...	*/
	const int charoffset = '0';
	char *string;
	char restart1[10], restart2[10], char_exp[9], hex_res[17];
	char in_line[1024];	/* This needs to be large enough to hold an entire line of the .stat file	*/
	char *char_addr;

/*...initialize logicals...	*/
	int restart=FALSE, timing_test=FALSE, start_run=TRUE, relax_err;

/*...allocatable data array...	*/
	static int32 nalloc = 0, *arrtmp = 0x0;
	static double *a = 0x0, *a_ptmp = 0x0;

/*...file pointers	*/
	FILE *fp, *fq;

/*...time-related stuff. clock_t is typically an int (signed 32-bit) and a typical value of CLOCKS_PER_SEC (e.g. as
     defined in <machine/machtime.h> on Alpha TruUnix) is 1000000, so we should accumulate clock_t-stored time
     differences at least roughly every half hour (preferably every second or so) to avoid weirdness due to flipping
     of the sign bit or integer overflow.
*/
	/*clock_t clock1, clock2;	Moved these to mers_mod_square */
	double tdiff;

/*...check that various data types are of the assumed length...	*/

	check_nbits_in_types();

/*...check qfloat routines (this call is also needed to init various qfloat global constants): */

	qtest();

/*...entry point for one or more Lucas-Lehmer tests is here.	*/

	INTERACT=FALSE;

  RANGE_BEG:
	restart1[0] = restart2[0] = STATFILE[0] = '\0';
	if(restart) restart=FALSE;
	relax_err=TRUE;		/* This variable determines whether roundoff error checking may be disabled after	*/
				/* a trial number of error-checked iterations completes with no error warnings.		*/

/*  ...If multithreading enabled, look for an nthreads.ini file: */

#if defined(USE_THREADS)
	if (start_run) printf(" looking for %s file...\n", ntfile);

	fp = fopen(ntfile, "r");
	if ((fp))
	{
	  fscanf(fp,"%d", &NTHREADS);
	  fclose(fp);

	  if(NTHREADS > MAX_THREADS)
	  {
	    if(start_run) printf(" NTHREADS = %d specified in %s file exceeds maximum allowed - reducing to %d\n", NTHREADS, ntfile, MAX_THREADS);
	    NTHREADS = MAX_THREADS;
	  }
	  if(start_run) printf(" NTHREADS = %d\n", NTHREADS);
	}
	else
	{
	  if(start_run)           printf(   " nthreads.ini file not found - running single-threaded.\n");
	  fp = fopen(OFILE,"a"); fprintf(fp," nthreads.ini file not found - running single-threaded.\n"); fclose(fp);
	  NTHREADS = 1;
	}
#else
	  NTHREADS = 1;
#endif

/*  ...look for worktodo.ini file...	*/

	if (start_run) printf(" looking for %s file...\n", RANGEFILE);

	fp = fopen(RANGEFILE, "r");
	if ((fp) && (exponent==0))
	{
	  if (start_run) printf(" worktodo.ini file found...checking next exponent in range...\n");

	  /* 7/06/04: need to handle both exponent-only (i.e. pre-networked Mlucas) and Prime95 ini file formats:
	  fscanf(fp,"%u", &p);
	  fclose(fp);
	  */
	  /* Read first line of worktodo.ini file into 1K character array... */
	  fgets(in_line, 1024, fp);

	  /* Skip any whitespace at beginning of the line: */
	  i = 0;
	  while(isspace(in_line[i]))
	  {
		  ++i;
	  }
	  /* If leading char is numeric, assume it's exponent-only format: */
	  if(isdigit(in_line[i]))
	  {
		/* Extract the exponent... */
		p = 0;
		while(isdigit(in_line[i]))
		{
			p = 10*p + (in_line[i]-charoffset);
			++i;
		}
	  }
	  /* Otherwise assume Prime95 ini file format: */
	  else
	  {
		/* Currently only LL test (of either the first-time or double-check variety is supported: */
		char_addr = strstr(in_line, "Test=");
		if(!char_addr)
			char_addr = strstr(in_line, "DoubleCheck=");
		if(!char_addr)
		{
			if(start_run)           printf(   "ERROR: Unrecognized/unsupported option. The ini file entry was\n%s\n", in_line);
			fp = fopen(OFILE,"a"); fprintf(fp,"ERROR: Unrecognized/unsupported option. The ini file entry was\n%s\n", in_line); fclose(fp);
			goto L99;
		}
		char_addr = strstr(in_line, "=");
		++char_addr;
		/* Extract the exponent... */
		p = 0;
		while(isdigit(*char_addr))
		{
			p = 10*p + (*char_addr-charoffset);
			++char_addr;
		}
	  }

	  if(start_run) printf(" p = %u\n", p);

	  if (p < 1024)
	  {
	    if(start_run)           printf(   "ERROR: p must be at least 1024. The ini file entry was\n%s\n", in_line);
	    fp = fopen(OFILE,"a"); fprintf(fp,"ERROR: p must be at least 1024. The ini file entry was\n%s\n", in_line); fclose(fp);
	    goto L99;
	  }

	  if(p > PMAX)
	  {
	    if(start_run)           printf(   "ERROR: p exceeds maximum allowable of %u. The ini file entry was\n%s\n", PMAX,in_line);
	    fp = fopen(OFILE,"a"); fprintf(fp,"ERROR: p exceeds maximum allowable of %u. The ini file entry was\n%s\n", PMAX,in_line); fclose(fp);
	    goto L99;
	  }

	  if(!(pprimeF(p,2) && pprimeF(p,3) && pprimeF(p,5) && pprimeF(p,7) && pprimeF(p,11) && pprimeF(p,13)))
	  {
	    if(start_run)           printf(   "ERROR: p is not prime! The ini file entry was\n%s\n", in_line);
	    fp = fopen(OFILE,"a"); fprintf(fp,"ERROR: p is not prime! The ini file entry was\n%s\n", in_line); fclose(fp);
	    goto L99;
	  }

/*...If a valid worktodo assignment found, look for primary restart file...
     this uses the fact that a decimal digit d (ranging from 0 to 9) is 48+d in the ascii collating sequence	*/

	  i=p;
	  char_exp[8]='\0';
	  for(j=7; j >= 0; j--)
	  {
	    char_exp[j]=i%10+charoffset;
	    i=i/10;
	  }

	/* gcc with optimizations turned on wasn't initing all elements of restart file names = \0, so insert one manually after the p and q. */
/*printf("STATFILE0= %s\n",STATFILE);	*/
	  restart1[0] = 'p'; restart1[1] = '\0'; strcat(restart1, char_exp);
/*printf("restart1 = %s\n",restart1);	*/
	  restart2[0] = 'q'; restart2[1] = '\0'; strcat(restart2, char_exp);
/*printf("restart2 = %s\n",restart2);	*/
	  strcat(STATFILE, restart1);
	  string = ".stat";
	  strcat(STATFILE, string);
/*printf("STATFILE = %s\n",STATFILE);	*/

	if (strcmp(&restart1[1],&restart2[1]))
	{
		printf("ERROR 10 in Mlucas.c\n"); exit(EXIT_FAILURE);
	}

	  fp = fopen(restart1,"rb");
	  if(fp)
	  {
	    if(start_run)printf(" restart file %s found...reading...\n",restart1);

	    i = fread(&p, sizeof(int), 1, fp);
	    if(i != 1)  { printf(" Error reading p.\n")                                   ; fclose(fp); goto SEG; }
	    if(feof(fp)){ printf(" End-of-file encountered while attemptimg to read p.\n"); fclose(fp); goto SEG; }

	    i = fread(&ilo, sizeof(int), 1, fp);
	    if(i != 1)  { printf(" Error reading ilo.\n")                                   ; fclose(fp); goto SEG; }
	    if(feof(fp)){ printf(" End-of-file encountered while attemptimg to read ilo.\n"); fclose(fp); goto SEG; }

	    i = fread(&n, sizeof(int), 1, fp);
	    if(i != 1)  { printf(" Error reading n.\n")                                   ; fclose(fp); goto SEG; }
	    if(feof(fp)){ printf(" End-of-file encountered while attemptimg to read n.\n"); fclose(fp); goto SEG; }
	    kblocks = n >> 10;	/* Vector length in units of K.	*/

	    if(!(n == (kblocks << 10))) { printf("ERROR 20 in Mlucas.c\n"); goto SEG; }

	    DAT_BITS = DAT_BITS_DEF;
	    PAD_BITS = PAD_BITS_DEF;
	    npad = n + ( (n >> DAT_BITS) << PAD_BITS );	/* length of padded data array.	*/

	/*
	   Initially allocate double the needed memory so (e.g. for sets of 100-iteration timing tests)
	   we're not constantly reallocating memory - some OSes are poor at re-using recently-freed memory:
	*/
	    if(nalloc > 0 && npad > nalloc)
	    {
		assert(a != 0x0 && a_ptmp != 0x0 && arrtmp != 0x0);
		/*	free((void *)a); a=0x0;	*/
		free((void *)a_ptmp); a_ptmp=0x0; a=0x0;
		free((void *)arrtmp); arrtmp=0x0;
		nalloc = 0;
	    }
	    if(nalloc == 0)
	    {
		nalloc = 2*npad;
		assert(a == 0x0 && a_ptmp == 0x0 && arrtmp == 0x0);
		/*	a = (double *)calloc(nalloc, sizeof(double)); */
		a_ptmp = ALLOC_DOUBLE(nalloc);				if(!a_ptmp){ printf("FATAL: unable to allocate array A in main.\n"); exit(EXIT_FAILURE); }
		a      = ALIGN_DOUBLE(a_ptmp);
		arrtmp = (int    *)malloc(nalloc* sizeof(int   ));	if(!arrtmp){ printf("FATAL: unable to allocate array ARRTMP in main.\n"); exit(EXIT_FAILURE); }
	    }

	    i = fread(arrtmp, sizeof(int), n, fp);		/* Read integer residues...	*/
	    if(i != n)  { printf(" Error reading integer residue array.\n")                                   ; fclose(fp); goto SEG; }
	    if(feof(fp)){ printf(" End-of-file encountered while attemptimg to read integer residue array.\n"); fclose(fp); goto SEG; }

	    i = fread(&sum1, sizeof(double), 1, fp);		/* ...and checksum...	*/
	    if(i != 1)  { printf(" Error reading checksum1.\n")                                   ; fclose(fp); goto SEG; }
	    if(feof(fp)){ printf(" End-of-file encountered while attemptimg to read checksum1.\n"); fclose(fp); goto SEG; }

	    fclose(fp);

	    sum2=0.0;
	    for(i=0; i < n; i++)				/* ...then move elements of residue into their normal doubles slots.	*/
	    {
	      j = i + ( (i >> DAT_BITS) << PAD_BITS );
	      a[j] = arrtmp[i];
	      sum2 += a[j];
	    }
	    /* Don't deallocate arrtmp here, since we'll need it later for savefile writes. */

/*...we can re-use I/O unit number 1 here since if program execution gets to this point, the savefile previously opened on unit 1 will have been closed.	*/

	    if(sum1 != sum2)
	    {
	      if(start_run)              printf(   "ERROR: Checksum error for primary restart file...looking for secondary...\n");
	      fp = fopen(   OFILE,"a"); fprintf(fp,"ERROR: Checksum error for primary restart file...looking for secondary...\n"); fclose(fp);
	      fp = fopen(STATFILE,"a"); fprintf(fp,"ERROR: Checksum error for primary restart file...looking for secondary...\n"); fclose(fp);
	      goto SEG;
	    }

	    hex_res64(a,n,p,&nbits,hex_res);

	    if(start_run)              printf(   "Restarting M%u at iteration %u\n",p,ilo);
	    fp = fopen(STATFILE,"a"); fprintf(fp,"Restarting M%u at iteration %u; Current Res%2d: %s\n",p,ilo,nbits,hex_res); fclose(fp);
	    restart=TRUE;
	    ihi = ilo + incr;
	    goto SAVFILE_END; /* Successfully read primary savefile.	*/
	  }

SEG:	  fp = fopen(restart2,"rb");
	  if(fp)
	  {
	    if(start_run)printf(" secondary restart file found...reading...\n");

	    i = fread(&p, sizeof(int), 1, fp);
	    if(i != 1)  { printf(" Error reading p.\n")                                   ; fclose(fp); goto L99; }
	    if(feof(fp)){ printf(" End-of-file encountered while attemptimg to read p.\n"); fclose(fp); goto L99; }

	    i = fread(&ilo, sizeof(int), 1, fp);
	    if(i != 1)  { printf(" Error reading ilo.\n")                                   ; fclose(fp); goto L99; }
	    if(feof(fp)){ printf(" End-of-file encountered while attemptimg to read ilo.\n"); fclose(fp); goto L99; }

	    i = fread(&n, sizeof(int), 1, fp);
	    if(i != 1)  { printf(" Error reading n.\n")                                   ; fclose(fp); goto L99; }
	    if(feof(fp)){ printf(" End-of-file encountered while attemptimg to read n.\n"); fclose(fp); goto L99; }
	    kblocks = n >> 10;	/* Vector length in units of K.	*/

	    if(!(n == (kblocks << 10))) { printf("ERROR 30 in Mlucas.c\n"); exit(EXIT_FAILURE); }

	    DAT_BITS = DAT_BITS_DEF;
	    PAD_BITS = PAD_BITS_DEF;
	    npad = n + ( (n >> DAT_BITS) << PAD_BITS );	/* length of padded data array.	*/

	/*
	   Initially allocate double the needed memory so (e.g. for sets of 100-iteration timing tests)
	   we're not constantly reallocating memory - some OSes are poor at re-using recently-freed memory:
	*/
	    if(nalloc > 0 && npad > nalloc)
	    {
		assert(a != 0x0 && a_ptmp != 0x0 && arrtmp != 0x0);
		/*	free((void *)a); a=0x0;	*/
		free((void *)a_ptmp); a_ptmp=0x0; a=0x0;
		free((void *)arrtmp); arrtmp=0x0;
		nalloc = 0;
	    }
	    if(nalloc == 0)
	    {
		nalloc = 2*npad;
		assert(a == 0x0 && a_ptmp == 0x0 && arrtmp == 0x0);
		/*	a = (double *)calloc(nalloc, sizeof(double)); */
		a_ptmp = ALLOC_DOUBLE(nalloc);				if(!a_ptmp){ printf("FATAL: unable to allocate array A in main.\n"); exit(EXIT_FAILURE); }
		a      = ALIGN_DOUBLE(a_ptmp);
		arrtmp = (int    *)malloc(nalloc* sizeof(int   ));	if(!arrtmp){ printf("FATAL: unable to allocate array ARRTMP in main.\n"); exit(EXIT_FAILURE); }
	    }

	    i = fread(arrtmp, sizeof(int), n, fp);		/* Read integer residues...	*/
	    if(i != n)  { printf(" Error reading integer residue array.\n")                                   ; fclose(fp); goto L99; }
	    if(feof(fp)){ printf(" End-of-file encountered while attemptimg to read integer residue array.\n"); fclose(fp); goto L99; }

	    i = fread(&sum1, sizeof(double), 1, fp);		/* ...and checksum...	*/
	    if(i != 1)  { printf(" Error reading checksum1.\n")                                   ; fclose(fp); goto L99; }
	    if(feof(fp)){ printf(" End-of-file encountered while attemptimg to read checksum1.\n"); fclose(fp); goto L99; }

	    fclose(fp);

	    sum2=0.0;
	    for(i=0; i < n; i++)				/* ...then move elements of residue into their normal doubles slots.	*/
	    {
	      j = i + ( (i >> DAT_BITS) << PAD_BITS );
	      a[j] = arrtmp[i];
	      sum2 += a[j];
	    }
	    /* Don't deallocate arrtmp here, since we'll need it later for savefile writes. */

	    if(sum1 != sum2)
	    {
	      if(start_run)              printf(   "ERROR: Checksum error for secondary restart file...skipping this exponent.\n");
	      fp = fopen(   OFILE,"a"); fprintf(fp,"ERROR: Checksum error for secondary restart file...skipping this exponent.\n"); fclose(fp);
	      fp = fopen(STATFILE,"a"); fprintf(fp,"ERROR: Checksum error for secondary restart file...skipping this exponent.\n"); fclose(fp);
	      goto L99;
	    }

	    hex_res64(a,n,p,&nbits,hex_res);

	    if(start_run)              printf(   "Restarting M%u at iteration %u\n",p,ilo);
	    fp = fopen(STATFILE,"a"); fprintf(fp,"Restarting M%u at iteration %u; Current Res%2d: %s\n",p,ilo,nbits,hex_res); fclose(fp);
	    restart=TRUE;
	    ihi = ilo + incr;
	    goto SAVFILE_END; /* Successfully read secondary savefile.	*/
	  }
	  else
	  {
	    goto SAVFILE_END;
	  }

	  SAVFILE_END:
	  dum = 0;/* This is just so we have a statement following the above lable	*/
	}
	else	/* elseif((found RANGEFILE) = FALSE)	*/
	{
/*...Interactive mode: read in Mersenne exponent and FFT length	*/

	  printf(" no worktodo.ini file found...switching to interactive mode.\n");
	  INTERACT=TRUE;

	  IN_BEG:

	    printf(" Enter exponent >");
	    if (exponent == 0) scanf("%u", &p);
	    else
	    {
	      p = exponent;
	      printf(" %u\n", p);
	    }
	    if(p < 1024){ printf(" p must be at least 1024.\n"); goto IN_BEG; }

	    printf(" Enter FFT length in K (set K = 0 for default FFT length) >");
	    if (exponent == 0) scanf("%u", &kblocks);
	    else
	    {
	      kblocks = fft_length;
	      printf(" %u\n", kblocks);
	    }
	    n = kblocks << 10;

	    printf(" Enter 0 to run a full LL test, any other integer for a self-test >");
/*	    printf(" Enter 'y' to run a self-test, <return> for a full LL test >");	*/
	    if (exponent == 0) scanf ("%u", &i);
	    else
	    {
	      i = selftest;
	      printf(" %u\n", i);
	    }
/*	    scanf ("%c",&inchar);	*/
	    if (i)
/*	    if(tolower(inchar) == 'y')	*/
	    {
	      timing_test=TRUE;
	      AME = 0.0;	/* Init Avg. Max. Error */
	      printf(" Enter number of iterations for timing test >");
	      if (exponent == 0) scanf ("%u", &timing_test_iters);
	      else
	      {
	        timing_test_iters = iterations;
	        printf(" %u\n", timing_test_iters);
	      }

	      if(timing_test_iters > p-2)
	      {
	        printf(" This exceeds the Lucas-Lehmer limit; will perform %u iterations for timing test.\n",p-2);
	        timing_test_iters = p-2;
	      }
	    }

	    if(n != 0)
	    {
	      printf(" Enter index of radix set to be used for the FFT: \n");
	      printf("(See file fft_radix.txt for a list of available choices; enter -1 to get the default)  >");
	      if (exponent == 0) scanf ("%d", &RADIX_SET);
	      else
	      {
	        RADIX_SET = radixset;
	        printf(" %d\n", RADIX_SET);
	      }
	    }

	    printf(" Enter 1 to enable per-iteration error checking, 0 for no error checking >");
/*	    printf(" Enter 'y' to enable per-iteration error checking, <return> for no error checking >");	*/
	    if (exponent == 0) scanf ("%u", &i);
	    else
	    {
	      i = error_checking;
	      printf(" %u\n", i);
	    }
/*	    scanf ("%c",&inchar);	*/
	    if (i)
/*	    if(tolower(inchar) == 'y')	*/
	    {
	      err_iter = p-2;
	    }
	    else
	    {
	      err_iter = 0;
	    }

/*  ...make sure p is likely prime...	*/

	    if(pprimeF(p,2) && pprimeF(p,3) && pprimeF(p,5) && pprimeF(p,7) && pprimeF(p,11) && pprimeF(p,13))
	    {
	      printf(" p is prime...proceeding with Lucas-Lehmer test...\n");
	      goto IN_END;
	    }
	    else
	    {
	      printf(" p is not prime!\n"); goto IN_BEG;
	    }

	  IN_END:
	  dum = 0;/* This is just so we have a statement following the above lable	*/
	} /* endif(found RANGEFILE?)	*/

     /*...If it's not a restart, need to do a bunch of stuff:	*/

	if(!restart)
	{

/*...If they haven't yet been named, name the stats and savefiles...	*/

	  if(restart1[0] != 'p')
	  {
	    i=p;
	    char_exp[8]='\0';
	    for(j=7; j >= 0; j--)
	    {
	      char_exp[j]=i%10+charoffset;
	      i=i/10;
	    }
	    restart1[0] = 'p'; restart1[1] = '\0'; strcat(restart1, char_exp);
	    restart2[0] = 'q'; restart2[1] = '\0'; strcat(restart2, char_exp);
	    strcat(STATFILE, restart1);
	    string = ".stat";
	    strcat(STATFILE, string);
}

/*...determine size of FFT and dimension arrays accordingly...	*/

	  if((INTERACT && n==0) || (!INTERACT))
	  {
		kblocks = get_default_fft_length(p);
		if((int)kblocks < 0)
		{
			printf("ERROR detected in get_default_fft_length for p = %d.\n", p); goto L99;
			goto L99;
		}

/*...If the FFT length appears in the config file, read in the corresponding radix set index.
     If the length does not appear in the file, find the next-smallest FFT length which does,
     subject to the constraint that it be no more than 50% larger than the originally set length.
     This allows us to null out certain runlengths, if they prove unadvantageous on a particular
     platform (since we"ll have customized config files for all the major platforms.)	*/

	  if(!INTERACT)
	  {
	    RADIX_SET = -1;	/* If fail to open CONFIGFILE, signal mers_mod_square to use default radix	*/
	    string = "#";
	    fp = fopen(CONFIGFILE,"r");
	    if(fp)
	    {
	      for(i = 0; i < 3; i++)
	      {
	        if(fgets(in_line, 1024, fp))		/* Skip over the first block of 3 comment lines, which begin with #	*/
	        {
/*printf("%s\n",in_line);*/
	          if(!strstr(in_line,string)) { printf("ERROR: Line %d of %s lacks a # prefix.\n", i+1, CONFIGFILE); goto L5; }
	        }
	        else
	        {
/*printf("ERROR reading in_line %d", i+1);*/
	          goto L5;
	        }
	      }

	      fscanf(fp,"%u",&err_iter);		/* Line 4 contains the iteration number beyond which to turm of RO error checking	*/
/*printf("err_iter = %u\n",err_iter);*/
	      if(feof(fp) || ferror(fp))
	      {
/*printf("ERROR reading err_iter\n");*/
	        goto L5;
	      }

	      fgets(in_line, 1024, fp);		/* This positions the input stream at the start of the next line.	*/

	      if(fgets(in_line, 1024, fp))		/* Skip over one more comment line.	*/
	      {
/*printf("%s\n",in_line);*/
	        if(!strstr(in_line,string)) { printf("ERROR: Line %d of %s lacks a # prefix.\n", i+2, CONFIGFILE); goto L5; }
	      }
	      else
	      {
/*printf("ERROR reading in_line %d", i+1);*/
	        goto L5;
	      }

	      for(;;)				/* Each line beyond line 5 contains a pair of ints: {FFT length(in K) | FFT radix set}	*/
	      {
	        fscanf(fp,"%u%u",&iu,&ju);
	        if(feof(fp) || ferror(fp)) goto L5;
/*printf("%u %u\n",iu,ju);*/
	        if(iu >= kblocks)
		{
		  if(iu <= kblocks + (kblocks >> 1))
		  {
		    kblocks = iu; RADIX_SET = ju; break;
		  }
		}
	        if(!fgets(in_line, 1024, fp)) break;		/* This positions the input stream at the start of the next line.	*/
	      }
	      fclose(fp);
	      if(iu != kblocks) printf("INFO: Configuration file has no entry for FFT length in [%uK, %uK] - Using default FFT radices.\n",kblocks,kblocks + (kblocks >> 1));
	      goto L6;

L5:	      fclose(fp);
	      printf("INFO: Configuration file has invalid format - Using default FFT radices.\n");
	      err_iter = 200000;
	    }
	    else
	    {
	      printf("INFO: Configuration file %s not found - Using default FFT radices.\n",CONFIGFILE);
	      err_iter = 200000;
	    }

L6:	    dum = 0;
	  }

/*...calculate the unpadded runlength...	*/

	    n = kblocks << 10;

	  }  /* end {if((INTERACT.and.n==0).or.(.not.INTERACT))}	*/

/*...Set the array padding parameters - only use array padding elements for runlengths > 32K. */
	  if(kblocks > 32)
	  {
	    DAT_BITS = DAT_BITS_DEF;
	    PAD_BITS = PAD_BITS_DEF;
	  }
	  else
	  {
	    DAT_BITS =31;	/* This causes the padding to go away */
	    PAD_BITS = 0;
	  }

/*...If array padding turned on, check that the blocklength divides the unpadded runlength...	*/

	if((DAT_BITS < 31) && ((n >> DAT_BITS) << DAT_BITS) != n)
	{
	  if(start_run)              printf(   "ERROR: blocklength does not divide runlength!\n");
	  fp = fopen(   OFILE,"a"); fprintf(fp,"ERROR: blocklength does not divide runlength!\n"); fclose(fp);
	  fp = fopen(STATFILE,"a"); fprintf(fp,"ERROR: blocklength does not divide runlength!\n"); fclose(fp);
	  goto L99;
	}

/*...Find padded array length...	*/

	  npad = n + ( (n >> DAT_BITS) << PAD_BITS );	/* length of padded data array.	*/

/*...FFT info.	*/

	  if(INTERACT)
	  {
	     printf(   "M%u: using FFT length %uK = %u 8-byte floats.\n",p,kblocks,n);
	     printf(   " this gives an average %20.15f bits per digit\n",1.0*p/n);
	     printf(   " INFO: Using real*%2u for FFT sincos and DWT weights tables inits.\n",(int)sizeof(long double));
	  }
	  else
	  {
	    fp = fopen(STATFILE,"a");
	    fprintf(fp,"M%u: using FFT length %uK = %u 8-byte floats.\n",p,kblocks,n);
	    fprintf(fp," this gives an average %20.15f bits per digit\n",1.0*p/n);
	    fprintf(fp," INFO: Using real*%2u for FFT sincos and DWT weights tables inits.\n",(int)sizeof(long double));
	    fclose(fp);
	  }

/*...set initial iteration loop parameters...	*/

	  ilo=0;
	  if(timing_test)
	    ihi=timing_test_iters;
	  else
	    ihi=incr;

/*...and allocate the data array.	*/

	/*
	   Initially allocate double the needed memory so (e.g. for sets of 100-iteration timing tests)
	   we're not constantly reallocating memory - some OSes are poor at re-using recently-freed memory:
	*/
	    if(nalloc > 0 && npad > nalloc)
	    {
		assert(a != 0x0 && a_ptmp != 0x0 && arrtmp != 0x0);
		/*	free((void *)a); a=0x0;	*/
		free((void *)a_ptmp); a_ptmp=0x0; a=0x0;
		free((void *)arrtmp); arrtmp=0x0;
		nalloc = 0;
	    }
	    if(nalloc == 0)
	    {
		nalloc = 2*npad;
		assert(a == 0x0 && a_ptmp == 0x0 && arrtmp == 0x0);
		/*	a = (double *)calloc(nalloc, sizeof(double)); */
		a_ptmp = ALLOC_DOUBLE(nalloc);				if(!a_ptmp){ printf("FATAL: unable to allocate array A in main.\n"); exit(EXIT_FAILURE); }
		a      = ALIGN_DOUBLE(a_ptmp);
		arrtmp = (int    *)malloc(nalloc* sizeof(int   ));	if(!arrtmp){ printf("FATAL: unable to allocate array ARRTMP in main.\n"); exit(EXIT_FAILURE); }
	    }
	}
	else	/* If a restart, parse the .stat file to see if any RO warnings were previously issued.	*/
	{	/* If so, will require RO error checking to be performed on each iteration, even if iter > err_iter.	*/

	    fp = fopen(STATFILE,"r");
	    if(fp)
	    {
	      for(;;)
	      {
	        if(fgets(in_line, 1024, fp))
	        {
	          if(strstr(in_line,"Roundoff"))	/* If any RO warnings were issued during initial 200K iterations, leave error checking on... */
	          {
	          	if(!strstr(in_line,"0.40625"))	/* ...but only if the RO errors were > 0.40625 . */
	          	{
	          		printf("INFO: RO warnings > 0.40625 detected in first 200K iterations;\n");
	          		printf("Leaving error checking on for rest of the LL test of this exponent.\n");
	          		relax_err = FALSE; break;
	          	}
	          }
	        }
	        else
	        {
	          break;
	        }
	      }
	      fclose(fp);
	    }

	}  /* end {if(.not.restart)}	*/

/*...If a restart and RO warnings were previously issued to the .stat file, require error checking
     to be done on each iteration, even if iter > err_iter. We implement this simply by making err_iter > p-2.	*/

	if(!relax_err) err_iter = p;

/*...Initial seed for Lucas-Lehmer test. Could also use other values (e.g. 10), say for independent verification using the same hardware.	*/

	if(ilo == 0)
	{
		memset(a, 0, npad*sizeof(double));
		a[0]=4;
	}

/*...main loop...	*/

	for(;;)
	{
	  if(ihi > p-2) ihi=p-2;

/*...If the FFT length appears in the config file, read in the corresponding radix set index.
     Placing this second copy inside the iteration loop means this file gets read every (incr)th iteration,
     allowing th user the play with the radix set and attempt to find the optimum for the
     platform while the job is running. The difference between this and the earlier version of the
     same program segment is that here we don't allow kblocks (and hence the vector length n) to be
     modified, since we've already allocated the data arrays.

     If the length does not appear in the file, use the default radix set.)	*/

	  if(!INTERACT)
	  {
	    RADIX_SET = -1;	/* If fail to open CONFIGFILE, signal mers_mod_square to use default radix	*/
	    string = "#";
	    fp = fopen(CONFIGFILE,"r");
	    fq = fopen(STATFILE  ,"a");
	    if(fp)
	    {
	      for(i = 0; i < 3; i++)
	      {
	        if(fgets(in_line, 1024, fp))		/* Skip over the first block of 3 comment lines, which begin with #	*/
	        {
/*printf("%s\n",in_line);	*/
	          if(!strstr(in_line,string)) { if(fq) fprintf(fq, "ERROR: Line %d of %s lacks a # prefix.\n", i+1, CONFIGFILE); goto L7; }
	        }
	        else
	        {
	          goto L7;
	        }
	      }

	      fscanf(fp,"%u",&err_iter);		/* Line 4 contains the iteration number beyond which to turm of RO error checking	*/
	      if(feof(fp) || ferror(fp)) goto L7;
/*printf("%u\n",err_iter);	*/

	      fgets(in_line, 1024, fp);		/* This positions the input stream at the start of the next line.	*/

	      if(fgets(in_line, 1024, fp))		/* Skip over one more comment line.	*/
	      {
/*printf("%s\n",in_line);	*/
	        if(!strstr(in_line,string)) { if(fq) fprintf(fq, "ERROR: Line %d of %s lacks a # prefix.\n", i+2, CONFIGFILE); goto L7; }
	      }
	      else
	      {
	        goto L7;
	      }

	      for(;;)				/* Each line beyond line 5 contains a pair of ints: {FFT length(in K) | FFT radix set}	*/
	      {
	        fscanf(fp,"%u%u",&iu,&ju);
	        if(feof(fp) || ferror(fp)) goto L7;
/*printf("%u %u\n",iu,ju);	*/
	        if(iu == kblocks){ RADIX_SET = ju; break; }
	        if(!fgets(in_line, 1024, fp)) break;		/* This positions the input stream at the start of the next line.	*/
	      }
	      fclose(fp);
	      if(iu != kblocks && fq) fprintf(fq, "INFO: Configuration file has no entry for FFT length %uK - Using default FFT radices.\n",kblocks);
	      fclose(fq);
	      goto L8;

L7:	      fclose(fp);
	      fprintf(fq, "INFO: Configuration file has invalid format - Using default FFT radices.\n");
	      fclose(fq);
	      err_iter = 200000;
	    }
	    else if(fq)
	    {
	      fprintf(fq, "INFO: Configuration file %s not found - Using default FFT radices.\n",CONFIGFILE);
	      err_iter = 200000;
	      fclose(fq);
	    }

L8:	    dum = 0;
	  }

/*...here"s the big one - (incr) squaring steps. mers_mod_square returns 0 if no errors detected during this iteration cycle.	*/
/*printf("STATFILE = %s\n",STATFILE);*/
	  if(mers_mod_square(a, n, ilo, ihi, p, &err_iter, scrnFlag, &tdiff)) goto L99;
				/* If fatal error was encountered, skip to next exponent in worktodo.ini file
				   but keep restart files for current exponent around (can finish using hiacc code.)	*/

/*...Done?	*/

	  if(ihi==p-2 || timing_test) break;	/* C version uses zero-offset iteration counter. */

/*...every (incr)th iteration, write residue to restart files and timings to std out or STATFILE.	*/

	  sum1=0.0;
	  for(i=0; i < n; i++)
	  {
	    j = i + ( (i >> DAT_BITS) << PAD_BITS );
	    arrtmp[i]=(int)a[j];
	    sum1 += a[j];
	  }

	  fp = fopen(restart1,"wb");
	  if(fp)
	  {
	    i = fwrite(&p, sizeof(int), 1, fp);
	    if(i != 1)  { printf(" primary restart file: Error writing p.\n")                    ; fclose(fp); break; }

	    i = fwrite(&ihi, sizeof(int), 1, fp);
	    if(i != 1)  { printf(" primary restart file: Error writing ihi.\n")                  ; fclose(fp); break; }

	    i = fwrite(&n, sizeof(int), 1, fp);
	    if(i != 1)  { printf(" primary restart file: Error writing n.\n")                    ; fclose(fp); break; }

	    i = fwrite(arrtmp, sizeof(int), n, fp);		/* Write integer residue...	*/
	    if(i != n)  { printf(" primary restart file: Error writing residue.\n")              ; fclose(fp); break; }

	    i = fwrite(&sum1, sizeof(double), 1, fp);		/* ...and checksum.	*/
	    if(i != 1)  { printf(" primary restart file: Error writing checksum1.\n")            ; fclose(fp); break; }

	    fclose(fp);
	  }
	  else
	  {
	    fp = fopen(STATFILE,"a");
	    if(fp)
	    {
	      fprintf(fp,"WARNING: unable to open primary restart file for write of checkpoint data.\n");
	      fclose(fp);
	    }
	    if (scrnFlag)                               /* Echo output to stddev */
	    {
	      printf("WARNING: unable to open primary restart file for write of checkpoint data.\n");
	    }
	  }

	  fp = fopen(restart2,"wb");
	  if(fp)
	  {
	    i = fwrite(&p, sizeof(int), 1, fp);
	    if(i != 1)  { printf(" secondary restart file: Error writing p.\n")                    ; fclose(fp); break; }

	    i = fwrite(&ihi, sizeof(int), 1, fp);
	    if(i != 1)  { printf(" secondary restart file: Error writing ihi.\n")                  ; fclose(fp); break; }

	    i = fwrite(&n, sizeof(int), 1, fp);
	    if(i != 1)  { printf(" secondary restart file: Error writing n.\n")                    ; fclose(fp); break; }

	    i = fwrite(arrtmp, sizeof(int), n, fp);		/* Write integer residue...	*/
	    if(i != n)  { printf(" secondary restart file: Error writing residue.\n")              ; fclose(fp); break; }

	    i = fwrite(&sum1, sizeof(double), 1, fp);		/* ...and checksum.	*/
	    if(i != 1)  { printf(" secondary restart file: Error writing checksum1.\n")            ; fclose(fp); break; }

	    fclose(fp);
	  }
	  else
	  {
	    fp = fopen(STATFILE,"a");
	    if(fp)
	    {
	      fprintf(fp,"WARNING: unable to open secondary restart file for write of checkpoint data.\n");
	      fclose(fp);
	    }
            if (scrnFlag)                               /* Echo output to stddev */
            {
	          printf("WARNING: unable to open secondary restart file for write of checkpoint data.\n");
            }
	  }

/*...get a timing...	*/

	  hex_res64(a,n,p,&nbits,hex_res);

/*	  time2 = time(0x0);	*/
/*	  tdiff = difftime(time2, time1);	*/
/*	  time1 = time2;	*/

	/*
	  clock1 = clock2;
	  clock2 = clock();
	  tdiff = (double)(clock2 - clock1)/CLOCKS_PER_SEC;
	*/

/*...print runtime in hh:mm:ss format.	*/
	  if(INTERACT)
	     printf("M%u iteration = %u clocks =%2d%1d:%1d%1d:%1d%1d.%1d%1d%1d. Res%2d: %s\n",p,ihi
		,(int)tdiff/36000,((int)tdiff%36000)/3600
		,((int)tdiff%3600)/600,((int)tdiff%600)/60
		,((int)tdiff%60)/10,(int)tdiff%10
		,(int)(10*(tdiff-(int)tdiff)),(int)(100*(tdiff-(int)tdiff))%10,(int)(1000*(tdiff-(int)tdiff))%10
		,nbits,hex_res);
	  else
	  {
	    fp = fopen(STATFILE,"a");
	    if(fp)
	    {
	      fprintf(fp,"M%u iteration = %u clocks =%2d%1d:%1d%1d:%1d%1d.%1d%1d%1d. Res%2d: %s\n",p,ihi
		,(int)tdiff/36000,((int)tdiff%36000)/3600
		,((int)tdiff%3600)/600,((int)tdiff%600)/60
		,((int)tdiff%60)/10,(int)tdiff%10
		,(int)(10*(tdiff-(int)tdiff)),(int)(100*(tdiff-(int)tdiff))%10,(int)(1000*(tdiff-(int)tdiff))%10
		,nbits,hex_res);
	      fclose(fp);
	    }
            if (scrnFlag)                               /* Echo output to stddev */
            {
	      printf("M%u iteration = %u clocks =%2d%1d:%1d%1d:%1d%1d.%1d%1d%1d. Res%2d: %s\n",p,ihi
		,(int)tdiff/36000,((int)tdiff%36000)/3600
		,((int)tdiff%3600)/600,((int)tdiff%600)/60
		,((int)tdiff%60)/10,(int)tdiff%10
		,(int)(10*(tdiff-(int)tdiff)),(int)(100*(tdiff-(int)tdiff))%10,(int)(1000*(tdiff-(int)tdiff))%10
		,nbits,hex_res);
	    }
	  }

/*...reset loop parameters and begin next iteration cycle...	*/

	  ilo=ihi;
	  ihi=ilo+incr;

	}

/*...For timing tests, print timing and 64-bit residue and return timing via arglist runtime ptr...	*/

	*runtime = tdiff;
	hex_res64(a,n,p,&nbits,hex_res);

	if(timing_test)
	{
	  if(timing_test_iters > AME_ITER_START)
		AME /= (timing_test_iters - AME_ITER_START);
	  else
	  	AME = 0.0;

	  printf("%u iterations of M%u with FFT length %u\n",timing_test_iters,p,n);
	  if(nbits < 64) printf("WARNING: this residue contains only %u bits\n",nbits);
	  printf("Res%2d: %s. AvgMaxErr = %10.9f. Program: E%s\n", nbits, hex_res, AME, VERSION);

	  if (strlen(res64) == 16)                           /* If Selftest mode... */
	  {
	    if (strcmp(hex_res, res64) == 0) resFlag = 0;    /* Compare the calculated Res64 to the reference value. */
	    else
	    {
	      resFlag = 1;                       /* False */
	      printf("  ***   Res64 Error   ***\n");
	      printf(" current   = %s\n", hex_res);
	      printf(" should be = %s\n", res64);
	    }
	  }

/*...print runtime in hh:mm:ss format.	*/
	  printf("Clocks =%2d%1d:%1d%1d:%1d%1d.%1d%1d%1d\n"
		,(int)tdiff/36000,((int)tdiff%36000)/3600
		,((int)tdiff%3600)/600,((int)tdiff%600)/60
		,((int)tdiff%60)/10,(int)tdiff%10
		,(int)(10*(tdiff-(int)tdiff)),(int)(100*(tdiff-(int)tdiff))%10,(int)(1000*(tdiff-(int)tdiff))%10 );

	/*exit(EXIT_SUCCESS);*/
	  return(resFlag);
	}

/*...Full test results:	*/

	isprime = 1;
	for(i=0; i < n; i++)
	{
	  j = i + ( (i >> DAT_BITS) << PAD_BITS );
	  if(a[j] != 0.0) { isprime = 0; break; }
	}

/*...Unbelievable - I must be hallucinating.	*/
	if(isprime)
	{
	  int pknown = 0;
	  for(i=0; i < nknown; i++)
	  {
	    if(p == knowns[i]) { pknown = 1; break; }
	  }

	  if(pknown)
	  {
	    if(INTERACT)
	    {
	      printf("M%u is a known MERSENNE PRIME.\n",p);
	    }
	    else
	    {
	      fp = fopen(STATFILE,"a");
	      fprintf(fp,"M%u is a known MERSENNE PRIME.\n",p);
	      fclose(fp);

	      fp = fopen(   OFILE,"a");
	      fprintf(fp,"M%u is a known MERSENNE PRIME.\n",p);
	      fclose(fp);
	    }
            if (scrnFlag)                               /* Echo output to stddev */
            {
	      printf("M%u is a known MERSENNE PRIME.\n",p);
	    }
	  }
	  else
	  {

/*... this gets written both to file and to stdout, the latter irrespective of whether the run is in interactive mode...	*/
	    if(!INTERACT)
	    {
	      fp = fopen(STATFILE,"a");
	      fprintf(fp,"M%u is a new MERSENNE PRIME!!!\n",p);
	      fprintf(fp,"Please send e-mail to ewmayer@aol.com and woltman@alum.mit.edu.\n");
	      fclose(fp);

	      fp = fopen(   OFILE,"a");
	      fprintf(fp,"M%u is a new MERSENNE PRIME!!!\n",p);
	      fprintf(fp,"Please send e-mail to ewmayer@aol.com and woltman@alum.mit.edu.\n");
	      fclose(fp);
	    }
	    printf("M%u is a new MERSENNE PRIME!!!\n",p);
	    printf("Please send e-mail to ewmayer@aol.com and woltman@alum.mit.edu.\n");
	  }
	}
/*...The most likely scenario - it"s not prime, so we form a 64-bit residue and write that. If residue has < 64 bits, print a warning.	*/
	else
	{
/*...otherwise, write the 64-bit hex residue.	*/
	  if(INTERACT)
	  {
	    if(nbits < 64) printf("WARNING: this residue contains only %u bits\n",nbits);
	    printf("M%u is not prime. Res%2d: %s. Program: E%s\n",p,nbits,hex_res,VERSION);
	  }
	  else
	  {
	    fp = fopen(STATFILE,"a");
	    if(nbits < 64) fprintf(fp,"WARNING: this residue contains only %u bits\n",nbits);
	    fprintf(fp,"M%u is not prime. Res%2d: %s. Program: E%s\n",p,nbits,hex_res,VERSION);
	    fclose(fp);

	    fp = fopen(   OFILE,"a");
	    if(nbits < 64) fprintf(fp,"WARNING: this residue contains only %u bits\n",nbits);
	    fprintf(fp,"M%u is not prime. Res%2d: %s. Program: E%s\n",p,nbits,hex_res,VERSION);
	    fclose(fp);

            if (scrnFlag)                               /* Echo output to stddev */
            {
	      if(nbits < 64) printf("WARNING: this residue contains only %u bits\n",nbits);
	      printf("M%u is not prime. Res%2d: %s. Program: E%s\n",p,nbits,hex_res,VERSION);
	    }
	  }
	}

/*...If succesful completion, delete the restart files...	*/

	if(remove(restart1)) printf("Unable to delete primary restart file %s\n",restart1);
	if(remove(restart2)) printf("Unable to delete secondary restart file %s\n",restart2);

	start_run=FALSE;	/* Put this above 99 so start_run only gets switched off after an exponent gets completed successfully.	*/

/*...If in non-interactive (range test) mode, delete the just-completed exponent from the rangefile...	*/
/*
	Copy all the currently assigned exponents (except the first) from the worktodo.ini file to a temporary file...
*/
L99:	if(!INTERACT)
	{
	  fp = fopen(RANGEFILE,"r");
	  if(!fp)
	  {
	    fp = fopen(OFILE,"a");
	    if(fp)
	    {
	      fprintf(fp,"ERROR: unable to open %s file for reading.\n",RANGEFILE);
	      fclose(fp);
	    }
            if (scrnFlag)                               /* Echo output to stddev */
            {
	      printf("ERROR: unable to open %s file for reading.\n",RANGEFILE);
	    }
	    exit(EXIT_FAILURE);
	  }

	/* Remove any WINI.TMP file that may be present: */
	  remove("WINI.TMP");
	  fq = fopen("WINI.TMP", "w");
	  if(!fq)
	  {
		printf("Unable to open WINI.TMP file for writing.\n");
		exit(EXIT_FAILURE);
	  }
	  i=0;
	  while(fgets(in_line, 1024, fp))
	  {
		i++;
		if(i > ntestmax) break;
		if(i>1) fputs(in_line, fq);
	  }
	  fclose(fp);
	  fclose(fq);
	  if(i > ntestmax)
	  {
	    fp = fopen(OFILE, "a");
	    if(fp)
	    {
	      fprintf(fp, "WARNING: %s file contains > %u entries...truncating.\n", RANGEFILE, ntestmax);
	      fclose(fp);
	      if(INTERACT)
	        printf("WARNING: %s file contains > %u entries...truncating.\n", RANGEFILE, ntestmax);
	    }
	    else
	    {
	        printf("WARNING: %s file contains > %u entries...truncating.\n", RANGEFILE, ntestmax);
	    }
	  }

/* Now blow away the old worktodo.ini file and rename WINI.TMP ==> worktodo.ini...	*/

	  remove(RANGEFILE);
	  /* If attempting to simply rename the TMP file fails, do it the hard way: */
	  if(rename("WINI.TMP", RANGEFILE))
	  {
		fp = fopen(RANGEFILE,"w");
		if(!fp)
		{
			fp = fopen(OFILE,"a");
			if(fp)
			{
				fprintf(fp,"ERROR: unable to open %s file for writing.\n",RANGEFILE);
				fclose(fp);
			}
			if(scrnFlag)		/* Echo output to stddev */
			{
				printf("ERROR: unable to open %s file for writing.\n",RANGEFILE);
			}
			exit(EXIT_FAILURE);
		}

		fq = fopen("WINI.TMP", "r");
		if(!fq)
		{
			printf("Unable to remove WINI.TMP file\n");
			exit(EXIT_FAILURE);
		}
		while(fgets(in_line, 1024, fq))
		{
			fputs(in_line, fp);
		}
		fclose(fp);
		fclose(fq);

		/*...Then remove the WINI.TMP file:	*/

		remove("WINI.TMP");
	  }

/* if one or more exponents left in rangefile, go back for more... */

	  if (i > 1) goto RANGE_BEG;   /* CYCLE RANGE */
	}

/* If it was an interactive run or the last exponent in the rangefile, we're done. */

  return(0);
}

int get_default_fft_length(int p)
{
/*
!...Set vector length, based on number to be tested.
!   Default bounds assumes real*16 or similar extended-precision real data type is available
!   for the precomputation of the weight factors and FFT trig functions. The rule of thumb
!   is that on an IEEE64-compliant platform (DP floats having 53 mantissa bits), we can test
!   exponents up to ~98% as large as George Woltman"s Prime95 can test on an x86-compliant
!   platform at the same FFT length. (Prime95 gets the extra accuracy due to the 64 mantissa bits
!   DP floats enjoy while in registers.) See http://www.mersenne.org/status.htm for the Prime95
!   PMAX vs. FFT length table.
*/
	if( p >= 1025 && p <= 22500)			/* 1K	*/
	  return(   1);
	else if( p >= 22501 && p <= 44500)		/* 2K	*/
	  return(   2);
	else if( p >= 44501 && p <= 66250)		/* 3K	*/
	  return(   3);
	else if( p >= 66251 && p <= 88000)		/* 4K	*/
	  return(   4);

	else if( p >= 88001 && p <= 109325)		/* 5K	*/
	  return(   5);
	else if( p >= 109326 && p <= 130650)		/* 6K	*/
	  return(   6);
	else if( p >= 130651 && p <= 151975)		/* 7K	*/
	  return(   7);
	else if( p >= 151976 && p <= 173300)		/* 8K	*/
	  return(   8);

	else if( p >= 173301 && p <= 194450)		/* 9K	*/
	  return(   9);
	else if( p >= 194451 && p <= 215600)		/* 10K	*/
	  return(  10);
	else if( p >= 215601 && p <= 257900)		/* 12K	*/
	  return(  12);
	else if( p >= 257901 && p <= 300200)		/* 14K	*/
	  return(  14);
	else if( p >= 300201 && p <= 342500)		/* 16K	*/
	  return(  16);

	else if( p >= 342501 && p <= 383500)		/* 18K	*/
	  return(  18);
	else if( p >= 383501 && p <= 424375)		/* 20K	*/
	  return(  20);
	else if( p >= 424376 && p <= 506250)		/* 24K	*/
	  return(  24);
	else if( p >= 506251 && p <= 588125)		/* 28K	*/
	  return(  28);
	else if( p >= 588126 && p <= 670000)		/* 32K	*/
	  return(  32);

	else if( p >= 670001 && p <= 752500)		/* 36K	*/
	  return(  36);
	else if( p >= 752501 && p <= 835000)		/* 40K	*/
	  return(  40);
	else if( p >= 835001 && p <= 1000000)		/* 48K	*/
	  return(  48);
	else if( p >= 1000001 && p <= 1160000)	/* 56K	*/
	  return(  56);
	else if( p >= 1160001 && p <= 1300000)	/* 64K	*/
	  return(  64);

	else if( p >= 1300001 && p <= 1460000)	/* 72K	*/
	  return(  72);
	else if( p >= 1460001 && p <= 1620000)	/* 80K	*/
	  return(  80);
	else if( p >= 1620001 && p <= 1950000)	/* 96K	*/
	  return(  96);
	else if( p >= 1950001 && p <= 2280000)	/* 112K	*/
	  return( 112);
	else if( p >= 2280001 && p <= 2600000)	/* 128K	*/
	  return( 128);

	else if( p >= 2600001 && p <= 2910000)	/* 144K	*/
	  return( 144);
	else if( p >= 2910001 && p <= 3220000)	/* 160K	*/
	  return( 160);
	else if( p >= 3220001 && p <= 3540000)	/* 176K	*/
	  return( 176);
	else if( p >= 3540001 && p <= 3860000)	/* 192K	*/
	  return( 192);
	else if( p >= 3860001 && p <= 4510000)	/* 224K	*/
	  return( 224);
	else if( p >= 4510001 && p <= 5150000)	/* 256K	*/
	  return( 256);

	else if( p >= 5150001 && p <= 5770000)	/* 288K	*/
	  return( 288);
	else if( p >= 5770001 && p <= 6400000)	/* 320K	*/
	  return( 320);
	else if( p >= 6400001 && p <= 7000000)	/* 352K	*/
	  return( 352);
	else if( p >= 7000001 && p <= 7630000)	/* 384K	*/
	  return( 384);
	else if( p >= 7630001 && p <= 8330000)	/* 416K	*/
	  return( 416);
	else if( p >= 8330001 && p <= 8870000)	/* 448K	*/
	  return( 448);
	else if( p >= 8870001 && p <= 9490000)	/* 480K	*/
	  return( 480);
	else if( p >= 9490001 && p <= 10150000)	/* 512K	*/
	  return( 512);

	else if( p >= 10150001 && p <= 11350000)	/* 576K	*/
	  return( 576);
	else if( p >= 11350001 && p <= 12600000)	/* 640K	*/
	  return( 640);
	else if( p >= 12600001 && p <= 13750000)	/* 704K	*/
	  return( 704);
	else if( p >= 13750001 && p <= 15200000)	/* 768K	*/
	  return( 768);
	else if( p >= 15100001 && p <= 16250000)	/* 832K	*/
	  return( 832);
	else if( p >= 16250001 && p <= 17600000)	/* 896K	*/
	  return( 896);
	else if( p >= 17600001 && p <= 18850000)	/* 960K	*/
	  return( 960);
	else if( p >= 18850001 && p <= 20100000)	/* 1024K	*/
	  return(1024);

	else if( p >= 20100001 && p <= 22100000)	/* 1152K	*/
	  return(1152);
	else if( p >= 22100001 && p <= 24600000)	/* 1280K	*/
	  return(1280);
	else if( p >= 24600001 && p <= 27000000)	/* 1408K	*/
	  return(1408);
	else if( p >= 27000001 && p <= 29600000)	/* 1536K	*/
	  return(1536);
	else if( p >= 29600001 && p <= 32000000)	/* 1664K	*/
	  return(1664);
	else if( p >= 32000001 && p <= 34600000)	/* 1792K	*/
	  return(1792);
	else if( p >= 34600001 && p <= 37100000)	/* 1920K	*/
	  return(1920);
	else if( p >= 37100001 && p <= 39700000)	/* 2048K	*/
	  return(2048);

	else if( p >= 39700001 && p <= 43050000)	/* 2304K	*/
	  return(2304);
	else if( p >= 43050001 && p <= 48050000)	/* 2560K	*/
	  return(2560);
	else if( p >= 48050001 && p <= 52800000)	/* 2816K	*/
	  return(2816);
	else if( p >= 52800001 && p <= 58050000)	/* 3072K	*/
	  return(3072);
	else if( p >= 58050001 && p <= 62850000)	/* 3328K	*/
	  return(3328);
	else if( p >= 62850001 && p <= 68050000)	/* 3584K	*/
	  return(3584);
	else if( p >= 68050001 && p <= 73100000)	/* 3840K	*/
	  return(3840);
	else if( p >= 73100001 && p <= 78050000)	/* 4096K	*/
	  return(4096);

	else if( p >= 78050001 && p <= 83760000)	/* 4608K	*/
	  return(4608);
	else if( p >= 83760001 && p <= 93760000)	/* 5120K	*/
	  return(5120);
	else if( p >= 93760001 && p <=103200000)	/* 5632K	*/
	  return(5632);
	else if( p >=103200001 && p <=113800000)	/* 6144K	*/
	  return(6144);
	else if( p >=113800001 && p <=123400000)	/* 6656K	*/
	  return(6656);
	else if( p >=123400001 && p <=133800000)	/* 7168K	*/
	  return(7168);
	else if( p >=133800001 && p <=143800000)	/* 7680K	*/
	  return(7680);
	else if( p >=143800001 && p <=154000000)	/* 8192K	*/
	  return(8192);

	else if(                  p <=154000000)	/* Make sure nothing slipped through the cracks. */
	{
	  printf("ERROR: gap in exponent/FFT-length table detected. Please notify ewmayer@aol.com\n"); return -1;
	}
	else if( p == 2147483647)		/* Special case for F31 feasibility tests: MM31	*/
	{
/*	  return(114688);*/		/* 112M	*/
/*	  return(122880);*/		/* 120M	*/
	  return(131072);		/* 128M	*/
	}
	else
	{
	  printf(" invalid value for exponent...halting execution.\n"); return -1;
	}
}

/******************Thanks to Tom Cage for the initial version of this: ************************/

#ifdef macintosh
#include <console.h>                                                                /* Macintosh */
#endif

int main(int argc, char *argv[])
{
  char stFlag[5];
  int  i, lo, hi, finish, errCheck, errFlag, radset, res64Flag, scrnFlag, start, userSetExponent = 0, xNum = 0;
  double rtime, rtime_best, tdiff;
  int radix_best;
  /* mlucas.cfg (configuration) file pointer: */
  FILE *cfp;

/* Number of distinct FFT lengths supported for self-tests: */
#define numTest		47
/* Number of FFT lengths in the  small-exponent self-test suite: */
#define numSmall	15
/* Number of FFT lengths in the medium-exponent self-test suite: */
#define numMedium	16
/* Number of FFT lengths in the  large-exponent self-test suite: */
#define numLarge	8
/* Number of FFT lengths in the   huge-exponent self-test suite: */
#define numHuge		8

#if(numSmall +numMedium+numLarge +numHuge != numTest)
	#error (numSmall +numMedium+numLarge +numHuge) != numTest in main!
#endif

struct testCase{
	int fftLength;	/* FFT length in K (i.e. 4 means an array of 4K doubles) */
	int maxRadix;	/* maximum radix set index (indices assumed to be a contiguous set starting from zero)
				supported for this FFT length. See mers_mod_square.c for more details. */
	int exponent;	/* Test exponent for this FFT length. */
	char *res64;	/* 100-iteration Res64s (assuming initial LL test seed = 4) for this exponent. */
};

/* Array of distinct test cases for self-tests. Add one extra slot to vector for user-specified self-test exponents: */
struct testCase testVec[numTest+1] =
{
	/* FFT #radices  p  100-iter Res64             #bits per digit  FFT radices      AvgMaxErr  */
	/* Small:                                                                         x86  alfa */
	{ 128, 3,  2550001,"CB6030D5790E2460"},/* testVec[ 0]   19.455  16,16,16,16     .1034 .1334 */
	{ 144, 2,  2920013,"7CC1B41482BCB7C0"},/* testVec[ 1]   19.803   9,16,16,32     .1508 .2113 */
	{ 160, 6,  3265007,"B912804D7FE4A9E5"},/* testVec[ 2]   19.928  10,16,16,32     .2020 .2656 */
	{ 176, 3,  3550007,"5059094E256FB886"},/* testVec[ 3]   19.698  11,16,16,32     .1686 .2403 */
	{ 192, 6,  3900067,"4744CB8E5287DA60"},/* testVec[ 4]   19.837  12,16,16,32     .1885 .2523 */
	{ 224, 6,  4540007,"1DA37E1FAC27BC68"},/* testVec[ 5]   19.793  14,16,16,32     .2097 .2929 */
	{ 256, 7,  5190001,"15216788A374E144"},/* testVec[ 6]   19.798  16,16,16,32     .2563 .3086 */
	{ 288, 2,  5780087,"ADB1333A531F6EED"},/* testVec[ 7]   19.599   9,16,32,32     .1774 .2384 */
	{ 320, 3,  6400013,"6B2DF2F4FD779CBC"},/* testVec[ 8]   19.531  10,16,32,32     .1846 .2392 */
	{ 352, 2,  7010011,"4FC7B9144100998F"},/* testVec[ 9]   19.448  11,16,32,32     .1756 .2585 */
	{ 384, 3,  7600013,"2AFA7C90899B583E"},/* testVec[10]   19.328  12,16,32,32     .1383 .1872 */
	{ 416, 2,  8330009,"74AB1D925A0E7DB7"},/* testVec[11]   19.555  13,16,32,32     .2488 .3152 */
	{ 448, 3,  8950001,"7D9DD642E10F2525"},/* testVec[12]   19.509  14,16,32,32     .2041 .2906 */
	{ 480, 2,  9490001,"01A4E738255C522B"},/* testVec[13]   19.307  15,16,32,32     .1642 .2186 */
	{ 512, 3, 10110007,"24AAC84A6CD400BE"},/* testVec[14]   19.283  16,16,32,32     .1884 .2260 */
	/* Medium:                                                                                  */
	{ 576, 2, 11350013,"7087EA4B45F416A6"},/* testVec[15]   19.243   9,32,32,32     .1657 .2181 */
	{ 640, 3, 12590009,"93E43FC168EAF6BF"},/* testVec[16]   19.211  10,32,32,32     .1885 .2382 */
	{ 704, 2, 13799939,"7A8B6F72D5F3A862"},/* testVec[17]   19.143  11,32,32,32     .1747 .2542 */
	{ 768, 3, 15099979,"D731A6D76D99F3F5"},/* testVec[18]   19.201  12,32,32,32     .1692 .2304 */
	{ 832, 2, 16299979,"39AB362A15AF832C"},/* testVec[19]   19.132  13,32,32,32     .2154 .2632 */
	{ 896, 3, 17599997,"EDF99B1D21DE8835"},/* testVec[20]   19.182  14,32,32,32     .2041 .2773 */
	{ 960, 2, 18899999,"AF0F81144A3372A4"},/* testVec[21]   19.226  15,32,32,32     .2186 .2915 */
	{1024, 6, 20099983,"119B2956917D0CC1"},/* testVec[22]   19.169  16,32,32,32     .2457 .2934 */
	{1152, 5, 22500011,"3D81D5C9CC3D1C65"},/* testVec[23]   19.073   9,16,16,16,16  .1845 .2582 */
	{1280, 3, 25000009,"B4A3AF6909228279"},/* testVec[24]   19.073  10,16,16,16,16  .2534 .3083 */
	{1408, 3, 27300019,"9390DCE3398819EB"},/* testVec[25]   18.935  11,16,16,16,16  .2003 .2873 */
	{1536, 3, 29800003,"FBD4E6168183223D"},/* testVec[26]   18.946  12,16,16,16,16  .1808 .2426 */
	{1664, 3, 32100017,"998B572CB0AD53BA"},/* testVec[27]   18.839  13,16,16,16,16  .2026 .2488 */
	{1792, 3, 34500001,"55B023B8B7B84664"},/* testVec[28]   18.801  14,16,16,16,16  .1907 .2571 */
	{1920, 3, 37200001,"6EC0FF28D1E1C7CE"},/* testVec[29]   18.921  15,16,16,16,16  .2120 .2816 */
	{2048, 4, 39500003,"B07A522435EB5042"},/* testVec[30]   18.835  16,16,16,16,16  .2420 .2819 */
	/* Large:                                                                                   */
	{2304, 4, 44500021,"5D83FDDF4B9BBDE5"},/* testVec[31]   18.862   9,16,16,16,32  .2061 .2844 */
	{2560, 3, 48200003,"5D5189D7896700FA"},/* testVec[32]   18.387  10,16,16,16,32  .1461 .1839 */
	{2816, 3, 53700001,"1DA8E82BF72BE18B"},/* testVec[33]   18.623  11,16,16,16,32  .1978 .2830 */
	{3072, 3, 59000003,"4F0A573D580F3F6D"},/* testVec[34]   18.756  12,16,16,16,32  .2125 .2817 */
	{3328, 3, 63000037,"94B5C99BD3AA8708"},/* testVec[35]   18.487  13,16,16,16,32  .1781 .2223 */
	{3584, 3, 68500037,"47D9748158C04D3F"},/* testVec[36]   18.665  14,16,16,16,32  .2383 .3113 */
	{3840, 3, 73200037,"58E6E958B27DA8D7"},/* testVec[37]   18.616  15,16,16,16,32  .2059 .2769 */
	{4096, 5, 78200033,"E59AEF9086084A84"},/* testVec[38]   18.644  16,16,16,16,32  .1733 .2138 */
	/* Huge:                                                                                    */
	{4608, 4, 84000013,"C7323B8F92DCB618"},/* testVec[39]   17.802   9,16,16,32,32  .0716 .0969 */	/* exponent much too small!!*/
	{5120, 3, 93800011,"B3BF1872CC7863A6"},/* testVec[40]   17.891  10,16,16,32,32  .1056 .1305 */
	{5632, 3,103200017,"29FE5A4D5338B87F"},/* testVec[41]   17.894  11,16,16,32,32  .1068 .1542 */
	{6144, 3,113800003,"82E7165D2BBC9CF8"},/* testVec[42]   18.088  12,16,16,32,32  .1331 .1695 */
	{6656, 3,123400021,"CFEBD4E2BCE99605"},/* testVec[43]   18.105  13,16,16,32,32  .1684 .2037 */
	{7168, 3,133800011,"B884661290CD3931"},/* testVec[44]   18.229  14,16,16,32,32        .2464 */
	{7680, 3,143800003,"6770951E17678D2C"},/* testVec[45]   18.285  15,16,16,32,32        .2608 */
	{8192, 5,154000001,"6C0BD705D2FC646A"},/* testVec[46]   18.358  16,16,16,32,32        .3165 */
	{   0, 0,        0,                ""}
};

  printf("\n    Mlucas %s\n", VERSION);
  printf("\n    http://hogranch.com//mayer/README.html#2\n\n");

#ifdef macintosh
  argc = ccommand(&argv);                                                        /* Macintosh CW */
#endif

  scrnFlag = 0;                                                  /* Do Not echo output to stddev */

  if (argc == 1)                                                                 /* Default Mode */
  {
    res64Flag = ernstMain(0,0,0,0,0,0,"\0",scrnFlag,&rtime);
    printf("\n  Done ...\n\n");
    exit(EXIT_SUCCESS);
  }

  strcpy(stFlag, argv[1]);
  if (stFlag[0] != '-')                                                 /* Trap for 2nd argument */
  {
    printf("\n ***   Command Line Entry Error 1   ***   %s\n\n", stFlag);
    return(1);
  }

  if (stFlag[1] == 'd')
  {
    scrnFlag = 1;                                                       /* Echo output to stddev */
    res64Flag = ernstMain(0,0,0,0,0,0,"\0",scrnFlag,&rtime);
    printf("\a\n  Done ...\n\n");
    exit(EXIT_SUCCESS);
  }

  if (stFlag[1] == 'h')
  {
    printf(" Mlucas command line options ...\n");
    printf(" <CR>        Default mode: looks for a %s file in the local;\n", RANGEFILE);
    printf("             directory; if none found, prompts for manual keyboard entry\n");
    printf("\n");
    printf(" -h          Prints this help file and exits\n");
    printf("\n");
    printf(" *** NOTE: *** The following options cause a new mlucas.cfg file, containing\n");
    printf("     the optimal FFT radix set for the runlength(s) tested, to be written.\n");
    printf("     This is important for tuning code parameters to your particular platform.\n");
    printf("\n");
    printf("   FOR BEST RESULTS, RUN THE SELF-TEST UNDER ZERO- OR CONSTANT-LOAD CONDITIONS\n");
    printf("\n");
    printf("     If you wish to keep your old mlucas.cfg file for any reason,\n");
    printf("     please save a copy of it under a different name first!\n");
    printf("\n");
    lo = 0; hi = numSmall;
    printf(" -s [s]mall  Runs 1st set of %3d exponents, ranging from %d to %d\n", numSmall, testVec[lo].exponent, testVec[hi-1].exponent);
    printf("             This will take from ~10 minutes to ~1 hour, depending on CPU speed.\n");
    printf("\n");
    lo =hi; hi+= numMedium;
    printf("************* THIS IS THE ONLY SELF-TEST ORDINARY USERS NEED TO DO: ************\n");
    printf("*                                                                              *\n");
    printf("* -s [m]edium Runs 2nd set of %3d exponents, ranging from %d to %d *\n", numMedium,testVec[lo].exponent, testVec[hi-1].exponent);
    printf("*             This will take from ~1-4 hours, depending on CPU speed.          *\n");
    printf("*                                                                              *\n");
    printf("********************************************************************************\n");
    printf("\n");
    lo =hi; hi+= numLarge;
    printf(" -s [l]arge  Runs 3rd set of %3d exponents, ranging from %d to %d\n", numLarge, testVec[lo].exponent, testVec[hi-1].exponent);
    printf("             This will take from ~2-8 hours, depending on CPU speed.\n");
    printf("\n");
    lo =hi; hi+= numHuge;
    printf(" -s [x]large Runs 4th set of %3d exponents, ranging from %d to %d\n", numHuge,  testVec[lo].exponent, testVec[hi-1].exponent);
    printf("             This will take from ~3-12 hours, depending on CPU speed.\n");
    printf("\n");
    printf(" -s [a]ll    Runs all test exponents and all FFT radix sets.\n");
    printf("             This will take from ~6-24 hours, depending on CPU speed.\n");
    printf("\n");
    printf(" -s {num}    If {num} is one of the available FFT lengths (in K), this runs all\n");
    printf("             available FFT radices available at that length, using the default\n");
    printf("             self-test exponent for that length (see table in Mlucas.c main().)\n");
    printf("             \n");
    printf("             If {num} > 8192, assumes that {num} is an exponent, not FFT length.\n");
    printf("             In this case the program runs LL test on exponent {num}, using all\n");
    printf("             sets of FFT radices available at that exponent's default FFT length.\n");
    printf("             Use this to find the optimal radix set for a single given FFT\n");
    printf("             length on your hardware, and to modify the corresponding entry\n");
    printf("             in the mlucas.cfg file correspondingly.)\n");
    printf("\n");
    printf(" -s [option] [on/off]    Turn per-iteration roundoff error checking on/off.\n");
    printf("             Program will run 3-7%% slower with checking enabled. Default is OFF.\n");
    return(0);
  }

  if (stFlag[1] != 's')                                                 /* Trap for 2nd argument */
  {
    printf("\n ***   Command Line Entry Error 2   ***   %s\n\n", stFlag);
    return(2);
  }

  if ((argc < 3) || (argc > 4))                                         /* Trap for 3rd argument */
  {
    printf("\n ***   Command Line Entry Error 3   ***   %s\n\n", stFlag);
    return(3);
  }

  strcpy(stFlag, argv[2]);
  switch (stFlag[0])                                                    /* Trap for 3rd argument */
  {
     case 'a': start = 0                            ; finish = numTest                                ; break;	/* all    */
     case 's': start = 0                            ; finish = numSmall                               ; break;	/* small  */
     case 'm': start = numSmall                     ; finish = numSmall +numMedium                    ; break;	/* medium */
     case 'l': start = numSmall +numMedium          ; finish = numSmall +numMedium+numLarge           ; break;	/* large  */
     case 'x': start = numSmall +numMedium+numLarge ; finish = numSmall +numMedium+numLarge +numHuge  ; break;	/* huge   */
     case '0': case '1': case '2': case '3': case '4': case '5': case '6': case '7': case '8': case '9':
		i = atoi(stFlag);
		if(i > 8192)	/* Treat I as the Mersenne exponent, use default FFT length: */
		{
			if(i > PMAX)
			{
				/* Print error message and allow execution to fall through to case default: */
				printf("ERROR: p exceeds maximum allowable of %u.\n",PMAX);
			}
			else
			{
				userSetExponent = 1;
				testVec[numTest].exponent = i;
				i = get_default_fft_length(i);
				testVec[numTest].fftLength= i;
				/* Find the number of radices available for this FFT length: */
				for(lo = 0; lo < numTest; lo++)
				{
					if(testVec[lo].fftLength == i)
					{
						testVec[numTest].maxRadix = testVec[lo].maxRadix;
						break;
					}
				}
				if(lo >= numTest)
					printf("ERROR: unable to find FFT length %d K in the testVec table.\n", i);
				else
				{
					start = numTest; finish = start+1;
					break;
				}
			}
		}
		else	/* Treat I as FFT length (in K), find matching entry of testVec: */
		{
			/* I must be at least 128 and must have a non-power-of-2 component < 16: */
			if(i < 128)
				printf("ERROR: FFT length in K must be at least 128 - you entered %d.\n", i);
			else if((i >> trailz(i)) > 15)
				printf("ERROR: FFT length must have a non-power-of-2 component < 16 - your entry has %d.\n", (i >> trailz(i)));
			else	/* Find the corresponding entry of testVec: */
			{
				for(lo = 0; lo < numTest; lo++)
				{
					if(testVec[lo].fftLength == i)
					{
						start = lo; finish = start+1;
						break;
					}
				}
				if(lo >= numTest)
					printf("ERROR: unable to find FFT length %d K in the testVec table.\n", i);
				else
					break;
			}
		}
     default : printf("\n ***   Command Line Entry Error 4   ***   %s\n\n", stFlag); return(4);
  }

  if ((start < 0) || (finish > numTest + userSetExponent))		/* Trap for 3rd argument */
  {
    printf("\n ***   Command Line Entry Error 5   ***  start = %u  finish = %u\n\n", start, finish);
    return(5);
  }

  if (argc == 4)		/* Trap for 4th argument */
  {
    strcpy(stFlag, argv[3]);
      if (strcmp(stFlag, "on") == 0) { errFlag = 1; errCheck = 1; }
      else
        if (strcmp(stFlag, "off") == 0) { errFlag = 0; errCheck = 1; }
        else { printf("\n ***   Command Line Entry Error 6   ***   %s\n\n", stFlag); return(6); }
  }
  else { errFlag = 0; errCheck = 0; }

  printf("\n           Mlucas selftest running.....\n\n");
  if(userSetExponent)
  {
	printf("\n     You will need to manually verify that the Res64s output\n");
	printf("  for this user-set p match for all FFT radix combinations\n\n\n");
  }
  printf("/****************************************************************************/\n\n");

/* exponent, fft_length, selftest, iterations, radixset, error_checking, res64, scrnFlag */

  cfp = fopen(CONFIGFILE,"w");
  if(cfp)
  {
  	fprintf(cfp, "#\n");
	fprintf(cfp, "# mlucas.cfg file\n");
	fprintf(cfp, "# Insert comments as desired in lines beginning with a # symbol.\n");
	fprintf(cfp, "#\n");
  	fprintf(cfp, "# First non-comment line contains number of iterations to do before\n");
  	fprintf(cfp, "# turning per-iteration roundoff error checking off. Default = 200000.\n");
  	fprintf(cfp, "#\n");
  	fprintf(cfp, "200000\n");
  	fprintf(cfp, "#\n");
  	fprintf(cfp, "# Remaining non-comment lines contain data about the optimal\n");
  	fprintf(cfp, "# FFT parameters at each runlength on the host platform.\n");
  	fprintf(cfp, "# This file is updated automatically as needed by the program.\n");
  	fprintf(cfp, "#\n");
  	fprintf(cfp, "# Each line below contains a pair of integers. The first is the FFT length\n");
  	fprintf(cfp, "# (in units of 1K), i.e. the number of 8-byte floats used to store the\n");
  	fprintf(cfp, "# LL test residues for the exponent being tested. The program does\n");
  	fprintf(cfp, "# a complex-data FFT, i.e. treats this vector of reals as a complex\n");
  	fprintf(cfp, "# vector having half the length. The second integer on each line is\n");
  	fprintf(cfp, "# the index of the optimal set of FFT radices with respect to the\n");
  	fprintf(cfp, "# supported FFT radix sets for the particular FFT length. The available\n");
  	fprintf(cfp, "# FFT radices can be found in the large table at the bottom (specifically\n");
  	fprintf(cfp, "# the switch(n){} structure in the function get_fft_radices()) of the\n");
  	fprintf(cfp, "# mers_mod_square.c source file. The optimal set of complex radices\n");
  	fprintf(cfp, "# and the per-iteration timing it yields on the host platform are written\n");
  	fprintf(cfp, "# to the comment field right of each FFT-length/optimal-radix-set pair.\n");
  	fprintf(cfp, "#\n");
  	fclose(cfp);
  }

  for (xNum = start; xNum < finish; xNum++)    /* Step through the exponents */
  {
    /* Init best-radix-set and best-runtime for this FFT length: */
    rtime_best = 0.0;
    radix_best = 0;
    for (radset = 0; radset <= testVec[xNum].maxRadix; radset++)	/* Try all the radix sets available for this FFT length. */
    {
      /* errFlag = 0/1: Error Checking is OFF/ON */
      res64Flag = ernstMain(testVec[xNum].exponent,testVec[xNum].fftLength,1,100,radset,errFlag,testVec[xNum].res64,scrnFlag,&rtime);

      /* The best-radix stuff is only done with error checking OFF: */
      if(!errFlag && rtime_best == 0.0 || rtime < rtime_best)
      {
	      rtime_best = rtime;
	      radix_best = radset;
      }

      if (!userSetExponent && (res64Flag > 0) && (errCheck == 0))  /* Incorrect Res64 - rerun with error checking ON. */
      {
        printf("Incorrect Res64 - rerunning this case with error checking ON.\n");
        res64Flag = ernstMain(testVec[xNum].exponent,testVec[xNum].fftLength,1,100,radset,1,testVec[xNum].res64,scrnFlag,&rtime);
      }
      printf("\n");
    }
    /* If get a nonzero best-runtime, write the corresponding radix set index to the mlucas.cfg file: */
    if(rtime_best!=0.0)
    {
		/* Divide by the number of iterations done in the self-test: */
		tdiff = rtime_best/100;

		cfp = fopen(CONFIGFILE,"a");
		if(cfp)
		{
			/*...print per-iteration time in sec and best FFT radix set into comment field: */
			fprintf(cfp, "%d\t%d\t# sec/iter = %8.3f; radices = ", testVec[xNum].fftLength, radix_best, tdiff);
			/*
			fprintf(cfp, "Clocks =%2d%1d:%1d%1d:%1d%1d.%1d%1d%1d\n"
			,(int)tdiff/36000,((int)tdiff%36000)/3600
			,((int)tdiff%3600)/600,((int)tdiff%600)/60
			,((int)tdiff%60)/10,(int)tdiff%10
			,(int)(10*(tdiff-(int)tdiff)),(int)(100*(tdiff-(int)tdiff))%10,(int)(1000*(tdiff-(int)tdiff))%10 );
			*/
			/*
			  This call sets NRADICES and the first (NRADICES) elements of RADIX_VEC:
			*/
			get_fft_radices(testVec[xNum].fftLength << 10, radix_best, scrnFlag);
			for(i = 0; i < NRADICES; i++){ fprintf(cfp,"%3u",RADIX_VEC[i]); }; fprintf(cfp,"\n");
			fclose(cfp);
		}
    }

    printf("/ **************************************************************************** /\n\n");
  }

  printf("\n  Done ...\n\n");
  return(0);
}

