/*******************************************************************************
*                                                                              *
*   (C) 1997-2002 by Ernst W. Mayer.                                           *
*                                                                              *
*  This program is free software; you can redistribute it and/or modify it     *
*  under the terms of the GNU General Public License as published by the       *
*  Free Software Foundation; either version 2 of the License, or (at your      *
*  option) any later version.                                                  *
*                                                                              *
*  This program is distributed in the hope that it will be useful, but WITHOUT *
*  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or       *
*  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for   *
*  more details.                                                               *
*                                                                              *
*  You should have received a copy of the GNU General Public License along     *
*  with this program; see the file GPL.txt.  If not, you may view one at       *
*  http://www.fsf.org/licenses/licenses.html, or obtain one by writing to the  *
*  Free Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA     *
*  02111-1307, USA.                                                            *
*                                                                              *
*******************************************************************************/

/****************************************************************************
 * We now include this header file if it was not included before.
 ****************************************************************************/
#ifndef Mlucas_h_included
#define Mlucas_h_included

#include "Mdata.h"
#include "align.h"
#include "carry.h"
#include "dft_macro.h"
#include "prefetch.h"

/* Useful structs: */

struct complex{
	double re;
	double im;
};

/* Mlucas.c: */
int ernstMain(long exponent, int fft_length, int selftest, int iterations, int radixset, int error_checking, char *res64, int scrnFlag, double *runtime);
int get_default_fft_length(int p);

/* util.c: */
void	check_nbits_in_types();
int	leadz	(int x);
int	trailz	(int x);
int	pprimeF	(uint32 p, uint32 z);
void	errprint_sincos	(double *x, double *y, double theta);
void	hex_res64	(double a[], int n, int p, int *nbits, char *hex_res);
struct complex cmul(struct complex *, struct complex *);

/* br.c: */
int	reverse(uint32 i, uint32 n);
void	bit_reverse_int(int vec[], int n, int nradices, int radix[], int incr);

int	mers_mod_square(double a[], int n, int ilo, int ihi, int p, uint32 *err_iter, int scrnFlag, double *tdiff);
void	get_fft_radices(int n, int radset, int scrnFlag);

void	radix5_dif_pass1	(double a[], int n);
void	radix6_dif_pass1	(double a[], int n);
void	radix6_dif_pass1A	(double a[], int n);
void	radix6_dif_pass1B	(double a[], int n);
void	radix7_dif_pass1	(double a[], int n);
void	radix8_dif_pass1	(double a[], int n);
void	radix9_dif_pass1	(double a[], int n);
void	radix10_dif_pass1	(double a[], int n);
void	radix11_dif_pass1	(double a[], int n);
void	radix12_dif_pass1	(double a[], int n);
void	radix13_dif_pass1	(double a[], int n);
void	radix14_dif_pass1	(double a[], int n);
void	radix15_dif_pass1	(double a[], int n);
void	radix16_dif_pass1	(double a[], int n);
void	radix18_dif_pass1	(double a[], int n);
void	radix20_dif_pass1	(double a[], int n);
void	radix22_dif_pass1	(double a[], int n);
void	radix24_dif_pass1	(double a[], int n);
void	radix26_dif_pass1	(double a[], int n);
void	radix28_dif_pass1	(double a[], int n);
void	radix30_dif_pass1	(double a[], int n);
void	radix32_dif_pass1	(double a[], int n);
void	radix36_dif_pass1	(double a[], int n);

void	radix5_dit_pass1	(double a[], int n);
void	radix6_dit_pass1	(double a[], int n);
void	radix7_dit_pass1	(double a[], int n);
void	radix8_dit_pass1	(double a[], int n);
void	radix9_dit_pass1	(double a[], int n);
void	radix10_dit_pass1	(double a[], int n);
void	radix11_dit_pass1	(double a[], int n);
void	radix12_dit_pass1	(double a[], int n);
void	radix13_dit_pass1	(double a[], int n);
void	radix13_dit_pass1A	(double a[], int n);
void	radix13_dit_pass1B	(double a[], int n);
void	radix14_dit_pass1	(double a[], int n);
void	radix15_dit_pass1	(double a[], int n);
void	radix16_dit_pass1	(double a[], int n);
void	radix18_dit_pass1	(double a[], int n);
void	radix20_dit_pass1	(double a[], int n);
void	radix22_dit_pass1	(double a[], int n);
void	radix24_dit_pass1	(double a[], int n);
void	radix26_dit_pass1	(double a[], int n);
void	radix28_dit_pass1	(double a[], int n);
void	radix30_dit_pass1	(double a[], int n);
void	radix32_dit_pass1	(double a[], int n);
void	radix36_dit_pass1	(double a[], int n);

void	process_chunk	(double a[], int n, int nrt, int nrt_bits, struct complex rt0[], struct complex rt1[], int index[], int block_index[], int ii, int nradices_prim, int radix_prim[], int ws_i[], int ws_j1[], int ws_j2[], int ws_j2_start[], int ws_k[], int ws_m[], int ws_blocklen[], int ws_blocklen_sum[]);

void	radix8_dif_pass		(double a[], int n, int nrt, int nrt_bits, struct complex rt0[], struct complex rt1[], int index[], int nloops, int incr);
void	radix16_dif_pass	(double a[], int n, int nrt, int nrt_bits, struct complex rt0[], struct complex rt1[], int index[], int nloops, int incr);
void	radix32_dif_pass	(double a[], int n, int nrt, int nrt_bits, struct complex rt0[], struct complex rt1[], int index[], int nloops, int incr);
void	radix64_dif_pass	(double a[], int n, int nrt, int nrt_bits, struct complex rt0[], struct complex rt1[], int index[], int nloops, int incr);

void	radix8_dit_pass		(double a[], int n, int nrt, int nrt_bits, struct complex rt0[], struct complex rt1[], int nloops, int incr);
void	radix16_dit_pass	(double a[], int n, int nrt, int nrt_bits, struct complex rt0[], struct complex rt1[], int nloops, int incr);
void	radix32_dit_pass	(double a[], int n, int nrt, int nrt_bits, struct complex rt0[], struct complex rt1[], int nloops, int incr);
void	radix64_dit_pass	(double a[], int n, int nrt, int nrt_bits, struct complex rt0[], struct complex rt1[], int nloops, int incr);
void	radix8_dit_passB	(double a[], int n, int nrt, int nrt_bits, struct complex rt0[], struct complex rt1[], int index[], int nloops, int incr);
void	radix16_dit_passB	(double a[], int n, int nrt, int nrt_bits, struct complex rt0[], struct complex rt1[], int index[], int nloops, int incr);
void	radix32_dit_passB	(double a[], int n, int nrt, int nrt_bits, struct complex rt0[], struct complex rt1[], int index[], int nloops, int incr);
void	radix64_dit_passB	(double a[], int n, int nrt, int nrt_bits, struct complex rt0[], struct complex rt1[], int index[], int nloops, int incr);

void pair_square2 (double *x1, double *y1, double *x2, double *y2, double c, double s);
void pair_square2A(double *x1, double *y1, double *x2, double *y2,
                   double *x3, double *y3, double *x4, double *y4,
                   double c, double s);
void pair_square2B(double *x1, double *y1, double *x2, double *y2,
                   double *x3, double *y3, double *x4, double *y4,
                   double c, double s);

void	radix16_wrapper_ini(int n, int radix0, int iblock, int nradices_prim, int radix_prim[], int i[], int j1[], int j2[], int j2_start[], int k[], int m[], int blocklen[], int blocklen_sum[]);
void	radix32_wrapper_ini(int n, int radix0, int iblock, int nradices_prim, int radix_prim[], int i[], int j1[], int j2[], int j2_start[], int k[], int m[], int blocklen[], int blocklen_sum[]);
void	radix64_wrapper_ini(int n, int radix0, int iblock, int nradices_prim, int radix_prim[], int i[], int j1[], int j2[], int j2_start[], int k[], int m[], int blocklen[], int blocklen_sum[]);

void	radix16_wrapper_square	(double a[], int n, int radix0, int nrt, int nrt_bits, struct complex rt0[], struct complex rt1[], int nradices_prim, int radix_prim[], int ws_i, int ws_j1, int ws_j2, int ws_j2_start, int ws_k, int ws_m, int ws_blocklen, int ws_blocklen_sum);
void	radix32_wrapper_square	(double a[], int n, int radix0, int nrt, int nrt_bits, struct complex rt0[], struct complex rt1[], int nradices_prim, int radix_prim[], int ws_i, int ws_j1, int ws_j2, int ws_j2_start, int ws_k, int ws_m, int ws_blocklen, int ws_blocklen_sum);
void	radix64_wrapper_square	(double a[], int n, int radix0, int nrt, int nrt_bits, struct complex rt0[], struct complex rt1[], int nradices_prim, int radix_prim[], int ws_i, int ws_j1, int ws_j2, int ws_j2_start, int ws_k, int ws_m, int ws_blocklen, int ws_blocklen_sum);

void	radix16_dyadic_square	(double a[], int n, int radix0, int nrt, int nrt_bits, struct complex rt0[], struct complex rt1[], int index[]);

/* Each of the final-DIT-pass/propagate-carries/initial-DIF-pass routines comes in both an error-checking and non-error-checking version: */

int	radix5_ditN_cy_dif1	(double a[], int n, int nwt, int nwt_bits, double wt0[], double wt1[], int si[], int nrt, int nrt_bits, struct complex rt0[],
			struct complex rt1[], double base[] , double baseinv[] , int iter, double *fracmax, int p);
int	radix6_ditN_cy_dif1	(double a[], int n, int nwt, int nwt_bits, double wt0[], double wt1[], int si[], int nrt, int nrt_bits, struct complex rt0[],
			struct complex rt1[], double base[] , double baseinv[] , int iter, double *fracmax, int p);
int	radix7_ditN_cy_dif1	(double a[], int n, int nwt, int nwt_bits, double wt0[], double wt1[], int si[], int nrt, int nrt_bits, struct complex rt0[],
			struct complex rt1[], double base[] , double baseinv[] , int iter, double *fracmax, int p);
int	radix8_ditN_cy_dif1	(double a[], int n, int nwt, int nwt_bits, double wt0[], double wt1[], int si[], int nrt, int nrt_bits, struct complex rt0[],
			struct complex rt1[], double base[] , double baseinv[] , int iter, double *fracmax, int p);
int	radix9_ditN_cy_dif1	(double a[], int n, int nwt, int nwt_bits, double wt0[], double wt1[], int si[], int nrt, int nrt_bits, struct complex rt0[],
			struct complex rt1[], double base[] , double baseinv[] , int iter, double *fracmax, int p);
int	radix10_ditN_cy_dif1	(double a[], int n, int nwt, int nwt_bits, double wt0[], double wt1[], int si[], int nrt, int nrt_bits, struct complex rt0[],
			struct complex rt1[], double base[] , double baseinv[] , int iter, double *fracmax, int p);
int	radix11_ditN_cy_dif1	(double a[], int n, int nwt, int nwt_bits, double wt0[], double wt1[], int si[], int nrt, int nrt_bits, struct complex rt0[],
			struct complex rt1[], double base[] , double baseinv[] , int iter, double *fracmax, int p);
int	radix12_ditN_cy_dif1	(double a[], int n, int nwt, int nwt_bits, double wt0[], double wt1[], int si[], int nrt, int nrt_bits, struct complex rt0[],
			struct complex rt1[], double base[] , double baseinv[] , int iter, double *fracmax, int p);
int	radix13_ditN_cy_dif1	(double a[], int n, int nwt, int nwt_bits, double wt0[], double wt1[], int si[], int nrt, int nrt_bits, struct complex rt0[],
			struct complex rt1[], double base[] , double baseinv[] , int iter, double *fracmax, int p);
int	radix14_ditN_cy_dif1	(double a[], int n, int nwt, int nwt_bits, double wt0[], double wt1[], int si[], int nrt, int nrt_bits, struct complex rt0[],
			struct complex rt1[], double base[] , double baseinv[] , int iter, double *fracmax, int p);
int	radix15_ditN_cy_dif1	(double a[], int n, int nwt, int nwt_bits, double wt0[], double wt1[], int si[], int nrt, int nrt_bits, struct complex rt0[],
			struct complex rt1[], double base[] , double baseinv[] , int iter, double *fracmax, int p);
int	radix16_ditN_cy_dif1	(double a[], int n, int nwt, int nwt_bits, double wt0[], double wt1[], int si[], int nrt, int nrt_bits, struct complex rt0[],
			struct complex rt1[], double base[] , double baseinv[] , int iter, double *fracmax, int p);
#if USE_NEW_DIT
int	radix18_ditN_cy_dif1	(double a[], int n, int nwt, int nwt_bits, double wt0[], double wt1[], int si[], double base[] , double baseinv[] , int iter, double *fracmax, int p);
int	radix20_ditN_cy_dif1	(double a[], int n, int nwt, int nwt_bits, double wt0[], double wt1[], int si[], double base[] , double baseinv[] , int iter, double *fracmax, int p);
int	radix22_ditN_cy_dif1	(double a[], int n, int nwt, int nwt_bits, double wt0[], double wt1[], int si[], double base[] , double baseinv[] , int iter, double *fracmax, int p);
int	radix24_ditN_cy_dif1	(double a[], int n, int nwt, int nwt_bits, double wt0[], double wt1[], int si[], double base[] , double baseinv[] , int iter, double *fracmax, int p);
int	radix26_ditN_cy_dif1	(double a[], int n, int nwt, int nwt_bits, double wt0[], double wt1[], int si[], double base[] , double baseinv[] , int iter, double *fracmax, int p);
int	radix28_ditN_cy_dif1	(double a[], int n, int nwt, int nwt_bits, double wt0[], double wt1[], int si[], double base[] , double baseinv[] , int iter, double *fracmax, int p);
int	radix30_ditN_cy_dif1	(double a[], int n, int nwt, int nwt_bits, double wt0[], double wt1[], int si[], double base[] , double baseinv[] , int iter, double *fracmax, int p);
int	radix36_ditN_cy_dif1	(double a[], int n, int nwt, int nwt_bits, double wt0[], double wt1[], int si[], double base[] , double baseinv[] , int iter, double *fracmax, int p);
#endif
int	radix32_ditN_cy_dif1	(double a[], int n, int nwt, int nwt_bits, double wt0[], double wt1[], int si[], int nrt, int nrt_bits, struct complex rt0[],
			struct complex rt1[], double base[] , double baseinv[] , int iter, double *fracmax, int p);

int	radix5_ditN_cy_dif1_nochk(double a[], int n, int nwt, int nwt_bits, double wt0[], double wt1[], int si[], int nrt, int nrt_bits, struct complex rt0[],
			struct complex rt1[], double base[] , double baseinv[] , int iter		   , int p);
int	radix6_ditN_cy_dif1_nochk(double a[], int n, int nwt, int nwt_bits, double wt0[], double wt1[], int si[], int nrt, int nrt_bits, struct complex rt0[],
			struct complex rt1[], double base[] , double baseinv[] , int iter		   , int p);
int	radix7_ditN_cy_dif1_nochk(double a[], int n, int nwt, int nwt_bits, double wt0[], double wt1[], int si[], int nrt, int nrt_bits, struct complex rt0[],
			struct complex rt1[], double base[] , double baseinv[] , int iter		   , int p);
int	radix8_ditN_cy_dif1_nochk(double a[], int n, int nwt, int nwt_bits, double wt0[], double wt1[], int si[], int nrt, int nrt_bits, struct complex rt0[],
			struct complex rt1[], double base[] , double baseinv[] , int iter		   , int p);
int	radix9_ditN_cy_dif1_nochk(double a[], int n, int nwt, int nwt_bits, double wt0[], double wt1[], int si[], int nrt, int nrt_bits, struct complex rt0[],
			struct complex rt1[], double base[] , double baseinv[] , int iter		   , int p);
int	radix10_ditN_cy_dif1_nochk(double a[], int n, int nwt, int nwt_bits, double wt0[], double wt1[], int si[], int nrt, int nrt_bits, struct complex rt0[],
			struct complex rt1[], double base[] , double baseinv[] , int iter		   , int p);
int	radix11_ditN_cy_dif1_nochk(double a[], int n, int nwt, int nwt_bits, double wt0[], double wt1[], int si[], int nrt, int nrt_bits, struct complex rt0[],
			struct complex rt1[], double base[] , double baseinv[] , int iter		   , int p);
int	radix12_ditN_cy_dif1_nochk(double a[], int n, int nwt, int nwt_bits, double wt0[], double wt1[], int si[], int nrt, int nrt_bits, struct complex rt0[],
			struct complex rt1[], double base[] , double baseinv[] , int iter		   , int p);
int	radix13_ditN_cy_dif1_nochk(double a[], int n, int nwt, int nwt_bits, double wt0[], double wt1[], int si[], int nrt, int nrt_bits, struct complex rt0[],
			struct complex rt1[], double base[] , double baseinv[] , int iter		   , int p);
int	radix14_ditN_cy_dif1_nochk(double a[], int n, int nwt, int nwt_bits, double wt0[], double wt1[], int si[], int nrt, int nrt_bits, struct complex rt0[],
			struct complex rt1[], double base[] , double baseinv[] , int iter		   , int p);
int	radix15_ditN_cy_dif1_nochk(double a[], int n, int nwt, int nwt_bits, double wt0[], double wt1[], int si[], int nrt, int nrt_bits, struct complex rt0[],
			struct complex rt1[], double base[] , double baseinv[] , int iter		   , int p);
int	radix16_ditN_cy_dif1_nochk(double a[], int n, int nwt, int nwt_bits, double wt0[], double wt1[], int si[], int nrt, int nrt_bits, struct complex rt0[],
			struct complex rt1[], double base[] , double baseinv[] , int iter		   , int p);
#if USE_NEW_DIT
int	radix18_ditN_cy_dif1_nochk(double a[], int n, int nwt, int nwt_bits, double wt0[], double wt1[], int si[], double base[] , double baseinv[] , int iter, int p);
int	radix20_ditN_cy_dif1_nochk(double a[], int n, int nwt, int nwt_bits, double wt0[], double wt1[], int si[], double base[] , double baseinv[] , int iter, int p);
int	radix22_ditN_cy_dif1_nochk(double a[], int n, int nwt, int nwt_bits, double wt0[], double wt1[], int si[], double base[] , double baseinv[] , int iter, int p);
int	radix24_ditN_cy_dif1_nochk(double a[], int n, int nwt, int nwt_bits, double wt0[], double wt1[], int si[], double base[] , double baseinv[] , int iter, int p);
int	radix26_ditN_cy_dif1_nochk(double a[], int n, int nwt, int nwt_bits, double wt0[], double wt1[], int si[], double base[] , double baseinv[] , int iter, int p);
int	radix28_ditN_cy_dif1_nochk(double a[], int n, int nwt, int nwt_bits, double wt0[], double wt1[], int si[], double base[] , double baseinv[] , int iter, int p);
int	radix30_ditN_cy_dif1_nochk(double a[], int n, int nwt, int nwt_bits, double wt0[], double wt1[], int si[], double base[] , double baseinv[] , int iter, int p);
int	radix36_ditN_cy_dif1_nochk(double a[], int n, int nwt, int nwt_bits, double wt0[], double wt1[], int si[], double base[] , double baseinv[] , int iter, int p);
#endif
int	radix32_ditN_cy_dif1_nochk(double a[], int n, int nwt, int nwt_bits, double wt0[], double wt1[], int si[], int nrt, int nrt_bits, struct complex rt0[],
			struct complex rt1[], double base[] , double baseinv[] , int iter		   , int p);

void	pair_square(double *x1, double *y1, double *x2, double *y2, double c, double s);

#endif	/* Mlucas_h_included */
