/*******************************************************************************
*                                                                              *
*   (C) 1997-2002 by Ernst W. Mayer.                                           *
*                                                                              *
*  This program is free software; you can redistribute it and/or modify it     *
*  under the terms of the GNU General Public License as published by the       *
*  Free Software Foundation; either version 2 of the License, or (at your      *
*  option) any later version.                                                  *
*                                                                              *
*  This program is distributed in the hope that it will be useful, but WITHOUT *
*  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or       *
*  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for   *
*  more details.                                                               *
*                                                                              *
*  You should have received a copy of the GNU General Public License along     *
*  with this program; see the file GPL.txt.  If not, you may view one at       *
*  http://www.fsf.org/licenses/licenses.html, or obtain one by writing to the  *
*  Free Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA     *
*  02111-1307, USA.                                                            *
*                                                                              *
*******************************************************************************/

/****************************************************************************
 * We now include this header file if it was not included before.
 ****************************************************************************/
#ifndef align_h_included
#define align_h_included

/* These are to force alignment of array data on 64-byte boundaries. We cast pointers to 64-bit ints
   to accommodate architectures which use 64-bit address arithmetic. */

#define ALLOC_INT(_n)		(memset((int           *)malloc((_n)*sizeof(int           )+64), 0, (_n)*sizeof(int           )+64))
#define ALIGN_INT(_p)		(       (int           *)(((long)(_p) | 63)+1))

#define ALLOC_DOUBLE(_n)	(memset((double        *)malloc((_n)*sizeof(double        )+64), 0, (_n)*sizeof(double        )+64))
#define ALIGN_DOUBLE(_p)	(       (double        *)(((long)(_p) | 63)+1))

#define ALLOC_COMPLEX(_n)	(memset((struct complex*)malloc((_n)*sizeof(struct complex)+64), 0, (_n)*sizeof(struct complex)+64))
#define ALIGN_COMPLEX(_p)	(       (struct complex*)(((long)(_p) | 63)+1))


/*
 On the x86 family, alignment of the stack is very important
 This uses the GNU gcc  __builtin_alloca function to align doubles properly
 This is taken from GNU/FFTW package
*/
#if(defined(__GNUC__) || defined(__GNUG__))
#	if (defined(__i386))
#		define HACK_ALIGN_STACK_EVEN(){					\
		if( (((long) (__builtin_alloca(0))) & 0x7)) __builtin_alloca(4);\
		}

#		define HACK_ALIGN_STACK_ODD() {					\
		if(!(((long) (__builtin_alloca(0))) & 0x7)) __builtin_alloca(4);\
		}
#	else
#		define HACK_ALIGN_STACK_EVEN() /* */
#		define HACK_ALIGN_STACK_ODD() /* */
#	endif
#else
#	define HACK_ALIGN_STACK_EVEN() /* */
#	define HACK_ALIGN_STACK_ODD() /* */
#endif


#endif	/* align_h_included */
