/*******************************************************************************
*                                                                              *
*   (C) 1997-2002 by Ernst W. Mayer.                                           *
*                                                                              *
*  This program is free software; you can redistribute it and/or modify it     *
*  under the terms of the GNU General Public License as published by the       *
*  Free Software Foundation; either version 2 of the License, or (at your      *
*  option) any later version.                                                  *
*                                                                              *
*  This program is distributed in the hope that it will be useful, but WITHOUT *
*  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or       *
*  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for   *
*  more details.                                                               *
*                                                                              *
*  You should have received a copy of the GNU General Public License along     *
*  with this program; see the file GPL.txt.  If not, you may view one at       *
*  http://www.fsf.org/licenses/licenses.html, or obtain one by writing to the  *
*  Free Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA     *
*  02111-1307, USA.                                                            *
*                                                                              *
*******************************************************************************/

/****************************************************************************
 * We now include this header file if it was not included before.
 ****************************************************************************/
#ifndef dft_macro_included
#define dft_macro_included

/* All these radix-N macros are structured so that any 2 of the 3 sets of arguments (A, t, B) may point to the same set of
   addresses, i.e. so the DFT may be done in place (A == B) or re-using a single set of temporaries (A == t or B == t).
   By "in place" we mean that A == B up to an arbitrary permutation of array indices (or element subscripts). */

/****** RADIX = 3: ALLOWS IN-PLACE ******/

#define RADIX_03_DFT(\
	__Ar0,__Ai0,\
	__Ar1,__Ai1,\
	__Ar2,__Ai2,\
	__tr0,__ti0,\
	__tr1,__ti1,\
	__tr2,__ti2,\
	__Br0,__Bi0,\
	__Br1,__Bi1,\
	__Br2,__Bi2,\
	__rt,__it)\
{\
	  __rt  = __Ar2;		__it  =__Ai2;		\
	  __tr2 =__Ar1 - __rt;		__ti2 =__Ai1 - __it;	\
	  __tr1 =__Ar1 + __rt;		__ti1 =__Ai1 + __it;	\
	  __Br0 =__Ar0 + __tr1;		__Bi0 =__Ai0 + __ti1;	\
	  __Br1 =__Br0 + __tr1*c3m1;	__Bi1 =__Bi0 + __ti1*c3m1;\
	  __rt  =c*__tr2;		__it  =c*__ti2;		\
	  __Br2 =__Br1 + __it;		__Bi2 =__Bi1 - __rt;	\
	  __Br1 =__Br1 - __it;		__Bi1 =__Bi1 + __rt;	\
}

#define RADIX_03_DFT_PFETCH(\
	__Ar0,__Ai0,\
	__Ar1,__Ai1,\
	__Ar2,__Ai2,\
	__tr0,__ti0,\
	__tr1,__ti1,\
	__tr2,__ti2,\
	__Br0,__Bi0,\
	__Br1,__Bi1,\
	__Br2,__Bi2,\
	__rt,__it,\
	__add_offset)\
{\
	  __rt  = __Ar2;		__it  =__Ai2;		\
	  __tr2 =__Ar1 - __rt;		__ti2 =__Ai1 - __it;	\
	  __tr1 =__Ar1 + __rt;		__ti1 =__Ai1 + __it;	\
	  __Br0 =__Ar0 + __tr1;		__Bi0 =__Ai0 + __ti1;	\
	  __Br1 =__Br0 + __tr1*c3m1;	__Bi1 =__Bi0 + __ti1*c3m1;\
\
add = add0 + __add_offset;	prefetch_p_doubles(add);	\
\
	  __rt  =c*__tr2;		__it  =c*__ti2;		\
	  __Br2 =__Br1 + __it;		__Bi2 =__Bi1 - __rt;	\
	  __Br1 =__Br1 - __it;		__Bi1 =__Bi1 + __rt;	\
}

/****** RADIX = 4: ******/

#define RADIX_04_DIF(\
	__Ar0,__Ai0,\
	__Ar1,__Ai1,\
	__Ar2,__Ai2,\
	__Ar3,__Ai3,\
	__Br0,__Bi0,\
	__Br1,__Bi1,\
	__Br2,__Bi2,\
	__Br3,__Bi3,\
	__rt,__it)\
{\
	__rt = __Ar2;	__Ar2 =__Ar0 - __rt;	__Ar0 = __Ar0 + __rt;	\
	__it = __Ai2;	__Ai2 =__Ai0 - __it;	__Ai0 = __Ai0 + __it;	\
	__rt = __Ar3;	__Ar3 =__Ar1 - __rt;	__Ar1 = __Ar1 + __rt;	\
	__it = __Ai3;	__Ai3 =__Ai1 - __it;	__Ai1 = __Ai1 + __it;	\
\
	__rt  = __Ar1;	__Br1 = __Ar0 - __rt;	__Br0 = __Ar0 + __rt;	\
	__it  = __Ai1;	__Bi1 = __Ai0 - __it;	__Bi0 = __Ai0 + __it;	\
	/* mpy by I is inlined here...	*/\
	__rt  = __Ar3;	__Br3 = __Ar2 + __Ai3;	__Br2 = __Ar2 - __Ai3;	\
			__Bi3 = __Ai2 - __rt;	__Bi2 = __Ai2 + __rt;	\
}

#define RADIX_04_DIF_PFETCH(\
	__Ar0,__Ai0,\
	__Ar1,__Ai1,\
	__Ar2,__Ai2,\
	__Ar3,__Ai3,\
	__Br0,__Bi0,\
	__Br1,__Bi1,\
	__Br2,__Bi2,\
	__Br3,__Bi3,\
	__rt,__it,\
	__add_offset0,__add_offset1)\
{\
	__rt = __Ar2;	__Ar2 =__Ar0 - __rt;	__Ar0 = __Ar0 + __rt;	\
	__it = __Ai2;	__Ai2 =__Ai0 - __it;	__Ai0 = __Ai0 + __it;	\
	__rt = __Ar3;	__Ar3 =__Ar1 - __rt;	__Ar1 = __Ar1 + __rt;	\
	__it = __Ai3;	__Ai3 =__Ai1 - __it;	__Ai1 = __Ai1 + __it;	\
\
add = add0 + __add_offset0;	prefetch_p_doubles(add);	\
\
	__rt  = __Ar1;	__Br1 = __Ar0 - __rt;	__Br0 = __Ar0 + __rt;	\
	__it  = __Ai1;	__Bi1 = __Ai0 - __it;	__Bi0 = __Ai0 + __it;	\
	/* mpy by I is inlined here...	*/\
	__rt  = __Ar3;	__Br3 = __Ar2 + __Ai3;	__Br2 = __Ar2 - __Ai3;	\
			__Bi3 = __Ai2 - __rt;	__Bi2 = __Ai2 + __rt;	\
\
add = add0 + __add_offset1;	prefetch_p_doubles(add);	\
}

#define RADIX_04_DIT(\
	__Ar0,__Ai0,\
	__Ar1,__Ai1,\
	__Ar2,__Ai2,\
	__Ar3,__Ai3,\
	__Br0,__Bi0,\
	__Br1,__Bi1,\
	__Br2,__Bi2,\
	__Br3,__Bi3,\
	__rt,__it)\
{\
	__rt = __Ar1;	__Ar1 =__Ar0 - __rt;	__Ar0 = __Ar0 + __rt;	\
	__it = __Ai1;	__Ai1 =__Ai0 - __it;	__Ai0 = __Ai0 + __it;	\
	__rt = __Ar3;	__Ar3 =__Ar2 - __rt;	__Ar2 = __Ar2 + __rt;	\
	__it = __Ai3;	__Ai3 =__Ai2 - __it;	__Ai2 = __Ai2 + __it;	\
\
	__rt  = __Ar2;	__Br2 = __Ar0 - __rt;	__Br0 = __Ar0 + __rt;	\
	__it  = __Ai2;	__Bi2 = __Ai0 - __it;	__Bi0 = __Ai0 + __it;	\
	/* mpy by I is inlined here...	*/\
	__rt  = __Ar3;	__Br3 = __Ar1 - __Ai3;	__Br1 = __Ar1 + __Ai3;	\
			__Bi3 = __Ai1 + __rt;	__Bi1 = __Ai1 - __rt;	\
}

/****** RADIX = 5: ******/

#define RADIX_05_DFT(\
	__Ar0,__Ai0,\
	__Ar1,__Ai1,\
	__Ar2,__Ai2,\
	__Ar3,__Ai3,\
	__Ar4,__Ai4,\
	__Br0,__Bi0,\
	__Br1,__Bi1,\
	__Br2,__Bi2,\
	__Br3,__Bi3,\
	__Br4,__Bi4,\
	__rt,__it)\
{\
	  __Br0 = __Ar0;		__Bi0 = __Ai0;				\
	  __rt  = __Ar4;		__it  = __Ai4;				\
	  __Br4 = __Ar1 - __rt;		__Bi4 = __Ai1 - __it;			\
	  __Br1 = __Ar1 + __rt;		__Bi1 = __Ai1 + __it;			\
	  __rt  = __Ar3;		__it  = __Ai3;				\
	  __Br3 = __Ar2 - __rt;		__Bi3 = __Ai2 - __it;			\
	  __Br2 = __Ar2 + __rt;		__Bi2 = __Ai2 + __it;			\
	  __rt  = __Br1 + __Br2;	__it  = __Bi1 + __Bi2;			\
	  __Br0 = __Br0 + __rt;		__Bi0 = __Bi0 + __it;		/* y0 */\
	  __rt  = __Br0 + __rt  *cc1;	__it  = __Bi0 + __it  *cc1;		\
	  __Br2 =(__Br1 - __Br2)*cc2;	__Bi2 =(__Bi1 - __Bi2)*cc2;		\
	  __Br1 = __rt  + __Br2;	__Bi1 = __it  + __Bi2;			\
	  __Br2 = __rt  - __Br2;	__Bi2 = __it  - __Bi2;			\
	  __rt  =(__Br4 - __Br3)* s2;	__it  =(__Bi4 - __Bi3)* s2;		\
	  __Br3 = __rt  + __Br3 *ss1;	__Bi3 = __it  + __Bi3 *ss1;		\
	  __Br4 = __rt  - __Br4 *ss2;	__Bi4 = __it  - __Bi4 *ss2;		\
	  __rt  = __Br4;		__it  = __Bi4;				\
	  __Br4 = __Br1 + __Bi3;	__Bi4 = __Bi1 - __Br3;		/* y4 - note swap with y3 below! */\
	  __Br1 = __Br1 - __Bi3;	__Bi1 = __Bi1 + __Br3;		/* y1 */\
	  __Br3 = __Br2 + __it;		__Bi3 = __Bi2 - __rt;		/* y3 */\
	  __Br2 = __Br2 - __it;		__Bi2 = __Bi2 + __rt;		/* y2 */\
}

#define RADIX_05_DFT_PFETCH(\
	__Ar0,__Ai0,\
	__Ar1,__Ai1,\
	__Ar2,__Ai2,\
	__Ar3,__Ai3,\
	__Ar4,__Ai4,\
	__Br0,__Bi0,\
	__Br1,__Bi1,\
	__Br2,__Bi2,\
	__Br3,__Bi3,\
	__Br4,__Bi4,\
	__rt,__it,\
	__add_offset0,__add_offset1)\
{\
	  __Br0 = __Ar0;		__Bi0 = __Ai0;				\
	  __rt  = __Ar4;		__it  = __Ai4;				\
	  __Br4 = __Ar1 - __rt;		__Bi4 = __Ai1 - __it;			\
	  __Br1 = __Ar1 + __rt;		__Bi1 = __Ai1 + __it;			\
	  __rt  = __Ar3;		__it  = __Ai3;				\
	  __Br3 = __Ar2 - __rt;		__Bi3 = __Ai2 - __it;			\
	  __Br2 = __Ar2 + __rt;		__Bi2 = __Ai2 + __it;			\
	  __rt  = __Br1 + __Br2;	__it  = __Bi1 + __Bi2;			\
	  __Br0 = __Br0 + __rt;		__Bi0 = __Bi0 + __it;		/* y0 */\
	  __rt  = __Br0 + __rt  *cc1;	__it  = __Bi0 + __it  *cc1;		\
	  __Br2 =(__Br1 - __Br2)*cc2;	__Bi2 =(__Bi1 - __Bi2)*cc2;		\
\
add = add0 + __add_offset0;	prefetch_p_doubles(add);	\
\
	  __Br1 = __rt  + __Br2;	__Bi1 = __it  + __Bi2;			\
	  __Br2 = __rt  - __Br2;	__Bi2 = __it  - __Bi2;			\
	  __rt  =(__Br4 - __Br3)* s2;	__it  =(__Bi4 - __Bi3)* s2;		\
	  __Br3 = __rt  + __Br3 *ss1;	__Bi3 = __it  + __Bi3 *ss1;		\
	  __Br4 = __rt  - __Br4 *ss2;	__Bi4 = __it  - __Bi4 *ss2;		\
	  __rt  = __Br3;		__it  = __Bi3;				\
	  __Br3 = __Br1 + __it;		__Bi3 = __Bi1 - __rt;		/* y4 - note swap with y3 below! */\
	  __Br1 = __Br1 - __it;		__Bi1 = __Bi1 + __rt;		/* y1 */\
	  __rt  = __Br4;		__it  = __Bi4;				\
	  __Br4 = __Br2 + __it;		__Bi4 = __Bi2 - __rt;		/* y3 */\
	  __Br2 = __Br2 - __it;		__Bi2 = __Bi2 + __rt;		/* y2 */\
\
add = add0 + __add_offset1;	prefetch_p_doubles(add);	\
\
}

/****** RADIX = 7: ******/

#define RADIX_07_DFT(\
	__Ar0,__Ai0,\
	__Ar1,__Ai1,\
	__Ar2,__Ai2,\
	__Ar3,__Ai3,\
	__Ar4,__Ai4,\
	__Ar5,__Ai5,\
	__Ar6,__Ai6,\
	__tr0,__ti0,\
	__tr1,__ti1,\
	__tr2,__ti2,\
	__tr3,__ti3,\
	__tr4,__ti4,\
	__tr5,__ti5,\
	__tr6,__ti6,\
	__Br0,__Bi0,\
	__Br1,__Bi1,\
	__Br2,__Bi2,\
	__Br3,__Bi3,\
	__Br4,__Bi4,\
	__Br5,__Bi5,\
	__Br6,__Bi6,\
	__rt,__it,\
	__re,__im)\
{\
	  __tr0 = __Ar0;					__ti0 = __Ai0;				\
	  __tr6 = __Ar1 - __Ar6;				__ti6 = __Ai1 - __Ai6;	/* x1 - x6 */	\
	  __tr1 = __Ar1 + __Ar6;				__ti1 = __Ai1 + __Ai6;	/* x1 + x6 */	\
\
	  __tr5 = __Ar2 - __Ar5;				__ti5 = __Ai2 - __Ai5;	/* x2 - x5 */	\
	  __tr2 = __Ar2 + __Ar5;				__ti2 = __Ai2 + __Ai5;	/* x2 + x5 */	\
\
	  __tr4 = __Ar3 - __Ar4;				__ti4 = __Ai3 - __Ai4;	/* x3 - x4 */	\
	  __tr3 = __Ar3 + __Ar4;				__ti3 = __Ai3 + __Ai4;	/* x3 + x4 */	\
\
	  __Br0 = __tr0 + __tr1 + __tr2 + __tr3;		__Bi0 = __ti0 + __ti1 + __ti2 + __ti3;			\
	  __rt  = __tr0 + cc1*__tr1 + cc2*__tr2 + cc3*__tr3;	__it  = __ti0 + cc1*__ti1 + cc2*__ti2 + cc3*__ti3;	\
	  __re  = __tr0 + cc2*__tr1 + cc3*__tr2 + cc1*__tr3;	__im  = __ti0 + cc2*__ti1 + cc3*__ti2 + cc1*__ti3;	\
	  __tr0 = __tr0 + cc3*__tr1 + cc1*__tr2 + cc2*__tr3;	__ti0 = __ti0 + cc3*__ti1 + cc1*__ti2 + cc2*__ti3;	\
\
	  __tr1 = ss1*__tr6 + ss2*__tr5 + ss3*__tr4;		__ti1 = ss1*__ti6 + ss2*__ti5 + ss3*__ti4;	\
	  __tr2 = ss2*__tr6 - ss3*__tr5 - ss1*__tr4;		__ti2 = ss2*__ti6 - ss3*__ti5 - ss1*__ti4;	\
	  __tr3 = ss3*__tr6 - ss1*__tr5 + ss2*__tr4;		__ti3 = ss3*__ti6 - ss1*__ti5 + ss2*__ti4;	\
	/* Output permutation causes signs to get flipped here: */\
	  __Br1 = __rt  - __ti1;				__Bi1 = __it  + __tr1;	\
	  __Br2 = __re  - __ti2;				__Bi2 = __im  + __tr2;	\
	  __Br3 = __tr0 - __ti3;				__Bi3 = __ti0 + __tr3;	\
	  __Br4 = __tr0 + __ti3;				__Bi4 = __ti0 - __tr3;	\
	  __Br5 = __re  + __ti2;				__Bi5 = __im  - __tr2;	\
	  __Br6 = __rt  + __ti1;				__Bi6 = __it  - __tr1;	\
}

#define RADIX_07_DFT_PFETCH(\
	__Ar0,__Ai0,\
	__Ar1,__Ai1,\
	__Ar2,__Ai2,\
	__Ar3,__Ai3,\
	__Ar4,__Ai4,\
	__Ar5,__Ai5,\
	__Ar6,__Ai6,\
	__tr0,__ti0,\
	__tr1,__ti1,\
	__tr2,__ti2,\
	__tr3,__ti3,\
	__tr4,__ti4,\
	__tr5,__ti5,\
	__tr6,__ti6,\
	__Br0,__Bi0,\
	__Br1,__Bi1,\
	__Br2,__Bi2,\
	__Br3,__Bi3,\
	__Br4,__Bi4,\
	__Br5,__Bi5,\
	__Br6,__Bi6,\
	__rt,__it,\
	__re,__im,\
	__add_offset0,__add_offset1,__add_offset2)\
{\
	  __tr0 = __Ar0;					__ti0 = __Ai0;				\
	  __tr6 = __Ar1 - __Ar6;				__ti6 = __Ai1 - __Ai6;	/* x1 - x6 */	\
	  __tr1 = __Ar1 + __Ar6;				__ti1 = __Ai1 + __Ai6;	/* x1 + x6 */	\
\
	  __tr5 = __Ar2 - __Ar5;				__ti5 = __Ai2 - __Ai5;	/* x2 - x5 */	\
	  __tr2 = __Ar2 + __Ar5;				__ti2 = __Ai2 + __Ai5;	/* x2 + x5 */	\
\
	  __tr4 = __Ar3 - __Ar4;				__ti4 = __Ai3 - __Ai4;	/* x3 - x4 */	\
	  __tr3 = __Ar3 + __Ar4;				__ti3 = __Ai3 + __Ai4;	/* x3 + x4 */	\
\
add = add0 + __add_offset0;	prefetch_p_doubles(add);	\
\
	  __Br0 = __tr0 + __tr1 + __tr2 + __tr3;		__Bi0 = __ti0 + __ti1 + __ti2 + __ti3;			\
	  __rt  = __tr0 + cc1*__tr1 + cc2*__tr2 + cc3*__tr3;	__it  = __ti0 + cc1*__ti1 + cc2*__ti2 + cc3*__ti3;	\
	  __re  = __tr0 + cc2*__tr1 + cc3*__tr2 + cc1*__tr3;	__im  = __ti0 + cc2*__ti1 + cc3*__ti2 + cc1*__ti3;	\
	  __tr0 = __tr0 + cc3*__tr1 + cc1*__tr2 + cc2*__tr3;	__ti0 = __ti0 + cc3*__ti1 + cc1*__ti2 + cc2*__ti3;	\
\
add = add0 + __add_offset1;	prefetch_p_doubles(add);	\
\
	  __tr1 = ss1*__tr6 + ss2*__tr5 + ss3*__tr4;		__ti1 = ss1*__ti6 + ss2*__ti5 + ss3*__ti4;	\
	  __tr2 = ss2*__tr6 - ss3*__tr5 - ss1*__tr4;		__ti2 = ss2*__ti6 - ss3*__ti5 - ss1*__ti4;	\
	  __tr3 = ss3*__tr6 - ss1*__tr5 + ss2*__tr4;		__ti3 = ss3*__ti6 - ss1*__ti5 + ss2*__ti4;	\
\
add = add0 + __add_offset2;	prefetch_p_doubles(add);	\
\
	/* Output permutation causes signs to get flipped here: */\
	  __Br1 = __rt  - __ti1;				__Bi1 = __it  + __tr1;	\
	  __Br2 = __re  - __ti2;				__Bi2 = __im  + __tr2;	\
	  __Br3 = __tr0 - __ti3;				__Bi3 = __ti0 + __tr3;	\
	  __Br4 = __tr0 + __ti3;				__Bi4 = __ti0 - __tr3;	\
	  __Br5 = __re  + __ti2;				__Bi5 = __im  - __tr2;	\
	  __Br6 = __rt  + __ti1;				__Bi6 = __it  - __tr1;	\
}

/****** RADIX = 8: ******/

#define RADIX_08_DIF(\
	__Ar0,__Ai0,\
	__Ar1,__Ai1,\
	__Ar2,__Ai2,\
	__Ar3,__Ai3,\
	__Ar4,__Ai4,\
	__Ar5,__Ai5,\
	__Ar6,__Ai6,\
	__Ar7,__Ai7,\
	__tr0,__ti0,\
	__tr1,__ti1,\
	__tr2,__ti2,\
	__tr3,__ti3,\
	__tr4,__ti4,\
	__tr5,__ti5,\
	__tr6,__ti6,\
	__tr7,__ti7,\
	__Br0,__Bi0,\
	__Br1,__Bi1,\
	__Br2,__Bi2,\
	__Br3,__Bi3,\
	__Br4,__Bi4,\
	__Br5,__Bi5,\
	__Br6,__Bi6,\
	__Br7,__Bi7,\
	__rt,__it)\
{\
	__tr0 = __Ar0;			__ti0 = __Ai0;			\
	__rt  = __Ar4;			__it  = __Ai4;			\
	__tr1 = __tr0 - __rt;		__ti1 = __ti0 - __it;		\
	__tr0 = __tr0 + __rt;		__ti0 = __ti0 + __it;		\
\
	__tr2 = __Ar2;			__ti2 = __Ai2;			\
	__rt  = __Ar6;			__it  = __Ai6;			\
	__tr3 = __tr2 - __rt;		__ti3 = __ti2 - __it;		\
	__tr2 = __tr2 + __rt;		__ti2 = __ti2 + __it;		\
\
	__tr4 = __Ar1;			__ti4 = __Ai1;			\
	__rt  = __Ar5;			__it  = __Ai5;			\
	__tr5 = __tr4 - __rt;		__ti5 = __ti4 - __it;		\
	__tr4 = __tr4 + __rt;		__ti4 = __ti4 + __it;		\
\
	__tr6 = __Ar3;			__ti6 = __Ai3;			\
	__rt  = __Ar7;			__it  = __Ai7;			\
	__tr7 = __tr6 - __rt;		__ti7 = __ti6 - __it;		\
	__tr6 = __tr6 + __rt;		__ti6 = __ti6 + __it;		\
/*       combine to get the 2 length-4 transforms...	*/		\
	__rt  = __tr2;			__it  = __ti2;			\
	__tr2 = __tr0 - __rt;		__ti2 = __ti0 - __it;		\
	__tr0 = __tr0 + __rt;		__ti0 = __ti0 + __it;		\
\
	__rt  = __tr3;			__it  = __ti3;			\
	__tr3 = __tr1 + __it;		__ti3 = __ti1 - __rt;		\
	__tr1 = __tr1 - __it;		__ti1 = __ti1 + __rt;		\
\
	__rt  = __tr6;			__it  = __ti6;			\
	__tr6 = __tr4 - __rt;		__ti6 = __ti4 - __it;		\
	__tr4 = __tr4 + __rt;		__ti4 = __ti4 + __it;		\
\
	__rt  = __tr7;			__it  = __ti7;			\
	__tr7 = __tr5 + __it;		__ti7 = __ti5 - __rt;		\
	__tr5 = __tr5 - __it;		__ti5 = __ti5 + __rt;		\
/*       now combine the two half-transforms	*/			\
	__Br0 = __tr0 + __tr4;		__Bi0 = __ti0 + __ti4;		\
	__Br1 = __tr0 - __tr4;		__Bi1 = __ti0 - __ti4;		\
\
	__Br2 = __tr2 - __ti6;		__Bi2 = __ti2 + __tr6;		\
	__Br3 = __tr2 + __ti6;		__Bi3 = __ti2 - __tr6;		\
\
	__rt = (__tr5 - __ti5)*ISRT2;	__it = (__tr5 + __ti5)*ISRT2;	\
	__Br4 = __tr1 + __rt;		__Bi4 = __ti1 + __it;		\
	__Br5 = __tr1 - __rt;		__Bi5 = __ti1 - __it;		\
\
	__rt = (__tr7 + __ti7)*ISRT2;	__it = (__ti7 - __tr7)*ISRT2;	\
	__Br6 = __tr3 - __rt;		__Bi6 = __ti3 - __it;		\
	__Br7 = __tr3 + __rt;		__Bi7 = __ti3 + __it;		\
}

#define RADIX_08_DIF_PFETCH(\
	__Ar0,__Ai0,\
	__Ar1,__Ai1,\
	__Ar2,__Ai2,\
	__Ar3,__Ai3,\
	__Ar4,__Ai4,\
	__Ar5,__Ai5,\
	__Ar6,__Ai6,\
	__Ar7,__Ai7,\
	__tr0,__ti0,\
	__tr1,__ti1,\
	__tr2,__ti2,\
	__tr3,__ti3,\
	__tr4,__ti4,\
	__tr5,__ti5,\
	__tr6,__ti6,\
	__tr7,__ti7,\
	__Br0,__Bi0,\
	__Br1,__Bi1,\
	__Br2,__Bi2,\
	__Br3,__Bi3,\
	__Br4,__Bi4,\
	__Br5,__Bi5,\
	__Br6,__Bi6,\
	__Br7,__Bi7,\
	__rt,__it,\
	__add_offset0,__add_offset1,__add_offset2,__add_offset3,__add_offset4)\
{\
	__tr0 = __Ar0;			__ti0 = __Ai0;			\
	__rt  = __Ar4;			__it  = __Ai4;			\
	__tr1 = __tr0 - __rt;		__ti1 = __ti0 - __it;		\
	__tr0 = __tr0 + __rt;		__ti0 = __ti0 + __it;		\
\
	__tr2 = __Ar2;			__ti2 = __Ai2;			\
	__rt  = __Ar6;			__it  = __Ai6;			\
	__tr3 = __tr2 - __rt;		__ti3 = __ti2 - __it;		\
	__tr2 = __tr2 + __rt;		__ti2 = __ti2 + __it;		\
\
add = add0 + __add_offset0;	prefetch_p_doubles(add);	\
\
	__tr4 = __Ar1;			__ti4 = __Ai1;			\
	__rt  = __Ar5;			__it  = __Ai5;			\
	__tr5 = __tr4 - __rt;		__ti5 = __ti4 - __it;		\
	__tr4 = __tr4 + __rt;		__ti4 = __ti4 + __it;		\
\
	__tr6 = __Ar3;			__ti6 = __Ai3;			\
	__rt  = __Ar7;			__it  = __Ai7;			\
	__tr7 = __tr6 - __rt;		__ti7 = __ti6 - __it;		\
	__tr6 = __tr6 + __rt;		__ti6 = __ti6 + __it;		\
\
add = add0 + __add_offset1;	prefetch_p_doubles(add);	\
\
/*       combine to get the 2 length-4 transforms...	*/		\
	__rt  = __tr2;			__it  = __ti2;			\
	__tr2 = __tr0 - __rt;		__ti2 = __ti0 - __it;		\
	__tr0 = __tr0 + __rt;		__ti0 = __ti0 + __it;		\
\
	__rt  = __tr3;			__it  = __ti3;			\
	__tr3 = __tr1 + __it;		__ti3 = __ti1 - __rt;		\
	__tr1 = __tr1 - __it;		__ti1 = __ti1 + __rt;		\
\
add = add0 + __add_offset2;	prefetch_p_doubles(add);	\
\
	__rt  = __tr6;			__it  = __ti6;			\
	__tr6 = __tr4 - __rt;		__ti6 = __ti4 - __it;		\
	__tr4 = __tr4 + __rt;		__ti4 = __ti4 + __it;		\
\
	__rt  = __tr7;			__it  = __ti7;			\
	__tr7 = __tr5 + __it;		__ti7 = __ti5 - __rt;		\
	__tr5 = __tr5 - __it;		__ti5 = __ti5 + __rt;		\
\
add = add0 + __add_offset3;	prefetch_p_doubles(add);	\
\
/*       now combine the two half-transforms	*/			\
	__Br0 = __tr0 + __tr4;		__Bi0 = __ti0 + __ti4;		\
	__Br1 = __tr0 - __tr4;		__Bi1 = __ti0 - __ti4;		\
\
	__Br2 = __tr2 - __ti6;		__Bi2 = __ti2 + __tr6;		\
	__Br3 = __tr2 + __ti6;		__Bi3 = __ti2 - __tr6;		\
\
add = add0 + __add_offset4;	prefetch_p_doubles(add);	\
\
	__rt = (__tr5 - __ti5)*ISRT2;	__it = (__tr5 + __ti5)*ISRT2;	\
	__Br4 = __tr1 + __rt;		__Bi4 = __ti1 + __it;		\
	__Br5 = __tr1 - __rt;		__Bi5 = __ti1 - __it;		\
\
	__rt = (__tr7 + __ti7)*ISRT2;	__it = (__ti7 - __tr7)*ISRT2;	\
	__Br6 = __tr3 - __rt;		__Bi6 = __ti3 - __it;		\
	__Br7 = __tr3 + __rt;		__Bi7 = __ti3 + __it;		\
}

#define RADIX_08_DIT(\
	__Ar0,__Ai0,\
	__Ar1,__Ai1,\
	__Ar2,__Ai2,\
	__Ar3,__Ai3,\
	__Ar4,__Ai4,\
	__Ar5,__Ai5,\
	__Ar6,__Ai6,\
	__Ar7,__Ai7,\
	__tr0,__ti0,\
	__tr1,__ti1,\
	__tr2,__ti2,\
	__tr3,__ti3,\
	__tr4,__ti4,\
	__tr5,__ti5,\
	__tr6,__ti6,\
	__tr7,__ti7,\
	__Br0,__Bi0,\
	__Br1,__Bi1,\
	__Br2,__Bi2,\
	__Br3,__Bi3,\
	__Br4,__Bi4,\
	__Br5,__Bi5,\
	__Br6,__Bi6,\
	__Br7,__Bi7,\
	__rt,__it)\
{\
	__tr0 = __Ar0;			__ti0 = __Ai0;			\
	__rt  = __Ar1;			__it  = __Ai1;			\
	__tr1 = __tr0 - __rt;		__ti1 = __ti0 - __it;		\
	__tr0 = __tr0 + __rt;		__ti0 = __ti0 + __it;		\
\
	__tr2 = __Ar2;			__ti2 = __Ai2;			\
	__rt  = __Ar3;			__it  = __Ai3;			\
	__tr3 = __tr2 - __rt;		__ti3 = __ti2 - __it;		\
	__tr2 = __tr2 + __rt;		__ti2 = __ti2 + __it;		\
\
	__tr4 = __Ar4;			__ti4 = __Ai4;			\
	__rt  = __Ar5;			__it  = __Ai5;			\
	__tr5 = __tr4 - __rt;		__ti5 = __ti4 - __it;		\
	__tr4 = __tr4 + __rt;		__ti4 = __ti4 + __it;		\
\
	__tr6 = __Ar6;			__ti6 = __Ai6;			\
	__rt  = __Ar7;			__it  = __Ai7;			\
	__tr7 = __tr6 - __rt;		__ti7 = __ti6 - __it;		\
	__tr6 = __tr6 + __rt;		__ti6 = __ti6 + __it;		\
/*       combine to get the 2 length-4 transforms...	*/		\
	__rt  = __tr2;			__it  = __ti2;			\
	__tr2 = __tr0 - __rt;		__ti2 = __ti0 - __it;		\
	__tr0 = __tr0 + __rt;		__ti0 = __ti0 + __it;		\
\
	__rt  = __tr3;			__it  = __ti3;			\
	__tr3 = __tr1 + __it;		__ti3 = __ti1 - __rt;		\
	__tr1 = __tr1 - __it;		__ti1 = __ti1 + __rt;		\
\
	__rt  = __tr6;			__it  = __ti6;			\
	__tr6 = __tr4 - __rt;		__ti6 = __ti4 - __it;		\
	__tr4 = __tr4 + __rt;		__ti4 = __ti4 + __it;		\
\
	__rt  = __tr7;			__it  = __ti7;			\
	__tr7 = __tr5 + __it;		__ti7 = __ti5 - __rt;		\
	__tr5 = __tr5 - __it;		__ti5 = __ti5 + __rt;		\
/*       now combine the two half-transforms	*/			\
	__Br0 = __tr0 + __tr4;		__Bi0 = __ti0 + __ti4;		\
	__Br4 = __tr0 - __tr4;		__Bi4 = __ti0 - __ti4;		\
\
	__Br2 = __tr2 - __ti6;		__Bi2 = __ti2 + __tr6;		\
	__Br6 = __tr2 + __ti6;		__Bi6 = __ti2 - __tr6;		\
\
	__rt = (__tr5 - __ti5)*ISRT2;	__it = (__tr5 + __ti5)*ISRT2;	\
	__Br1 = __tr1 + __rt;		__Bi1 = __ti1 + __it;		\
	__Br5 = __tr1 - __rt;		__Bi5 = __ti1 - __it;		\
\
	__rt = (__tr7 + __ti7)*ISRT2;	__it = (__ti7 - __tr7)*ISRT2;	\
	__Br3 = __tr3 - __rt;		__Bi3 = __ti3 - __it;		\
	__Br7 = __tr3 + __rt;		__Bi7 = __ti3 + __it;		\
}


/****** RADIX = 9: ******/

#ifdef PFETCH
/*...Radix-9 DIF: A/Bs are in/outputs and t's are temporaries (all doubles),
	add0 and add are the (double *) base and temporary address and pj's are integer index offsets for prefetch:	*/
#define RADIX_9_DIF(Ar0,Ar1,Ar2,Ar3,Ar4,Ar5,Ar6,Ar7,Ar8,Ai0,Ai1,Ai2,Ai3,Ai4,Ai5,Ai6,Ai7,Ai8,\
                    t1 ,t2 ,t3 ,t4 ,t5 ,t6 ,t7 ,t8 ,t9 ,t10,t11,t12,t13,t14,t15,t16,t17,t18,\
                    Br0,Br1,Br2,Br3,Br4,Br5,Br6,Br7,Br8,Bi0,Bi1,Bi2,Bi3,Bi4,Bi5,Bi6,Bi7,Bi8,\
                    add0, add, p0, p1, p2, p3, p4, p5, p6, p7, p8)\
{\
add = add0+p0;						\
prefetch_p_doubles(add);				\
/*...gather the needed data (9 64-bit complex, i.e. 18 64-bit reals) and do three radix-3 transforms...	*/\
	t1 =Ar0;		t2 =Ai0;			\
	t3 =Ar3+Ar6;		t4 =Ai3+Ai6;			\
	t5 =Ar3-Ar6;		t6 =Ai3-Ai6;			\
	t1 =t1+t3;		t2 =t2+t4;			\
	t3 =t1+c3m1*t3;		t4 =t2+c3m1*t4;			\
	rt =s3*t5;		it =s3*t6;			\
	t5 =t3+it;		t6 =t4-rt;			\
	t3 =t3-it;		t4 =t4+rt;			\
add = add0+p1;						\
prefetch_p_doubles(add);				\
\
	t7 =Ar1;		t8 =Ai1;			\
	t9 =Ar4+Ar7;		t10=Ai4+Ai7;			\
	t11=Ar4-Ar7;		t12=Ai4-Ai7;			\
	t7 =t7+t9;		t8 =t8+t10;			\
	t9 =t7+c3m1*t9;		t10=t8+c3m1*t10;		\
	rt =s3*t11;		it =s3*t12;			\
	t11=t9+it;		t12=t10-rt;			\
	t9 =t9-it;		t10=t10+rt;			\
add = add0+p2;						\
prefetch_p_doubles(add);				\
\
	t13=Ar2;		t14=Ai2;			\
	t15=Ar5+Ar8;		t16=Ai5+Ai8;			\
	t17=Ar5-Ar8;		t18=Ai5-Ai8;			\
	t13=t13+t15;		t14=t14+t16;			\
	t15=t13+c3m1*t15;	t16=t14+c3m1*t16;		\
	rt =s3*t17;		it =s3*t18;			\
	t17=t15+it;		t18=t16-rt;			\
	t15=t15-it;		t16=t16+rt;			\
add = add0+p3;						\
prefetch_p_doubles(add);				\
/*...and now do three more radix-3 transforms, including the twiddle factors:	*/\
	rt =t7;			it =t8;				\
	t7 =rt+t13;		t8 =it+t14;			\
	t13=rt-t13;		t14=it-t14;			\
	t1 =t1+t7;		t2 =t2+t8;			\
	Br0=t1;			Bi0=t2;				\
	t7 =t1+c3m1*t7;		t8 =t2+c3m1*t8;			\
	rt =s3*t13;		it =s3*t14;			\
	Br1=t7-it;		Bi1=t8+rt;			\
	Br2=t7+it;		Bi2=t8-rt;			\
add = add0+p4;						\
prefetch_p_doubles(add);				\
\
	rt =t9 *c -t10*s;	it =t9 *s +t10*c;		\
	re =t15*c2-t16*s2;	t16=t15*s2+t16*c2;	t15=re;	\
	t9 =rt+t15;		t10=it+t16;			\
	t15=rt-t15;		t16=it-t16;			\
	t3 =t3+t9;		t4 =t4+t10;			\
add = add0+p5;						\
prefetch_p_doubles(add);				\
	Br3=t3;			Bi3=t4;				\
	t9 =t3+c3m1*t9;		t10=t4+c3m1*t10;		\
	rt =s3*t15;		it =s3*t16;			\
	Br4=t9-it;		Bi4=t10+rt;			\
	Br5=t9+it;		Bi5=t10-rt;			\
add = add0+p6;						\
prefetch_p_doubles(add);				\
\
	rt =t11*c2-t12*s2;	it =t11*s2+t12*c2;		\
	re =t17*c4-t18*s4;	t18=t17*s4+t18*c4;	t17=re;	\
	t11=rt+t17;		t12=it+t18;			\
	t17=rt-t17;		t18=it-t18;			\
	t5 =t5+t11;		t6 =t6+t12;			\
add = add0+p7;						\
prefetch_p_doubles(add);				\
	Br6=t5;			Bi6=t6;				\
	t11=t5+c3m1*t11;	t12=t6+c3m1*t12;		\
	rt =s3*t17;		it =s3*t18;			\
	Br7=t11-it;		Bi7=t12+rt;			\
	Br8=t11+it;		Bi8=t12-rt;			\
add = add0+p8;						\
prefetch_p_doubles(add);				\
}

#else

/*...Radix-9 DIF: A/Bs are in/outputs and t's are temporaries (all doubles):	*/
#define RADIX_9_DIF(Ar0,Ar1,Ar2,Ar3,Ar4,Ar5,Ar6,Ar7,Ar8,Ai0,Ai1,Ai2,Ai3,Ai4,Ai5,Ai6,Ai7,Ai8,\
                    t1 ,t2 ,t3 ,t4 ,t5 ,t6 ,t7 ,t8 ,t9 ,t10,t11,t12,t13,t14,t15,t16,t17,t18,\
                    Br0,Br1,Br2,Br3,Br4,Br5,Br6,Br7,Br8,Bi0,Bi1,Bi2,Bi3,Bi4,Bi5,Bi6,Bi7,Bi8)\
{\
/*...gather the needed data (9 64-bit complex, i.e. 18 64-bit reals) and do three radix-3 transforms...	*/\
	t1 =Ar0;		t2 =Ai0;			\
	t3 =Ar3+Ar6;		t4 =Ai3+Ai6;			\
	t5 =Ar3-Ar6;		t6 =Ai3-Ai6;			\
	t1 =t1+t3;		t2 =t2+t4;			\
	t3 =t1+c3m1*t3;		t4 =t2+c3m1*t4;			\
	rt =s3*t5;		it =s3*t6;			\
	t5 =t3+it;		t6 =t4-rt;			\
	t3 =t3-it;		t4 =t4+rt;			\
\
	t7 =Ar1;		t8 =Ai1;			\
	t9 =Ar4+Ar7;		t10=Ai4+Ai7;			\
	t11=Ar4-Ar7;		t12=Ai4-Ai7;			\
	t7 =t7+t9;		t8 =t8+t10;			\
	t9 =t7+c3m1*t9;		t10=t8+c3m1*t10;		\
	rt =s3*t11;		it =s3*t12;			\
	t11=t9+it;		t12=t10-rt;			\
	t9 =t9-it;		t10=t10+rt;			\
\
	t13=Ar2;		t14=Ai2;			\
	t15=Ar5+Ar8;		t16=Ai5+Ai8;			\
	t17=Ar5-Ar8;		t18=Ai5-Ai8;			\
	t13=t13+t15;		t14=t14+t16;			\
	t15=t13+c3m1*t15;	t16=t14+c3m1*t16;		\
	rt =s3*t17;		it =s3*t18;			\
	t17=t15+it;		t18=t16-rt;			\
	t15=t15-it;		t16=t16+rt;			\
/*...and now do three more radix-3 transforms, including the twiddle factors:	*/\
	rt =t7;			it =t8;				\
	t7 =rt+t13;		t8 =it+t14;			\
	t13=rt-t13;		t14=it-t14;			\
	t1 =t1+t7;		t2 =t2+t8;			\
	Br0=t1;			Bi0=t2;				\
	t7 =t1+c3m1*t7;		t8 =t2+c3m1*t8;			\
	rt =s3*t13;		it =s3*t14;			\
	Br1=t7-it;		Bi1=t8+rt;			\
	Br2=t7+it;		Bi2=t8-rt;			\
\
	rt =t9 *c -t10*s;	it =t9 *s +t10*c;		\
	re =t15*c2-t16*s2;	t16=t15*s2+t16*c2;	t15=re;	\
	t9 =rt+t15;		t10=it+t16;			\
	t15=rt-t15;		t16=it-t16;			\
	t3 =t3+t9;		t4 =t4+t10;			\
	Br3=t3;			Bi3=t4;				\
	t9 =t3+c3m1*t9;		t10=t4+c3m1*t10;		\
	rt =s3*t15;		it =s3*t16;			\
	Br4=t9-it;		Bi4=t10+rt;			\
	Br5=t9+it;		Bi5=t10-rt;			\
\
	rt =t11*c2-t12*s2;	it =t11*s2+t12*c2;		\
	re =t17*c4-t18*s4;	t18=t17*s4+t18*c4;	t17=re;	\
	t11=rt+t17;		t12=it+t18;			\
	t17=rt-t17;		t18=it-t18;			\
	t5 =t5+t11;		t6 =t6+t12;			\
	Br6=t5;			Bi6=t6;				\
	t11=t5+c3m1*t11;	t12=t6+c3m1*t12;		\
	rt =s3*t17;		it =s3*t18;			\
	Br7=t11-it;		Bi7=t12+rt;			\
	Br8=t11+it;		Bi8=t12-rt;			\
}

#endif

/*...Radix-9 DIT:	*/
#define RADIX_9_DIT(Ar0,Ar1,Ar2,Ar3,Ar4,Ar5,Ar6,Ar7,Ar8,Ai0,Ai1,Ai2,Ai3,Ai4,Ai5,Ai6,Ai7,Ai8,\
                    t1 ,t2 ,t3 ,t4 ,t5 ,t6 ,t7 ,t8 ,t9 ,t10,t11,t12,t13,t14,t15,t16,t17,t18,\
                    Br0,Br1,Br2,Br3,Br4,Br5,Br6,Br7,Br8,Bi0,Bi1,Bi2,Bi3,Bi4,Bi5,Bi6,Bi7,Bi8)\
{\
/*...gather the needed data (9 64-bit complex, i.e. 18 64-bit reals) and do three radix-3 transforms...	*/\
	t1 =Ar0;		t2 =Ai0;			\
	t3 =Ar1+Ar2;		t4 =Ai1+Ai2;			\
	t5 =Ar1-Ar2;		t6 =Ai1-Ai2;			\
	t1 =t1+t3;		t2 =t2+t4;			\
	t3 =t1+c3m1*t3;		t4 =t2+c3m1*t4;			\
	rt =s3*t5;		it =s3*t6;			\
	t5 =t3-it;		t6 =t4+rt;			\
	t3 =t3+it;		t4 =t4-rt;			\
\
	t7 =Ar3;		t8 =Ai3;			\
	t9 =Ar4+Ar5;		t10=Ai4+Ai5;			\
	t11=Ar4-Ar5;		t12=Ai4-Ai5;			\
	t7 =t7+t9;		t8 =t8+t10;			\
	t9 =t7+c3m1*t9;		t10=t8+c3m1*t10;		\
	rt =s3*t11;		it =s3*t12;			\
	t11=t9-it;		t12=t10+rt;			\
	t9 =t9+it;		t10=t10-rt;			\
\
	t13=Ar6;		t14=Ai6;			\
	t15=Ar7+Ar8;		t16=Ai7+Ai8;			\
	t17=Ar7-Ar8;		t18=Ai7-Ai8;			\
	t13=t13+t15;		t14=t14+t16;			\
	t15=t13+c3m1*t15;	t16=t14+c3m1*t16;		\
	rt =s3*t17;		it =s3*t18;			\
	t17=t15-it;		t18=t16+rt;			\
	t15=t15+it;		t16=t16-rt;			\
/*...and now do three more radix-3 transforms, including the twiddle factors:	*/\
	rt =t7;			it =t8;				\
	t7 =rt+t13;		t8 =it+t14;			\
	t13=rt-t13;		t14=it-t14;			\
	t1 =t1+t7;		t2 =t2+t8;			\
	Ar0=t1;			Ai0=t2;				\
	t7 =t1+c3m1*t7;		t8 =t2+c3m1*t8;			\
	rt =s3*t13;		it =s3*t14;			\
	Ar3=t7+it;		Ai3=t8-rt;			\
	Ar6=t7-it;		Ai6=t8+rt;			\
\
	rt =t9 *c +t10*s;	it =t10*c -t9 *s;		\
	re =t15*c2+t16*s2;	t16=t16*c2-t15*s2;	t15=re;	\
	t9 =rt+t15;		t10=it+t16;			\
	t15=rt-t15;		t16=it-t16;			\
	t3 =t3+t9;		t4 =t4+t10;			\
	Ar1=t3;			Ai1=t4;				\
	t9 =t3+c3m1*t9;		t10=t4+c3m1*t10;		\
	rt =s3*t15;		it =s3*t16;			\
	Ar4=t9+it;		Ai4=t10-rt;			\
	Ar7=t9-it;		Ai7=t10+rt;			\
\
	rt =t11*c2+t12*s2;	it =t12*c2-t11*s2;		\
	re =t17*c4+t18*s4;	t18=t18*c4-t17*s4;	t17=re;	\
	t11=rt+t17;		t12=it+t18;			\
	t17=rt-t17;		t18=it-t18;			\
	t5 =t5+t11;		t6 =t6+t12;			\
	Ar2=t5;			Ai2=t6;				\
	t11=t5+c3m1*t11;	t12=t6+c3m1*t12;		\
	rt =s3*t17;		it =s3*t18;			\
	Ar5=t11+it;		Ai5=t12-rt;			\
	Ar8=t11-it;		Ai8=t12+rt;			\
}

/****** RADIX = 15: ******/

#define RADIX_15_DIF(\
	__Ar00,__Ai00,\
	__Ar01,__Ai01,\
	__Ar02,__Ai02,\
	__Ar03,__Ai03,\
	__Ar04,__Ai04,\
	__Ar05,__Ai05,\
	__Ar06,__Ai06,\
	__Ar07,__Ai07,\
	__Ar08,__Ai08,\
	__Ar09,__Ai09,\
	__Ar10,__Ai10,\
	__Ar11,__Ai11,\
	__Ar12,__Ai12,\
	__Ar13,__Ai13,\
	__Ar14,__Ai14,\
	__tr00,__ti00,\
	__tr01,__ti01,\
	__tr02,__ti02,\
	__tr03,__ti03,\
	__tr04,__ti04,\
	__tr05,__ti05,\
	__tr06,__ti06,\
	__tr07,__ti07,\
	__tr08,__ti08,\
	__tr09,__ti09,\
	__tr10,__ti10,\
	__tr11,__ti11,\
	__tr12,__ti12,\
	__tr13,__ti13,\
	__tr14,__ti14,\
	__Br00,__Bi00,\
	__Br01,__Bi01,\
	__Br02,__Bi02,\
	__Br03,__Bi03,\
	__Br04,__Bi04,\
	__Br05,__Bi05,\
	__Br06,__Bi06,\
	__Br07,__Bi07,\
	__Br08,__Bi08,\
	__Br09,__Bi09,\
	__Br10,__Bi10,\
	__Br11,__Bi11,\
	__Br12,__Bi12,\
	__Br13,__Bi13,\
	__Br14,__Bi14,\
	__rt,__it)\
{\
/*...gather the needed data (15 64-bit complex, i.e. 30 64-bit reals) and do 3 radix-5 transforms...*/\
/*...Block 1:	*/							\
	__tr00 = __Ar00;		__ti00 = __Ai00;		\
	__tr01 = __Ar03;		__ti01 = __Ai03;		\
	__rt   = __Ar12;		__it   = __Ai12;		\
	__tr03 = __tr01-__rt;		__ti03 = __ti01-__it;		\
	__tr01 = __tr01+__rt;		__ti01 = __ti01+__it;		\
	__tr02 = __Ar06;		__ti02 = __Ai06;		\
	__rt   = __Ar09;		__it   = __Ai09;		\
	__tr04 = __tr02-__rt;		__ti04 = __ti02-__it;		\
	__tr02 = __tr02+__rt;		__ti02 = __ti02+__it;		\
									\
	__rt   = __tr01+__tr02;		__it   = __ti01+__ti02;		\
	__tr00 = __tr00+__rt;		__ti00 = __ti00+__it;		\
	__rt   = __tr00+cn1*__rt;	__it   = __ti00+cn1*__it;	\
	__tr02 = cn2*(__tr01-__tr02);	__ti02 = cn2*(__ti01-__ti02);	\
	__tr01 = __rt+__tr02;		__ti01 = __it+__ti02;		\
	__tr02 = __rt-__tr02;		__ti02 = __it-__ti02;		\
	__rt   = ss3*(__tr03-__tr04);	__it   = ss3*(__ti03-__ti04);	\
	__tr04 = __rt+sn1*__tr04;	__ti04 = __it+sn1*__ti04;	\
	__tr03 = __rt-sn2*__tr03;	__ti03 = __it-sn2*__ti03;	\
	__rt   = __tr04;		__it   = __ti04;		\
	__tr04 = __tr01+__it;		__ti04 = __ti01-__rt;		/*<==prefer these to be stored in __tr,i04 */\
	__tr01 = __tr01-__it;		__ti01 = __ti01+__rt;		\
	__rt   = __tr03;		__it   = __ti03;		\
	__tr03 = __tr02+__it;		__ti03 = __ti02-__rt;		/*<==prefer these to be stored in __tr,i03 */\
	__tr02 = __tr02-__it;		__ti02 = __ti02+__rt;		\
									\
/*...Block 2:	*/							\
	__tr05 = __Ar01;		__ti05 = __Ai01;		\
	__tr06 = __Ar04;		__ti06 = __Ai04;		\
	__rt   = __Ar13;		__it   = __Ai13;		\
	__tr08 = __tr06 -__rt;		__ti08 = __ti06 -__it;		\
	__tr06 = __tr06 +__rt;		__ti06 = __ti06 +__it;		\
	__tr07 = __Ar07;		__ti07 = __Ai07;		\
	__rt   = __Ar10;		__it   = __Ai10;		\
	__tr09 = __tr07 -__rt;		__ti09 = __ti07 -__it;		\
	__tr07 = __tr07 +__rt;		__ti07 = __ti07 +__it;		\
									\
	__rt   = __tr06+__tr07;		__it   = __ti06+__ti07;		\
	__tr05 = __tr05+__rt;		__ti05 = __ti05+__it;		\
	__rt   = __tr05+cn1*__rt;	__it   = __ti05+cn1*__it;	\
	__tr07 = cn2*(__tr06-__tr07);	__ti07 = cn2*(__ti06-__ti07);	\
	__tr06 = __rt+__tr07;		__ti06 = __it+__ti07;		\
	__tr07 = __rt-__tr07;		__ti07 = __it-__ti07;		\
	__rt   = ss3*(__tr08-__tr09);	__it   = ss3*(__ti08-__ti09);	\
	__tr09 = __rt+sn1*__tr09;	__ti09 = __it+sn1*__ti09;	\
	__tr08 = __rt-sn2*__tr08;	__ti08 = __it-sn2*__ti08;	\
	__rt   = __tr09;		__it   = __ti09;		\
	__tr09 = __tr06+__it;		__ti09 = __ti06-__rt;		\
	__tr06 = __tr06-__it;		__ti06 = __ti06+__rt;		\
	__rt   = __tr08;		__it   = __ti08;		\
	__tr08 = __tr07+__it;		__ti08 = __ti07-__rt;		\
	__tr07 = __tr07-__it;		__ti07 = __ti07+__rt;		\
									\
/*...Block 3:	*/							\
	__tr10 = __Ar02;		__ti10 = __Ai02;		\
	__tr11 = __Ar05;		__ti11 = __Ai05;		\
	__rt   = __Ar14;		__it   = __Ai14;		\
	__tr13 = __tr11 -__rt;		__ti13 = __ti11 -__it;		\
	__tr11 = __tr11 +__rt;		__ti11 = __ti11 +__it;		\
	__tr12 = __Ar08;		__ti12 = __Ai08;		\
	__rt   = __Ar11;		__it   = __Ai11;		\
	__tr14 = __tr12 -__rt;		__ti14 = __ti12 -__it;		\
	__tr12 = __tr12 +__rt;		__ti12 = __ti12 +__it;		\
									\
	__rt   = __tr11+__tr12;		__it   = __ti11+__ti12;		\
	__tr10 = __tr10+__rt;		__ti10 = __ti10+__it;		\
	__rt   = __tr10+cn1*__rt;	__it   = __ti10+cn1*__it;	\
	__tr12 = cn2*(__tr11-__tr12);	__ti12 = cn2*(__ti11-__ti12);	\
	__tr11 = __rt+__tr12;		__ti11 = __it+__ti12;		\
	__tr12 = __rt-__tr12;		__ti12 = __it-__ti12;		\
	__rt   = ss3*(__tr13-__tr14);	__it   = ss3*(__ti13-__ti14);	\
	__tr14 = __rt+sn1*__tr14;	__ti14 = __it+sn1*__ti14;	\
	__tr13 = __rt-sn2*__tr13;	__ti13 = __it-sn2*__ti13;	\
	__rt   = __tr14;		__it   = __ti14;		\
	__tr14 = __tr11+__it;		__ti14 = __ti11-__rt;		\
	__tr11 = __tr11-__it;		__ti11 = __ti11+__rt;		\
	__rt   = __tr13;		__it   = __ti13;		\
	__tr13 = __tr12+__it;		__ti13 = __ti12-__rt;		\
	__tr12 = __tr12-__it;		__ti12 = __ti12+__rt;		\
									\
/*...and now do five radix-3 transforms:	*/			\
/*...Block 1:	*/							\
	__rt   = __tr10;		__it   = __ti10;		\
	__tr10 = __tr05-__rt;		__ti10 = __ti05-__it;		\
	__tr05 = __tr05+__rt;		__ti05 = __ti05+__it;		\
	__tr00 = __tr00+__tr05;		__ti00 = __ti00+__ti05;		\
	__Br00 = __tr00;		__Bi00 = __ti00;		\
	__tr05 = __tr00+c3m1*__tr05;	__ti05 = __ti00+c3m1*__ti05;	\
	__rt   = s*__tr10;		__it   = s*__ti10;		\
	__Br01 = __tr05-__it;		__Bi01 = __ti05+__rt;		\
	__Br02 = __tr05+__it;		__Bi02 = __ti05-__rt;		\
									\
/*...Block 2:	*/							\
	__rt   = __tr11;		__it   = __ti11;		\
	__tr11 = __tr06-__rt;		__ti11 = __ti06-__it;		\
	__tr06 = __tr06+__rt;		__ti06 = __ti06+__it;		\
	__tr01 = __tr01+__tr06;		__ti01 = __ti01+__ti06;		\
	__Br03 = __tr01;		__Bi03 = __ti01;		\
	__tr06 = __tr01+c3m1*__tr06;	__ti06 = __ti01+c3m1*__ti06;	\
	__rt   = s*__tr11;		__it   = s*__ti11;		\
	__Br04 = __tr06-__it;		__Bi04 = __ti06+__rt;		\
	__Br05 = __tr06+__it;		__Bi05 = __ti06-__rt;		\
									\
/*...Block 3:	*/							\
	__rt   = __tr12;		__it   = __ti12;		\
	__tr12 = __tr07-__rt;		__ti12 = __ti07-__it;		\
	__tr07 = __tr07+__rt;		__ti07 = __ti07+__it;		\
	__tr02 = __tr02+__tr07;		__ti02 = __ti02+__ti07;		\
	__Br06 = __tr02;		__Bi06 = __ti02;		\
	__tr07 = __tr02+c3m1*__tr07;	__ti07 = __ti02+c3m1*__ti07;	\
	__rt   = s*__tr12;		__it   = s*__ti12;		\
	__Br07 = __tr07-__it;		__Bi07 = __ti07+__rt;		\
	__Br08 = __tr07+__it;		__Bi08 = __ti07-__rt;		\
									\
/*...Block 4:	*/							\
	__rt   = __tr13;		__it   = __ti13;		\
	__tr13 = __tr08-__rt;		__ti13 = __ti08-__it;		\
	__tr08 = __tr08+__rt;		__ti08 = __ti08+__it;		\
	__tr03 = __tr03+__tr08;		__ti03 = __ti03+__ti08;		\
	__Br09 = __tr03;		__Bi09 = __ti03;		\
	__tr08 = __tr03+c3m1*__tr08;	__ti08 = __ti03+c3m1*__ti08;	\
	__rt   = s*__tr13;		__it   = s*__ti13;		\
	__Br10 = __tr08-__it;		__Bi10 = __ti08+__rt;		\
	__Br11 = __tr08+__it;		__Bi11 = __ti08-__rt;		\
									\
/*...Block 5:	*/							\
	__rt   = __tr14;		__it   = __ti14;		\
	__tr14 = __tr09-__rt;		__ti14 = __ti09-__it;		\
	__tr09 = __tr09+__rt;		__ti09 = __ti09+__it;		\
	__tr04 = __tr04+__tr09;		__ti04 = __ti04+__ti09;		\
	__Br12 = __tr04;		__Bi12 = __ti04;		\
	__tr09 = __tr04+c3m1*__tr09;	__ti09 = __ti04+c3m1*__ti09;	\
	__rt   = s*__tr14;		__it   = s*__ti14;		\
	__Br13 = __tr09-__it;		__Bi13 = __ti09+__rt;		\
	__Br14 = __tr09+__it;		__Bi14 = __ti09-__rt;		\
}

#define RADIX_15_DIT(\
	__Ar00,__Ai00,\
	__Ar01,__Ai01,\
	__Ar02,__Ai02,\
	__Ar03,__Ai03,\
	__Ar04,__Ai04,\
	__Ar05,__Ai05,\
	__Ar06,__Ai06,\
	__Ar07,__Ai07,\
	__Ar08,__Ai08,\
	__Ar09,__Ai09,\
	__Ar10,__Ai10,\
	__Ar11,__Ai11,\
	__Ar12,__Ai12,\
	__Ar13,__Ai13,\
	__Ar14,__Ai14,\
	__tr00,__ti00,\
	__tr01,__ti01,\
	__tr02,__ti02,\
	__tr03,__ti03,\
	__tr04,__ti04,\
	__tr05,__ti05,\
	__tr06,__ti06,\
	__tr07,__ti07,\
	__tr08,__ti08,\
	__tr09,__ti09,\
	__tr10,__ti10,\
	__tr11,__ti11,\
	__tr12,__ti12,\
	__tr13,__ti13,\
	__tr14,__ti14,\
	__Br00,__Bi00,\
	__Br01,__Bi01,\
	__Br02,__Bi02,\
	__Br03,__Bi03,\
	__Br04,__Bi04,\
	__Br05,__Bi05,\
	__Br06,__Bi06,\
	__Br07,__Bi07,\
	__Br08,__Bi08,\
	__Br09,__Bi09,\
	__Br10,__Bi10,\
	__Br11,__Bi11,\
	__Br12,__Bi12,\
	__Br13,__Bi13,\
	__Br14,__Bi14,\
	__rt,__it)\
{\
/*...gather the needed data (15 64-bit complex, i.e. 30 64-bit reals) and do 5 radix-3 transforms...*/\
/*...Block 1:	*/								\
	    __tr00 = __Ar00;			__ti00 = __Ai00;		\
	    __tr01 = __Ar01;			__ti01 = __Ai01;		\
	    __rt   = __Ar02;			__it   = __Ai02;		\
	    __tr02 = __tr01-__rt;		__ti02 = __ti01-__it;		\
	    __tr01 = __tr01+__rt;		__ti01 = __ti01+__it;		\
	    __tr00 = __tr00+__tr01;		__ti00 = __ti00+__ti01;		\
	    __tr01 = __tr00+c3m1*__tr01;	__ti01 = __ti00+c3m1*__ti01;	\
	    __rt   = s*__tr02;			__it   = s*__ti02;		\
	    __tr02 = __tr01-__it;		__ti02 = __ti01+__rt;		\
	    __tr01 = __tr01+__it;		__ti01 = __ti01-__rt;		\
										\
/*...Block 2:	*/								\
	    __tr03 = __Ar03;			__ti03 = __Ai03;		\
	    __tr04 = __Ar04;			__ti04 = __Ai04;		\
	    __rt   = __Ar05;			__it   = __Ai05;		\
	    __tr05 = __tr04-__rt;		__ti05 = __ti04-__it;		\
	    __tr04 = __tr04+__rt;		__ti04 = __ti04+__it;		\
	    __tr03 = __tr03+__tr04;		__ti03 = __ti03+__ti04;		\
	    __tr04 = __tr03+c3m1*__tr04;	__ti04 = __ti03+c3m1*__ti04;	\
	    __rt   = s*__tr05;			__it   = s*__ti05;		\
	    __tr05 = __tr04-__it;		__ti05 = __ti04+__rt;		\
	    __tr04 = __tr04+__it;		__ti04 = __ti04-__rt;		\
										\
/*...Block 3:	*/								\
	    __tr06 = __Ar06;			__ti06 = __Ai06;		\
	    __tr07 = __Ar07;			__ti07 = __Ai07;		\
	    __rt   = __Ar08;			__it   = __Ai08;		\
	    __tr08 = __tr07-__rt;		__ti08 = __ti07-__it;		\
	    __tr07 = __tr07+__rt;		__ti07 = __ti07+__it;		\
	    __tr06 = __tr06+__tr07;		__ti06 = __ti06+__ti07;		\
	    __tr07 = __tr06+c3m1*__tr07;	__ti07 = __ti06+c3m1*__ti07;	\
	    __rt   = s*__tr08;			__it   = s*__ti08;		\
	    __tr08 = __tr07-__it;		__ti08 = __ti07+__rt;		\
	    __tr07 = __tr07+__it;		__ti07 = __ti07-__rt;		\
										\
/*...Block 4:	*/								\
	    __tr09 = __Ar09;			__ti09 = __Ai09;		\
	    __tr10 = __Ar10;			__ti10 = __Ai10;		\
	    __rt   = __Ar11;			__it   = __Ai11;		\
	    __tr11 = __tr10-__rt;		__ti11 = __ti10-__it;		\
	    __tr10 = __tr10+__rt;		__ti10 = __ti10+__it;		\
	    __tr09 = __tr09+__tr10;		__ti09 = __ti09+__ti10;		\
	    __tr10 = __tr09+c3m1*__tr10;	__ti10 = __ti09+c3m1*__ti10;	\
	    __rt   = s*__tr11;			__it   = s*__ti11;		\
	    __tr11 = __tr10-__it;		__ti11 = __ti10+__rt;		\
	    __tr10 = __tr10+__it;		__ti10 = __ti10-__rt;		\
										\
/*...Block 5:	*/								\
	    __tr12 = __Ar12;			__ti12 = __Ai12;		\
	    __tr13 = __Ar13;			__ti13 = __Ai13;		\
	    __rt   = __Ar14;			__it   = __Ai14;		\
	    __tr14 = __tr13-__rt;		__ti14 = __ti13-__it;		\
	    __tr13 = __tr13+__rt;		__ti13 = __ti13+__it;		\
	    __tr12 = __tr12+__tr13;		__ti12 = __ti12+__ti13;		\
	    __tr13 = __tr12+c3m1*__tr13;	__ti13 = __ti12+c3m1*__ti13;	\
	    __rt   = s*__tr14;			__it   = s*__ti14;		\
	    __tr14 = __tr13-__it;		__ti14 = __ti13+__rt;		\
	    __tr13 = __tr13+__it;		__ti13 = __ti13-__rt;		\
										\
/*...and now do three radix-5 transforms:	*/				\
/*...Block 1:	*/								\
	    __rt   = __tr12;			__it   = __ti12;		\
	    __tr12 = __tr03-__rt;		__ti12 = __ti03-__it;		\
	    __tr03 = __tr03+__rt;		__ti03 = __ti03+__it;		\
	    __rt   = __tr09;			__it   = __ti09;		\
	    __tr09 = __tr06-__rt;		__ti09 = __ti06-__it;		\
	    __tr06 = __tr06+__rt;		__ti06 = __ti06+__it;		\
										\
	    __rt   = __tr03+__tr06;		__it   = __ti03+__ti06;		\
	    __tr00 = __tr00+__rt;		__ti00 = __ti00+__it;		\
	    __rt   = __tr00+cn1*__rt;		__it   = __ti00+cn1*__it;	\
	    __tr06 = cn2*(__tr03-__tr06);	__ti06 = cn2*(__ti03-__ti06);	\
	    __tr03 = __rt+__tr06;		__ti03 = __it+__ti06;		\
	    __tr06 = __rt-__tr06;		__ti06 = __it-__ti06;		\
	    __rt   = ss3*(__tr09-__tr12);	__it   = ss3*(__ti09-__ti12);	\
	    __tr09 = __rt-sn1*__tr09;		__ti09 = __it-sn1*__ti09;	\
	    __tr12 = __rt+sn2*__tr12;		__ti12 = __it+sn2*__ti12;	\
										\
	    __Br00 = __tr00;			__Bi00 = __ti00;		\
	    __Br03 = __tr03-__ti09;		__Bi03 = __ti03+__tr09;		\
	    __Br06 = __tr06-__ti12;		__Bi06 = __ti06+__tr12;		\
	    __Br09 = __tr06+__ti12;		__Bi09 = __ti06-__tr12;		\
	    __Br12 = __tr03+__ti09;		__Bi12 = __ti03-__tr09;		\
										\
/*...Block 2:	*/								\
	    __rt   = __tr13;			__it   = __ti13;		\
	    __tr13 = __tr04-__rt;		__ti13 = __ti04-__it;		\
	    __tr04 = __tr04+__rt;		__ti04 = __ti04+__it;		\
	    __rt   = __tr10;			__it   = __ti10;		\
	    __tr10 = __tr07-__rt;		__ti10 = __ti07-__it;		\
	    __tr07 = __tr07+__rt;		__ti07 = __ti07+__it;		\
										\
	    __rt   = __tr04+__tr07;		__it   = __ti04+__ti07;		\
	    __tr01 = __tr01+__rt;		__ti01 = __ti01+__it;		\
	    __rt   = __tr01+cn1*__rt;		__it   = __ti01+cn1*__it;	\
	    __tr07 = cn2*(__tr04-__tr07);	__ti07 = cn2*(__ti04-__ti07);	\
	    __tr04 = __rt+__tr07;		__ti04 = __it+__ti07;		\
	    __tr07 = __rt-__tr07;		__ti07 = __it-__ti07;		\
	    __rt   = ss3*(__tr10-__tr13);	__it   = ss3*(__ti10-__ti13);	\
	    __tr10 = __rt-sn1*__tr10;		__ti10 = __it-sn1*__ti10;	\
	    __tr13 = __rt+sn2*__tr13;		__ti13 = __it+sn2*__ti13;	\
										\
	    __Br01 = __tr01;			__Bi01 = __ti01;		\
	    __Br04 = __tr04-__ti10;		__Bi04 = __ti04+__tr10;		\
	    __Br07 = __tr07-__ti13;		__Bi07 = __ti07+__tr13;		\
	    __Br10 = __tr07+__ti13;		__Bi10 = __ti07-__tr13;		\
	    __Br13 = __tr04+__ti10;		__Bi13 = __ti04-__tr10;		\
										\
/*...Block 3:	*/								\
	    __rt   = __tr14;			__it   = __ti14;		\
	    __tr14 = __tr05-__rt;		__ti14 = __ti05-__it;		\
	    __tr05 = __tr05+__rt;		__ti05 = __ti05+__it;		\
	    __rt   = __tr11;			__it   = __ti11;		\
	    __tr11 = __tr08-__rt;		__ti11 = __ti08-__it;		\
	    __tr08 = __tr08+__rt;		__ti08 = __ti08+__it;		\
										\
	    __rt   = __tr05+__tr08;		__it   = __ti05+__ti08;		\
	    __tr02 = __tr02+__rt;		__ti02 = __ti02+__it;		\
	    __rt   = __tr02+cn1*__rt;		__it   = __ti02+cn1*__it;	\
	    __tr08 = cn2*(__tr05-__tr08);	__ti08 = cn2*(__ti05-__ti08);	\
	    __tr05 = __rt+__tr08;		__ti05 = __it+__ti08;		\
	    __tr08 = __rt-__tr08;		__ti08 = __it-__ti08;		\
	    __rt   = ss3*(__tr11-__tr14);	__it   = ss3*(__ti11-__ti14);	\
	    __tr11 = __rt-sn1*__tr11;		__ti11 = __it-sn1*__ti11;	\
	    __tr14 = __rt+sn2*__tr14;		__ti14 = __it+sn2*__ti14;	\
										\
	    __Br02 = __tr02;			__Bi02 = __ti02;		\
	    __Br05 = __tr05-__ti11;		__Bi05 = __ti05+__tr11;		\
	    __Br08 = __tr08-__ti14;		__Bi08 = __ti08+__tr14;		\
	    __Br11 = __tr08+__ti14;		__Bi11 = __ti08-__tr14;		\
	    __Br14 = __tr05+__ti11;		__Bi14 = __ti05-__tr11;		\
}








#define RADIX_15_DFT_PFETCH(\
	__add_offset0,__add_offset1,__add_offset2)\
{\
\
add = add0 + __add_offset0;	prefetch_p_doubles(add);	\
\
}









#endif	/* #ifndef dft_macro_included */
