/*******************************************************************************
*                                                                              *
*   (C) 1997-2002 by Ernst W. Mayer.                                           *
*                                                                              *
*  This program is free software; you can redistribute it and/or modify it     *
*  under the terms of the GNU General Public License as published by the       *
*  Free Software Foundation; either version 2 of the License, or (at your      *
*  option) any later version.                                                  *
*                                                                              *
*  This program is distributed in the hope that it will be useful, but WITHOUT *
*  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or       *
*  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for   *
*  more details.                                                               *
*                                                                              *
*  You should have received a copy of the GNU General Public License along     *
*  with this program; see the file GPL.txt.  If not, you may view one at       *
*  http://www.fsf.org/licenses/licenses.html, or obtain one by writing to the  *
*  Free Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA     *
*  02111-1307, USA.                                                            *
*                                                                              *
*******************************************************************************/

/****************************************************************************
 * We now include this header file if it was not included before.
 ****************************************************************************/
#ifndef prefetch_h_included
#define prefetch_h_included

/* These implement some basic prefetching on various architectures: */

#undef	PFETCH

/* Pentium3 family, Gnu C or C++ compiler */
#if(defined(__i386__) && (defined(__GNUC__) || defined(__GNUG__)))
/*
#if(defined(__i686__) && (defined(__GNUC__) || defined(__GNUG__)))
*/
# define PFETCH 1
# define CACHE_LINE_INTS    8
# define CACHE_LINE_DOUBLES 4
# define CACHE_LINE_COMPLEX 2
# define prefetch_data(_array,_k)	__asm__ volatile ("prefetchnta (%0)": :"r"(_array + _k));
# define prefetch_p_doubles(_pd)	__asm__ volatile ("prefetchnta (%0)": :"r"(_pd+2));

/* Pentium3 family, Metrowerks CodeWarrior C compiler */
#elif(defined(__i386__) && defined(__MWERKS__))

# define PFETCH 1
# define CACHE_LINE_INTS    8
# define CACHE_LINE_DOUBLES 4
# define CACHE_LINE_COMPLEX 2
# define prefetch_data(_array,_k)

# define prefetch_p_doubles(_pd)\
{\
	/* EWM - need the extra {} to get CW to inline these sans errors: */\
	{	__asm	mov edx, _pd	};	\
	{	__asm	prefetcht0 [edx]};	\
}

/* Pentium 3 family, Intel C or C++ compiler */
#elif(defined(_M_IX86) && (_M_IX86 == 600) && defined(__INTEL_COMPILER))

# define __i386__ 1
/*
# define PFETCH 1
# define CACHE_LINE_INTS    16
# define CACHE_LINE_DOUBLES  8
# define CACHE_LINE_COMPLEX  4
# define prefetch_data(_array,_k)	__asm__ volatile ("prefetchnta (%0)": :"r"(_array + _k));
# define prefetch_p_doubles(_pd)	__asm__ volatile ("prefetchnta (%0)": :"r"(_pd+2));
*/
# define prefetch_data(_array,_k)
# define prefetch_p_doubles(_pd)

/* AMD Athlon/Duron family, Gnu C or C++ compiler */
/* Since the x86 prefetch seems to work as well as any of the Athlon-specific ones, skip this for now.
#elif(defined(__i???__) && (defined(__GNUC__) || defined(__GNUG__)))

# define PFETCH 1
# define CACHE_LINE_INTS    8
# define CACHE_LINE_DOUBLES 4
# define CACHE_LINE_COMPLEX 2
# define prefetch_data(_array,_k)	__asm__ volatile ("prefetchw (%0)": :"r"(_array + _k));
# define prefetch_p_doubles(_pd)	__asm__ volatile ("prefetchw (%0)": :"r"(_pd));
*/

/* AMD Athlon/Duron family, Metrowerks CodeWarrior C compiler */
/* Since the x86 prefetch seems to work as well as any of the Athlon-specific ones, skip this for now.
#elif(defined(__i???__) && defined(__MWERKS__))

# define PFETCH 1
# define CACHE_LINE_INTS    8
# define CACHE_LINE_DOUBLES 4
# define CACHE_LINE_COMPLEX 2
# define prefetch_data(_array,_k)

# define prefetch_p_doubles(_pd)	\
	__asm	mov edx, _pd		\
	__asm	prefetchw [edx]
*/
/* Pentium4 family, Gnu C or C++ compiler */
/* NEED CPUID CODE HERE
#elif(defined(__i???__) && (defined(__GNUC__) || defined(__GNUG__)))

# define PFETCH 1
# define CACHE_LINE_INTS    8
# define CACHE_LINE_DOUBLES 4
# define CACHE_LINE_COMPLEX 2
# define prefetch_data(_array,_k)	__asm__ volatile ("prefetcht1 (%0)": :"r"(_array + _k));
# define prefetch_p_doubles(_pd)	__asm__ volatile ("prefetcht1 (%0)": :"r"(_pd + CACHE_LINE_DOUBLES));
*/

/* Pentium 4, Intel C or C++ compiler */
#elif(defined(_M_IX86) && (_M_IX86 == 700) && defined(__INTEL_COMPILER))

# define pentium4 1
/*
# define PFETCH 1
# define CACHE_LINE_INTS    16
# define CACHE_LINE_DOUBLES  8
# define CACHE_LINE_COMPLEX  4
*/
# define prefetch_data(_array,_k)
# define prefetch_p_doubles(_pd)

/* IA-64 (Itanium), HP C or C++ compiler for HPUX: */
#elif(defined(__ia64) && (defined(__HP_cc) || defined(__HP_aCC)))

# include <fenv.h>
# include <machine/sys/inline.h>

# define PFETCH 1
# define ia64_hpc
# define CACHE_LINE_INTS    16
# define CACHE_LINE_DOUBLES  8
# define CACHE_LINE_COMPLEX  4
/*
	From HP's "inline assembly for Itanium-based HP-UX":

	Table 1-14 _Asm_lftype  legal lfetch lftype completers:

	typedef enum {
		_LFTYPE_NONE  = 0,	// Ignore faults
		_LFTYPE_FAULT = 1	// Raise faults
	} _Asm_lftype;

	Table 1-15 _Asm_lfhint  legal lfetch lfhint completers:

	typedef enum {
		_LFHINT_NONE = 0,	// Temporal locality, level 1
		_LFHINT_NT1 = 1,	// No temporal locality, level 1
		_LFHINT_NT2 = 2,	// No temporal locality, level 2
		_LFHINT_NTA = 3		// No temporal locality, all levels
	} _Asm_lfhint;

	Table 1-31 Memory Management Opcodes:
	...
	void _Asm_lfetch	( _Asm_lftype, _Asm_lfhint, void * r3);
	void _Asm_lfetch_excl	( _Asm_lftype, _Asm_lfhint, void * r3);
*/
# define prefetch_data(_array,_k)	_Asm_lfetch_excl(_LFTYPE_NONE, _LFHINT_NT2 , (void *)(_array + _k))
# define prefetch_p_doubles(_pd)	_Asm_lfetch_excl(_LFTYPE_NONE, _LFHINT_NT2 , (void *)(_pd    +  4))

/* IA-64 (Itanium), Intel C compiler (ICC v7 or above). We need to put this before the IA64/GCC case
because, bizarrely, ICC also #defines __GNUC__ and __GNUG__ . */
#elif(defined(__ia64__) && defined(__INTEL_COMPILER))

# define PFETCH 0

# define ia64_icc
# define CACHE_LINE_INTS    16
# define CACHE_LINE_DOUBLES  8
# define CACHE_LINE_COMPLEX  4

/* For ecc, use the intrinsics defined in xmmintrin.h: */
# include <xmmintrin.h>
/* These don't work under ecc v6, because _MM_HINT_* are not defined; will have to wait until v7.
# define prefetch_data(_array,_k)	_mm_prefetch((char *)(_array + _k), _MM_HINT_T0)
# define prefetch_p_doubles(_pd)	_mm_prefetch((char *)(_pd + 2), _MM_HINT_T0)
*/
# define prefetch_data(_array,_k)	_mm_prefetch((char *)(_array + _k), (int)3);
# define prefetch_p_doubles(_pd)	_mm_prefetch((char *)(_pd    +  2), (int)3);

/* IA-64 (Itanium), Gnu C or C++ compiler */
#elif(defined(__ia64__) && (defined(__GNUC__) || defined(__GNUG__)))

# define PFETCH 1
# define ia64_gcc
# define CACHE_LINE_INTS    16
# define CACHE_LINE_DOUBLES  8
# define CACHE_LINE_COMPLEX  4
# define prefetch_data(_array,_k)	__asm__ volatile ("lfetch.nta [%0]": :"r"(_array + _k));
# define prefetch_p_doubles(_pd)	__asm__ volatile ("lfetch.nta [%0]": :"r"(_pd+2));

/*
	HP PA-RISC UNDER HPUX:

	From http://docs.hp.com/hpux/onlinedocs/2212/A-03-37relnotes.html:

	Gather/Scatter Prefetch pragma
	A pragma is now supported to prefetch specified cache lines. The behavior of this pragma is similar to +Odataprefetch but the
	prefetch pragma can access specific elements in indexed arrays that are stored in cache.  In addition, any valid lvalue can be used
	as an argument, but the intent of the pragma is to support array processing.

	Syntax:

	#pragma prefetch

	There can be only one argument per pragma.  The compiler generates instructions to prefetch the cache lines starting from the address
	given in the argument. The array element values prefetched must be valid.  Reading outside the boundaries of an array results
	in undefined behavior at runtime.

	Example:
	The function below will prefetch ia and b, but not a[ia[i]] when compiled with +O2 +Odataprefetch +DA2.0 (or +DA2.0W).

	void testprefc2(int n, double *a, int *ia, double *b)
	{
		for(int i=0; i<n, i++)
		{
			b[i]=a[ia[i]];
		}
	}

	Recording this routine as

	#define USER_SPECIFIED 30
	void testprefc2(int n, double *a, int *ia, double *b)
	{
		int dist=(int)USER_SPECIFIED;
		int nend=max(0,n_dist);	// so as not to read past the end of ia
		for(i=0;i<nend;i++)	// original loop is for (i=0;i<n;i++)
		{
			#pragma prefetch ia[i+4*dist]
			#pragma prefetch a[ia[i+dist]]
			b[i]=a[ia[i]];
		}
		// finish up last part with no prefetching
		for (int i=nend;i<n;i++)
		{
		   b[i]=a[ia[i]];
		}
	}

	The two pragma statements allow a[ia[i]] to be prefetched. Note that the compiler continues to unroll the loops as in the original code.

	There can be problems using the prefetch pragma when the kernel cannot allocate large pages.  Without large pages, there can be
	performance lost to Translation Lookaside Faults (TLB).  The optimal page size varies with different applications but 4MB page size is
	a good average.

	TLB faults occur when a particular page address does not reside in the TLB buffer.  This buffer contains the mapping of the
	virtual addresses to the absolute addresses of the pages recently fetched in the cache.  A TLB fault happens when a reference to a
	particular virtual page address cannot be translated to an absolute address in the buffer.

	Even when all the TLB and prefetch features are working, you are still limited by the memory bandwidth of the system.  The top
	bandwidth may be reduced by failing to load all the memory slots in some PA-RISC systems.  The memory controller depends on having all slots loaded to get the best bank interleaving.
*/
#elif defined(_PA_RISC1_0) || defined(_PA_RISC1_1) || defined(_PA_RISC2_0) || defined(hppa) || defined(__hppa__)

# define PFETCH 1
# define hppa_hpc
# define CACHE_LINE_INTS    16
# define CACHE_LINE_DOUBLES  8
# define CACHE_LINE_COMPLEX  4

# define prefetch_data(_array,_k)	/*#pragma prefetch _array[_k]*/
# define prefetch_p_doubles(_pd)

/* PowerPC, Gnu C or C++ compiler: */
#elif((defined(__ppc__) || defined(__powerpc__)) && (defined(__GNUC__) || defined(__GNUG__)))

# ifndef PPC
#  define PPC
# endif
# define PFETCH 1
# define CACHE_LINE_INTS    8
# define CACHE_LINE_DOUBLES 4
# define CACHE_LINE_COMPLEX 2

# define prefetch_data(_array,_k)	__asm__ volatile ("dcbt  0,%0": :"r"(_array + _k));
/* Use << 3 here rather than *sizeof(double). */
# define prefetch_p_doubles(_pd)	__asm__ volatile ("dcbt  0,%1": :"r"(_pd), "r"(CACHE_LINE_DOUBLES << 3));

/* PowerPC, Metrowerks CodeWarrior C compiler */
#elif(defined(__MWERKS__) && defined(__powerc))

# define PPC
# define PFETCH 1
# define CACHE_LINE_INTS    8
# define CACHE_LINE_DOUBLES 4
# define CACHE_LINE_COMPLEX 2

# define prefetch_data(_array,_k)	__dcbt(_array,_k);
/* Use << 3 here rather than *sizeof(double). */
# define prefetch_p_doubles(_pd)	__dcbt(_pd,CACHE_LINE_DOUBLES << 3);

/* Alpha system, DEC/Compaq C compiler */
#elif(defined(__alpha) && defined(__DECC) && !defined(__VMS))
 /* Alpha system under TruUNIX, formerly known as OSF */
/*#elif(defined(__alpha) && defined(__osf__)) */

# include <c_asm.h>	/* Needed for assembly code macros  */

# define PFETCH 1
# define ALPHA_DECC 1
# define CACHE_LINE_INTS    16
# define CACHE_LINE_DOUBLES  8
# define CACHE_LINE_COMPLEX  4
# define prefetch_data(_array,_k)	asm("ldt %f31, 0(%a0)",(_array + _k));
# define prefetch_block(_array,_k)	asm("fetch_m   0(%a0)",(_array + _k));
# define prefetch_p_doubles(_pd)	asm("ldt %f31,32(%a0)",(_pd));

 /* Alpha system, Gnu C or C++ compiler */
#elif(defined(__alpha__) && (defined(__GNUC__) || defined(__GNUG__)))
 /* Alpha system under Gnu/Linux */
/*#elif(defined(linux) && defined(__alpha__)) */

# define PFETCH 1
# define ALPHA_GNUC 1
# define CACHE_LINE_INTS    16
# define CACHE_LINE_DOUBLES  8
# define CACHE_LINE_COMPLEX  4
# define prefetch_data(_array,_k)	__asm__ volatile ("ldt $f31, 0(%0)": :"r"(_array + _k));
# define prefetch_p_doubles(_pd)	__asm__ volatile ("ldt $f31,32(%0)": :"r"(_pd));

/* Ultrasparc pre-v9 (no prefetch instruction per se), Sunpro compiler */
/* *** NEED CORRECT ASM SYNTAX HERE - JSP SAYS NATIVE C ALLOWS ONLY VERY LIMITED ASM INSERTION ***
#elif(defined(sparc) && !defined(__sparc_v9__) && defined(__SUNPRO_C))

# define SPARCV8 1
# define CACHE_LINE_INTS    16
# define CACHE_LINE_DOUBLES  8
# define CACHE_LINE_COMPLEX  4
# define prefetch_data(_array,_k) asm ("ldd [%0   ],%%f30": :"r"(_array+_k));
# define prefetch_p_doubles(_pd)  asm ("ldd [%0+64],%%f30": :"r"(_pd      ));
*/
/* Ultrasparc v8 and above, Sunpro compiler - these macros are defined in the prefetch.h file */
#elif(defined(__SUN_PREFETCH))

# include <sun_prefetch.h>
# define PFETCH 1
# define SPARCV9 1
# define CACHE_LINE_INTS    16
# define CACHE_LINE_DOUBLES  8
# define CACHE_LINE_COMPLEX  4
# define prefetch_data(_array,_k)	/* */
# define prefetch_p_doubles(_pd)	sparc_prefetch_write_many(_pd);	/* Alternatives here are {read|write}_{once|many} */

/* Ultrasparc pre-v8 (no prefetch instruction per se), GNU C or C++ compiler */
#elif(defined(__sparc) && (defined(__GNUC__) || defined(__GNUG__)))

# define PFETCH 1
# define SPARCV8 1
# define CACHE_LINE_INTS    16
# define CACHE_LINE_DOUBLES  8
# define CACHE_LINE_COMPLEX  4
# define prefetch_data(_array,_k)	__asm__ volatile ("ldd [%0   ],%%f30": :"r"(_array+_k):"%f30");
# define prefetch_p_doubles(_pd)	__asm__ volatile ("ldd [%0+16],%%f30": :"r"(_pd      ):"%f30");

/* Ultrasparc v8+, GNU C or C++ compiler */
#elif(defined(__sparc_v9__) && (defined(__GNUC__) || defined(__GNUG__)))

# define PFETCH 1
# define SPARCV9 1
# define CACHE_LINE_INTS    16
# define CACHE_LINE_DOUBLES  8
# define CACHE_LINE_COMPLEX  4
# define prefetch_data(_array,_k)	__asm__ volatile ("prefetch [%0   ],3": :"g"(_array+_k));
# define prefetch_p_doubles(_pd)	__asm__ volatile ("prefetch [%0+64],3": :"g"(_pd));

/* Something I'd like to try with Itanium, Intel v7 compiler, when I get access to it. Compile with -DTARGET=31: */
/*
#elif(TARGET==31)

# define PFETCH	1
# define HPUX	1
# define CACHE_LINE_INTS    16
# define CACHE_LINE_DOUBLES  8
# define CACHE_LINE_COMPLEX  4
# define prefetch_data(_array, _k)	#pragma prefetch _array[_k];
*/

/* MIPS R10000 and above, MipsPro V7+ C compiler. An excertp from the C/C++ compile documentation,
found at http://dmawww.epfl.ch/ebt-bin/nph-dweb/dynaweb/SGI_Developer/Pragmas/%40Generic__BookTextView/4649 :

8.12 #pragma prefetch

	The #pragma prefetch directive specifies prefetching for each level of the cache.

	The syntax of the #pragma prefetch directive is as follows:

		#pragma prefetch [(n1, n2)]

	n1 controls the level 1 cache; n2 controls level 2. n1 and n2 can have the following values:

	Value	Effect
	-----	---------
	0	Prefetching is off (default for all processors except R10000)
	1	Prefetching is on but conservative (default at -03 when prefetch is on)
	2	Prefetching on and aggressive

	The scope of this directive is the entire function that contains it.

8.13 #pragma prefetch_manual

	The #pragma prefetch_manual directive instructs the compiler as to whether manual prefetches
	(through #pragma directives) should be respected or ignored.

	The syntax of the #pragma prefetch_manual directive is as follows:

		#pragma prefetch_manual[(n)]

	Value	Effect
	-----	---------
	0	Compiler ignores manual prefetches (default for all processors except R10000)
	1	Compiler respects manual prefetches (default at -03 for R10000 and beyond)

	The scope of this directive is the entire function that contains it.

8.14 #pragma prefetch_ref

	The #pragma prefetch_ref directive generates a prefetch and connects it to the specified reference (if possible).

	The syntax of the #pragma prefetch_ref directive is as follows:

		pragma prefetch_ref = ref[, stride = num1[, num2] ]  [, level = [lev1] [, lev2] ]  [, kind = {rd|wr}]  [, size = sz]

	ref is the object you want prefetched.

	Table 9 describes each of the possible #pragma prefetch_ref clauses. These clauses are optional.

	Table 9. Clauses for #pragma prefetch_ref

		Clause	Effect											Default
		------	----------------------------------							-------
		stride	Prefetches every num iteration(s) of this loop.						1

		level	Specifies the level in memory hierarchy to prefetch. The possible values for level are
			1: prefetch from L2 to L1 cache
			2: prefetch from memory to L1 cache							2

		kind	Specifies read or write.								write

		size	Specifies the size (in KB) of the object referenced in this loop. Must be a constant.	N/A

	The #pragma prefetch_ref directive instructs the compiler to take the following actions:

	* Generate a prefetch and connect to the specified object (if possible).

	* Search for references in the current loop-nest that match the supplied object.

		* If such a reference is found, then the prefetch for that object is scheduled relative
		     to the prefetch node, based on the miss latency for the specified level of the cache.
		* If no such reference is found, the prefetch is generated at the start of the loop body.

	* Ignore all references by the automatic prefetcher (if enabled) to this variable in this loop-nest.

	* Have the automatic prefetcher (if enabled) use the supplied size (if specified) in its volume analysis for this object.

	This directive has no scope; it just generates a prefetch.

8.15 #pragma prefetch_ref_disable

	The #pragma prefetch_ref_disable directive explicitly disables prefetching for the specified reference (in the current loop nest).

	The syntax of the #pragma prefetch_ref_disable directive is as follows:

		#pragma prefetch_ref_disable = ref[, size = num]

	ref is the object for which you want to disable prefetching.

	num specifies the size (in KB) of the object referenced in this loop (optional).
	The size must be a constant. This explicitly disables the prefetching of all references
	to object ref in the current loop nest. If enabled, the auto-prefetcher runs but ignores ref.
	The size is used for volume analysis.

	The scope of this directive is the entire function containing it.
*/

#else

# define CACHE_LINE_INTS     0
# define CACHE_LINE_DOUBLES  0
# define CACHE_LINE_COMPLEX  0
# define prefetch_data(_array,_k)	/* */
# define prefetch_p_doubles(_pd)	/* */

#endif

#endif	/* prefetch_h_included */
