/*******************************************************************************
*                                                                              *
*   (C) 1997-2004 by Ernst W. Mayer.                                           *
*                                                                              *
*  This program is free software; you can redistribute it and/or modify it     *
*  under the terms of the GNU General Public License as published by the       *
*  Free Software Foundation; either version 2 of the License, or (at your      *
*  option) any later version.                                                  *
*                                                                              *
*  This program is distributed in the hope that it will be useful, but WITHOUT *
*  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or       *
*  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for   *
*  more details.                                                               *
*                                                                              *
*  You should have received a copy of the GNU General Public License along     *
*  with this program; see the file GPL.txt.  If not, you may view one at       *
*  http://www.fsf.org/licenses/licenses.html, or obtain one by writing to the  *
*  Free Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA     *
*  02111-1307, USA.                                                            *
*                                                                              *
*******************************************************************************/

#include "Mlucas.h"

/***************/

void radix24_dif_pass1(double a[], int n)
{
/*
!...Acronym: DIF = Decimation In Frequency
!
!...Subroutine to perform an initial radix-24 complex DIF FFT pass on the data in the length-N real vector A.
!
!   See the documentation in radix16_dif_pass for further details on storage and indexing.
!
!   See the documentation in radix3,8_dif_pass for details on the radix-3,8 subtransforms.
*/
	int j,j1;
	static int n24,p01,p02,p03,p04,p05,p06,p07,p08,p09,p10,p11,p12,p13,p14,p15,p16,p17,p18,p19,p20,p21,p22,p23, first_entry=TRUE;
	static double c = .86602540378443864676, c3m1 = -1.5;
	double rt,it
	,t1,t2,t3,t4,t5,t6,t7,t8,t9,t10,t11,t12,t13,t14,t15,t16
	,aj1p00r,aj1p01r,aj1p02r,aj1p03r,aj1p04r,aj1p05r,aj1p06r,aj1p07r,aj1p08r,aj1p09r,aj1p10r,aj1p11r,aj1p12r,aj1p13r,aj1p14r,aj1p15r,aj1p16r,aj1p17r,aj1p18r,aj1p19r,aj1p20r,aj1p21r,aj1p22r,aj1p23r
	,aj1p00i,aj1p01i,aj1p02i,aj1p03i,aj1p04i,aj1p05i,aj1p06i,aj1p07i,aj1p08i,aj1p09i,aj1p10i,aj1p11i,aj1p12i,aj1p13i,aj1p14i,aj1p15i,aj1p16i,aj1p17i,aj1p18i,aj1p19i,aj1p20i,aj1p21i,aj1p22i,aj1p23i;

	if(!first_entry && (n/24) != n24)	/* New runlength?	*/
	{
	  first_entry=TRUE;
	}

/*...initialize things upon first entry	*/

	if(first_entry)
	{
	  first_entry=FALSE;
	  n24=n/24;

/*   constant index offsets for array load/stores are here.	*/

	  p01 = n24;
	  p02 = p01 + p01;
	  p03 = p02 + p01;
	  p04 = p03 + p01;
	  p05 = p04 + p01;
	  p06 = p05 + p01;
	  p07 = p06 + p01;
	  p08 = p07 + p01;
	  p09 = p08 + p01;
	  p10 = p09 + p01;
	  p11 = p10 + p01;
	  p12 = p11 + p01;
	  p13 = p12 + p01;
	  p14 = p13 + p01;
	  p15 = p14 + p01;
	  p16 = p15 + p01;
	  p17 = p16 + p01;
	  p18 = p17 + p01;
	  p19 = p18 + p01;
	  p20 = p19 + p01;
	  p21 = p20 + p01;
	  p22 = p21 + p01;
	  p23 = p22 + p01;

	  p01 = p01 + ( (p01 >> DAT_BITS) << PAD_BITS );
	  p02 = p02 + ( (p02 >> DAT_BITS) << PAD_BITS );
	  p03 = p03 + ( (p03 >> DAT_BITS) << PAD_BITS );
	  p04 = p04 + ( (p04 >> DAT_BITS) << PAD_BITS );
	  p05 = p05 + ( (p05 >> DAT_BITS) << PAD_BITS );
	  p06 = p06 + ( (p06 >> DAT_BITS) << PAD_BITS );
	  p07 = p07 + ( (p07 >> DAT_BITS) << PAD_BITS );
	  p08 = p08 + ( (p08 >> DAT_BITS) << PAD_BITS );
	  p09 = p09 + ( (p09 >> DAT_BITS) << PAD_BITS );
	  p10 = p10 + ( (p10 >> DAT_BITS) << PAD_BITS );
	  p11 = p11 + ( (p11 >> DAT_BITS) << PAD_BITS );
	  p12 = p12 + ( (p12 >> DAT_BITS) << PAD_BITS );
	  p13 = p13 + ( (p13 >> DAT_BITS) << PAD_BITS );
	  p14 = p14 + ( (p14 >> DAT_BITS) << PAD_BITS );
	  p15 = p15 + ( (p15 >> DAT_BITS) << PAD_BITS );
	  p16 = p16 + ( (p16 >> DAT_BITS) << PAD_BITS );
	  p17 = p17 + ( (p17 >> DAT_BITS) << PAD_BITS );
	  p18 = p18 + ( (p18 >> DAT_BITS) << PAD_BITS );
	  p19 = p19 + ( (p19 >> DAT_BITS) << PAD_BITS );
	  p20 = p20 + ( (p20 >> DAT_BITS) << PAD_BITS );
	  p21 = p21 + ( (p21 >> DAT_BITS) << PAD_BITS );
	  p22 = p22 + ( (p22 >> DAT_BITS) << PAD_BITS );
	  p23 = p23 + ( (p23 >> DAT_BITS) << PAD_BITS );
	}

/*...The radix-24 pass is here.	*/

	for(j=0; j < n24; j += 2)
	{
	  j1 = j  + ( (j  >> DAT_BITS) << PAD_BITS );
/*
Twiddleless version requires us to swap inputs as follows:
indices  0, 1, 2, 3, 4, 5, 6, 7, 8, 9,10,11,12,13,14,15,16,17,18,19,20,21,22,23
      -> 0,21,18,15,12, 9, 6, 3,16,13,10, 7, 4, 1,22,19, 8, 5, 2,23,20,17,14,11
I.e. start out with first triplet of indices {0,8,16}, permute those according to
{0,8,16}*23%24 = {0,16,8}, then each is head of a length-8 list of indices with decrement 3.
*/

#if 1
/* EWM: 10/21/04: We swap the following outputs of the radix-3 transforms: {1,9,17}<=>{5,13,21}, {3,11,19}<=>{7,15,23}, so that the indexing
                  winds up being in-place. This allows us to properly re-use the ajp1 variables in the carry-pass version of this routine.
*/
/*...gather the needed data (24 64-bit complex, i.e. 48 64-bit reals) and do 8 radix-3 transforms...*/
                     /*                        inputs                               */ /*             intermediates                 */ /*                 outputs                   */
	RADIX_03_DFT(a[j1    ],a[j1+1    ],a[j1+p16],a[j1+1+p16],a[j1+p08],a[j1+1+p08],aj1p00r,aj1p00i,aj1p08r,aj1p08i,aj1p16r,aj1p16i,aj1p00r,aj1p00i,aj1p08r,aj1p08i,aj1p16r,aj1p16i,rt,it);
	RADIX_03_DFT(a[j1+p21],a[j1+1+p21],a[j1+p13],a[j1+1+p13],a[j1+p05],a[j1+1+p05],aj1p05r,aj1p05i,aj1p13r,aj1p13i,aj1p21r,aj1p21i,aj1p05r,aj1p05i,aj1p13r,aj1p13i,aj1p21r,aj1p21i,rt,it);
	RADIX_03_DFT(a[j1+p18],a[j1+1+p18],a[j1+p10],a[j1+1+p10],a[j1+p02],a[j1+1+p02],aj1p02r,aj1p02i,aj1p10r,aj1p10i,aj1p18r,aj1p18i,aj1p02r,aj1p02i,aj1p10r,aj1p10i,aj1p18r,aj1p18i,rt,it);
	RADIX_03_DFT(a[j1+p15],a[j1+1+p15],a[j1+p07],a[j1+1+p07],a[j1+p23],a[j1+1+p23],aj1p07r,aj1p07i,aj1p15r,aj1p15i,aj1p23r,aj1p23i,aj1p07r,aj1p07i,aj1p15r,aj1p15i,aj1p23r,aj1p23i,rt,it);
	RADIX_03_DFT(a[j1+p12],a[j1+1+p12],a[j1+p04],a[j1+1+p04],a[j1+p20],a[j1+1+p20],aj1p04r,aj1p04i,aj1p12r,aj1p12i,aj1p20r,aj1p20i,aj1p04r,aj1p04i,aj1p12r,aj1p12i,aj1p20r,aj1p20i,rt,it);
	RADIX_03_DFT(a[j1+p09],a[j1+1+p09],a[j1+p01],a[j1+1+p01],a[j1+p17],a[j1+1+p17],aj1p01r,aj1p01i,aj1p09r,aj1p09i,aj1p17r,aj1p17i,aj1p01r,aj1p01i,aj1p09r,aj1p09i,aj1p17r,aj1p17i,rt,it);
	RADIX_03_DFT(a[j1+p06],a[j1+1+p06],a[j1+p22],a[j1+1+p22],a[j1+p14],a[j1+1+p14],aj1p06r,aj1p06i,aj1p14r,aj1p14i,aj1p22r,aj1p22i,aj1p06r,aj1p06i,aj1p14r,aj1p14i,aj1p22r,aj1p22i,rt,it);
	RADIX_03_DFT(a[j1+p03],a[j1+1+p03],a[j1+p19],a[j1+1+p19],a[j1+p11],a[j1+1+p11],aj1p03r,aj1p03i,aj1p11r,aj1p11i,aj1p19r,aj1p19i,aj1p03r,aj1p03i,aj1p11r,aj1p11i,aj1p19r,aj1p19i,rt,it);

/*...and now do 3 radix-8 transforms:	*/
                     /*                                                          inputs                                                           */ /*                 intermediates                    */ /*                 outputs                   */
	RADIX_08_DIF(aj1p00r,aj1p00i,aj1p05r,aj1p05i,aj1p02r,aj1p02i,aj1p07r,aj1p07i,aj1p04r,aj1p04i,aj1p01r,aj1p01i,aj1p06r,aj1p06i,aj1p03r,aj1p03i,t1,t2,t3,t4,t5,t6,t7,t8,t9,t10,t11,t12,t13,t14,t15,t16,a[j1    ],a[j1+1    ],a[j1+p01],a[j1+1+p01],a[j1+p02],a[j1+1+p02],a[j1+p03],a[j1+1+p03],a[j1+p05],a[j1+1+p05],a[j1+p04],a[j1+1+p04],a[j1+p07],a[j1+1+p07],a[j1+p06],a[j1+1+p06],rt,it);
	RADIX_08_DIF(aj1p08r,aj1p08i,aj1p13r,aj1p13i,aj1p10r,aj1p10i,aj1p15r,aj1p15i,aj1p12r,aj1p12i,aj1p09r,aj1p09i,aj1p14r,aj1p14i,aj1p11r,aj1p11i,t1,t2,t3,t4,t5,t6,t7,t8,t9,t10,t11,t12,t13,t14,t15,t16,a[j1+p18],a[j1+1+p18],a[j1+p19],a[j1+1+p19],a[j1+p17],a[j1+1+p17],a[j1+p16],a[j1+1+p16],a[j1+p23],a[j1+1+p23],a[j1+p22],a[j1+1+p22],a[j1+p20],a[j1+1+p20],a[j1+p21],a[j1+1+p21],rt,it);
	RADIX_08_DIF(aj1p16r,aj1p16i,aj1p21r,aj1p21i,aj1p18r,aj1p18i,aj1p23r,aj1p23i,aj1p20r,aj1p20i,aj1p17r,aj1p17i,aj1p22r,aj1p22i,aj1p19r,aj1p19i,t1,t2,t3,t4,t5,t6,t7,t8,t9,t10,t11,t12,t13,t14,t15,t16,a[j1+p13],a[j1+1+p13],a[j1+p12],a[j1+1+p12],a[j1+p15],a[j1+1+p15],a[j1+p14],a[j1+1+p14],a[j1+p10],a[j1+1+p10],a[j1+p11],a[j1+1+p11],a[j1+p09],a[j1+1+p09],a[j1+p08],a[j1+1+p08],rt,it);
#else

/*...gather the needed data (24 64-bit complex, i.e. 48 64-bit reals) and do 8 radix-3 transforms...*/
/*...1st radix-3 block uses aj1p[0:16:8] as inputs:	*/
	  aj1p00r =a[j1    ];			aj1p00i =a[j1+1    ];
	  aj1p08r =a[j1+p16]+a[j1+p08];		aj1p08i =a[j1+1+p16]+a[j1+1+p08];
	  aj1p16r =a[j1+p16]-a[j1+p08];		aj1p16i =a[j1+1+p16]-a[j1+1+p08];
	  aj1p00r =aj1p00r+aj1p08r;		aj1p00i =aj1p00i+aj1p08i;
	  aj1p08r =aj1p00r+aj1p08r*c3m1;	aj1p08i =aj1p00i+aj1p08i*c3m1;
	  rt =c*aj1p16r;			it =c*aj1p16i;
	  aj1p16r =aj1p08r+it;			aj1p16i =aj1p08i-rt;
	  aj1p08r =aj1p08r-it;			aj1p08i =aj1p08i+rt;

/*...2nd radix-3 block uses aj1p[1:17:8] as inputs:	*/
	  aj1p01r =a[j1+p21];			aj1p01i =a[j1+1+p21];
	  aj1p09r =a[j1+p13]+a[j1+p05];		aj1p09i =a[j1+1+p13]+a[j1+1+p05];
	  aj1p17r =a[j1+p13]-a[j1+p05];		aj1p17i =a[j1+1+p13]-a[j1+1+p05];
	  aj1p01r =aj1p01r+aj1p09r;		aj1p01i =aj1p01i+aj1p09i;
	  aj1p09r =aj1p01r+aj1p09r*c3m1;	aj1p09i =aj1p01i+aj1p09i*c3m1;
	  rt =c*aj1p17r;			it =c*aj1p17i;
	  aj1p17r =aj1p09r+it;			aj1p17i =aj1p09i-rt;
	  aj1p09r =aj1p09r-it;			aj1p09i =aj1p09i+rt;

/*...3rd radix-3 block uses aj1p[2:18:8] as inputs:	*/
	  aj1p02r =a[j1+p18];			aj1p02i =a[j1+1+p18];
	  aj1p10r =a[j1+p10]+a[j1+p02];		aj1p10i =a[j1+1+p10]+a[j1+1+p02];
	  aj1p18r =a[j1+p10]-a[j1+p02];		aj1p18i =a[j1+1+p10]-a[j1+1+p02];
	  aj1p02r =aj1p02r+aj1p10r;		aj1p02i =aj1p02i+aj1p10i;
	  aj1p10r =aj1p02r+aj1p10r*c3m1;	aj1p10i =aj1p02i+aj1p10i*c3m1;
	  rt =c*aj1p18r;			it =c*aj1p18i;
	  aj1p18r =aj1p10r+it;			aj1p18i =aj1p10i-rt;
	  aj1p10r =aj1p10r-it;			aj1p10i =aj1p10i+rt;

/*...4th radix-3 block uses aj1p[3:19:8] as inputs:	*/
	  aj1p03r =a[j1+p15];			aj1p03i =a[j1+1+p15];
	  aj1p11r =a[j1+p07]+a[j1+p23];		aj1p11i =a[j1+1+p07]+a[j1+1+p23];
	  aj1p19r =a[j1+p07]-a[j1+p23];		aj1p19i =a[j1+1+p07]-a[j1+1+p23];
	  aj1p03r =aj1p03r+aj1p11r;		aj1p03i =aj1p03i+aj1p11i;
	  aj1p11r =aj1p03r+aj1p11r*c3m1;	aj1p11i =aj1p03i+aj1p11i*c3m1;
	  rt =c*aj1p19r;			it =c*aj1p19i;
	  aj1p19r =aj1p11r+it;			aj1p19i =aj1p11i-rt;
	  aj1p11r =aj1p11r-it;			aj1p11i =aj1p11i+rt;

/*...5th radix-3 block uses aj1p[4:20:8] as inputs:	*/
	  aj1p04r =a[j1+p12];			aj1p04i =a[j1+1+p12];
	  aj1p12r =a[j1+p04]+a[j1+p20];		aj1p12i =a[j1+1+p04]+a[j1+1+p20];
	  aj1p20r =a[j1+p04]-a[j1+p20];		aj1p20i =a[j1+1+p04]-a[j1+1+p20];
	  aj1p04r =aj1p04r+aj1p12r;		aj1p04i =aj1p04i+aj1p12i;
	  aj1p12r =aj1p04r+aj1p12r*c3m1;	aj1p12i =aj1p04i+aj1p12i*c3m1;
	  rt =c*aj1p20r;			it =c*aj1p20i;
	  aj1p20r =aj1p12r+it;			aj1p20i =aj1p12i-rt;
	  aj1p12r =aj1p12r-it;			aj1p12i =aj1p12i+rt;

/*...6th radix-3 block uses aj1p[5:21:8] as inputs:	*/
	  aj1p05r =a[j1+p09];			aj1p05i =a[j1+1+p09];
	  aj1p13r =a[j1+p01]+a[j1+p17];		aj1p13i =a[j1+1+p01]+a[j1+1+p17];
	  aj1p21r =a[j1+p01]-a[j1+p17];		aj1p21i =a[j1+1+p01]-a[j1+1+p17];
	  aj1p05r =aj1p05r+aj1p13r;		aj1p05i =aj1p05i+aj1p13i;
	  aj1p13r =aj1p05r+aj1p13r*c3m1;	aj1p13i =aj1p05i+aj1p13i*c3m1;
	  rt =c*aj1p21r;			it =c*aj1p21i;
	  aj1p21r =aj1p13r+it;			aj1p21i =aj1p13i-rt;
	  aj1p13r =aj1p13r-it;			aj1p13i =aj1p13i+rt;

/*...7th radix-3 block uses aj1p[6:22:8] as inputs:	*/
	  aj1p06r =a[j1+p06];			aj1p06i =a[j1+1+p06];
	  aj1p14r =a[j1+p22]+a[j1+p14];		aj1p14i =a[j1+1+p22]+a[j1+1+p14];
	  aj1p22r =a[j1+p22]-a[j1+p14];		aj1p22i =a[j1+1+p22]-a[j1+1+p14];
	  aj1p06r =aj1p06r+aj1p14r;		aj1p06i =aj1p06i+aj1p14i;
	  aj1p14r =aj1p06r+aj1p14r*c3m1;	aj1p14i =aj1p06i+aj1p14i*c3m1;
	  rt =c*aj1p22r;			it =c*aj1p22i;
	  aj1p22r =aj1p14r+it;			aj1p22i =aj1p14i-rt;
	  aj1p14r =aj1p14r-it;			aj1p14i =aj1p14i+rt;

/*...8th radix-3 block uses aj1p[7:23:8] as inputs:	*/
	  aj1p07r =a[j1+p03];			aj1p07i =a[j1+1+p03];
	  aj1p15r =a[j1+p19]+a[j1+p11];		aj1p15i =a[j1+1+p19]+a[j1+1+p11];
	  aj1p23r =a[j1+p19]-a[j1+p11];		aj1p23i =a[j1+1+p19]-a[j1+1+p11];
	  aj1p07r =aj1p07r+aj1p15r;		aj1p07i =aj1p07i+aj1p15i;
	  aj1p15r =aj1p07r+aj1p15r*c3m1;	aj1p15i =aj1p07i+aj1p15i*c3m1;
	  rt =c*aj1p23r;			it =c*aj1p23i;
	  aj1p23r =aj1p15r+it;			aj1p23i =aj1p15i-rt;
	  aj1p15r =aj1p15r-it;			aj1p15i =aj1p15i+rt;

/*...and now do 3 radix-8 transforms:	*/
/*...Block 1: */
	  t1 =aj1p00r;		t2 =aj1p00i;
	  rt =aj1p04r;		it =aj1p04i;
	  t3 =t1 -rt;		t4 =t2 -it;
	  t1 =t1 +rt;		t2 =t2 +it;

	  t5 =aj1p02r;		t6 =aj1p02i;
	  rt =aj1p06r;		it =aj1p06i;
	  t7 =t5 -rt;		t8 =t6 -it;
	  t5 =t5 +rt;		t6 =t6 +it;

	  t9 =aj1p01r;		t10=aj1p01i;
	  rt =aj1p05r;		it =aj1p05i;
	  t11=t9 -rt;		t12=t10-it;
	  t9 =t9 +rt;		t10=t10+it;

	  t13=aj1p03r;		t14=aj1p03i;
	  rt =aj1p07r;		it =aj1p07i;
	  t15=t13-rt;		t16=t14-it;
	  t13=t13+rt;		t14=t14+it;
/*       combine to get the 2 length-4 transforms...	*/
	  rt =t5;		it =t6;
	  t5 =t1 -rt;		t6 =t2 -it;
	  t1 =t1 +rt;		t2 =t2 +it;

	  rt =t7;		it =t8;
	  t7 =t3 +it;		t8 =t4 -rt;
	  t3 =t3 -it;		t4 =t4 +rt;

	  rt =t13;		it =t14;
	  t13=t9 -rt;		t14=t10-it;
	  t9 =t9 +rt;		t10=t10+it;

	  rt =t15;		it =t16;
	  t15=t11+it;		t16=t12-rt;
	  t11=t11-it;		t12=t12+rt;
/*       now combine the two half-transforms	*/
	  a[j1    ]=t1+t9;	a[j1+1    ]=t2+t10;
	  a[j1+p01]=t1-t9;	a[j1+1+p01]=t2-t10;

	  a[j1+p02]=t5-t14;	a[j1+1+p02]=t6+t13;
	  a[j1+p03]=t5+t14;	a[j1+1+p03]=t6-t13;

	  rt =(t11-t12)*ISRT2;	it =(t11+t12)*ISRT2;
	  a[j1+p05]=t3+rt;	a[j1+1+p05]=t4+it;
	  a[j1+p04]=t3-rt;	a[j1+1+p04]=t4-it;

	  rt =(t15+t16)*ISRT2;	it =(t16-t15)*ISRT2;
	  a[j1+p07]=t7-rt;	a[j1+1+p07]=t8-it;
	  a[j1+p06]=t7+rt;	a[j1+1+p06]=t8+it;

/*...Block 2: */
	  t1 =aj1p08r;		t2 =aj1p08i;
	  rt =aj1p12r;		it =aj1p12i;
	  t3 =t1 -rt;		t4 =t2 -it;
	  t1 =t1 +rt;		t2 =t2 +it;

	  t5 =aj1p10r;		t6 =aj1p10i;
	  rt =aj1p14r;		it =aj1p14i;
	  t7 =t5 -rt;		t8 =t6 -it;
	  t5 =t5 +rt;		t6 =t6 +it;

	  t9 =aj1p09r;		t10=aj1p09i;
	  rt =aj1p13r;		it =aj1p13i;
	  t11=t9 -rt;		t12=t10-it;
	  t9 =t9 +rt;		t10=t10+it;

	  t13=aj1p11r;		t14=aj1p11i;
	  rt =aj1p15r;		it =aj1p15i;
	  t15=t13-rt;		t16=t14-it;
	  t13=t13+rt;		t14=t14+it;
/*       combine to get the 2 length-4 transforms...	*/
	  rt =t5;		it =t6;
	  t5 =t1 -rt;		t6 =t2 -it;
	  t1 =t1 +rt;		t2 =t2 +it;

	  rt =t7;		it =t8;
	  t7 =t3 +it;		t8 =t4 -rt;
	  t3 =t3 -it;		t4 =t4 +rt;

	  rt =t13;		it =t14;
	  t13=t9 -rt;		t14=t10-it;
	  t9 =t9 +rt;		t10=t10+it;

	  rt =t15;		it =t16;
	  t15=t11+it;		t16=t12-rt;
	  t11=t11-it;		t12=t12+rt;
/*       now combine the two half-transforms	*/
	  a[j1+p18]=t1+t9;	a[j1+1+p18]=t2+t10;
	  a[j1+p19]=t1-t9;	a[j1+1+p19]=t2-t10;

	  a[j1+p17]=t5-t14;	a[j1+1+p17]=t6+t13;
	  a[j1+p16]=t5+t14;	a[j1+1+p16]=t6-t13;

	  rt =(t11-t12)*ISRT2;	it =(t11+t12)*ISRT2;
	  a[j1+p23]=t3+rt;	a[j1+1+p23]=t4+it;
	  a[j1+p22]=t3-rt;	a[j1+1+p22]=t4-it;

	  rt =(t15+t16)*ISRT2;	it =(t16-t15)*ISRT2;
	  a[j1+p20]=t7-rt;	a[j1+1+p20]=t8-it;
	  a[j1+p21]=t7+rt;	a[j1+1+p21]=t8+it;

/*...Block 3: */

	  t1 =aj1p16r;		t2 =aj1p16i;
	  rt =aj1p20r;		it =aj1p20i;
	  t3 =t1 -rt;		t4 =t2 -it;
	  t1 =t1 +rt;		t2 =t2 +it;

	  t5 =aj1p18r;		t6 =aj1p18i;
	  rt =aj1p22r;		it =aj1p22i;
	  t7 =t5 -rt;		t8 =t6 -it;
	  t5 =t5 +rt;		t6 =t6 +it;

	  t9 =aj1p17r;		t10=aj1p17i;
	  rt =aj1p21r;		it =aj1p21i;
	  t11=t9 -rt;		t12=t10-it;
	  t9 =t9 +rt;		t10=t10+it;

	  t13=aj1p19r;		t14=aj1p19i;
	  rt =aj1p23r;		it =aj1p23i;
	  t15=t13-rt;		t16=t14-it;
	  t13=t13+rt;		t14=t14+it;
/*       combine to get the 2 length-4 transforms...	*/
	  rt =t5;		it =t6;
	  t5 =t1 -rt;		t6 =t2 -it;
	  t1 =t1 +rt;		t2 =t2 +it;

	  rt =t7;		it =t8;
	  t7 =t3 +it;		t8 =t4 -rt;
	  t3 =t3 -it;		t4 =t4 +rt;

	  rt =t13;		it =t14;
	  t13=t9 -rt;		t14=t10-it;
	  t9 =t9 +rt;		t10=t10+it;

	  rt =t15;		it =t16;
	  t15=t11+it;		t16=t12-rt;
	  t11=t11-it;		t12=t12+rt;
/*       now combine the two half-transforms	*/
	  a[j1+p13]=t1+t9;	a[j1+1+p13]=t2+t10;
	  a[j1+p12]=t1-t9;	a[j1+1+p12]=t2-t10;

	  a[j1+p15]=t5-t14;	a[j1+1+p15]=t6+t13;
	  a[j1+p14]=t5+t14;	a[j1+1+p14]=t6-t13;

	  rt =(t11-t12)*ISRT2;	it =(t11+t12)*ISRT2;
	  a[j1+p10]=t3+rt;	a[j1+1+p10]=t4+it;
	  a[j1+p11]=t3-rt;	a[j1+1+p11]=t4-it;

	  rt =(t15+t16)*ISRT2;	it =(t16-t15)*ISRT2;
	  a[j1+p09]=t7-rt;	a[j1+1+p09]=t8-it;
	  a[j1+p08]=t7+rt;	a[j1+1+p08]=t8+it;
#endif
	}
}

/*
   0 (   0)  129.0000000000   112.0000000000   129.0000000000   112.0000000000
   1 (  12)  -31.0000000000    -6.0000000000   -31.0000000000    -6.0000000000
   2 (   6)  -16.0000000000   -13.0000000000   -16.0000000000   -13.0000000000
   3 (  18)  -26.0000000000   -21.0000000000   -26.0000000000   -21.0000000000
   4 (   3)    7.7071067812   -17.3639610307     7.7071067812   -17.3639610307
   5 (  15)    6.2928932188    -4.6360389693     6.2928932188    -4.6360389693
   6 (   9)  -17.1213203436    -3.1213203436   -17.1213203436    -3.1213203436
   7 (  21)  -12.8786796564     1.1213203436   -12.8786796564     1.1213203436
   8 (   1)   25.0525588833     4.5884572681    -2.6087904079    11.1781322900
   9 (  13)   -1.4641016151    -4.2679491924    13.4651968684    24.8026298235
  10 (   7)   11.2942286341    13.8301270189     9.7818873666    -7.8061272737
  11 (  19)  -13.0262794416   -19.2224318643    -5.7100905968    -7.3900251490
  12 (   4)   -6.5600619596     8.4127289915     5.4641016151    -7.7320508076
  13 (  16)   24.4882651899   -13.2165765688   -13.0525588833   -26.5884572681
  14 (  10)   -1.6159544635     0.0466125814     6.0262794416    10.2224318643
  15 (  22)  -15.2404519970   -16.0273746949    -4.2942286341     5.1698729811
  16 (   2)  -13.0525588833   -26.5884572681   -13.0262794416   -19.2224318643
  17 (  14)    5.4641016151    -7.7320508076    11.2942286341    13.8301270189
  18 (   8)   -4.2942286341     5.1698729811    25.0525588833     4.5884572681
  19 (  20)    6.0262794416    10.2224318643    -1.4641016151    -4.2679491924
  20 (   5)    9.7818873666    -7.8061272737    -1.6159544635     0.0466125814
  21 (  17)   -5.7100905968    -7.3900251490   -15.2404519970   -16.0273746949
  22 (  11)   13.4651968684    24.8026298235    24.4882651899   -13.2165765688
  23 (  23)   -2.6087904079    11.1781322900    -6.5600619596     8.4127289915
*/

/**************/

int radix24_ditN_cy_dif1(double a[], int n, int nwt, int nwt_bits, double wt0[], double wt1[], int si[], double base[], double baseinv[], int iter, double *fracmax, int p)
{
/*
!...Acronym: DWT = Discrete Weighted Transform, DIT = Decimation In Time, DIF = Decimation In Frequency
!
!...Performs a final radix-24 complex DIT pass, an inverse DWT weighting, a carry propagation,
!   a forward DWT weighting, and an initial radix-24 complex DIF pass on the data in the length-N real vector A.
!
!   Data enter and are returned in the A-array.
!
!   See the documentation in mers_mod_square and radix16_dif_pass for further details on the array
!   storage scheme, and radix8_ditN_cy_dif1 for details on the reduced-length weights array scheme.
*/
	int n24, bjmodn00,bjmodn01,bjmodn02,bjmodn03,bjmodn04,bjmodn05,bjmodn06,bjmodn07,bjmodn08,bjmodn09,bjmodn10,bjmodn11,bjmodn12,bjmodn13,bjmodn14,bjmodn15,bjmodn16,bjmodn17,bjmodn18,bjmodn19,bjmodn20,bjmodn21,bjmodn22,bjmodn23
		,i,j,j1,jstart,jhi,iroot,root_incr,k,khi,l,outer;
	static int psave=0,bw,sw,bjmodnini,nrtm1,p01,p02,p03,p04,p05,p06,p07,p08,p09,p10,p11,p12,p13,p14,p15,p16,p17,p18,p19,p20,p21,p22,p23;
	const  double one_half[3] = {1.0, 0.5, 0.25};	/* Needed for small-weights-tables scheme */
	static double radix_inv, n2inv;
	static double c = .86602540378443864676, c3m1 = -1.5;
	double rt,it
	,t1,t2,t3,t4,t5,t6,t7,t8,t9,t10,t11,t12,t13,t14,t15,t16
	,aj1p00r,aj1p01r,aj1p02r,aj1p03r,aj1p04r,aj1p05r,aj1p06r,aj1p07r,aj1p08r,aj1p09r,aj1p10r,aj1p11r,aj1p12r,aj1p13r,aj1p14r,aj1p15r,aj1p16r,aj1p17r,aj1p18r,aj1p19r,aj1p20r,aj1p21r,aj1p22r,aj1p23r
	,aj1p00i,aj1p01i,aj1p02i,aj1p03i,aj1p04i,aj1p05i,aj1p06i,aj1p07i,aj1p08i,aj1p09i,aj1p10i,aj1p11i,aj1p12i,aj1p13i,aj1p14i,aj1p15i,aj1p16i,aj1p17i,aj1p18i,aj1p19i,aj1p20i,aj1p21i,aj1p22i,aj1p23i
	,cy00,cy01,cy02,cy03,cy04,cy05,cy06,cy07,cy08,cy09,cy10,cy11,cy12,cy13,cy14,cy15,cy16,cy17,cy18,cy19,cy20,cy21,cy22,cy23,temp,frac,scale;
	double maxerr = 0.0;
#if PFETCH
	double *add0, *add;
#endif
	int err;
	FILE *f1,*f2;
	static int first_entry=TRUE;

/*...stuff for the reduced-length DWT weights array is here:	*/
	int n_div_nwt;
	int col,co2,co3,m,m2,n_minus_sil,n_minus_silp1,sinwt,sinwtm1;
	double wt,wtinv,wtl,wtlp1,wtn,wtnm1,wtA,wtB,wtC;

/*...change n24 and n_div_wt to non-static to work around a gcc compiler bug. */
	n24   = n/24;
	n_div_nwt = n24 >> nwt_bits;

	if((n_div_nwt << nwt_bits) != n24)
	{
	  if(INTERACT)printf("FATAL: iter = %10d; NWT_BITS does not divide N/24 in radix24_ditN_cy_dif1.\n",iter);
	  f1 = fopen(   OFILE,"a");
	  f2 = fopen(STATFILE,"a");
	  fprintf(f1,"FATAL: iter = %10d; NWT_BITS does not divide N/24 in radix24_ditN_cy_dif1.\n",iter);
	  fprintf(f2,"FATAL: iter = %10d; NWT_BITS does not divide N/24 in radix24_ditN_cy_dif1.\n",iter);
	  fclose(f1);
	  fclose(f2);
	  err=1;
	  return(err);
	}

	if(p != psave)
	{
	  first_entry=TRUE;
	}

/*...initialize things upon first entry	*/

	if(first_entry)
	{
	  psave = p;
	  first_entry=FALSE;
	  radix_inv = (double)((long double)1.0L/(long double)24.0L);
	  n2inv     = (double)((long double)1.0L/(long double)(n/2));

	  bw    = p%n;		/* Number of bigwords in the Crandall/Fagin mixed-radix representation = (Mersenne exponent) mod (vector length).	*/
	  sw    = n - bw;	/* Number of smallwords.	*/

/*   constant index offsets for array load/stores are here.	*/

	  p01 = n24;
	  p02 = p01 + p01;
	  p03 = p02 + p01;
	  p04 = p03 + p01;
	  p05 = p04 + p01;
	  p06 = p05 + p01;
	  p07 = p06 + p01;
	  p08 = p07 + p01;
	  p09 = p08 + p01;
	  p10 = p09 + p01;
	  p11 = p10 + p01;
	  p12 = p11 + p01;
	  p13 = p12 + p01;
	  p14 = p13 + p01;
	  p15 = p14 + p01;
	  p16 = p15 + p01;
	  p17 = p16 + p01;
	  p18 = p17 + p01;
	  p19 = p18 + p01;
	  p20 = p19 + p01;
	  p21 = p20 + p01;
	  p22 = p21 + p01;
	  p23 = p22 + p01;

	  p01 = p01 + ( (p01 >> DAT_BITS) << PAD_BITS );
	  p02 = p02 + ( (p02 >> DAT_BITS) << PAD_BITS );
	  p03 = p03 + ( (p03 >> DAT_BITS) << PAD_BITS );
	  p04 = p04 + ( (p04 >> DAT_BITS) << PAD_BITS );
	  p05 = p05 + ( (p05 >> DAT_BITS) << PAD_BITS );
	  p06 = p06 + ( (p06 >> DAT_BITS) << PAD_BITS );
	  p07 = p07 + ( (p07 >> DAT_BITS) << PAD_BITS );
	  p08 = p08 + ( (p08 >> DAT_BITS) << PAD_BITS );
	  p09 = p09 + ( (p09 >> DAT_BITS) << PAD_BITS );
	  p10 = p10 + ( (p10 >> DAT_BITS) << PAD_BITS );
	  p11 = p11 + ( (p11 >> DAT_BITS) << PAD_BITS );
	  p12 = p12 + ( (p12 >> DAT_BITS) << PAD_BITS );
	  p13 = p13 + ( (p13 >> DAT_BITS) << PAD_BITS );
	  p14 = p14 + ( (p14 >> DAT_BITS) << PAD_BITS );
	  p15 = p15 + ( (p15 >> DAT_BITS) << PAD_BITS );
	  p16 = p16 + ( (p16 >> DAT_BITS) << PAD_BITS );
	  p17 = p17 + ( (p17 >> DAT_BITS) << PAD_BITS );
	  p18 = p18 + ( (p18 >> DAT_BITS) << PAD_BITS );
	  p19 = p19 + ( (p19 >> DAT_BITS) << PAD_BITS );
	  p20 = p20 + ( (p20 >> DAT_BITS) << PAD_BITS );
	  p21 = p21 + ( (p21 >> DAT_BITS) << PAD_BITS );
	  p22 = p22 + ( (p22 >> DAT_BITS) << PAD_BITS );
	  p23 = p23 + ( (p23 >> DAT_BITS) << PAD_BITS );

	  bjmodnini=0;
	  for(j=0; j < n24; j++)
	  {
	    bjmodnini -= sw; bjmodnini = bjmodnini + ( (-(int)((uint32)bjmodnini >> 31)) & n);
	  }
	}

/*...The radix-24 final DIT pass is here.	*/

	cy00=-2;		/* init carries	*/
	cy01= 0;
	cy02= 0;
	cy03= 0;
	cy04= 0;
	cy05= 0;
	cy06= 0;
	cy07= 0;
	cy08= 0;
	cy09= 0;
	cy10= 0;
	cy11= 0;
	cy12= 0;
	cy13= 0;
	cy14= 0;
	cy15= 0;
	cy16= 0;
	cy17= 0;
	cy18= 0;
	cy19= 0;
	cy20= 0;
	cy21= 0;
	cy22= 0;
	cy23= 0;

	*fracmax=0;	/* init max. fractional error	*/

	iroot = 0;	/* init sincos array index	*/
	root_incr = 1;	/* init sincos array index increment (set = 1 for normal carry pass, = 0 for wrapper pass)	*/
	scale = n2inv;	/* init inverse-weight scale factor  (set = 2/n for normal carry pass, = 1 for wrapper pass)	*/

	jstart = 0;
	jhi = jstart+nwt-1;
	khi = n_div_nwt;

for(outer=0; outer <= 1; outer++)
{
	i = 1;		/* Pointer to the BASE and BASEINV arrays. lowest-order digit is always a bigword (i = 1).	*/

	bjmodn00= 0;
	bjmodn01= bjmodnini;
	bjmodn02= bjmodn01+bjmodnini-n; bjmodn02= bjmodn02+ ( (-(int)((uint32)bjmodn02>> 31)) & n);
	bjmodn03= bjmodn02+bjmodnini-n; bjmodn03= bjmodn03+ ( (-(int)((uint32)bjmodn03>> 31)) & n);
	bjmodn04= bjmodn03+bjmodnini-n; bjmodn04= bjmodn04+ ( (-(int)((uint32)bjmodn04>> 31)) & n);
	bjmodn05= bjmodn04+bjmodnini-n; bjmodn05= bjmodn05+ ( (-(int)((uint32)bjmodn05>> 31)) & n);
	bjmodn06= bjmodn05+bjmodnini-n; bjmodn06= bjmodn06+ ( (-(int)((uint32)bjmodn06>> 31)) & n);
	bjmodn07= bjmodn06+bjmodnini-n; bjmodn07= bjmodn07+ ( (-(int)((uint32)bjmodn07>> 31)) & n);
	bjmodn08= bjmodn07+bjmodnini-n; bjmodn08= bjmodn08+ ( (-(int)((uint32)bjmodn08>> 31)) & n);
	bjmodn09= bjmodn08+bjmodnini-n; bjmodn09= bjmodn09+ ( (-(int)((uint32)bjmodn09>> 31)) & n);
	bjmodn10= bjmodn09+bjmodnini-n; bjmodn10= bjmodn10+ ( (-(int)((uint32)bjmodn10>> 31)) & n);
	bjmodn11= bjmodn10+bjmodnini-n; bjmodn11= bjmodn11+ ( (-(int)((uint32)bjmodn11>> 31)) & n);
	bjmodn12= bjmodn11+bjmodnini-n; bjmodn12= bjmodn12+ ( (-(int)((uint32)bjmodn12>> 31)) & n);
	bjmodn13= bjmodn12+bjmodnini-n; bjmodn13= bjmodn13+ ( (-(int)((uint32)bjmodn13>> 31)) & n);
	bjmodn14= bjmodn13+bjmodnini-n; bjmodn14= bjmodn14+ ( (-(int)((uint32)bjmodn14>> 31)) & n);
	bjmodn15= bjmodn14+bjmodnini-n; bjmodn15= bjmodn15+ ( (-(int)((uint32)bjmodn15>> 31)) & n);
	bjmodn16= bjmodn15+bjmodnini-n; bjmodn16= bjmodn16+ ( (-(int)((uint32)bjmodn16>> 31)) & n);
	bjmodn17= bjmodn16+bjmodnini-n; bjmodn17= bjmodn17+ ( (-(int)((uint32)bjmodn17>> 31)) & n);
	bjmodn18= bjmodn17+bjmodnini-n; bjmodn18= bjmodn18+ ( (-(int)((uint32)bjmodn18>> 31)) & n);
	bjmodn19= bjmodn18+bjmodnini-n; bjmodn19= bjmodn19+ ( (-(int)((uint32)bjmodn19>> 31)) & n);
	bjmodn20= bjmodn19+bjmodnini-n; bjmodn20= bjmodn20+ ( (-(int)((uint32)bjmodn20>> 31)) & n);
	bjmodn21= bjmodn20+bjmodnini-n; bjmodn21= bjmodn21+ ( (-(int)((uint32)bjmodn21>> 31)) & n);
	bjmodn22= bjmodn21+bjmodnini-n; bjmodn22= bjmodn22+ ( (-(int)((uint32)bjmodn22>> 31)) & n);
	bjmodn23= bjmodn22+bjmodnini-n; bjmodn23= bjmodn23+ ( (-(int)((uint32)bjmodn23>> 31)) & n);

	col=0;
	co2=(n >> nwt_bits)-1+24;
	co3=co2-24;		/* At the start of each new j-loop, co3=co2-radix(1)	*/

	for(k=1; k <= khi; k++)	/* Do n/(radix(1)*nwt) outer loop executions...	*/
	{
	  for(j=jstart; j<jhi; j += 2)	/* Each inner loop execution processes (radix(1)*nwt) array data.	*/
	  {
	    j1 = j + ( (j >> DAT_BITS) << PAD_BITS );	/* padded-array fetch index is here	*/

/*...The radix-24 DIT pass is here:	*/
/*...gather the needed data (24 64-bit complex, i.e. 48 64-bit reals) and do 3 radix-8 transforms,	*/
                     /*                                                                                 inputs                                                                                    */ /*                 intermediates                    */ /*                                                         outputs                                                           */
	RADIX_08_DIT(a[j1    ],a[j1+1    ],a[j1+p01],a[j1+1+p01],a[j1+p03],a[j1+1+p03],a[j1+p02],a[j1+1+p02],a[j1+p07],a[j1+1+p07],a[j1+p06],a[j1+1+p06],a[j1+p05],a[j1+1+p05],a[j1+p04],a[j1+1+p04],t1,t2,t3,t4,t5,t6,t7,t8,t9,t10,t11,t12,t13,t14,t15,t16,aj1p00r,aj1p00i,aj1p01r,aj1p01i,aj1p02r,aj1p02i,aj1p03r,aj1p03i,aj1p04r,aj1p04i,aj1p05r,aj1p05i,aj1p06r,aj1p06i,aj1p07r,aj1p07i,rt,it);
	RADIX_08_DIT(a[j1+p13],a[j1+1+p13],a[j1+p12],a[j1+1+p12],a[j1+p14],a[j1+1+p14],a[j1+p15],a[j1+1+p15],a[j1+p09],a[j1+1+p09],a[j1+p08],a[j1+1+p08],a[j1+p10],a[j1+1+p10],a[j1+p11],a[j1+1+p11],t1,t2,t3,t4,t5,t6,t7,t8,t9,t10,t11,t12,t13,t14,t15,t16,aj1p08r,aj1p08i,aj1p09r,aj1p09i,aj1p10r,aj1p10i,aj1p11r,aj1p11i,aj1p12r,aj1p12i,aj1p13r,aj1p13i,aj1p14r,aj1p14i,aj1p15r,aj1p15i,rt,it);
	RADIX_08_DIT(a[j1+p18],a[j1+1+p18],a[j1+p19],a[j1+1+p19],a[j1+p16],a[j1+1+p16],a[j1+p17],a[j1+1+p17],a[j1+p20],a[j1+1+p20],a[j1+p21],a[j1+1+p21],a[j1+p23],a[j1+1+p23],a[j1+p22],a[j1+1+p22],t1,t2,t3,t4,t5,t6,t7,t8,t9,t10,t11,t12,t13,t14,t15,t16,aj1p16r,aj1p16i,aj1p17r,aj1p17i,aj1p18r,aj1p18i,aj1p19r,aj1p19i,aj1p20r,aj1p20i,aj1p21r,aj1p21i,aj1p22r,aj1p22i,aj1p23r,aj1p23i,rt,it);

/*...and now do 8 in-place radix-3 transforms.	*/
                     /*                  inputs                   */ /*             intermediates                 */ /*                       outputs                               */
	RADIX_03_DFT(aj1p00r,aj1p00i,aj1p08r,aj1p08i,aj1p16r,aj1p16i,t1,t2,t3,t4,t5,t6,aj1p00r,aj1p00i,aj1p16r,aj1p16i,aj1p08r,aj1p08i,rt,it);
	RADIX_03_DFT(aj1p01r,aj1p01i,aj1p09r,aj1p09i,aj1p17r,aj1p17i,t1,t2,t3,t4,t5,t6,aj1p09r,aj1p09i,aj1p01r,aj1p01i,aj1p17r,aj1p17i,rt,it);
	RADIX_03_DFT(aj1p02r,aj1p02i,aj1p10r,aj1p10i,aj1p18r,aj1p18i,t1,t2,t3,t4,t5,t6,aj1p18r,aj1p18i,aj1p10r,aj1p10i,aj1p02r,aj1p02i,rt,it);
	RADIX_03_DFT(aj1p03r,aj1p03i,aj1p11r,aj1p11i,aj1p19r,aj1p19i,t1,t2,t3,t4,t5,t6,aj1p03r,aj1p03i,aj1p19r,aj1p19i,aj1p11r,aj1p11i,rt,it);
	RADIX_03_DFT(aj1p04r,aj1p04i,aj1p12r,aj1p12i,aj1p20r,aj1p20i,t1,t2,t3,t4,t5,t6,aj1p12r,aj1p12i,aj1p04r,aj1p04i,aj1p20r,aj1p20i,rt,it);
	RADIX_03_DFT(aj1p05r,aj1p05i,aj1p13r,aj1p13i,aj1p21r,aj1p21i,t1,t2,t3,t4,t5,t6,aj1p21r,aj1p21i,aj1p13r,aj1p13i,aj1p05r,aj1p05i,rt,it);
	RADIX_03_DFT(aj1p06r,aj1p06i,aj1p14r,aj1p14i,aj1p22r,aj1p22i,t1,t2,t3,t4,t5,t6,aj1p06r,aj1p06i,aj1p22r,aj1p22i,aj1p14r,aj1p14i,rt,it);
	RADIX_03_DFT(aj1p07r,aj1p07i,aj1p15r,aj1p15i,aj1p23r,aj1p23i,t1,t2,t3,t4,t5,t6,aj1p15r,aj1p15i,aj1p07r,aj1p07i,aj1p23r,aj1p23i,rt,it);

/*...Now do the carries. Since the outputs would
    normally be getting dispatched to 24 separate blocks of the A-array, we need 24 separate carries.	*/

	    l= j & (nwt-1);
	    n_minus_sil   = n-si[l  ];
	    n_minus_silp1 = n-si[l+1];
	    sinwt   = si[nwt-l  ];
	    sinwtm1 = si[nwt-l-1];

	    wtl     =wt0[    l  ];
	    wtn     =wt0[nwt-l  ]*scale;	/* Include 1/(n/2) scale factor of inverse transform here...	*/
	    wtlp1   =wt0[    l+1];
	    wtnm1   =wt0[nwt-l-1]*scale;	/* ...and here.	*/

/*...set0 is slightly different from others:	*/
	   cmplx_carry_norm_errcheck0(aj1p00r,aj1p00i,cy00,bjmodn00   );
	    cmplx_carry_norm_errcheck(aj1p01r,aj1p01i,cy01,bjmodn01,1 );
	    cmplx_carry_norm_errcheck(aj1p02r,aj1p02i,cy02,bjmodn02,2 );
	    cmplx_carry_norm_errcheck(aj1p03r,aj1p03i,cy03,bjmodn03,3 );
	    cmplx_carry_norm_errcheck(aj1p04r,aj1p04i,cy04,bjmodn04,4 );
	    cmplx_carry_norm_errcheck(aj1p05r,aj1p05i,cy05,bjmodn05,5 );
	    cmplx_carry_norm_errcheck(aj1p06r,aj1p06i,cy06,bjmodn06,6 );
	    cmplx_carry_norm_errcheck(aj1p07r,aj1p07i,cy07,bjmodn07,7 );
	    cmplx_carry_norm_errcheck(aj1p08r,aj1p08i,cy08,bjmodn08,8 );
	    cmplx_carry_norm_errcheck(aj1p09r,aj1p09i,cy09,bjmodn09,9 );
	    cmplx_carry_norm_errcheck(aj1p10r,aj1p10i,cy10,bjmodn10,10);
	    cmplx_carry_norm_errcheck(aj1p11r,aj1p11i,cy11,bjmodn11,11);
	    cmplx_carry_norm_errcheck(aj1p12r,aj1p12i,cy12,bjmodn12,12);
	    cmplx_carry_norm_errcheck(aj1p13r,aj1p13i,cy13,bjmodn13,13);
	    cmplx_carry_norm_errcheck(aj1p14r,aj1p14i,cy14,bjmodn14,14);
	    cmplx_carry_norm_errcheck(aj1p15r,aj1p15i,cy15,bjmodn15,15);
	    cmplx_carry_norm_errcheck(aj1p16r,aj1p16i,cy16,bjmodn16,16);
	    cmplx_carry_norm_errcheck(aj1p17r,aj1p17i,cy17,bjmodn17,17);
	    cmplx_carry_norm_errcheck(aj1p18r,aj1p18i,cy18,bjmodn18,18);
	    cmplx_carry_norm_errcheck(aj1p19r,aj1p19i,cy19,bjmodn19,19);
	    cmplx_carry_norm_errcheck(aj1p20r,aj1p20i,cy20,bjmodn20,20);
	    cmplx_carry_norm_errcheck(aj1p21r,aj1p21i,cy21,bjmodn21,21);
	    cmplx_carry_norm_errcheck(aj1p22r,aj1p22i,cy22,bjmodn22,22);
	    cmplx_carry_norm_errcheck(aj1p23r,aj1p23i,cy23,bjmodn23,23);

	    i =((uint32)(sw - bjmodn00) >> 31);	/* get ready for the next set...	*/
	    co2=co3;	/* For all data but the first set in each j-block, co2=co3. Thus, after the first block of data is done
			   and only then: for all subsequent blocks it's superfluous), this assignment decrements co2 by radix(1).	*/

#if PFETCH
add0 = &a[j1];
prefetch_p_doubles(add0);
#endif

/*...The radix-24 DIF pass is here:	*/

/* EWM: 10/21/04: We swap the following outputs of the radix-3 transforms: {1,9,17}<=>{5,13,21}, {3,11,19}<=>{7,15,23}, so that the indexing
                  winds up being in-place. This allows us to properly re-use the ajp1 variables in the carry-pass version of this routine.
*/

/*...gather the needed data (24 64-bit complex, i.e. 48 64-bit reals) and do 8 in-place radix-3 transforms...*/
                     /*                        inputs                               */ /*             intermediates                 */ /*                 outputs                   */
#if PFETCH
	RADIX_03_DFT_PFETCH(aj1p00r,aj1p00i,aj1p16r,aj1p16i,aj1p08r,aj1p08i,t1,t2,t3,t4,t5,t6,aj1p00r,aj1p00i,aj1p08r,aj1p08i,aj1p16r,aj1p16i,rt,it,p01);
	RADIX_03_DFT_PFETCH(aj1p21r,aj1p21i,aj1p13r,aj1p13i,aj1p05r,aj1p05i,t1,t2,t3,t4,t5,t6,aj1p05r,aj1p05i,aj1p13r,aj1p13i,aj1p21r,aj1p21i,rt,it,p02);
	RADIX_03_DFT_PFETCH(aj1p18r,aj1p18i,aj1p10r,aj1p10i,aj1p02r,aj1p02i,t1,t2,t3,t4,t5,t6,aj1p02r,aj1p02i,aj1p10r,aj1p10i,aj1p18r,aj1p18i,rt,it,p03);
	RADIX_03_DFT_PFETCH(aj1p15r,aj1p15i,aj1p07r,aj1p07i,aj1p23r,aj1p23i,t1,t2,t3,t4,t5,t6,aj1p07r,aj1p07i,aj1p15r,aj1p15i,aj1p23r,aj1p23i,rt,it,p04);
	RADIX_03_DFT_PFETCH(aj1p12r,aj1p12i,aj1p04r,aj1p04i,aj1p20r,aj1p20i,t1,t2,t3,t4,t5,t6,aj1p04r,aj1p04i,aj1p12r,aj1p12i,aj1p20r,aj1p20i,rt,it,p05);
	RADIX_03_DFT_PFETCH(aj1p09r,aj1p09i,aj1p01r,aj1p01i,aj1p17r,aj1p17i,t1,t2,t3,t4,t5,t6,aj1p01r,aj1p01i,aj1p09r,aj1p09i,aj1p17r,aj1p17i,rt,it,p06);
	RADIX_03_DFT_PFETCH(aj1p06r,aj1p06i,aj1p22r,aj1p22i,aj1p14r,aj1p14i,t1,t2,t3,t4,t5,t6,aj1p06r,aj1p06i,aj1p14r,aj1p14i,aj1p22r,aj1p22i,rt,it,p07);
	RADIX_03_DFT_PFETCH(aj1p03r,aj1p03i,aj1p19r,aj1p19i,aj1p11r,aj1p11i,t1,t2,t3,t4,t5,t6,aj1p03r,aj1p03i,aj1p11r,aj1p11i,aj1p19r,aj1p19i,rt,it,p08);
#else
	RADIX_03_DFT       (aj1p00r,aj1p00i,aj1p16r,aj1p16i,aj1p08r,aj1p08i,t1,t2,t3,t4,t5,t6,aj1p00r,aj1p00i,aj1p08r,aj1p08i,aj1p16r,aj1p16i,rt,it);
	RADIX_03_DFT       (aj1p21r,aj1p21i,aj1p13r,aj1p13i,aj1p05r,aj1p05i,t1,t2,t3,t4,t5,t6,aj1p05r,aj1p05i,aj1p13r,aj1p13i,aj1p21r,aj1p21i,rt,it);
	RADIX_03_DFT       (aj1p18r,aj1p18i,aj1p10r,aj1p10i,aj1p02r,aj1p02i,t1,t2,t3,t4,t5,t6,aj1p02r,aj1p02i,aj1p10r,aj1p10i,aj1p18r,aj1p18i,rt,it);
	RADIX_03_DFT       (aj1p15r,aj1p15i,aj1p07r,aj1p07i,aj1p23r,aj1p23i,t1,t2,t3,t4,t5,t6,aj1p07r,aj1p07i,aj1p15r,aj1p15i,aj1p23r,aj1p23i,rt,it);
	RADIX_03_DFT       (aj1p12r,aj1p12i,aj1p04r,aj1p04i,aj1p20r,aj1p20i,t1,t2,t3,t4,t5,t6,aj1p04r,aj1p04i,aj1p12r,aj1p12i,aj1p20r,aj1p20i,rt,it);
	RADIX_03_DFT       (aj1p09r,aj1p09i,aj1p01r,aj1p01i,aj1p17r,aj1p17i,t1,t2,t3,t4,t5,t6,aj1p01r,aj1p01i,aj1p09r,aj1p09i,aj1p17r,aj1p17i,rt,it);
	RADIX_03_DFT       (aj1p06r,aj1p06i,aj1p22r,aj1p22i,aj1p14r,aj1p14i,t1,t2,t3,t4,t5,t6,aj1p06r,aj1p06i,aj1p14r,aj1p14i,aj1p22r,aj1p22i,rt,it);
	RADIX_03_DFT       (aj1p03r,aj1p03i,aj1p19r,aj1p19i,aj1p11r,aj1p11i,t1,t2,t3,t4,t5,t6,aj1p03r,aj1p03i,aj1p11r,aj1p11i,aj1p19r,aj1p19i,rt,it);
#endif

/*...and now do 3 radix-8 transforms:	*/
                     /*                                                          inputs                                                           */ /*                 intermediates                    */ /*                 outputs                   */
#if PFETCH
	RADIX_08_DIF_PFETCH(aj1p00r,aj1p00i,aj1p05r,aj1p05i,aj1p02r,aj1p02i,aj1p07r,aj1p07i,aj1p04r,aj1p04i,aj1p01r,aj1p01i,aj1p06r,aj1p06i,aj1p03r,aj1p03i,t1,t2,t3,t4,t5,t6,t7,t8,t9,t10,t11,t12,t13,t14,t15,t16,a[j1    ],a[j1+1    ],a[j1+p01],a[j1+1+p01],a[j1+p02],a[j1+1+p02],a[j1+p03],a[j1+1+p03],a[j1+p05],a[j1+1+p05],a[j1+p04],a[j1+1+p04],a[j1+p07],a[j1+1+p07],a[j1+p06],a[j1+1+p06],rt,it,p09,p10,p11,p12,p13);
	RADIX_08_DIF_PFETCH(aj1p08r,aj1p08i,aj1p13r,aj1p13i,aj1p10r,aj1p10i,aj1p15r,aj1p15i,aj1p12r,aj1p12i,aj1p09r,aj1p09i,aj1p14r,aj1p14i,aj1p11r,aj1p11i,t1,t2,t3,t4,t5,t6,t7,t8,t9,t10,t11,t12,t13,t14,t15,t16,a[j1+p18],a[j1+1+p18],a[j1+p19],a[j1+1+p19],a[j1+p17],a[j1+1+p17],a[j1+p16],a[j1+1+p16],a[j1+p23],a[j1+1+p23],a[j1+p22],a[j1+1+p22],a[j1+p20],a[j1+1+p20],a[j1+p21],a[j1+1+p21],rt,it,p14,p15,p16,p17,p18);
	RADIX_08_DIF_PFETCH(aj1p16r,aj1p16i,aj1p21r,aj1p21i,aj1p18r,aj1p18i,aj1p23r,aj1p23i,aj1p20r,aj1p20i,aj1p17r,aj1p17i,aj1p22r,aj1p22i,aj1p19r,aj1p19i,t1,t2,t3,t4,t5,t6,t7,t8,t9,t10,t11,t12,t13,t14,t15,t16,a[j1+p13],a[j1+1+p13],a[j1+p12],a[j1+1+p12],a[j1+p15],a[j1+1+p15],a[j1+p14],a[j1+1+p14],a[j1+p10],a[j1+1+p10],a[j1+p11],a[j1+1+p11],a[j1+p09],a[j1+1+p09],a[j1+p08],a[j1+1+p08],rt,it,p19,p20,p21,p22,p23);
#else
	RADIX_08_DIF       (aj1p00r,aj1p00i,aj1p05r,aj1p05i,aj1p02r,aj1p02i,aj1p07r,aj1p07i,aj1p04r,aj1p04i,aj1p01r,aj1p01i,aj1p06r,aj1p06i,aj1p03r,aj1p03i,t1,t2,t3,t4,t5,t6,t7,t8,t9,t10,t11,t12,t13,t14,t15,t16,a[j1    ],a[j1+1    ],a[j1+p01],a[j1+1+p01],a[j1+p02],a[j1+1+p02],a[j1+p03],a[j1+1+p03],a[j1+p05],a[j1+1+p05],a[j1+p04],a[j1+1+p04],a[j1+p07],a[j1+1+p07],a[j1+p06],a[j1+1+p06],rt,it);
	RADIX_08_DIF       (aj1p08r,aj1p08i,aj1p13r,aj1p13i,aj1p10r,aj1p10i,aj1p15r,aj1p15i,aj1p12r,aj1p12i,aj1p09r,aj1p09i,aj1p14r,aj1p14i,aj1p11r,aj1p11i,t1,t2,t3,t4,t5,t6,t7,t8,t9,t10,t11,t12,t13,t14,t15,t16,a[j1+p18],a[j1+1+p18],a[j1+p19],a[j1+1+p19],a[j1+p17],a[j1+1+p17],a[j1+p16],a[j1+1+p16],a[j1+p23],a[j1+1+p23],a[j1+p22],a[j1+1+p22],a[j1+p20],a[j1+1+p20],a[j1+p21],a[j1+1+p21],rt,it);
	RADIX_08_DIF       (aj1p16r,aj1p16i,aj1p21r,aj1p21i,aj1p18r,aj1p18i,aj1p23r,aj1p23i,aj1p20r,aj1p20i,aj1p17r,aj1p17i,aj1p22r,aj1p22i,aj1p19r,aj1p19i,t1,t2,t3,t4,t5,t6,t7,t8,t9,t10,t11,t12,t13,t14,t15,t16,a[j1+p13],a[j1+1+p13],a[j1+p12],a[j1+1+p12],a[j1+p15],a[j1+1+p15],a[j1+p14],a[j1+1+p14],a[j1+p10],a[j1+1+p10],a[j1+p11],a[j1+1+p11],a[j1+p09],a[j1+1+p09],a[j1+p08],a[j1+1+p08],rt,it);
#endif
	  iroot += root_incr;		/* increment sincos index.	*/

	  }

	  jstart += nwt;
	  jhi    += nwt;
	  col += 24;
	  co3 -= 24;

	}

	if(root_incr==0)break;

/*   Wraparound carry cleanup loop is here: ***
!
!   (1) Invert the radix-24 forward DIF FFT of the first block of 24 complex elements in A and unweight;
!   (2) Propagate cleanup carries among the real and imaginary parts of the 24 outputs of (1);
!   (3) Reweight and perform a radix-24 forward DIF FFT on the result of (2);
!   (4) If any of the exit carries from (2) are nonzero, advance to the next 24 elements and repeat (1-4).
*/
	t1  = cy23;
	cy23= cy22;
	cy22= cy21;
	cy21= cy20;
	cy20= cy19;
	cy19= cy18;
	cy18= cy17;
	cy17= cy16;
	cy16= cy15;
	cy15= cy14;
	cy14= cy13;
	cy13= cy12;
	cy12= cy11;
	cy11= cy10;
	cy10= cy09;
	cy09= cy08;
	cy08= cy07;
	cy07= cy06;
	cy06= cy05;
	cy05= cy04;
	cy04= cy03;
	cy03= cy02;
	cy02= cy01;
	cy01= cy00;
	cy00= t1;

	iroot = 0;
	root_incr = 0;
	scale = 1;

	jstart = 0;
	jhi = 3;
	khi = 1;

	for(j=0; j<=jhi; j++)
	{
	  a[j    ] *= radix_inv;
	  a[j+p01] *= radix_inv;
	  a[j+p02] *= radix_inv;
	  a[j+p03] *= radix_inv;
	  a[j+p04] *= radix_inv;
	  a[j+p05] *= radix_inv;
	  a[j+p06] *= radix_inv;
	  a[j+p07] *= radix_inv;
	  a[j+p08] *= radix_inv;
	  a[j+p09] *= radix_inv;
	  a[j+p10] *= radix_inv;
	  a[j+p11] *= radix_inv;
	  a[j+p12] *= radix_inv;
	  a[j+p13] *= radix_inv;
	  a[j+p14] *= radix_inv;
	  a[j+p15] *= radix_inv;
	  a[j+p16] *= radix_inv;
	  a[j+p17] *= radix_inv;
	  a[j+p18] *= radix_inv;
	  a[j+p19] *= radix_inv;
	  a[j+p20] *= radix_inv;
	  a[j+p21] *= radix_inv;
	  a[j+p22] *= radix_inv;
	  a[j+p23] *= radix_inv;
	}
}

	if(fabs(cy00)+fabs(cy01)+fabs(cy02)+fabs(cy03)+fabs(cy04)+fabs(cy05)+fabs(cy06)+fabs(cy07)+fabs(cy08)+fabs(cy09)+fabs(cy10)+fabs(cy11)+fabs(cy12)+fabs(cy13)+fabs(cy14)+fabs(cy15)+fabs(cy16)+fabs(cy17)+fabs(cy18)+fabs(cy19)+fabs(cy20)+fabs(cy21)+fabs(cy22)+fabs(cy23) != 0.0)
	{
	    if(INTERACT)printf("FATAL: iter = %10d; nonzero exit carry in radix24_ditN_cy_dif1.\n",iter);
	    f1 = fopen(   OFILE,"a");
	    f2 = fopen(STATFILE,"a");
	    fprintf(f1,"FATAL: iter = %10d; nonzero exit carry in radix24_ditN_cy_dif1.\n",iter);
	    fprintf(f2,"FATAL: iter = %10d; nonzero exit carry in radix24_ditN_cy_dif1.\n",iter);
	    fclose(f1);
	    fclose(f2);
	    err=1;
	    return(err);
	}
	*fracmax = maxerr;
	return(0);
}

/***************/

int radix24_ditN_cy_dif1_nochk(double a[], int n, int nwt, int nwt_bits, double wt0[], double wt1[], int si[], double base[], double baseinv[], int iter                 , int p)
{
/*
!...Acronym: DWT = Discrete Weighted Transform, DIT = Decimation In Time, DIF = Decimation In Frequency
!
!...Performs a final radix-24 complex DIT pass, an inverse DWT weighting, a carry propagation,
!   a forward DWT weighting, and an initial radix-24 complex DIF pass on the data in the length-N real vector A.
!
!   Data enter and are returned in the A-array.
!
!   See the documentation in mers_mod_square and radix16_dif_pass for further details on the array
!   storage scheme, and radix8_ditN_cy_dif1 for details on the reduced-length weights array scheme.
*/
	int n24, bjmodn00,bjmodn01,bjmodn02,bjmodn03,bjmodn04,bjmodn05,bjmodn06,bjmodn07,bjmodn08,bjmodn09,bjmodn10,bjmodn11,bjmodn12,bjmodn13,bjmodn14,bjmodn15,bjmodn16,bjmodn17,bjmodn18,bjmodn19,bjmodn20,bjmodn21,bjmodn22,bjmodn23
		,i,j,j1,jstart,jhi,iroot,root_incr,k,khi,l,outer;
	static int psave=0,bw,sw,bjmodnini,nrtm1,p01,p02,p03,p04,p05,p06,p07,p08,p09,p10,p11,p12,p13,p14,p15,p16,p17,p18,p19,p20,p21,p22,p23;
	const  double one_half[3] = {1.0, 0.5, 0.25};	/* Needed for small-weights-tables scheme */
	static double radix_inv, n2inv;
	static double c = .86602540378443864676, c3m1 = -1.5;
	double rt,it
	,t1,t2,t3,t4,t5,t6,t7,t8,t9,t10,t11,t12,t13,t14,t15,t16
	,aj1p00r,aj1p01r,aj1p02r,aj1p03r,aj1p04r,aj1p05r,aj1p06r,aj1p07r,aj1p08r,aj1p09r,aj1p10r,aj1p11r,aj1p12r,aj1p13r,aj1p14r,aj1p15r,aj1p16r,aj1p17r,aj1p18r,aj1p19r,aj1p20r,aj1p21r,aj1p22r,aj1p23r
	,aj1p00i,aj1p01i,aj1p02i,aj1p03i,aj1p04i,aj1p05i,aj1p06i,aj1p07i,aj1p08i,aj1p09i,aj1p10i,aj1p11i,aj1p12i,aj1p13i,aj1p14i,aj1p15i,aj1p16i,aj1p17i,aj1p18i,aj1p19i,aj1p20i,aj1p21i,aj1p22i,aj1p23i
	,cy00,cy01,cy02,cy03,cy04,cy05,cy06,cy07,cy08,cy09,cy10,cy11,cy12,cy13,cy14,cy15,cy16,cy17,cy18,cy19,cy20,cy21,cy22,cy23,temp,scale;
#if PFETCH
	double *add0, *add;
#endif
	int err;
	FILE *f1,*f2;
	static int first_entry=TRUE;

/*...stuff for the reduced-length DWT weights array is here:	*/
	int n_div_nwt;
	int col,co2,co3,m,m2,n_minus_sil,n_minus_silp1,sinwt,sinwtm1;
	double wt,wtinv,wtl,wtlp1,wtn,wtnm1,wtA,wtB,wtC;

/*...change n24 and n_div_wt to non-static to work around a gcc compiler bug. */
	n24   = n/24;
	n_div_nwt = n24 >> nwt_bits;

	if((n_div_nwt << nwt_bits) != n24)
	{
	  if(INTERACT)printf("FATAL: iter = %10d; NWT_BITS does not divide N/24 in radix24_ditN_cy_dif1.\n",iter);
	  f1 = fopen(   OFILE,"a");
	  f2 = fopen(STATFILE,"a");
	  fprintf(f1,"FATAL: iter = %10d; NWT_BITS does not divide N/24 in radix24_ditN_cy_dif1.\n",iter);
	  fprintf(f2,"FATAL: iter = %10d; NWT_BITS does not divide N/24 in radix24_ditN_cy_dif1.\n",iter);
	  fclose(f1);
	  fclose(f2);
	  err=1;
	  return(err);
	}

	if(p != psave)
	{
	  first_entry=TRUE;
	}

/*...initialize things upon first entry	*/

	if(first_entry)
	{
	  psave = p;
	  first_entry=FALSE;
	  radix_inv = (double)((long double)1.0L/(long double)24.0L);
	  n2inv     = (double)((long double)1.0L/(long double)(n/2));

	  bw    = p%n;		/* Number of bigwords in the Crandall/Fagin mixed-radix representation = (Mersenne exponent) mod (vector length).	*/
	  sw    = n - bw;	/* Number of smallwords.	*/

/*   constant index offsets for array load/stores are here.	*/

	  p01 = n24;
	  p02 = p01 + p01;
	  p03 = p02 + p01;
	  p04 = p03 + p01;
	  p05 = p04 + p01;
	  p06 = p05 + p01;
	  p07 = p06 + p01;
	  p08 = p07 + p01;
	  p09 = p08 + p01;
	  p10 = p09 + p01;
	  p11 = p10 + p01;
	  p12 = p11 + p01;
	  p13 = p12 + p01;
	  p14 = p13 + p01;
	  p15 = p14 + p01;
	  p16 = p15 + p01;
	  p17 = p16 + p01;
	  p18 = p17 + p01;
	  p19 = p18 + p01;
	  p20 = p19 + p01;
	  p21 = p20 + p01;
	  p22 = p21 + p01;
	  p23 = p22 + p01;

	  p01 = p01 + ( (p01 >> DAT_BITS) << PAD_BITS );
	  p02 = p02 + ( (p02 >> DAT_BITS) << PAD_BITS );
	  p03 = p03 + ( (p03 >> DAT_BITS) << PAD_BITS );
	  p04 = p04 + ( (p04 >> DAT_BITS) << PAD_BITS );
	  p05 = p05 + ( (p05 >> DAT_BITS) << PAD_BITS );
	  p06 = p06 + ( (p06 >> DAT_BITS) << PAD_BITS );
	  p07 = p07 + ( (p07 >> DAT_BITS) << PAD_BITS );
	  p08 = p08 + ( (p08 >> DAT_BITS) << PAD_BITS );
	  p09 = p09 + ( (p09 >> DAT_BITS) << PAD_BITS );
	  p10 = p10 + ( (p10 >> DAT_BITS) << PAD_BITS );
	  p11 = p11 + ( (p11 >> DAT_BITS) << PAD_BITS );
	  p12 = p12 + ( (p12 >> DAT_BITS) << PAD_BITS );
	  p13 = p13 + ( (p13 >> DAT_BITS) << PAD_BITS );
	  p14 = p14 + ( (p14 >> DAT_BITS) << PAD_BITS );
	  p15 = p15 + ( (p15 >> DAT_BITS) << PAD_BITS );
	  p16 = p16 + ( (p16 >> DAT_BITS) << PAD_BITS );
	  p17 = p17 + ( (p17 >> DAT_BITS) << PAD_BITS );
	  p18 = p18 + ( (p18 >> DAT_BITS) << PAD_BITS );
	  p19 = p19 + ( (p19 >> DAT_BITS) << PAD_BITS );
	  p20 = p20 + ( (p20 >> DAT_BITS) << PAD_BITS );
	  p21 = p21 + ( (p21 >> DAT_BITS) << PAD_BITS );
	  p22 = p22 + ( (p22 >> DAT_BITS) << PAD_BITS );
	  p23 = p23 + ( (p23 >> DAT_BITS) << PAD_BITS );

	  bjmodnini=0;
	  for(j=0; j < n24; j++)
	  {
	    bjmodnini -= sw; bjmodnini = bjmodnini + ( (-(int)((uint32)bjmodnini >> 31)) & n);
	  }
	}

/*...The radix-24 final DIT pass is here.	*/

	cy00=-2;		/* init carries	*/
	cy01= 0;
	cy02= 0;
	cy03= 0;
	cy04= 0;
	cy05= 0;
	cy06= 0;
	cy07= 0;
	cy08= 0;
	cy09= 0;
	cy10= 0;
	cy11= 0;
	cy12= 0;
	cy13= 0;
	cy14= 0;
	cy15= 0;
	cy16= 0;
	cy17= 0;
	cy18= 0;
	cy19= 0;
	cy20= 0;
	cy21= 0;
	cy22= 0;
	cy23= 0;

	iroot = 0;	/* init sincos array index	*/
	root_incr = 1;	/* init sincos array index increment (set = 1 for normal carry pass, = 0 for wrapper pass)	*/
	scale = n2inv;	/* init inverse-weight scale factor  (set = 2/n for normal carry pass, = 1 for wrapper pass)	*/

	jstart = 0;
	jhi = jstart+nwt-1;
	khi = n_div_nwt;

for(outer=0; outer <= 1; outer++)
{
	i = 1;		/* Pointer to the BASE and BASEINV arrays. lowest-order digit is always a bigword (i = 1).	*/

	bjmodn00= 0;
	bjmodn01= bjmodnini;
	bjmodn02= bjmodn01+bjmodnini-n; bjmodn02= bjmodn02+ ( (-(int)((uint32)bjmodn02>> 31)) & n);
	bjmodn03= bjmodn02+bjmodnini-n; bjmodn03= bjmodn03+ ( (-(int)((uint32)bjmodn03>> 31)) & n);
	bjmodn04= bjmodn03+bjmodnini-n; bjmodn04= bjmodn04+ ( (-(int)((uint32)bjmodn04>> 31)) & n);
	bjmodn05= bjmodn04+bjmodnini-n; bjmodn05= bjmodn05+ ( (-(int)((uint32)bjmodn05>> 31)) & n);
	bjmodn06= bjmodn05+bjmodnini-n; bjmodn06= bjmodn06+ ( (-(int)((uint32)bjmodn06>> 31)) & n);
	bjmodn07= bjmodn06+bjmodnini-n; bjmodn07= bjmodn07+ ( (-(int)((uint32)bjmodn07>> 31)) & n);
	bjmodn08= bjmodn07+bjmodnini-n; bjmodn08= bjmodn08+ ( (-(int)((uint32)bjmodn08>> 31)) & n);
	bjmodn09= bjmodn08+bjmodnini-n; bjmodn09= bjmodn09+ ( (-(int)((uint32)bjmodn09>> 31)) & n);
	bjmodn10= bjmodn09+bjmodnini-n; bjmodn10= bjmodn10+ ( (-(int)((uint32)bjmodn10>> 31)) & n);
	bjmodn11= bjmodn10+bjmodnini-n; bjmodn11= bjmodn11+ ( (-(int)((uint32)bjmodn11>> 31)) & n);
	bjmodn12= bjmodn11+bjmodnini-n; bjmodn12= bjmodn12+ ( (-(int)((uint32)bjmodn12>> 31)) & n);
	bjmodn13= bjmodn12+bjmodnini-n; bjmodn13= bjmodn13+ ( (-(int)((uint32)bjmodn13>> 31)) & n);
	bjmodn14= bjmodn13+bjmodnini-n; bjmodn14= bjmodn14+ ( (-(int)((uint32)bjmodn14>> 31)) & n);
	bjmodn15= bjmodn14+bjmodnini-n; bjmodn15= bjmodn15+ ( (-(int)((uint32)bjmodn15>> 31)) & n);
	bjmodn16= bjmodn15+bjmodnini-n; bjmodn16= bjmodn16+ ( (-(int)((uint32)bjmodn16>> 31)) & n);
	bjmodn17= bjmodn16+bjmodnini-n; bjmodn17= bjmodn17+ ( (-(int)((uint32)bjmodn17>> 31)) & n);
	bjmodn18= bjmodn17+bjmodnini-n; bjmodn18= bjmodn18+ ( (-(int)((uint32)bjmodn18>> 31)) & n);
	bjmodn19= bjmodn18+bjmodnini-n; bjmodn19= bjmodn19+ ( (-(int)((uint32)bjmodn19>> 31)) & n);
	bjmodn20= bjmodn19+bjmodnini-n; bjmodn20= bjmodn20+ ( (-(int)((uint32)bjmodn20>> 31)) & n);
	bjmodn21= bjmodn20+bjmodnini-n; bjmodn21= bjmodn21+ ( (-(int)((uint32)bjmodn21>> 31)) & n);
	bjmodn22= bjmodn21+bjmodnini-n; bjmodn22= bjmodn22+ ( (-(int)((uint32)bjmodn22>> 31)) & n);
	bjmodn23= bjmodn22+bjmodnini-n; bjmodn23= bjmodn23+ ( (-(int)((uint32)bjmodn23>> 31)) & n);

	col=0;
	co2=(n >> nwt_bits)-1+24;
	co3=co2-24;		/* At the start of each new j-loop, co3=co2-radix(1)	*/

	for(k=1; k <= khi; k++)	/* Do n/(radix(1)*nwt) outer loop executions...	*/
	{
	  for(j=jstart; j<jhi; j += 2)	/* Each inner loop execution processes (radix(1)*nwt) array data.	*/
	  {
	    j1 = j + ( (j >> DAT_BITS) << PAD_BITS );	/* padded-array fetch index is here	*/

/*...The radix-24 DIT pass is here:	*/
/*...gather the needed data (24 64-bit complex, i.e. 48 64-bit reals) and do 3 radix-8 transforms,	*/
                     /*                                                                                 inputs                                                                                    */ /*                 intermediates                    */ /*                                                         outputs                                                           */
	RADIX_08_DIT(a[j1    ],a[j1+1    ],a[j1+p01],a[j1+1+p01],a[j1+p03],a[j1+1+p03],a[j1+p02],a[j1+1+p02],a[j1+p07],a[j1+1+p07],a[j1+p06],a[j1+1+p06],a[j1+p05],a[j1+1+p05],a[j1+p04],a[j1+1+p04],t1,t2,t3,t4,t5,t6,t7,t8,t9,t10,t11,t12,t13,t14,t15,t16,aj1p00r,aj1p00i,aj1p01r,aj1p01i,aj1p02r,aj1p02i,aj1p03r,aj1p03i,aj1p04r,aj1p04i,aj1p05r,aj1p05i,aj1p06r,aj1p06i,aj1p07r,aj1p07i,rt,it);
	RADIX_08_DIT(a[j1+p13],a[j1+1+p13],a[j1+p12],a[j1+1+p12],a[j1+p14],a[j1+1+p14],a[j1+p15],a[j1+1+p15],a[j1+p09],a[j1+1+p09],a[j1+p08],a[j1+1+p08],a[j1+p10],a[j1+1+p10],a[j1+p11],a[j1+1+p11],t1,t2,t3,t4,t5,t6,t7,t8,t9,t10,t11,t12,t13,t14,t15,t16,aj1p08r,aj1p08i,aj1p09r,aj1p09i,aj1p10r,aj1p10i,aj1p11r,aj1p11i,aj1p12r,aj1p12i,aj1p13r,aj1p13i,aj1p14r,aj1p14i,aj1p15r,aj1p15i,rt,it);
	RADIX_08_DIT(a[j1+p18],a[j1+1+p18],a[j1+p19],a[j1+1+p19],a[j1+p16],a[j1+1+p16],a[j1+p17],a[j1+1+p17],a[j1+p20],a[j1+1+p20],a[j1+p21],a[j1+1+p21],a[j1+p23],a[j1+1+p23],a[j1+p22],a[j1+1+p22],t1,t2,t3,t4,t5,t6,t7,t8,t9,t10,t11,t12,t13,t14,t15,t16,aj1p16r,aj1p16i,aj1p17r,aj1p17i,aj1p18r,aj1p18i,aj1p19r,aj1p19i,aj1p20r,aj1p20i,aj1p21r,aj1p21i,aj1p22r,aj1p22i,aj1p23r,aj1p23i,rt,it);

/*...and now do 8 in-place radix-3 transforms.	*/
                     /*                  inputs                   */ /*             intermediates                 */ /*                       outputs                               */
	RADIX_03_DFT(aj1p00r,aj1p00i,aj1p08r,aj1p08i,aj1p16r,aj1p16i,t1,t2,t3,t4,t5,t6,aj1p00r,aj1p00i,aj1p16r,aj1p16i,aj1p08r,aj1p08i,rt,it);
	RADIX_03_DFT(aj1p01r,aj1p01i,aj1p09r,aj1p09i,aj1p17r,aj1p17i,t1,t2,t3,t4,t5,t6,aj1p09r,aj1p09i,aj1p01r,aj1p01i,aj1p17r,aj1p17i,rt,it);
	RADIX_03_DFT(aj1p02r,aj1p02i,aj1p10r,aj1p10i,aj1p18r,aj1p18i,t1,t2,t3,t4,t5,t6,aj1p18r,aj1p18i,aj1p10r,aj1p10i,aj1p02r,aj1p02i,rt,it);
	RADIX_03_DFT(aj1p03r,aj1p03i,aj1p11r,aj1p11i,aj1p19r,aj1p19i,t1,t2,t3,t4,t5,t6,aj1p03r,aj1p03i,aj1p19r,aj1p19i,aj1p11r,aj1p11i,rt,it);
	RADIX_03_DFT(aj1p04r,aj1p04i,aj1p12r,aj1p12i,aj1p20r,aj1p20i,t1,t2,t3,t4,t5,t6,aj1p12r,aj1p12i,aj1p04r,aj1p04i,aj1p20r,aj1p20i,rt,it);
	RADIX_03_DFT(aj1p05r,aj1p05i,aj1p13r,aj1p13i,aj1p21r,aj1p21i,t1,t2,t3,t4,t5,t6,aj1p21r,aj1p21i,aj1p13r,aj1p13i,aj1p05r,aj1p05i,rt,it);
	RADIX_03_DFT(aj1p06r,aj1p06i,aj1p14r,aj1p14i,aj1p22r,aj1p22i,t1,t2,t3,t4,t5,t6,aj1p06r,aj1p06i,aj1p22r,aj1p22i,aj1p14r,aj1p14i,rt,it);
	RADIX_03_DFT(aj1p07r,aj1p07i,aj1p15r,aj1p15i,aj1p23r,aj1p23i,t1,t2,t3,t4,t5,t6,aj1p15r,aj1p15i,aj1p07r,aj1p07i,aj1p23r,aj1p23i,rt,it);

/*...Now do the carries. Since the outputs would
    normally be getting dispatched to 24 separate blocks of the A-array, we need 24 separate carries.	*/

	    l= j & (nwt-1);
	    n_minus_sil   = n-si[l  ];
	    n_minus_silp1 = n-si[l+1];
	    sinwt   = si[nwt-l  ];
	    sinwtm1 = si[nwt-l-1];

	    wtl     =wt0[    l  ];
	    wtn     =wt0[nwt-l  ]*scale;	/* Include 1/(n/2) scale factor of inverse transform here...	*/
	    wtlp1   =wt0[    l+1];
	    wtnm1   =wt0[nwt-l-1]*scale;	/* ...and here.	*/

/*...set0 is slightly different from others:	*/
	   cmplx_carry_norm_nocheck0(aj1p00r,aj1p00i,cy00,bjmodn00   );
	    cmplx_carry_norm_nocheck(aj1p01r,aj1p01i,cy01,bjmodn01,1 );
	    cmplx_carry_norm_nocheck(aj1p02r,aj1p02i,cy02,bjmodn02,2 );
	    cmplx_carry_norm_nocheck(aj1p03r,aj1p03i,cy03,bjmodn03,3 );
	    cmplx_carry_norm_nocheck(aj1p04r,aj1p04i,cy04,bjmodn04,4 );
	    cmplx_carry_norm_nocheck(aj1p05r,aj1p05i,cy05,bjmodn05,5 );
	    cmplx_carry_norm_nocheck(aj1p06r,aj1p06i,cy06,bjmodn06,6 );
	    cmplx_carry_norm_nocheck(aj1p07r,aj1p07i,cy07,bjmodn07,7 );
	    cmplx_carry_norm_nocheck(aj1p08r,aj1p08i,cy08,bjmodn08,8 );
	    cmplx_carry_norm_nocheck(aj1p09r,aj1p09i,cy09,bjmodn09,9 );
	    cmplx_carry_norm_nocheck(aj1p10r,aj1p10i,cy10,bjmodn10,10);
	    cmplx_carry_norm_nocheck(aj1p11r,aj1p11i,cy11,bjmodn11,11);
	    cmplx_carry_norm_nocheck(aj1p12r,aj1p12i,cy12,bjmodn12,12);
	    cmplx_carry_norm_nocheck(aj1p13r,aj1p13i,cy13,bjmodn13,13);
	    cmplx_carry_norm_nocheck(aj1p14r,aj1p14i,cy14,bjmodn14,14);
	    cmplx_carry_norm_nocheck(aj1p15r,aj1p15i,cy15,bjmodn15,15);
	    cmplx_carry_norm_nocheck(aj1p16r,aj1p16i,cy16,bjmodn16,16);
	    cmplx_carry_norm_nocheck(aj1p17r,aj1p17i,cy17,bjmodn17,17);
	    cmplx_carry_norm_nocheck(aj1p18r,aj1p18i,cy18,bjmodn18,18);
	    cmplx_carry_norm_nocheck(aj1p19r,aj1p19i,cy19,bjmodn19,19);
	    cmplx_carry_norm_nocheck(aj1p20r,aj1p20i,cy20,bjmodn20,20);
	    cmplx_carry_norm_nocheck(aj1p21r,aj1p21i,cy21,bjmodn21,21);
	    cmplx_carry_norm_nocheck(aj1p22r,aj1p22i,cy22,bjmodn22,22);
	    cmplx_carry_norm_nocheck(aj1p23r,aj1p23i,cy23,bjmodn23,23);

	    i =((uint32)(sw - bjmodn00) >> 31);	/* get ready for the next set...	*/
	    co2=co3;	/* For all data but the first set in each j-block, co2=co3. Thus, after the first block of data is done
			   and only then: for all subsequent blocks it's superfluous), this assignment decrements co2 by radix(1).	*/

#if PFETCH
add0 = &a[j1];
prefetch_p_doubles(add0);
#endif

/*...The radix-24 DIF pass is here:	*/

/* EWM: 10/21/04: We swap the following outputs of the radix-3 transforms: {1,9,17}<=>{5,13,21}, {3,11,19}<=>{7,15,23}, so that the indexing
                  winds up being in-place. This allows us to properly re-use the ajp1 variables in the carry-pass version of this routine.
*/

/*...gather the needed data (24 64-bit complex, i.e. 48 64-bit reals) and do 8 in-place radix-3 transforms...*/
                     /*                        inputs                               */ /*             intermediates                 */ /*                 outputs                   */
#if PFETCH
	RADIX_03_DFT_PFETCH(aj1p00r,aj1p00i,aj1p16r,aj1p16i,aj1p08r,aj1p08i,t1,t2,t3,t4,t5,t6,aj1p00r,aj1p00i,aj1p08r,aj1p08i,aj1p16r,aj1p16i,rt,it,p01);
	RADIX_03_DFT_PFETCH(aj1p21r,aj1p21i,aj1p13r,aj1p13i,aj1p05r,aj1p05i,t1,t2,t3,t4,t5,t6,aj1p05r,aj1p05i,aj1p13r,aj1p13i,aj1p21r,aj1p21i,rt,it,p02);
	RADIX_03_DFT_PFETCH(aj1p18r,aj1p18i,aj1p10r,aj1p10i,aj1p02r,aj1p02i,t1,t2,t3,t4,t5,t6,aj1p02r,aj1p02i,aj1p10r,aj1p10i,aj1p18r,aj1p18i,rt,it,p03);
	RADIX_03_DFT_PFETCH(aj1p15r,aj1p15i,aj1p07r,aj1p07i,aj1p23r,aj1p23i,t1,t2,t3,t4,t5,t6,aj1p07r,aj1p07i,aj1p15r,aj1p15i,aj1p23r,aj1p23i,rt,it,p04);
	RADIX_03_DFT_PFETCH(aj1p12r,aj1p12i,aj1p04r,aj1p04i,aj1p20r,aj1p20i,t1,t2,t3,t4,t5,t6,aj1p04r,aj1p04i,aj1p12r,aj1p12i,aj1p20r,aj1p20i,rt,it,p05);
	RADIX_03_DFT_PFETCH(aj1p09r,aj1p09i,aj1p01r,aj1p01i,aj1p17r,aj1p17i,t1,t2,t3,t4,t5,t6,aj1p01r,aj1p01i,aj1p09r,aj1p09i,aj1p17r,aj1p17i,rt,it,p06);
	RADIX_03_DFT_PFETCH(aj1p06r,aj1p06i,aj1p22r,aj1p22i,aj1p14r,aj1p14i,t1,t2,t3,t4,t5,t6,aj1p06r,aj1p06i,aj1p14r,aj1p14i,aj1p22r,aj1p22i,rt,it,p07);
	RADIX_03_DFT_PFETCH(aj1p03r,aj1p03i,aj1p19r,aj1p19i,aj1p11r,aj1p11i,t1,t2,t3,t4,t5,t6,aj1p03r,aj1p03i,aj1p11r,aj1p11i,aj1p19r,aj1p19i,rt,it,p08);
#else
	RADIX_03_DFT       (aj1p00r,aj1p00i,aj1p16r,aj1p16i,aj1p08r,aj1p08i,t1,t2,t3,t4,t5,t6,aj1p00r,aj1p00i,aj1p08r,aj1p08i,aj1p16r,aj1p16i,rt,it);
	RADIX_03_DFT       (aj1p21r,aj1p21i,aj1p13r,aj1p13i,aj1p05r,aj1p05i,t1,t2,t3,t4,t5,t6,aj1p05r,aj1p05i,aj1p13r,aj1p13i,aj1p21r,aj1p21i,rt,it);
	RADIX_03_DFT       (aj1p18r,aj1p18i,aj1p10r,aj1p10i,aj1p02r,aj1p02i,t1,t2,t3,t4,t5,t6,aj1p02r,aj1p02i,aj1p10r,aj1p10i,aj1p18r,aj1p18i,rt,it);
	RADIX_03_DFT       (aj1p15r,aj1p15i,aj1p07r,aj1p07i,aj1p23r,aj1p23i,t1,t2,t3,t4,t5,t6,aj1p07r,aj1p07i,aj1p15r,aj1p15i,aj1p23r,aj1p23i,rt,it);
	RADIX_03_DFT       (aj1p12r,aj1p12i,aj1p04r,aj1p04i,aj1p20r,aj1p20i,t1,t2,t3,t4,t5,t6,aj1p04r,aj1p04i,aj1p12r,aj1p12i,aj1p20r,aj1p20i,rt,it);
	RADIX_03_DFT       (aj1p09r,aj1p09i,aj1p01r,aj1p01i,aj1p17r,aj1p17i,t1,t2,t3,t4,t5,t6,aj1p01r,aj1p01i,aj1p09r,aj1p09i,aj1p17r,aj1p17i,rt,it);
	RADIX_03_DFT       (aj1p06r,aj1p06i,aj1p22r,aj1p22i,aj1p14r,aj1p14i,t1,t2,t3,t4,t5,t6,aj1p06r,aj1p06i,aj1p14r,aj1p14i,aj1p22r,aj1p22i,rt,it);
	RADIX_03_DFT       (aj1p03r,aj1p03i,aj1p19r,aj1p19i,aj1p11r,aj1p11i,t1,t2,t3,t4,t5,t6,aj1p03r,aj1p03i,aj1p11r,aj1p11i,aj1p19r,aj1p19i,rt,it);
#endif

/*...and now do 3 radix-8 transforms:	*/
                     /*                                                          inputs                                                           */ /*                 intermediates                    */ /*                 outputs                   */
#if PFETCH
	RADIX_08_DIF_PFETCH(aj1p00r,aj1p00i,aj1p05r,aj1p05i,aj1p02r,aj1p02i,aj1p07r,aj1p07i,aj1p04r,aj1p04i,aj1p01r,aj1p01i,aj1p06r,aj1p06i,aj1p03r,aj1p03i,t1,t2,t3,t4,t5,t6,t7,t8,t9,t10,t11,t12,t13,t14,t15,t16,a[j1    ],a[j1+1    ],a[j1+p01],a[j1+1+p01],a[j1+p02],a[j1+1+p02],a[j1+p03],a[j1+1+p03],a[j1+p05],a[j1+1+p05],a[j1+p04],a[j1+1+p04],a[j1+p07],a[j1+1+p07],a[j1+p06],a[j1+1+p06],rt,it,p09,p10,p11,p12,p13);
	RADIX_08_DIF_PFETCH(aj1p08r,aj1p08i,aj1p13r,aj1p13i,aj1p10r,aj1p10i,aj1p15r,aj1p15i,aj1p12r,aj1p12i,aj1p09r,aj1p09i,aj1p14r,aj1p14i,aj1p11r,aj1p11i,t1,t2,t3,t4,t5,t6,t7,t8,t9,t10,t11,t12,t13,t14,t15,t16,a[j1+p18],a[j1+1+p18],a[j1+p19],a[j1+1+p19],a[j1+p17],a[j1+1+p17],a[j1+p16],a[j1+1+p16],a[j1+p23],a[j1+1+p23],a[j1+p22],a[j1+1+p22],a[j1+p20],a[j1+1+p20],a[j1+p21],a[j1+1+p21],rt,it,p14,p15,p16,p17,p18);
	RADIX_08_DIF_PFETCH(aj1p16r,aj1p16i,aj1p21r,aj1p21i,aj1p18r,aj1p18i,aj1p23r,aj1p23i,aj1p20r,aj1p20i,aj1p17r,aj1p17i,aj1p22r,aj1p22i,aj1p19r,aj1p19i,t1,t2,t3,t4,t5,t6,t7,t8,t9,t10,t11,t12,t13,t14,t15,t16,a[j1+p13],a[j1+1+p13],a[j1+p12],a[j1+1+p12],a[j1+p15],a[j1+1+p15],a[j1+p14],a[j1+1+p14],a[j1+p10],a[j1+1+p10],a[j1+p11],a[j1+1+p11],a[j1+p09],a[j1+1+p09],a[j1+p08],a[j1+1+p08],rt,it,p19,p20,p21,p22,p23);
#else
	RADIX_08_DIF       (aj1p00r,aj1p00i,aj1p05r,aj1p05i,aj1p02r,aj1p02i,aj1p07r,aj1p07i,aj1p04r,aj1p04i,aj1p01r,aj1p01i,aj1p06r,aj1p06i,aj1p03r,aj1p03i,t1,t2,t3,t4,t5,t6,t7,t8,t9,t10,t11,t12,t13,t14,t15,t16,a[j1    ],a[j1+1    ],a[j1+p01],a[j1+1+p01],a[j1+p02],a[j1+1+p02],a[j1+p03],a[j1+1+p03],a[j1+p05],a[j1+1+p05],a[j1+p04],a[j1+1+p04],a[j1+p07],a[j1+1+p07],a[j1+p06],a[j1+1+p06],rt,it);
	RADIX_08_DIF       (aj1p08r,aj1p08i,aj1p13r,aj1p13i,aj1p10r,aj1p10i,aj1p15r,aj1p15i,aj1p12r,aj1p12i,aj1p09r,aj1p09i,aj1p14r,aj1p14i,aj1p11r,aj1p11i,t1,t2,t3,t4,t5,t6,t7,t8,t9,t10,t11,t12,t13,t14,t15,t16,a[j1+p18],a[j1+1+p18],a[j1+p19],a[j1+1+p19],a[j1+p17],a[j1+1+p17],a[j1+p16],a[j1+1+p16],a[j1+p23],a[j1+1+p23],a[j1+p22],a[j1+1+p22],a[j1+p20],a[j1+1+p20],a[j1+p21],a[j1+1+p21],rt,it);
	RADIX_08_DIF       (aj1p16r,aj1p16i,aj1p21r,aj1p21i,aj1p18r,aj1p18i,aj1p23r,aj1p23i,aj1p20r,aj1p20i,aj1p17r,aj1p17i,aj1p22r,aj1p22i,aj1p19r,aj1p19i,t1,t2,t3,t4,t5,t6,t7,t8,t9,t10,t11,t12,t13,t14,t15,t16,a[j1+p13],a[j1+1+p13],a[j1+p12],a[j1+1+p12],a[j1+p15],a[j1+1+p15],a[j1+p14],a[j1+1+p14],a[j1+p10],a[j1+1+p10],a[j1+p11],a[j1+1+p11],a[j1+p09],a[j1+1+p09],a[j1+p08],a[j1+1+p08],rt,it);
#endif
	  iroot += root_incr;		/* increment sincos index.	*/

	  }

	  jstart += nwt;
	  jhi    += nwt;
	  col += 24;
	  co3 -= 24;

	}

	if(root_incr==0)break;

/*   Wraparound carry cleanup loop is here: ***
!
!   (1) Invert the radix-24 forward DIF FFT of the first block of 24 complex elements in A and unweight;
!   (2) Propagate cleanup carries among the real and imaginary parts of the 24 outputs of (1);
!   (3) Reweight and perform a radix-24 forward DIF FFT on the result of (2);
!   (4) If any of the exit carries from (2) are nonzero, advance to the next 24 elements and repeat (1-4).
*/
	t1  = cy23;
	cy23= cy22;
	cy22= cy21;
	cy21= cy20;
	cy20= cy19;
	cy19= cy18;
	cy18= cy17;
	cy17= cy16;
	cy16= cy15;
	cy15= cy14;
	cy14= cy13;
	cy13= cy12;
	cy12= cy11;
	cy11= cy10;
	cy10= cy09;
	cy09= cy08;
	cy08= cy07;
	cy07= cy06;
	cy06= cy05;
	cy05= cy04;
	cy04= cy03;
	cy03= cy02;
	cy02= cy01;
	cy01= cy00;
	cy00= t1;

	iroot = 0;
	root_incr = 0;
	scale = 1;

	jstart = 0;
	jhi = 3;
	khi = 1;

	for(j=0; j<=jhi; j++)
	{
	  a[j    ] *= radix_inv;
	  a[j+p01] *= radix_inv;
	  a[j+p02] *= radix_inv;
	  a[j+p03] *= radix_inv;
	  a[j+p04] *= radix_inv;
	  a[j+p05] *= radix_inv;
	  a[j+p06] *= radix_inv;
	  a[j+p07] *= radix_inv;
	  a[j+p08] *= radix_inv;
	  a[j+p09] *= radix_inv;
	  a[j+p10] *= radix_inv;
	  a[j+p11] *= radix_inv;
	  a[j+p12] *= radix_inv;
	  a[j+p13] *= radix_inv;
	  a[j+p14] *= radix_inv;
	  a[j+p15] *= radix_inv;
	  a[j+p16] *= radix_inv;
	  a[j+p17] *= radix_inv;
	  a[j+p18] *= radix_inv;
	  a[j+p19] *= radix_inv;
	  a[j+p20] *= radix_inv;
	  a[j+p21] *= radix_inv;
	  a[j+p22] *= radix_inv;
	  a[j+p23] *= radix_inv;
	}
}

	if(fabs(cy00)+fabs(cy01)+fabs(cy02)+fabs(cy03)+fabs(cy04)+fabs(cy05)+fabs(cy06)+fabs(cy07)+fabs(cy08)+fabs(cy09)+fabs(cy10)+fabs(cy11)+fabs(cy12)+fabs(cy13)+fabs(cy14)+fabs(cy15)+fabs(cy16)+fabs(cy17)+fabs(cy18)+fabs(cy19)+fabs(cy20)+fabs(cy21)+fabs(cy22)+fabs(cy23) != 0.0)
	{
	    if(INTERACT)printf("FATAL: iter = %10d; nonzero exit carry in radix24_ditN_cy_dif1.\n",iter);
	    f1 = fopen(   OFILE,"a");
	    f2 = fopen(STATFILE,"a");
	    fprintf(f1,"FATAL: iter = %10d; nonzero exit carry in radix24_ditN_cy_dif1.\n",iter);
	    fprintf(f2,"FATAL: iter = %10d; nonzero exit carry in radix24_ditN_cy_dif1.\n",iter);
	    fclose(f1);
	    fclose(f2);
	    err=1;
	    return(err);
	}

	return(0);
}

/***************/

void radix24_dit_pass1(double a[], int n)
{
/*
!...Acronym: DIF = Decimation In Frequency
!
!...Subroutine to perform an initial radix-24 complex DIT FFT pass on the data in the length-N real vector A.
!
!   See the documentation in radix16_dif_pass for further details on storage and indexing.
!
!   See the documentation in radix7_dif_pass for details on the radix-15 subtransforms.
*/
	int j,j1;
	static int n24,p01,p02,p03,p04,p05,p06,p07,p08,p09,p10,p11,p12,p13,p14,p15,p16,p17,p18,p19,p20,p21,p22,p23, first_entry=TRUE;
	static double c = .86602540378443864676, c3m1 = -1.5;
	double rt,it
	,t1,t2,t3,t4,t5,t6,t7,t8,t9,t10,t11,t12,t13,t14,t15,t16
	,aj1p00r,aj1p01r,aj1p02r,aj1p03r,aj1p04r,aj1p05r,aj1p06r,aj1p07r,aj1p08r,aj1p09r,aj1p10r,aj1p11r,aj1p12r,aj1p13r,aj1p14r,aj1p15r,aj1p16r,aj1p17r,aj1p18r,aj1p19r,aj1p20r,aj1p21r,aj1p22r,aj1p23r
	,aj1p00i,aj1p01i,aj1p02i,aj1p03i,aj1p04i,aj1p05i,aj1p06i,aj1p07i,aj1p08i,aj1p09i,aj1p10i,aj1p11i,aj1p12i,aj1p13i,aj1p14i,aj1p15i,aj1p16i,aj1p17i,aj1p18i,aj1p19i,aj1p20i,aj1p21i,aj1p22i,aj1p23i;

	if(!first_entry && (n/24) != n24)	/* New runlength?	*/
	{
	  first_entry=TRUE;
	}

/*...initialize things upon first entry	*/

	if(first_entry)
	{
	  first_entry=FALSE;
	  n24=n/24;

/*   constant index offsets for array load/stores are here.	*/

	  p01 = n24;
	  p02 = p01 + p01;
	  p03 = p02 + p01;
	  p04 = p03 + p01;
	  p05 = p04 + p01;
	  p06 = p05 + p01;
	  p07 = p06 + p01;
	  p08 = p07 + p01;
	  p09 = p08 + p01;
	  p10 = p09 + p01;
	  p11 = p10 + p01;
	  p12 = p11 + p01;
	  p13 = p12 + p01;
	  p14 = p13 + p01;
	  p15 = p14 + p01;
	  p16 = p15 + p01;
	  p17 = p16 + p01;
	  p18 = p17 + p01;
	  p19 = p18 + p01;
	  p20 = p19 + p01;
	  p21 = p20 + p01;
	  p22 = p21 + p01;
	  p23 = p22 + p01;

	  p01 = p01 + ( (p01 >> DAT_BITS) << PAD_BITS );
	  p02 = p02 + ( (p02 >> DAT_BITS) << PAD_BITS );
	  p03 = p03 + ( (p03 >> DAT_BITS) << PAD_BITS );
	  p04 = p04 + ( (p04 >> DAT_BITS) << PAD_BITS );
	  p05 = p05 + ( (p05 >> DAT_BITS) << PAD_BITS );
	  p06 = p06 + ( (p06 >> DAT_BITS) << PAD_BITS );
	  p07 = p07 + ( (p07 >> DAT_BITS) << PAD_BITS );
	  p08 = p08 + ( (p08 >> DAT_BITS) << PAD_BITS );
	  p09 = p09 + ( (p09 >> DAT_BITS) << PAD_BITS );
	  p10 = p10 + ( (p10 >> DAT_BITS) << PAD_BITS );
	  p11 = p11 + ( (p11 >> DAT_BITS) << PAD_BITS );
	  p12 = p12 + ( (p12 >> DAT_BITS) << PAD_BITS );
	  p13 = p13 + ( (p13 >> DAT_BITS) << PAD_BITS );
	  p14 = p14 + ( (p14 >> DAT_BITS) << PAD_BITS );
	  p15 = p15 + ( (p15 >> DAT_BITS) << PAD_BITS );
	  p16 = p16 + ( (p16 >> DAT_BITS) << PAD_BITS );
	  p17 = p17 + ( (p17 >> DAT_BITS) << PAD_BITS );
	  p18 = p18 + ( (p18 >> DAT_BITS) << PAD_BITS );
	  p19 = p19 + ( (p19 >> DAT_BITS) << PAD_BITS );
	  p20 = p20 + ( (p20 >> DAT_BITS) << PAD_BITS );
	  p21 = p21 + ( (p21 >> DAT_BITS) << PAD_BITS );
	  p22 = p22 + ( (p22 >> DAT_BITS) << PAD_BITS );
	  p23 = p23 + ( (p23 >> DAT_BITS) << PAD_BITS );
	}

/*...The radix-24 pass is here.	*/

	for(j=0; j < n24; j += 2)
	{
	  j1 = j  + ( (j  >> DAT_BITS) << PAD_BITS );
/*
Twiddleless version requires us to swap inputs as follows:
indices  0, 1, 2, 3, 4, 5, 6, 7, 8, 9,10,11,12,13,14,15,16,17,18,19,20,21,22,23
      -> 0,16, 8|21,13, 5|18,10, 2|15, 7,23|12, 4,20| 9, 1,17| 6,22,14| 3,19,11

I.e. start out with first octet of indices {0,3,6,9,12,15,18,21}, permute those according to
  {0,3,6,9,12,15,18,21}*23%24 = {0,21,18,15,12,9,6,3}, then each is head of a length-3 list of indices with decrement 8

Remember, inputs to DIT are bit-reversed, so
a[0, 1, 2, 3, 4, 5, 6, 7, 8, 9,10,11,12,13,14,15,16,17,18,19,20,21,22,23] contain
x[0,12, 6,18, 3,15, 9,21, 1,13, 7,19, 4,16,10,22, 2,14, 8,20, 5,17,11,23], which get swapped to
x[0,12,18, 6,21, 9,15, 3,16, 4,10,22,13, 1, 7,19, 8,20, 2,14, 5,17,23,11], which means the a-indices get swapped as
a[0, 1, 3, 2, 7, 6, 5, 4,13,12,14,15, 9, 8,10,11,18,19,16,17,20,21,23,22].
*/
/*...gather the needed data (24 64-bit complex, i.e. 48 64-bit reals) and do 3 radix-8 transforms,	*/
                     /*                                                                                 inputs                                                                                    */ /*                 intermediates                    */ /*                                                         outputs                                                           */
	RADIX_08_DIT(a[j1    ],a[j1+1    ],a[j1+p01],a[j1+1+p01],a[j1+p03],a[j1+1+p03],a[j1+p02],a[j1+1+p02],a[j1+p07],a[j1+1+p07],a[j1+p06],a[j1+1+p06],a[j1+p05],a[j1+1+p05],a[j1+p04],a[j1+1+p04],t1,t2,t3,t4,t5,t6,t7,t8,t9,t10,t11,t12,t13,t14,t15,t16,aj1p00r,aj1p00i,aj1p01r,aj1p01i,aj1p02r,aj1p02i,aj1p03r,aj1p03i,aj1p04r,aj1p04i,aj1p05r,aj1p05i,aj1p06r,aj1p06i,aj1p07r,aj1p07i,rt,it);
	RADIX_08_DIT(a[j1+p13],a[j1+1+p13],a[j1+p12],a[j1+1+p12],a[j1+p14],a[j1+1+p14],a[j1+p15],a[j1+1+p15],a[j1+p09],a[j1+1+p09],a[j1+p08],a[j1+1+p08],a[j1+p10],a[j1+1+p10],a[j1+p11],a[j1+1+p11],t1,t2,t3,t4,t5,t6,t7,t8,t9,t10,t11,t12,t13,t14,t15,t16,aj1p08r,aj1p08i,aj1p09r,aj1p09i,aj1p10r,aj1p10i,aj1p11r,aj1p11i,aj1p12r,aj1p12i,aj1p13r,aj1p13i,aj1p14r,aj1p14i,aj1p15r,aj1p15i,rt,it);
	RADIX_08_DIT(a[j1+p18],a[j1+1+p18],a[j1+p19],a[j1+1+p19],a[j1+p16],a[j1+1+p16],a[j1+p17],a[j1+1+p17],a[j1+p20],a[j1+1+p20],a[j1+p21],a[j1+1+p21],a[j1+p23],a[j1+1+p23],a[j1+p22],a[j1+1+p22],t1,t2,t3,t4,t5,t6,t7,t8,t9,t10,t11,t12,t13,t14,t15,t16,aj1p16r,aj1p16i,aj1p17r,aj1p17i,aj1p18r,aj1p18i,aj1p19r,aj1p19i,aj1p20r,aj1p20i,aj1p21r,aj1p21i,aj1p22r,aj1p22i,aj1p23r,aj1p23i,rt,it);

/*...and now do 8 radix-3 transforms.	*/
                     /*                  inputs                   */ /*             intermediates                 */ /*                       outputs                               */
	RADIX_03_DFT(aj1p00r,aj1p00i,aj1p08r,aj1p08i,aj1p16r,aj1p16i,aj1p00r,aj1p00i,aj1p08r,aj1p08i,aj1p16r,aj1p16i,a[j1    ],a[j1+1    ],a[j1+p16],a[j1+1+p16],a[j1+p08],a[j1+1+p08],rt,it);
	RADIX_03_DFT(aj1p01r,aj1p01i,aj1p09r,aj1p09i,aj1p17r,aj1p17i,aj1p01r,aj1p01i,aj1p09r,aj1p09i,aj1p17r,aj1p17i,a[j1+p09],a[j1+1+p09],a[j1+p01],a[j1+1+p01],a[j1+p17],a[j1+1+p17],rt,it);
	RADIX_03_DFT(aj1p02r,aj1p02i,aj1p10r,aj1p10i,aj1p18r,aj1p18i,aj1p02r,aj1p02i,aj1p10r,aj1p10i,aj1p18r,aj1p18i,a[j1+p18],a[j1+1+p18],a[j1+p10],a[j1+1+p10],a[j1+p02],a[j1+1+p02],rt,it);
	RADIX_03_DFT(aj1p03r,aj1p03i,aj1p11r,aj1p11i,aj1p19r,aj1p19i,aj1p03r,aj1p03i,aj1p11r,aj1p11i,aj1p19r,aj1p19i,a[j1+p03],a[j1+1+p03],a[j1+p19],a[j1+1+p19],a[j1+p11],a[j1+1+p11],rt,it);
	RADIX_03_DFT(aj1p04r,aj1p04i,aj1p12r,aj1p12i,aj1p20r,aj1p20i,aj1p04r,aj1p04i,aj1p12r,aj1p12i,aj1p20r,aj1p20i,a[j1+p12],a[j1+1+p12],a[j1+p04],a[j1+1+p04],a[j1+p20],a[j1+1+p20],rt,it);
	RADIX_03_DFT(aj1p05r,aj1p05i,aj1p13r,aj1p13i,aj1p21r,aj1p21i,aj1p05r,aj1p05i,aj1p13r,aj1p13i,aj1p21r,aj1p21i,a[j1+p21],a[j1+1+p21],a[j1+p13],a[j1+1+p13],a[j1+p05],a[j1+1+p05],rt,it);
	RADIX_03_DFT(aj1p06r,aj1p06i,aj1p14r,aj1p14i,aj1p22r,aj1p22i,aj1p06r,aj1p06i,aj1p14r,aj1p14i,aj1p22r,aj1p22i,a[j1+p06],a[j1+1+p06],a[j1+p22],a[j1+1+p22],a[j1+p14],a[j1+1+p14],rt,it);
	RADIX_03_DFT(aj1p07r,aj1p07i,aj1p15r,aj1p15i,aj1p23r,aj1p23i,aj1p07r,aj1p07i,aj1p15r,aj1p15i,aj1p23r,aj1p23i,a[j1+p15],a[j1+1+p15],a[j1+p07],a[j1+1+p07],a[j1+p23],a[j1+1+p23],rt,it);

	}
}
/*
   0   129.0000000000  -112.0000000000   129.0000000000  -112.0000000000
   1    -2.6087904079     2.4246212025    -2.6087904079   -11.1781322900 ERR=    0.0000000000    13.6027534925
   2   -13.0262794416    -7.0000000000   -13.0262794416    19.2224318643 ERR=    0.0000000000    26.2224318643
   3     7.7071067812    17.3639610307     7.7071067812    17.3639610307
   4     5.4641016151     6.0000000000     5.4641016151     7.7320508076 ERR=    0.0000000000     1.7320508076
   5    -1.6159544635    -5.0000000000    -1.6159544635    -0.0466125814 ERR=    0.0000000000     4.9533874186
   6   -16.0000000000    13.0000000000   -16.0000000000    13.0000000000
   7     9.7818873666    -0.3033008589     9.7818873666     7.8061272737 ERR=    0.0000000000     8.1094281326
   8    25.0525588833    -1.0000000000    25.0525588833    -4.5884572681 ERR=    0.0000000000     3.5884572681
   9   -17.1213203436     3.1213203436   -17.1213203436     3.1213203436
  10     6.0262794416     4.5000000000     6.0262794416   -10.2224318643 ERR=    0.0000000000    14.7224318643
  11    24.4882651899    -7.0000000000    24.4882651899    13.2165765688 ERR=    0.0000000000    20.2165765688
  12   -31.0000000000     6.0000000000   -31.0000000000     6.0000000000
  13    13.4651968684   -12.4246212025    13.4651968684   -24.8026298235 ERR=    0.0000000000    12.3780086211
  14    11.2942286341    -2.0000000000    11.2942286341   -13.8301270189 ERR=    0.0000000000    11.8301270189
  15     6.2928932188     4.6360389693     6.2928932188     4.6360389693
  16   -13.0525588833    11.0000000000   -13.0525588833    26.5884572681 ERR=    0.0000000000    15.5884572681
  17   -15.2404519970    -7.0000000000   -15.2404519970    16.0273746949 ERR=    0.0000000000    23.0273746949
  18   -26.0000000000    21.0000000000   -26.0000000000    21.0000000000
  19    -5.7100905968    10.3033008589    -5.7100905968     7.3900251490 ERR=    0.0000000000     2.9132757099
  20    -1.4641016151    -8.0000000000    -1.4641016151     4.2679491924 ERR=    0.0000000000    12.2679491924
  21   -12.8786796564    -1.1213203436   -12.8786796564    -1.1213203436
  22    -4.2942286341    -9.5000000000    -4.2942286341    -5.1698729811 ERR=    0.0000000000     4.3301270189
  23    -6.5600619596    -7.0000000000    -6.5600619596    -8.4127289915 ERR=    0.0000000000     1.4127289915
*/