/*******************************************************************************
*                                                                              *
*   (C) 1997-2004 by Ernst W. Mayer.                                           *
*                                                                              *
*  This program is free software; you can redistribute it and/or modify it     *
*  under the terms of the GNU General Public License as published by the       *
*  Free Software Foundation; either version 2 of the License, or (at your      *
*  option) any later version.                                                  *
*                                                                              *
*  This program is distributed in the hope that it will be useful, but WITHOUT *
*  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or       *
*  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for   *
*  more details.                                                               *
*                                                                              *
*  You should have received a copy of the GNU General Public License along     *
*  with this program; see the file GPL.txt.  If not, you may view one at       *
*  http://www.fsf.org/licenses/licenses.html, or obtain one by writing to the  *
*  Free Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA     *
*  02111-1307, USA.                                                            *
*                                                                              *
*******************************************************************************/

#include "Mlucas.h"

/***************/

void radix28_dif_pass1(double a[], int n)
{
/*
!...Acronym: DIF = Decimation In Frequency
!
!...Subroutine to perform an initial radix-28 complex DIF FFT pass on the data in the length-N real vector A.
!
!   See the documentation in radix16_dif_pass for further details on storage and indexing.
!
!   See the documentation in radix7_dif_pass for details on the radix-15 subtransforms.
*/
	int j,j1;
	static int n28,p01,p02,p03,p04,p05,p06,p07,p08,p09,p10,p11,p12,p13,p14,p15,p16,p17,p18,p19,p20,p21,p22,p23,p24,p25,p26,p27, first_entry=TRUE;
	static double cc1 = .62348980185873353053,	 /* cos(u) = Real part of exp(i*2*pi/7), the radix-7 fundamental sincos datum	*/
		      ss1 = .78183148246802980870,	 /* sin(u) = Imag part of exp(i*2*pi/7).	*/
		      cc2 =-.22252093395631440426,	 /* cos(2u)	*/
		      ss2 = .97492791218182360702,	 /* sin(2u)	*/
		      cc3 =-.90096886790241912622,	 /* cos(3u)	*/
		      ss3 = .43388373911755812050;	 /* sin(3u)	*/
	double re,im,rt,it
	,t1,t2,t3,t4,t5,t6,t7,t8,t9,t10,t11,t12,t13,t14
	,aj1p00r,aj1p01r,aj1p02r,aj1p03r,aj1p04r,aj1p05r,aj1p06r,aj1p07r,aj1p08r,aj1p09r,aj1p10r,aj1p11r,aj1p12r,aj1p13r,aj1p14r,aj1p15r,aj1p16r,aj1p17r,aj1p18r,aj1p19r,aj1p20r,aj1p21r,aj1p22r,aj1p23r,aj1p24r,aj1p25r,aj1p26r,aj1p27r
	,aj1p00i,aj1p01i,aj1p02i,aj1p03i,aj1p04i,aj1p05i,aj1p06i,aj1p07i,aj1p08i,aj1p09i,aj1p10i,aj1p11i,aj1p12i,aj1p13i,aj1p14i,aj1p15i,aj1p16i,aj1p17i,aj1p18i,aj1p19i,aj1p20i,aj1p21i,aj1p22i,aj1p23i,aj1p24i,aj1p25i,aj1p26i,aj1p27i;

	if(!first_entry && (n/28) != n28)	/* New runlength?	*/
	{
	  first_entry=TRUE;
	}

/*...initialize things upon first entry	*/

	if(first_entry)
	{
	  first_entry=FALSE;
	  n28=n/28;

/*   constant index offsets for array load/stores are here.	*/

	  p01 = n28;
	  p02 = p01 + p01;
	  p03 = p02 + p01;
	  p04 = p03 + p01;
	  p05 = p04 + p01;
	  p06 = p05 + p01;
	  p07 = p06 + p01;
	  p08 = p07 + p01;
	  p09 = p08 + p01;
	  p10 = p09 + p01;
	  p11 = p10 + p01;
	  p12 = p11 + p01;
	  p13 = p12 + p01;
	  p14 = p13 + p01;
	  p15 = p14 + p01;
	  p16 = p15 + p01;
	  p17 = p16 + p01;
	  p18 = p17 + p01;
	  p19 = p18 + p01;
	  p20 = p19 + p01;
	  p21 = p20 + p01;
	  p22 = p21 + p01;
	  p23 = p22 + p01;
	  p24 = p23 + p01;
	  p25 = p24 + p01;
	  p26 = p25 + p01;
	  p27 = p26 + p01;

	  p01 = p01 + ( (p01 >> DAT_BITS) << PAD_BITS );
	  p02 = p02 + ( (p02 >> DAT_BITS) << PAD_BITS );
	  p03 = p03 + ( (p03 >> DAT_BITS) << PAD_BITS );
	  p04 = p04 + ( (p04 >> DAT_BITS) << PAD_BITS );
	  p05 = p05 + ( (p05 >> DAT_BITS) << PAD_BITS );
	  p06 = p06 + ( (p06 >> DAT_BITS) << PAD_BITS );
	  p07 = p07 + ( (p07 >> DAT_BITS) << PAD_BITS );
	  p08 = p08 + ( (p08 >> DAT_BITS) << PAD_BITS );
	  p09 = p09 + ( (p09 >> DAT_BITS) << PAD_BITS );
	  p10 = p10 + ( (p10 >> DAT_BITS) << PAD_BITS );
	  p11 = p11 + ( (p11 >> DAT_BITS) << PAD_BITS );
	  p12 = p12 + ( (p12 >> DAT_BITS) << PAD_BITS );
	  p13 = p13 + ( (p13 >> DAT_BITS) << PAD_BITS );
	  p14 = p14 + ( (p14 >> DAT_BITS) << PAD_BITS );
	  p15 = p15 + ( (p15 >> DAT_BITS) << PAD_BITS );
	  p16 = p16 + ( (p16 >> DAT_BITS) << PAD_BITS );
	  p17 = p17 + ( (p17 >> DAT_BITS) << PAD_BITS );
	  p18 = p18 + ( (p18 >> DAT_BITS) << PAD_BITS );
	  p19 = p19 + ( (p19 >> DAT_BITS) << PAD_BITS );
	  p20 = p20 + ( (p20 >> DAT_BITS) << PAD_BITS );
	  p21 = p21 + ( (p21 >> DAT_BITS) << PAD_BITS );
	  p22 = p22 + ( (p22 >> DAT_BITS) << PAD_BITS );
	  p23 = p23 + ( (p23 >> DAT_BITS) << PAD_BITS );
	  p24 = p24 + ( (p24 >> DAT_BITS) << PAD_BITS );
	  p25 = p25 + ( (p25 >> DAT_BITS) << PAD_BITS );
	  p26 = p26 + ( (p26 >> DAT_BITS) << PAD_BITS );
	  p27 = p27 + ( (p27 >> DAT_BITS) << PAD_BITS );
	}

/*...The radix-28 pass is here.	*/

	for(j=0; j < n28; j += 2)
	{
	  j1 = j  + ( (j  >> DAT_BITS) << PAD_BITS );

/*...gather the needed data (28 64-bit complex, i.e. 56 64-bit reals) and do 4 radix-7 transforms...*/

/*
Twiddleless version requires us to swap inputs as follows:
indices  0, 1, 2, 3, 4, 5, 6, 7, 8, 9,10,11,12,13,14,15,16,17,18,19,20,21,22,23,24,25,26,27
      -> 0,21,14, 7,24,17,10, 3,20,13, 6,27,16, 9, 2,23,12, 5,26,19, 8, 1,22,15, 4,25,18,11
I.e. start out with first septet of indices {0,4,8,12,16,20,24}, permute those according to
{0,4,8,12,16,20,24}*27%28 = {0,24,20,16,12,8,4}, then each is head of a length-4 list of indices with decrement 7.
*/
/*...gather the needed data (28 64-bit complex, i.e. 56 64-bit reals) and do 4 radix-7 transforms...*/
#if 1
                     /*                                                                      inputs                                                                         */ /*               intermediates              */ /*                                                  outputs                                                  */
	RADIX_07_DFT(a[j1    ],a[j1+1    ],a[j1+p24],a[j1+1+p24],a[j1+p20],a[j1+1+p20],a[j1+p16],a[j1+1+p16],a[j1+p12],a[j1+1+p12],a[j1+p08],a[j1+1+p08],a[j1+p04],a[j1+1+p04],t1,t2,t3,t4,t5,t6,t7,t8,t9,t10,t11,t12,t13,t14,aj1p00r,aj1p00i,aj1p04r,aj1p04i,aj1p08r,aj1p08i,aj1p12r,aj1p12i,aj1p16r,aj1p16i,aj1p20r,aj1p20i,aj1p24r,aj1p24i,rt,it,re,im);
	RADIX_07_DFT(a[j1+p21],a[j1+1+p21],a[j1+p17],a[j1+1+p17],a[j1+p13],a[j1+1+p13],a[j1+p09],a[j1+1+p09],a[j1+p05],a[j1+1+p05],a[j1+p01],a[j1+1+p01],a[j1+p25],a[j1+1+p25],t1,t2,t3,t4,t5,t6,t7,t8,t9,t10,t11,t12,t13,t14,aj1p01r,aj1p01i,aj1p05r,aj1p05i,aj1p09r,aj1p09i,aj1p13r,aj1p13i,aj1p17r,aj1p17i,aj1p21r,aj1p21i,aj1p25r,aj1p25i,rt,it,re,im);
	RADIX_07_DFT(a[j1+p14],a[j1+1+p14],a[j1+p10],a[j1+1+p10],a[j1+p06],a[j1+1+p06],a[j1+p02],a[j1+1+p02],a[j1+p26],a[j1+1+p26],a[j1+p22],a[j1+1+p22],a[j1+p18],a[j1+1+p18],t1,t2,t3,t4,t5,t6,t7,t8,t9,t10,t11,t12,t13,t14,aj1p02r,aj1p02i,aj1p06r,aj1p06i,aj1p10r,aj1p10i,aj1p14r,aj1p14i,aj1p18r,aj1p18i,aj1p22r,aj1p22i,aj1p26r,aj1p26i,rt,it,re,im);
	RADIX_07_DFT(a[j1+p07],a[j1+1+p07],a[j1+p03],a[j1+1+p03],a[j1+p27],a[j1+1+p27],a[j1+p23],a[j1+1+p23],a[j1+p19],a[j1+1+p19],a[j1+p15],a[j1+1+p15],a[j1+p11],a[j1+1+p11],t1,t2,t3,t4,t5,t6,t7,t8,t9,t10,t11,t12,t13,t14,aj1p03r,aj1p03i,aj1p07r,aj1p07i,aj1p11r,aj1p11i,aj1p15r,aj1p15i,aj1p19r,aj1p19i,aj1p23r,aj1p23i,aj1p27r,aj1p27i,rt,it,re,im);

/*...and now do 7 radix-4 transforms...*/
                     /*                          inputs                           */ /*                                      outputs                                      */
	RADIX_04_DIF(aj1p00r,aj1p00i,aj1p01r,aj1p01i,aj1p02r,aj1p02i,aj1p03r,aj1p03i,a[j1    ],a[j1+1    ],a[j1+p01],a[j1+1+p01],a[j1+p02],a[j1+1+p02],a[j1+p03],a[j1+1+p03],rt,it);
	RADIX_04_DIF(aj1p04r,aj1p04i,aj1p05r,aj1p05i,aj1p06r,aj1p06i,aj1p07r,aj1p07i,a[j1+p25],a[j1+1+p25],a[j1+p24],a[j1+1+p24],a[j1+p27],a[j1+1+p27],a[j1+p26],a[j1+1+p26],rt,it);
	RADIX_04_DIF(aj1p08r,aj1p08i,aj1p09r,aj1p09i,aj1p10r,aj1p10i,aj1p11r,aj1p11i,a[j1+p22],a[j1+1+p22],a[j1+p23],a[j1+1+p23],a[j1+p21],a[j1+1+p21],a[j1+p20],a[j1+1+p20],rt,it);
	RADIX_04_DIF(aj1p12r,aj1p12i,aj1p13r,aj1p13i,aj1p14r,aj1p14i,aj1p15r,aj1p15i,a[j1+p16],a[j1+1+p16],a[j1+p17],a[j1+1+p17],a[j1+p18],a[j1+1+p18],a[j1+p19],a[j1+1+p19],rt,it);
	RADIX_04_DIF(aj1p16r,aj1p16i,aj1p17r,aj1p17i,aj1p18r,aj1p18i,aj1p19r,aj1p19i,a[j1+p15],a[j1+1+p15],a[j1+p14],a[j1+1+p14],a[j1+p12],a[j1+1+p12],a[j1+p13],a[j1+1+p13],rt,it);
	RADIX_04_DIF(aj1p20r,aj1p20i,aj1p21r,aj1p21i,aj1p22r,aj1p22i,aj1p23r,aj1p23i,a[j1+p09],a[j1+1+p09],a[j1+p08],a[j1+1+p08],a[j1+p11],a[j1+1+p11],a[j1+p10],a[j1+1+p10],rt,it);
	RADIX_04_DIF(aj1p24r,aj1p24i,aj1p25r,aj1p25i,aj1p26r,aj1p26i,aj1p27r,aj1p27i,a[j1+p06],a[j1+1+p06],a[j1+p07],a[j1+1+p07],a[j1+p05],a[j1+1+p05],a[j1+p04],a[j1+1+p04],rt,it);
#else

/*...1st radix-7 block uses aj1p[0:24:4] as inputs:	*/

	  t1 =a[j1    ];		t2 =a[j1+1    ];
	  t3 =a[j1+p24]+a[j1+p04];	t4 =a[j1+1+p24]+a[j1+1+p04];	/* x1+x6 */
	  t5 =a[j1+p20]+a[j1+p08];	t6 =a[j1+1+p20]+a[j1+1+p08];	/* x2+x5 */
	  t7 =a[j1+p16]+a[j1+p12];	t8 =a[j1+1+p16]+a[j1+1+p12];	/* x3+x4 */
	  t9 =a[j1+p16]-a[j1+p12];	t10=a[j1+1+p16]-a[j1+1+p12];	/* x3-x4 */
	  t11=a[j1+p20]-a[j1+p08];	t12=a[j1+1+p20]-a[j1+1+p08];	/* x2-x5 */
	  t13=a[j1+p24]-a[j1+p04];	t14=a[j1+1+p24]-a[j1+1+p04];	/* x1-x6 */

	  aj1p00r = t1+t3+t5+t7;	aj1p00i = t2+t4+t6+t8;
	  rt = t1+cc1*t3+cc2*t5+cc3*t7;	it = t2+cc1*t4+cc2*t6+cc3*t8;
	  re = t1+cc2*t3+cc3*t5+cc1*t7;	im = t2+cc2*t4+cc3*t6+cc1*t8;
	  t1 = t1+cc3*t3+cc1*t5+cc2*t7;	t2 = t2+cc3*t4+cc1*t6+cc2*t8;

	  t3 = ss1*t13+ss2*t11+ss3*t9;	t4 = ss1*t14+ss2*t12+ss3*t10;
	  t5 = ss2*t13-ss3*t11-ss1*t9;	t6 = ss2*t14-ss3*t12-ss1*t10;
	  t7 = ss3*t13-ss1*t11+ss2*t9;	t8 = ss3*t14-ss1*t12+ss2*t10;
	/* Output permutation causes signs to get flipped here: */
	  aj1p04r=rt-t4;		aj1p04i=it+t3;
	  aj1p08r=re-t6;		aj1p08i=im+t5;
	  aj1p12r=t1-t8;		aj1p12i=t2+t7;
	  aj1p16r=t1+t8;		aj1p16i=t2-t7;
	  aj1p20r=re+t6;		aj1p20i=im-t5;
	  aj1p24r=rt+t4;		aj1p24i=it-t3;

/*...2nd radix-7 block uses aj1p[1:25:4] as inputs:	*/

	  t1 =a[j1+p21];		t2 =a[j1+1+p21];
	  t3 =a[j1+p17]+a[j1+p25];	t4 =a[j1+1+p17]+a[j1+1+p25];	/* x1+x6 */
	  t5 =a[j1+p13]+a[j1+p01];	t6 =a[j1+1+p13]+a[j1+1+p01];	/* x2+x5 */
	  t7 =a[j1+p09]+a[j1+p05];	t8 =a[j1+1+p09]+a[j1+1+p05];	/* x3+x4 */
	  t9 =a[j1+p09]-a[j1+p05];	t10=a[j1+1+p09]-a[j1+1+p05];	/* x3-x4 */
	  t11=a[j1+p13]-a[j1+p01];	t12=a[j1+1+p13]-a[j1+1+p01];	/* x2-x5 */
	  t13=a[j1+p17]-a[j1+p25];	t14=a[j1+1+p17]-a[j1+1+p25];	/* x1-x6 */

	  aj1p01r = t1+t3+t5+t7;	aj1p01i = t2+t4+t6+t8;
	  rt = t1+cc1*t3+cc2*t5+cc3*t7;	it = t2+cc1*t4+cc2*t6+cc3*t8;
	  re = t1+cc2*t3+cc3*t5+cc1*t7;	im = t2+cc2*t4+cc3*t6+cc1*t8;
	  t1 = t1+cc3*t3+cc1*t5+cc2*t7;	t2 = t2+cc3*t4+cc1*t6+cc2*t8;

	  t3 = ss1*t13+ss2*t11+ss3*t9;	t4 = ss1*t14+ss2*t12+ss3*t10;
	  t5 = ss2*t13-ss3*t11-ss1*t9;	t6 = ss2*t14-ss3*t12-ss1*t10;
	  t7 = ss3*t13-ss1*t11+ss2*t9;	t8 = ss3*t14-ss1*t12+ss2*t10;
	/* Output permutation causes signs to get flipped here: */
	  aj1p05r=rt-t4;		aj1p05i=it+t3;
	  aj1p09r=re-t6;		aj1p09i=im+t5;
	  aj1p13r=t1-t8;		aj1p13i=t2+t7;
	  aj1p17r=t1+t8;		aj1p17i=t2-t7;
	  aj1p21r=re+t6;		aj1p21i=im-t5;
	  aj1p25r=rt+t4;		aj1p25i=it-t3;

/*...3rd radix-7 block uses aj1p[2:26:4] as inputs:	*/

	  t1 =a[j1+p14];		t2 =a[j1+1+p14];
	  t3 =a[j1+p10]+a[j1+p18];	t4 =a[j1+1+p10]+a[j1+1+p18];	/* x1+x6 */
	  t5 =a[j1+p06]+a[j1+p22];	t6 =a[j1+1+p06]+a[j1+1+p22];	/* x2+x5 */
	  t7 =a[j1+p02]+a[j1+p26];	t8 =a[j1+1+p02]+a[j1+1+p26];	/* x3+x4 */
	  t9 =a[j1+p02]-a[j1+p26];	t10=a[j1+1+p02]-a[j1+1+p26];	/* x3-x4 */
	  t11=a[j1+p06]-a[j1+p22];	t12=a[j1+1+p06]-a[j1+1+p22];	/* x2-x5 */
	  t13=a[j1+p10]-a[j1+p18];	t14=a[j1+1+p10]-a[j1+1+p18];	/* x1-x6 */

	  aj1p02r = t1+t3+t5+t7;	aj1p02i = t2+t4+t6+t8;
	  rt = t1+cc1*t3+cc2*t5+cc3*t7;	it = t2+cc1*t4+cc2*t6+cc3*t8;
	  re = t1+cc2*t3+cc3*t5+cc1*t7;	im = t2+cc2*t4+cc3*t6+cc1*t8;
	  t1 = t1+cc3*t3+cc1*t5+cc2*t7;	t2 = t2+cc3*t4+cc1*t6+cc2*t8;

	  t3 = ss1*t13+ss2*t11+ss3*t9;	t4 = ss1*t14+ss2*t12+ss3*t10;
	  t5 = ss2*t13-ss3*t11-ss1*t9;	t6 = ss2*t14-ss3*t12-ss1*t10;
	  t7 = ss3*t13-ss1*t11+ss2*t9;	t8 = ss3*t14-ss1*t12+ss2*t10;
	/* Output permutation causes signs to get flipped here: */
	  aj1p06r=rt-t4;		aj1p06i=it+t3;
	  aj1p10r=re-t6;		aj1p10i=im+t5;
	  aj1p14r=t1-t8;		aj1p14i=t2+t7;
	  aj1p18r=t1+t8;		aj1p18i=t2-t7;
	  aj1p22r=re+t6;		aj1p22i=im-t5;
	  aj1p26r=rt+t4;		aj1p26i=it-t3;

/*...4th radix-7 block uses aj1p[3:27:4] as inputs:	*/

	  t1 =a[j1+p07];		t2 =a[j1+1+p07];
	  t3 =a[j1+p03]+a[j1+p11];	t4 =a[j1+1+p03]+a[j1+1+p11];	/* x1+x6 */
	  t5 =a[j1+p27]+a[j1+p15];	t6 =a[j1+1+p27]+a[j1+1+p15];	/* x2+x5 */
	  t7 =a[j1+p23]+a[j1+p19];	t8 =a[j1+1+p23]+a[j1+1+p19];	/* x3+x4 */
	  t9 =a[j1+p23]-a[j1+p19];	t10=a[j1+1+p23]-a[j1+1+p19];	/* x3-x4 */
	  t11=a[j1+p27]-a[j1+p15];	t12=a[j1+1+p27]-a[j1+1+p15];	/* x2-x5 */
	  t13=a[j1+p03]-a[j1+p11];	t14=a[j1+1+p03]-a[j1+1+p11];	/* x1-x6 */

	  aj1p03r = t1+t3+t5+t7;	aj1p03i = t2+t4+t6+t8;
	  rt = t1+cc1*t3+cc2*t5+cc3*t7;	it = t2+cc1*t4+cc2*t6+cc3*t8;
	  re = t1+cc2*t3+cc3*t5+cc1*t7;	im = t2+cc2*t4+cc3*t6+cc1*t8;
	  t1 = t1+cc3*t3+cc1*t5+cc2*t7;	t2 = t2+cc3*t4+cc1*t6+cc2*t8;

	  t3 = ss1*t13+ss2*t11+ss3*t9;	t4 = ss1*t14+ss2*t12+ss3*t10;
	  t5 = ss2*t13-ss3*t11-ss1*t9;	t6 = ss2*t14-ss3*t12-ss1*t10;
	  t7 = ss3*t13-ss1*t11+ss2*t9;	t8 = ss3*t14-ss1*t12+ss2*t10;
	/* Output permutation causes signs to get flipped here: */
	  aj1p07r=rt-t4;		aj1p07i=it+t3;
	  aj1p11r=re-t6;		aj1p11i=im+t5;
	  aj1p15r=t1-t8;		aj1p15i=t2+t7;
	  aj1p19r=t1+t8;		aj1p19i=t2-t7;
	  aj1p23r=re+t6;		aj1p23i=im-t5;
	  aj1p27r=rt+t4;		aj1p27i=it-t3;

/*...and now do 7 radix-4 transforms:	*/
/*...Block 1: */
	    rt =aj1p02r;	aj1p02r=aj1p00r-rt;	aj1p00r=aj1p00r+rt;
	    it =aj1p02i;	aj1p02i=aj1p00i-it;	aj1p00i=aj1p00i+it;

	    rt =aj1p03r;	aj1p03r=aj1p01r-rt;	aj1p01r=aj1p01r+rt;
	    it =aj1p03i;	aj1p03i=aj1p01i-it;	aj1p01i=aj1p01i+it;

	    a[j1    ]=aj1p00r+aj1p01r;	a[j1+1    ]=aj1p00i+aj1p01i;
	    a[j1+p01]=aj1p00r-aj1p01r;	a[j1+1+p01]=aj1p00i-aj1p01i;
	    a[j1+p02]=aj1p02r-aj1p03i;	a[j1+1+p02]=aj1p02i+aj1p03r;	/* mpy by I is inlined here...	*/
	    a[j1+p03]=aj1p02r+aj1p03i;	a[j1+1+p03]=aj1p02i-aj1p03r;

/*...Block 2: */
	    rt =aj1p06r;	aj1p06r=aj1p04r-rt;	aj1p04r=aj1p04r+rt;
	    it =aj1p06i;	aj1p06i=aj1p04i-it;	aj1p04i=aj1p04i+it;

	    rt =aj1p07r;	aj1p07r=aj1p05r-rt;	aj1p05r=aj1p05r+rt;
	    it =aj1p07i;	aj1p07i=aj1p05i-it;	aj1p05i=aj1p05i+it;

	    a[j1+p25]=aj1p04r+aj1p05r;	a[j1+1+p25]=aj1p04i+aj1p05i;
	    a[j1+p24]=aj1p04r-aj1p05r;	a[j1+1+p24]=aj1p04i-aj1p05i;
	    a[j1+p27]=aj1p06r-aj1p07i;	a[j1+1+p27]=aj1p06i+aj1p07r;	/* mpy by I is inlined here...	*/
	    a[j1+p26]=aj1p06r+aj1p07i;	a[j1+1+p26]=aj1p06i-aj1p07r;

/*...Block 3: */
	    rt =aj1p10r;	aj1p10r=aj1p08r-rt;	aj1p08r=aj1p08r+rt;
	    it =aj1p10i;	aj1p10i=aj1p08i-it;	aj1p08i=aj1p08i+it;

	    rt =aj1p11r;	aj1p11r=aj1p09r-rt;	aj1p09r=aj1p09r+rt;
	    it =aj1p11i;	aj1p11i=aj1p09i-it;	aj1p09i=aj1p09i+it;

	    a[j1+p22]=aj1p08r+aj1p09r;	a[j1+1+p22]=aj1p08i+aj1p09i;
	    a[j1+p23]=aj1p08r-aj1p09r;	a[j1+1+p23]=aj1p08i-aj1p09i;
	    a[j1+p21]=aj1p10r-aj1p11i;	a[j1+1+p21]=aj1p10i+aj1p11r;	/* mpy by I is inlined here...	*/
	    a[j1+p20]=aj1p10r+aj1p11i;	a[j1+1+p20]=aj1p10i-aj1p11r;

/*...Block 4: */
	    rt =aj1p14r;	aj1p14r=aj1p12r-rt;	aj1p12r=aj1p12r+rt;
	    it =aj1p14i;	aj1p14i=aj1p12i-it;	aj1p12i=aj1p12i+it;

	    rt =aj1p15r;	aj1p15r=aj1p13r-rt;	aj1p13r=aj1p13r+rt;
	    it =aj1p15i;	aj1p15i=aj1p13i-it;	aj1p13i=aj1p13i+it;

	    a[j1+p16]=aj1p12r+aj1p13r;	a[j1+1+p16]=aj1p12i+aj1p13i;
	    a[j1+p17]=aj1p12r-aj1p13r;	a[j1+1+p17]=aj1p12i-aj1p13i;
	    a[j1+p18]=aj1p14r-aj1p15i;	a[j1+1+p18]=aj1p14i+aj1p15r;	/* mpy by I is inlined here...	*/
	    a[j1+p19]=aj1p14r+aj1p15i;	a[j1+1+p19]=aj1p14i-aj1p15r;

/*...Block 5: */
	    rt =aj1p18r;	aj1p18r=aj1p16r-rt;	aj1p16r=aj1p16r+rt;
	    it =aj1p18i;	aj1p18i=aj1p16i-it;	aj1p16i=aj1p16i+it;

	    rt =aj1p19r;	aj1p19r=aj1p17r-rt;	aj1p17r=aj1p17r+rt;
	    it =aj1p19i;	aj1p19i=aj1p17i-it;	aj1p17i=aj1p17i+it;

	    a[j1+p15]=aj1p16r+aj1p17r;	a[j1+1+p15]=aj1p16i+aj1p17i;
	    a[j1+p14]=aj1p16r-aj1p17r;	a[j1+1+p14]=aj1p16i-aj1p17i;
	    a[j1+p12]=aj1p18r-aj1p19i;	a[j1+1+p12]=aj1p18i+aj1p19r;	/* mpy by I is inlined here...	*/
	    a[j1+p13]=aj1p18r+aj1p19i;	a[j1+1+p13]=aj1p18i-aj1p19r;

/*...Block 6: */
	    rt =aj1p22r;	aj1p22r=aj1p20r-rt;	aj1p20r=aj1p20r+rt;
	    it =aj1p22i;	aj1p22i=aj1p20i-it;	aj1p20i=aj1p20i+it;

	    rt =aj1p23r;	aj1p23r=aj1p21r-rt;	aj1p21r=aj1p21r+rt;
	    it =aj1p23i;	aj1p23i=aj1p21i-it;	aj1p21i=aj1p21i+it;

	    a[j1+p09]=aj1p20r+aj1p21r;	a[j1+1+p09]=aj1p20i+aj1p21i;
	    a[j1+p08]=aj1p20r-aj1p21r;	a[j1+1+p08]=aj1p20i-aj1p21i;
	    a[j1+p11]=aj1p22r-aj1p23i;	a[j1+1+p11]=aj1p22i+aj1p23r;	/* mpy by I is inlined here...	*/
	    a[j1+p10]=aj1p22r+aj1p23i;	a[j1+1+p10]=aj1p22i-aj1p23r;

/*...Block 7: */
	    rt =aj1p26r;	aj1p26r=aj1p24r-rt;	aj1p24r=aj1p24r+rt;
	    it =aj1p26i;	aj1p26i=aj1p24i-it;	aj1p24i=aj1p24i+it;

	    rt =aj1p27r;	aj1p27r=aj1p25r-rt;	aj1p25r=aj1p25r+rt;
	    it =aj1p27i;	aj1p27i=aj1p25i-it;	aj1p25i=aj1p25i+it;

	    a[j1+p06]=aj1p24r+aj1p25r;	a[j1+1+p06]=aj1p24i+aj1p25i;
	    a[j1+p07]=aj1p24r-aj1p25r;	a[j1+1+p07]=aj1p24i-aj1p25i;
	    a[j1+p05]=aj1p26r-aj1p27i;	a[j1+1+p05]=aj1p26i+aj1p27r;	/* mpy by I is inlined here...	*/
	    a[j1+p04]=aj1p26r+aj1p27i;	a[j1+1+p04]=aj1p26i-aj1p27r;
#endif
	}
}

/*
   2 (   7)  -33.0000000000   -22.0000000000   -15.0000000000   -14.0000000000
   3 (  21)  -15.0000000000   -14.0000000000   -33.0000000000   -22.0000000000
   4 (   1)  -11.6595587366   -19.7679307866    -7.6577999302     8.5976313922
   5 (  15)    2.8294700831     0.5106027780    32.4689557080    -0.3465505441
   6 (   8)   46.1480382688    -2.2866689587     9.9796188501    -9.6408663259
   7 (  22)  -17.5691832932     0.3833862984    -7.6241120460    -1.5581921548
   8 (   2)    3.4127732919    10.2586355896   -30.9999555547    -5.6414488172
   9 (  16)   -8.0030283504    -8.3212414840     2.0104983118   -20.7200709104
  10 (   9)   13.2364948412     1.8767883671    17.5130677056    -5.0313101030
  11 (  23)    1.0018763035   -12.2220878772     1.3671011496     0.7498989339
  12 (   3)    9.4157637378     2.9550563366    -6.8274600541   -10.0985523545
  13 (  17)  -25.6563211655    16.0852626944   -31.4472791768    -9.5901194084
  14 (  10)    2.2074782914   -13.5663212892    -3.5460529665    29.9250169837
  15 (  24)   11.5587101862    -6.4660944566     5.8409045449     5.9151760968
  16 (   4)    5.8409045449     5.9151760968     9.4157637378     2.9550563366
  17 (  18)   -3.5460529665    29.9250169837   -25.6563211655    16.0852626944
  18 (  11)  -31.4472791768    -9.5901194084    11.5587101862    -6.4660944566
  19 (  25)   -6.8274600541   -10.0985523545     2.2074782914   -13.5663212892
  20 (   5)    2.0104983118   -20.7200709104    13.2364948412     1.8767883671
  21 (  19)  -30.9999555547    -5.6414488172     1.0018763035   -12.2220878772
  22 (  12)   17.5130677056    -5.0313101030     3.4127732919    10.2586355896
  23 (  26)    1.3671011496     0.7498989339    -8.0030283504    -8.3212414840
  24 (   6)    9.9796188501    -9.6408663259     2.8294700831     0.5106027780
  25 (  20)   -7.6241120460    -1.5581921548   -11.6595587366   -19.7679307866
  26 (  13)   -7.6577999302     8.5976313922    46.1480382688    -2.2866689587
  27 (  27)   32.4689557080    -0.3465505441   -17.5691832932     0.3833862984
*/

/**************/

int radix28_ditN_cy_dif1(double a[], int n, int nwt, int nwt_bits, double wt0[], double wt1[], int si[], double base[], double baseinv[], int iter, double *fracmax, int p)
{
/*
!...Acronym: DWT = Discrete Weighted Transform, DIT = Decimation In Time, DIF = Decimation In Frequency
!
!...Performs a final radix-28 complex DIT pass, an inverse DWT weighting, a carry propagation,
!   a forward DWT weighting, and an initial radix-28 complex DIF pass on the data in the length-N real vector A.
!
!   Data enter and are returned in the A-array.
!
!   See the documentation in mers_mod_square and radix16_dif_pass for further details on the array
!   storage scheme, and radix8_ditN_cy_dif1 for details on the reduced-length weights array scheme.
*/
	int n28, bjmodn00,bjmodn01,bjmodn02,bjmodn03,bjmodn04,bjmodn05,bjmodn06,bjmodn07,bjmodn08,bjmodn09,bjmodn10,bjmodn11,bjmodn12,bjmodn13,bjmodn14,bjmodn15,bjmodn16,bjmodn17,bjmodn18,bjmodn19,bjmodn20,bjmodn21,bjmodn22,bjmodn23,bjmodn24,bjmodn25,bjmodn26,bjmodn27
		,i,j,j1,jstart,jhi,iroot,root_incr,k,khi,l,outer;
	static int psave=0,bw,sw,bjmodnini,nrtm1,p01,p02,p03,p04,p05,p06,p07,p08,p09,p10,p11,p12,p13,p14,p15,p16,p17,p18,p19,p20,p21,p22,p23,p24,p25,p26,p27;
	const  double one_half[3] = {1.0, 0.5, 0.25};	/* Needed for small-weights-tables scheme */
	static double radix_inv, n2inv;
	static double cc1 = .62348980185873353053,	 /* cos(u) = Real part of exp(i*2*pi/7), the radix-7 fundamental sincos datum	*/
		      ss1 = .78183148246802980870,	 /* sin(u) = Imag part of exp(i*2*pi/7).	*/
		      cc2 =-.22252093395631440426,	 /* cos(2u)	*/
		      ss2 = .97492791218182360702,	 /* sin(2u)	*/
		      cc3 =-.90096886790241912622,	 /* cos(3u)	*/
		      ss3 = .43388373911755812050;	 /* sin(3u)	*/
	double re,im,rt,it
	,t1,t2,t3,t4,t5,t6,t7,t8,t9,t10,t11,t12,t13,t14
	,aj1p00r,aj1p01r,aj1p02r,aj1p03r,aj1p04r,aj1p05r,aj1p06r,aj1p07r,aj1p08r,aj1p09r,aj1p10r,aj1p11r,aj1p12r,aj1p13r,aj1p14r,aj1p15r,aj1p16r,aj1p17r,aj1p18r,aj1p19r,aj1p20r,aj1p21r,aj1p22r,aj1p23r,aj1p24r,aj1p25r,aj1p26r,aj1p27r
	,aj1p00i,aj1p01i,aj1p02i,aj1p03i,aj1p04i,aj1p05i,aj1p06i,aj1p07i,aj1p08i,aj1p09i,aj1p10i,aj1p11i,aj1p12i,aj1p13i,aj1p14i,aj1p15i,aj1p16i,aj1p17i,aj1p18i,aj1p19i,aj1p20i,aj1p21i,aj1p22i,aj1p23i,aj1p24i,aj1p25i,aj1p26i,aj1p27i
	,cy00,cy01,cy02,cy03,cy04,cy05,cy06,cy07,cy08,cy09,cy10,cy11,cy12,cy13,cy14,cy15,cy16,cy17,cy18,cy19,cy20,cy21,cy22,cy23,cy24,cy25,cy26,cy27,temp,frac,scale;
	double maxerr = 0.0;
#if PFETCH
	double *add0, *add;
#endif
	int err;
	FILE *f1,*f2;
	static int first_entry=TRUE;

/*...stuff for the reduced-length DWT weights array is here:	*/
	int n_div_nwt;
	int col,co2,co3,m,m2,n_minus_sil,n_minus_silp1,sinwt,sinwtm1;
	double wt,wtinv,wtl,wtlp1,wtn,wtnm1,wtA,wtB,wtC;

/*...change n28 and n_div_wt to non-static to work around a gcc compiler bug. */
	n28   = n/28;
	n_div_nwt = n28 >> nwt_bits;

	if((n_div_nwt << nwt_bits) != n28)
	{
	  if(INTERACT)printf("FATAL: iter = %10d; NWT_BITS does not divide N/28 in radix28_ditN_cy_dif1.\n",iter);
	  f1 = fopen(   OFILE,"a");
	  f2 = fopen(STATFILE,"a");
	  fprintf(f1,"FATAL: iter = %10d; NWT_BITS does not divide N/28 in radix28_ditN_cy_dif1.\n",iter);
	  fprintf(f2,"FATAL: iter = %10d; NWT_BITS does not divide N/28 in radix28_ditN_cy_dif1.\n",iter);
	  fclose(f1);
	  fclose(f2);
	  err=1;
	  return(err);
	}

	if(p != psave)
	{
	  first_entry=TRUE;
	}

/*...initialize things upon first entry	*/

	if(first_entry)
	{
	  psave = p;
	  first_entry=FALSE;
	  radix_inv = (double)((long double)1.0L/(long double)28.0L);
	  n2inv     = (double)((long double)1.0L/(long double)(n/2));

	  bw    = p%n;		/* Number of bigwords in the Crandall/Fagin mixed-radix representation = (Mersenne exponent) mod (vector length).	*/
	  sw    = n - bw;	/* Number of smallwords.	*/

/*   constant index offsets for array load/stores are here.	*/

	  p01 = n28;
	  p02 = p01 + p01;
	  p03 = p02 + p01;
	  p04 = p03 + p01;
	  p05 = p04 + p01;
	  p06 = p05 + p01;
	  p07 = p06 + p01;
	  p08 = p07 + p01;
	  p09 = p08 + p01;
	  p10 = p09 + p01;
	  p11 = p10 + p01;
	  p12 = p11 + p01;
	  p13 = p12 + p01;
	  p14 = p13 + p01;
	  p15 = p14 + p01;
	  p16 = p15 + p01;
	  p17 = p16 + p01;
	  p18 = p17 + p01;
	  p19 = p18 + p01;
	  p20 = p19 + p01;
	  p21 = p20 + p01;
	  p22 = p21 + p01;
	  p23 = p22 + p01;
	  p24 = p23 + p01;
	  p25 = p24 + p01;
	  p26 = p25 + p01;
	  p27 = p26 + p01;

	  p01 = p01 + ( (p01 >> DAT_BITS) << PAD_BITS );
	  p02 = p02 + ( (p02 >> DAT_BITS) << PAD_BITS );
	  p03 = p03 + ( (p03 >> DAT_BITS) << PAD_BITS );
	  p04 = p04 + ( (p04 >> DAT_BITS) << PAD_BITS );
	  p05 = p05 + ( (p05 >> DAT_BITS) << PAD_BITS );
	  p06 = p06 + ( (p06 >> DAT_BITS) << PAD_BITS );
	  p07 = p07 + ( (p07 >> DAT_BITS) << PAD_BITS );
	  p08 = p08 + ( (p08 >> DAT_BITS) << PAD_BITS );
	  p09 = p09 + ( (p09 >> DAT_BITS) << PAD_BITS );
	  p10 = p10 + ( (p10 >> DAT_BITS) << PAD_BITS );
	  p11 = p11 + ( (p11 >> DAT_BITS) << PAD_BITS );
	  p12 = p12 + ( (p12 >> DAT_BITS) << PAD_BITS );
	  p13 = p13 + ( (p13 >> DAT_BITS) << PAD_BITS );
	  p14 = p14 + ( (p14 >> DAT_BITS) << PAD_BITS );
	  p15 = p15 + ( (p15 >> DAT_BITS) << PAD_BITS );
	  p16 = p16 + ( (p16 >> DAT_BITS) << PAD_BITS );
	  p17 = p17 + ( (p17 >> DAT_BITS) << PAD_BITS );
	  p18 = p18 + ( (p18 >> DAT_BITS) << PAD_BITS );
	  p19 = p19 + ( (p19 >> DAT_BITS) << PAD_BITS );
	  p20 = p20 + ( (p20 >> DAT_BITS) << PAD_BITS );
	  p21 = p21 + ( (p21 >> DAT_BITS) << PAD_BITS );
	  p22 = p22 + ( (p22 >> DAT_BITS) << PAD_BITS );
	  p23 = p23 + ( (p23 >> DAT_BITS) << PAD_BITS );
	  p24 = p24 + ( (p24 >> DAT_BITS) << PAD_BITS );
	  p25 = p25 + ( (p25 >> DAT_BITS) << PAD_BITS );
	  p26 = p26 + ( (p26 >> DAT_BITS) << PAD_BITS );
	  p27 = p27 + ( (p27 >> DAT_BITS) << PAD_BITS );

	  bjmodnini=0;
	  for(j=0; j < n28; j++)
	  {
	    bjmodnini -= sw; bjmodnini = bjmodnini + ( (-(int)((uint32)bjmodnini >> 31)) & n);
	  }
	}

/*...The radix-28 final DIT pass is here.	*/

	cy00=-2;		/* init carries	*/
	cy01= 0;
	cy02= 0;
	cy03= 0;
	cy04= 0;
	cy05= 0;
	cy06= 0;
	cy07= 0;
	cy08= 0;
	cy09= 0;
	cy10= 0;
	cy11= 0;
	cy12= 0;
	cy13= 0;
	cy14= 0;
	cy15= 0;
	cy16= 0;
	cy17= 0;
	cy18= 0;
	cy19= 0;
	cy20= 0;
	cy21= 0;
	cy22= 0;
	cy23= 0;
	cy24= 0;
	cy25= 0;
	cy26= 0;
	cy27= 0;

	*fracmax=0;	/* init max. fractional error	*/

	iroot = 0;	/* init sincos array index	*/
	root_incr = 1;	/* init sincos array index increment (set = 1 for normal carry pass, = 0 for wrapper pass)	*/
	scale = n2inv;	/* init inverse-weight scale factor  (set = 2/n for normal carry pass, = 1 for wrapper pass)	*/

	jstart = 0;
	jhi = jstart+nwt-1;
	khi = n_div_nwt;

for(outer=0; outer <= 1; outer++)
{
	i = 1;		/* Pointer to the BASE and BASEINV arrays. lowest-order digit is always a bigword (i = 1).	*/

	bjmodn00= 0;
	bjmodn01= bjmodnini;
	bjmodn02= bjmodn01+bjmodnini-n; bjmodn02= bjmodn02+ ( (-(int)((uint32)bjmodn02>> 31)) & n);
	bjmodn03= bjmodn02+bjmodnini-n; bjmodn03= bjmodn03+ ( (-(int)((uint32)bjmodn03>> 31)) & n);
	bjmodn04= bjmodn03+bjmodnini-n; bjmodn04= bjmodn04+ ( (-(int)((uint32)bjmodn04>> 31)) & n);
	bjmodn05= bjmodn04+bjmodnini-n; bjmodn05= bjmodn05+ ( (-(int)((uint32)bjmodn05>> 31)) & n);
	bjmodn06= bjmodn05+bjmodnini-n; bjmodn06= bjmodn06+ ( (-(int)((uint32)bjmodn06>> 31)) & n);
	bjmodn07= bjmodn06+bjmodnini-n; bjmodn07= bjmodn07+ ( (-(int)((uint32)bjmodn07>> 31)) & n);
	bjmodn08= bjmodn07+bjmodnini-n; bjmodn08= bjmodn08+ ( (-(int)((uint32)bjmodn08>> 31)) & n);
	bjmodn09= bjmodn08+bjmodnini-n; bjmodn09= bjmodn09+ ( (-(int)((uint32)bjmodn09>> 31)) & n);
	bjmodn10= bjmodn09+bjmodnini-n; bjmodn10= bjmodn10+ ( (-(int)((uint32)bjmodn10>> 31)) & n);
	bjmodn11= bjmodn10+bjmodnini-n; bjmodn11= bjmodn11+ ( (-(int)((uint32)bjmodn11>> 31)) & n);
	bjmodn12= bjmodn11+bjmodnini-n; bjmodn12= bjmodn12+ ( (-(int)((uint32)bjmodn12>> 31)) & n);
	bjmodn13= bjmodn12+bjmodnini-n; bjmodn13= bjmodn13+ ( (-(int)((uint32)bjmodn13>> 31)) & n);
	bjmodn14= bjmodn13+bjmodnini-n; bjmodn14= bjmodn14+ ( (-(int)((uint32)bjmodn14>> 31)) & n);
	bjmodn15= bjmodn14+bjmodnini-n; bjmodn15= bjmodn15+ ( (-(int)((uint32)bjmodn15>> 31)) & n);
	bjmodn16= bjmodn15+bjmodnini-n; bjmodn16= bjmodn16+ ( (-(int)((uint32)bjmodn16>> 31)) & n);
	bjmodn17= bjmodn16+bjmodnini-n; bjmodn17= bjmodn17+ ( (-(int)((uint32)bjmodn17>> 31)) & n);
	bjmodn18= bjmodn17+bjmodnini-n; bjmodn18= bjmodn18+ ( (-(int)((uint32)bjmodn18>> 31)) & n);
	bjmodn19= bjmodn18+bjmodnini-n; bjmodn19= bjmodn19+ ( (-(int)((uint32)bjmodn19>> 31)) & n);
	bjmodn20= bjmodn19+bjmodnini-n; bjmodn20= bjmodn20+ ( (-(int)((uint32)bjmodn20>> 31)) & n);
	bjmodn21= bjmodn20+bjmodnini-n; bjmodn21= bjmodn21+ ( (-(int)((uint32)bjmodn21>> 31)) & n);
	bjmodn22= bjmodn21+bjmodnini-n; bjmodn22= bjmodn22+ ( (-(int)((uint32)bjmodn22>> 31)) & n);
	bjmodn23= bjmodn22+bjmodnini-n; bjmodn23= bjmodn23+ ( (-(int)((uint32)bjmodn23>> 31)) & n);
	bjmodn24= bjmodn23+bjmodnini-n; bjmodn24= bjmodn24+ ( (-(int)((uint32)bjmodn24>> 31)) & n);
	bjmodn25= bjmodn24+bjmodnini-n; bjmodn25= bjmodn25+ ( (-(int)((uint32)bjmodn25>> 31)) & n);
	bjmodn26= bjmodn25+bjmodnini-n; bjmodn26= bjmodn26+ ( (-(int)((uint32)bjmodn26>> 31)) & n);
	bjmodn27= bjmodn26+bjmodnini-n; bjmodn27= bjmodn27+ ( (-(int)((uint32)bjmodn27>> 31)) & n);

	col=0;
	co2=(n >> nwt_bits)-1+28;
	co3=co2-28;		/* At the start of each new j-loop, co3=co2-radix(1)	*/

	for(k=1; k <= khi; k++)	/* Do n/(radix(1)*nwt) outer loop executions...	*/
	{
	  for(j=jstart; j<jhi; j += 2)	/* Each inner loop execution processes (radix(1)*nwt) array data.	*/
	  {
	    j1 = j + ( (j >> DAT_BITS) << PAD_BITS );	/* padded-array fetch index is here	*/

/*...The radix-28 DIT pass is here:	*/

/* EWM: 10/18/04: We swap the odd-index outputs of each of the radix-4 DIT transforms (1<=>3, 5<=>7, etc.) so that the indexing
                  of all the radix-7 transforms (really just the 2nd and 4th of these) winds up being in-place. This allows us
                  to properly re-use the ajp1 variables in the carry-pass version of this routine.
*/

/*...gather the needed data (28 64-bit complex, i.e. 56 64-bit reals) and do 7 radix-4 transforms...*/
                     /*                                      outputs                                      */ /*                          inputs                           */
	RADIX_04_DIT(a[j1    ],a[j1+1    ],a[j1+p01],a[j1+1+p01],a[j1+p03],a[j1+1+p03],a[j1+p02],a[j1+1+p02],aj1p00r,aj1p00i,aj1p03r,aj1p03i,aj1p02r,aj1p02i,aj1p01r,aj1p01i,rt,it);
	RADIX_04_DIT(a[j1+p15],a[j1+1+p15],a[j1+p14],a[j1+1+p14],a[j1+p13],a[j1+1+p13],a[j1+p12],a[j1+1+p12],aj1p04r,aj1p04i,aj1p07r,aj1p07i,aj1p06r,aj1p06i,aj1p05r,aj1p05i,rt,it);
	RADIX_04_DIT(a[j1+p25],a[j1+1+p25],a[j1+p24],a[j1+1+p24],a[j1+p26],a[j1+1+p26],a[j1+p27],a[j1+1+p27],aj1p08r,aj1p08i,aj1p11r,aj1p11i,aj1p10r,aj1p10i,aj1p09r,aj1p09i,rt,it);
	RADIX_04_DIT(a[j1+p09],a[j1+1+p09],a[j1+p08],a[j1+1+p08],a[j1+p10],a[j1+1+p10],a[j1+p11],a[j1+1+p11],aj1p12r,aj1p12i,aj1p15r,aj1p15i,aj1p14r,aj1p14i,aj1p13r,aj1p13i,rt,it);
	RADIX_04_DIT(a[j1+p22],a[j1+1+p22],a[j1+p23],a[j1+1+p23],a[j1+p20],a[j1+1+p20],a[j1+p21],a[j1+1+p21],aj1p16r,aj1p16i,aj1p19r,aj1p19i,aj1p18r,aj1p18i,aj1p17r,aj1p17i,rt,it);
	RADIX_04_DIT(a[j1+p06],a[j1+1+p06],a[j1+p07],a[j1+1+p07],a[j1+p04],a[j1+1+p04],a[j1+p05],a[j1+1+p05],aj1p20r,aj1p20i,aj1p23r,aj1p23i,aj1p22r,aj1p22i,aj1p21r,aj1p21i,rt,it);
	RADIX_04_DIT(a[j1+p16],a[j1+1+p16],a[j1+p17],a[j1+1+p17],a[j1+p19],a[j1+1+p19],a[j1+p18],a[j1+1+p18],aj1p24r,aj1p24i,aj1p27r,aj1p27i,aj1p26r,aj1p26i,aj1p25r,aj1p25i,rt,it);

/*...and now do 4 radix-4 transforms...*/
                     /*                                                   inputs                                                  */ /*               intermediates              */ /*                                                                     outputs                                                                         */
	RADIX_07_DFT(aj1p00r,aj1p00i,aj1p04r,aj1p04i,aj1p08r,aj1p08i,aj1p12r,aj1p12i,aj1p16r,aj1p16i,aj1p20r,aj1p20i,aj1p24r,aj1p24i,t1,t2,t3,t4,t5,t6,t7,t8,t9,t10,t11,t12,t13,t14,aj1p00r,aj1p00i,aj1p08r,aj1p08i,aj1p16r,aj1p16i,aj1p24r,aj1p24i,aj1p04r,aj1p04i,aj1p12r,aj1p12i,aj1p20r,aj1p20i,rt,it,re,im);
	RADIX_07_DFT(aj1p03r,aj1p03i,aj1p07r,aj1p07i,aj1p11r,aj1p11i,aj1p15r,aj1p15i,aj1p19r,aj1p19i,aj1p23r,aj1p23i,aj1p27r,aj1p27i,t1,t2,t3,t4,t5,t6,t7,t8,t9,t10,t11,t12,t13,t14,aj1p07r,aj1p07i,aj1p15r,aj1p15i,aj1p23r,aj1p23i,aj1p03r,aj1p03i,aj1p11r,aj1p11i,aj1p19r,aj1p19i,aj1p27r,aj1p27i,rt,it,re,im);
	RADIX_07_DFT(aj1p02r,aj1p02i,aj1p06r,aj1p06i,aj1p10r,aj1p10i,aj1p14r,aj1p14i,aj1p18r,aj1p18i,aj1p22r,aj1p22i,aj1p26r,aj1p26i,t1,t2,t3,t4,t5,t6,t7,t8,t9,t10,t11,t12,t13,t14,aj1p14r,aj1p14i,aj1p22r,aj1p22i,aj1p02r,aj1p02i,aj1p10r,aj1p10i,aj1p18r,aj1p18i,aj1p26r,aj1p26i,aj1p06r,aj1p06i,rt,it,re,im);
	RADIX_07_DFT(aj1p01r,aj1p01i,aj1p05r,aj1p05i,aj1p09r,aj1p09i,aj1p13r,aj1p13i,aj1p17r,aj1p17i,aj1p21r,aj1p21i,aj1p25r,aj1p25i,t1,t2,t3,t4,t5,t6,t7,t8,t9,t10,t11,t12,t13,t14,aj1p21r,aj1p21i,aj1p01r,aj1p01i,aj1p09r,aj1p09i,aj1p17r,aj1p17i,aj1p25r,aj1p25i,aj1p05r,aj1p05i,aj1p13r,aj1p13i,rt,it,re,im);

/*...Now do the carries. Since the outputs would
    normally be getting dispatched to 28 separate blocks of the A-array, we need 28 separate carries.	*/

	    l= j & (nwt-1);
	    n_minus_sil   = n-si[l  ];
	    n_minus_silp1 = n-si[l+1];
	    sinwt   = si[nwt-l  ];
	    sinwtm1 = si[nwt-l-1];

	    wtl     =wt0[    l  ];
	    wtn     =wt0[nwt-l  ]*scale;	/* Include 1/(n/2) scale factor of inverse transform here...	*/
	    wtlp1   =wt0[    l+1];
	    wtnm1   =wt0[nwt-l-1]*scale;	/* ...and here.	*/

/*...set0 is slightly different from others:	*/

	   cmplx_carry_norm_errcheck0(aj1p00r,aj1p00i,cy00,bjmodn00   );
	    cmplx_carry_norm_errcheck(aj1p01r,aj1p01i,cy01,bjmodn01,1 );
	    cmplx_carry_norm_errcheck(aj1p02r,aj1p02i,cy02,bjmodn02,2 );
	    cmplx_carry_norm_errcheck(aj1p03r,aj1p03i,cy03,bjmodn03,3 );
	    cmplx_carry_norm_errcheck(aj1p04r,aj1p04i,cy04,bjmodn04,4 );
	    cmplx_carry_norm_errcheck(aj1p05r,aj1p05i,cy05,bjmodn05,5 );
	    cmplx_carry_norm_errcheck(aj1p06r,aj1p06i,cy06,bjmodn06,6 );
	    cmplx_carry_norm_errcheck(aj1p07r,aj1p07i,cy07,bjmodn07,7 );
	    cmplx_carry_norm_errcheck(aj1p08r,aj1p08i,cy08,bjmodn08,8 );
	    cmplx_carry_norm_errcheck(aj1p09r,aj1p09i,cy09,bjmodn09,9 );
	    cmplx_carry_norm_errcheck(aj1p10r,aj1p10i,cy10,bjmodn10,10);
	    cmplx_carry_norm_errcheck(aj1p11r,aj1p11i,cy11,bjmodn11,11);
	    cmplx_carry_norm_errcheck(aj1p12r,aj1p12i,cy12,bjmodn12,12);
	    cmplx_carry_norm_errcheck(aj1p13r,aj1p13i,cy13,bjmodn13,13);
	    cmplx_carry_norm_errcheck(aj1p14r,aj1p14i,cy14,bjmodn14,14);
	    cmplx_carry_norm_errcheck(aj1p15r,aj1p15i,cy15,bjmodn15,15);
	    cmplx_carry_norm_errcheck(aj1p16r,aj1p16i,cy16,bjmodn16,16);
	    cmplx_carry_norm_errcheck(aj1p17r,aj1p17i,cy17,bjmodn17,17);
	    cmplx_carry_norm_errcheck(aj1p18r,aj1p18i,cy18,bjmodn18,18);
	    cmplx_carry_norm_errcheck(aj1p19r,aj1p19i,cy19,bjmodn19,19);
	    cmplx_carry_norm_errcheck(aj1p20r,aj1p20i,cy20,bjmodn20,20);
	    cmplx_carry_norm_errcheck(aj1p21r,aj1p21i,cy21,bjmodn21,21);
	    cmplx_carry_norm_errcheck(aj1p22r,aj1p22i,cy22,bjmodn22,22);
	    cmplx_carry_norm_errcheck(aj1p23r,aj1p23i,cy23,bjmodn23,23);
	    cmplx_carry_norm_errcheck(aj1p24r,aj1p24i,cy24,bjmodn24,24);
	    cmplx_carry_norm_errcheck(aj1p25r,aj1p25i,cy25,bjmodn25,25);
	    cmplx_carry_norm_errcheck(aj1p26r,aj1p26i,cy26,bjmodn26,26);
	    cmplx_carry_norm_errcheck(aj1p27r,aj1p27i,cy27,bjmodn27,27);

	    i =((uint32)(sw - bjmodn00) >> 31);	/* get ready for the next set...	*/
	    co2=co3;	/* For all data but the first set in each j-block, co2=co3. Thus, after the first block of data is done
			   and only then: for all subsequent blocks it's superfluous), this assignment decrements co2 by radix(1).	*/

/*...The radix-28 DIF pass is here:	*/
#if PFETCH
add0 = &a[j1];
prefetch_p_doubles(add0);
#endif

/*...gather the needed data (28 64-bit complex, i.e. 56 64-bit reals) and do 4 radix-7 transforms...*/
                     /*                                                                      inputs                                                                         */ /*               intermediates              */ /*                                                  outputs                                                  */
#if PFETCH
	RADIX_07_DFT_PFETCH(aj1p00r,aj1p00i,aj1p24r,aj1p24i,aj1p20r,aj1p20i,aj1p16r,aj1p16i,aj1p12r,aj1p12i,aj1p08r,aj1p08i,aj1p04r,aj1p04i,t1,t2,t3,t4,t5,t6,t7,t8,t9,t10,t11,t12,t13,t14,aj1p00r,aj1p00i,aj1p04r,aj1p04i,aj1p08r,aj1p08i,aj1p12r,aj1p12i,aj1p16r,aj1p16i,aj1p20r,aj1p20i,aj1p24r,aj1p24i,rt,it,re,im,p01,p02,p03);
	RADIX_07_DFT_PFETCH(aj1p21r,aj1p21i,aj1p17r,aj1p17i,aj1p13r,aj1p13i,aj1p09r,aj1p09i,aj1p05r,aj1p05i,aj1p01r,aj1p01i,aj1p25r,aj1p25i,t1,t2,t3,t4,t5,t6,t7,t8,t9,t10,t11,t12,t13,t14,aj1p01r,aj1p01i,aj1p05r,aj1p05i,aj1p09r,aj1p09i,aj1p13r,aj1p13i,aj1p17r,aj1p17i,aj1p21r,aj1p21i,aj1p25r,aj1p25i,rt,it,re,im,p04,p05,p06);
	RADIX_07_DFT_PFETCH(aj1p14r,aj1p14i,aj1p10r,aj1p10i,aj1p06r,aj1p06i,aj1p02r,aj1p02i,aj1p26r,aj1p26i,aj1p22r,aj1p22i,aj1p18r,aj1p18i,t1,t2,t3,t4,t5,t6,t7,t8,t9,t10,t11,t12,t13,t14,aj1p02r,aj1p02i,aj1p06r,aj1p06i,aj1p10r,aj1p10i,aj1p14r,aj1p14i,aj1p18r,aj1p18i,aj1p22r,aj1p22i,aj1p26r,aj1p26i,rt,it,re,im,p07,p08,p09);
	RADIX_07_DFT_PFETCH(aj1p07r,aj1p07i,aj1p03r,aj1p03i,aj1p27r,aj1p27i,aj1p23r,aj1p23i,aj1p19r,aj1p19i,aj1p15r,aj1p15i,aj1p11r,aj1p11i,t1,t2,t3,t4,t5,t6,t7,t8,t9,t10,t11,t12,t13,t14,aj1p03r,aj1p03i,aj1p07r,aj1p07i,aj1p11r,aj1p11i,aj1p15r,aj1p15i,aj1p19r,aj1p19i,aj1p23r,aj1p23i,aj1p27r,aj1p27i,rt,it,re,im,p10,p11,p12);
#else
	RADIX_07_DFT       (aj1p00r,aj1p00i,aj1p24r,aj1p24i,aj1p20r,aj1p20i,aj1p16r,aj1p16i,aj1p12r,aj1p12i,aj1p08r,aj1p08i,aj1p04r,aj1p04i,t1,t2,t3,t4,t5,t6,t7,t8,t9,t10,t11,t12,t13,t14,aj1p00r,aj1p00i,aj1p04r,aj1p04i,aj1p08r,aj1p08i,aj1p12r,aj1p12i,aj1p16r,aj1p16i,aj1p20r,aj1p20i,aj1p24r,aj1p24i,rt,it,re,im);
	RADIX_07_DFT       (aj1p21r,aj1p21i,aj1p17r,aj1p17i,aj1p13r,aj1p13i,aj1p09r,aj1p09i,aj1p05r,aj1p05i,aj1p01r,aj1p01i,aj1p25r,aj1p25i,t1,t2,t3,t4,t5,t6,t7,t8,t9,t10,t11,t12,t13,t14,aj1p01r,aj1p01i,aj1p05r,aj1p05i,aj1p09r,aj1p09i,aj1p13r,aj1p13i,aj1p17r,aj1p17i,aj1p21r,aj1p21i,aj1p25r,aj1p25i,rt,it,re,im);
	RADIX_07_DFT       (aj1p14r,aj1p14i,aj1p10r,aj1p10i,aj1p06r,aj1p06i,aj1p02r,aj1p02i,aj1p26r,aj1p26i,aj1p22r,aj1p22i,aj1p18r,aj1p18i,t1,t2,t3,t4,t5,t6,t7,t8,t9,t10,t11,t12,t13,t14,aj1p02r,aj1p02i,aj1p06r,aj1p06i,aj1p10r,aj1p10i,aj1p14r,aj1p14i,aj1p18r,aj1p18i,aj1p22r,aj1p22i,aj1p26r,aj1p26i,rt,it,re,im);
	RADIX_07_DFT       (aj1p07r,aj1p07i,aj1p03r,aj1p03i,aj1p27r,aj1p27i,aj1p23r,aj1p23i,aj1p19r,aj1p19i,aj1p15r,aj1p15i,aj1p11r,aj1p11i,t1,t2,t3,t4,t5,t6,t7,t8,t9,t10,t11,t12,t13,t14,aj1p03r,aj1p03i,aj1p07r,aj1p07i,aj1p11r,aj1p11i,aj1p15r,aj1p15i,aj1p19r,aj1p19i,aj1p23r,aj1p23i,aj1p27r,aj1p27i,rt,it,re,im);
#endif

/*...and now do 7 radix-4 transforms...*/
                     /*                          inputs                           */ /*                                      outputs                                      */
#if PFETCH
	add = add0+p13;
	prefetch_p_doubles(add);

	RADIX_04_DIF_PFETCH(aj1p00r,aj1p00i,aj1p01r,aj1p01i,aj1p02r,aj1p02i,aj1p03r,aj1p03i,a[j1    ],a[j1+1    ],a[j1+p01],a[j1+1+p01],a[j1+p02],a[j1+1+p02],a[j1+p03],a[j1+1+p03],rt,it,p14,p15);
	RADIX_04_DIF_PFETCH(aj1p04r,aj1p04i,aj1p05r,aj1p05i,aj1p06r,aj1p06i,aj1p07r,aj1p07i,a[j1+p25],a[j1+1+p25],a[j1+p24],a[j1+1+p24],a[j1+p27],a[j1+1+p27],a[j1+p26],a[j1+1+p26],rt,it,p16,p17);
	RADIX_04_DIF_PFETCH(aj1p08r,aj1p08i,aj1p09r,aj1p09i,aj1p10r,aj1p10i,aj1p11r,aj1p11i,a[j1+p22],a[j1+1+p22],a[j1+p23],a[j1+1+p23],a[j1+p21],a[j1+1+p21],a[j1+p20],a[j1+1+p20],rt,it,p18,p19);
	RADIX_04_DIF_PFETCH(aj1p12r,aj1p12i,aj1p13r,aj1p13i,aj1p14r,aj1p14i,aj1p15r,aj1p15i,a[j1+p16],a[j1+1+p16],a[j1+p17],a[j1+1+p17],a[j1+p18],a[j1+1+p18],a[j1+p19],a[j1+1+p19],rt,it,p20,p21);
	RADIX_04_DIF_PFETCH(aj1p16r,aj1p16i,aj1p17r,aj1p17i,aj1p18r,aj1p18i,aj1p19r,aj1p19i,a[j1+p15],a[j1+1+p15],a[j1+p14],a[j1+1+p14],a[j1+p12],a[j1+1+p12],a[j1+p13],a[j1+1+p13],rt,it,p22,p23);
	RADIX_04_DIF_PFETCH(aj1p20r,aj1p20i,aj1p21r,aj1p21i,aj1p22r,aj1p22i,aj1p23r,aj1p23i,a[j1+p09],a[j1+1+p09],a[j1+p08],a[j1+1+p08],a[j1+p11],a[j1+1+p11],a[j1+p10],a[j1+1+p10],rt,it,p24,p25);
	RADIX_04_DIF_PFETCH(aj1p24r,aj1p24i,aj1p25r,aj1p25i,aj1p26r,aj1p26i,aj1p27r,aj1p27i,a[j1+p06],a[j1+1+p06],a[j1+p07],a[j1+1+p07],a[j1+p05],a[j1+1+p05],a[j1+p04],a[j1+1+p04],rt,it,p26,p27);
#else
	RADIX_04_DIF       (aj1p00r,aj1p00i,aj1p01r,aj1p01i,aj1p02r,aj1p02i,aj1p03r,aj1p03i,a[j1    ],a[j1+1    ],a[j1+p01],a[j1+1+p01],a[j1+p02],a[j1+1+p02],a[j1+p03],a[j1+1+p03],rt,it);
	RADIX_04_DIF       (aj1p04r,aj1p04i,aj1p05r,aj1p05i,aj1p06r,aj1p06i,aj1p07r,aj1p07i,a[j1+p25],a[j1+1+p25],a[j1+p24],a[j1+1+p24],a[j1+p27],a[j1+1+p27],a[j1+p26],a[j1+1+p26],rt,it);
	RADIX_04_DIF       (aj1p08r,aj1p08i,aj1p09r,aj1p09i,aj1p10r,aj1p10i,aj1p11r,aj1p11i,a[j1+p22],a[j1+1+p22],a[j1+p23],a[j1+1+p23],a[j1+p21],a[j1+1+p21],a[j1+p20],a[j1+1+p20],rt,it);
	RADIX_04_DIF       (aj1p12r,aj1p12i,aj1p13r,aj1p13i,aj1p14r,aj1p14i,aj1p15r,aj1p15i,a[j1+p16],a[j1+1+p16],a[j1+p17],a[j1+1+p17],a[j1+p18],a[j1+1+p18],a[j1+p19],a[j1+1+p19],rt,it);
	RADIX_04_DIF       (aj1p16r,aj1p16i,aj1p17r,aj1p17i,aj1p18r,aj1p18i,aj1p19r,aj1p19i,a[j1+p15],a[j1+1+p15],a[j1+p14],a[j1+1+p14],a[j1+p12],a[j1+1+p12],a[j1+p13],a[j1+1+p13],rt,it);
	RADIX_04_DIF       (aj1p20r,aj1p20i,aj1p21r,aj1p21i,aj1p22r,aj1p22i,aj1p23r,aj1p23i,a[j1+p09],a[j1+1+p09],a[j1+p08],a[j1+1+p08],a[j1+p11],a[j1+1+p11],a[j1+p10],a[j1+1+p10],rt,it);
	RADIX_04_DIF       (aj1p24r,aj1p24i,aj1p25r,aj1p25i,aj1p26r,aj1p26i,aj1p27r,aj1p27i,a[j1+p06],a[j1+1+p06],a[j1+p07],a[j1+1+p07],a[j1+p05],a[j1+1+p05],a[j1+p04],a[j1+1+p04],rt,it);
#endif

	  iroot += root_incr;		/* increment sincos index.	*/

	  }

	  jstart += nwt;
	  jhi    += nwt;
	  col += 28;
	  co3 -= 28;

	}

	if(root_incr==0)break;

/*   Wraparound carry cleanup loop is here: ***
!
!   (1) Invert the radix-28 forward DIF FFT of the first block of 28 complex elements in A and unweight;
!   (2) Propagate cleanup carries among the real and imaginary parts of the 28 outputs of (1);
!   (3) Reweight and perform a radix-28 forward DIF FFT on the result of (2);
!   (4) If any of the exit carries from (2) are nonzero, advance to the next 28 elements and repeat (1-4).
*/
	t1  = cy27;
	cy27= cy26;
	cy26= cy25;
	cy25= cy24;
	cy24= cy23;
	cy23= cy22;
	cy22= cy21;
	cy21= cy20;
	cy20= cy19;
	cy19= cy18;
	cy18= cy17;
	cy17= cy16;
	cy16= cy15;
	cy15= cy14;
	cy14= cy13;
	cy13= cy12;
	cy12= cy11;
	cy11= cy10;
	cy10= cy09;
	cy09= cy08;
	cy08= cy07;
	cy07= cy06;
	cy06= cy05;
	cy05= cy04;
	cy04= cy03;
	cy03= cy02;
	cy02= cy01;
	cy01= cy00;
	cy00= t1  ;

	iroot = 0;
	root_incr = 0;
	scale = 1;

	jstart = 0;
	jhi = 3;
	khi = 1;

	for(j=0; j<=jhi; j++)
	{
	  a[j    ] *= radix_inv;
	  a[j+p01] *= radix_inv;
	  a[j+p02] *= radix_inv;
	  a[j+p03] *= radix_inv;
	  a[j+p04] *= radix_inv;
	  a[j+p05] *= radix_inv;
	  a[j+p06] *= radix_inv;
	  a[j+p07] *= radix_inv;
	  a[j+p08] *= radix_inv;
	  a[j+p09] *= radix_inv;
	  a[j+p10] *= radix_inv;
	  a[j+p11] *= radix_inv;
	  a[j+p12] *= radix_inv;
	  a[j+p13] *= radix_inv;
	  a[j+p14] *= radix_inv;
	  a[j+p15] *= radix_inv;
	  a[j+p16] *= radix_inv;
	  a[j+p17] *= radix_inv;
	  a[j+p18] *= radix_inv;
	  a[j+p19] *= radix_inv;
	  a[j+p20] *= radix_inv;
	  a[j+p21] *= radix_inv;
	  a[j+p22] *= radix_inv;
	  a[j+p23] *= radix_inv;
	  a[j+p24] *= radix_inv;
	  a[j+p25] *= radix_inv;
	  a[j+p26] *= radix_inv;
	  a[j+p27] *= radix_inv;
	}
}

	if(fabs(cy00)+fabs(cy01)+fabs(cy02)+fabs(cy03)+fabs(cy04)+fabs(cy05)+fabs(cy06)+fabs(cy07)+fabs(cy08)+fabs(cy09)+fabs(cy10)+fabs(cy11)+fabs(cy12)+fabs(cy13)+fabs(cy14)+fabs(cy15)+fabs(cy16)+fabs(cy17)+fabs(cy18)+fabs(cy19)+fabs(cy20)+fabs(cy21)+fabs(cy22)+fabs(cy23)+fabs(cy24)+fabs(cy25)+fabs(cy26)+fabs(cy27) != 0.0)
	{
	    if(INTERACT)printf("FATAL: iter = %10d; nonzero exit carry in radix28_ditN_cy_dif1.\n",iter);
	    f1 = fopen(   OFILE,"a");
	    f2 = fopen(STATFILE,"a");
	    fprintf(f1,"FATAL: iter = %10d; nonzero exit carry in radix28_ditN_cy_dif1.\n",iter);
	    fprintf(f2,"FATAL: iter = %10d; nonzero exit carry in radix28_ditN_cy_dif1.\n",iter);
	    fclose(f1);
	    fclose(f2);
	    err=1;
	    return(err);
	}
	*fracmax = maxerr;
	return(0);
}

/***************/

int radix28_ditN_cy_dif1_nochk(double a[], int n, int nwt, int nwt_bits, double wt0[], double wt1[], int si[], double base[], double baseinv[], int iter                 , int p)
{
/*
!...Acronym: DWT = Discrete Weighted Transform, DIT = Decimation In Time, DIF = Decimation In Frequency
!
!...Performs a final radix-28 complex DIT pass, an inverse DWT weighting, a carry propagation,
!   a forward DWT weighting, and an initial radix-28 complex DIF pass on the data in the length-N real vector A.
!
!   Data enter and are returned in the A-array.
!
!   See the documentation in mers_mod_square and radix16_dif_pass for further details on the array
!   storage scheme, and radix8_ditN_cy_dif1 for details on the reduced-length weights array scheme.
*/
	int n28, bjmodn00,bjmodn01,bjmodn02,bjmodn03,bjmodn04,bjmodn05,bjmodn06,bjmodn07,bjmodn08,bjmodn09,bjmodn10,bjmodn11,bjmodn12,bjmodn13,bjmodn14,bjmodn15,bjmodn16,bjmodn17,bjmodn18,bjmodn19,bjmodn20,bjmodn21,bjmodn22,bjmodn23,bjmodn24,bjmodn25,bjmodn26,bjmodn27
		,i,j,j1,jstart,jhi,iroot,root_incr,k,khi,l,outer;
	static int psave=0,bw,sw,bjmodnini,nrtm1,p01,p02,p03,p04,p05,p06,p07,p08,p09,p10,p11,p12,p13,p14,p15,p16,p17,p18,p19,p20,p21,p22,p23,p24,p25,p26,p27;
	const  double one_half[3] = {1.0, 0.5, 0.25};	/* Needed for small-weights-tables scheme */
	static double radix_inv, n2inv;
	static double cc1 = .62348980185873353053,	 /* cos(u) = Real part of exp(i*2*pi/7), the radix-7 fundamental sincos datum	*/
		      ss1 = .78183148246802980870,	 /* sin(u) = Imag part of exp(i*2*pi/7).	*/
		      cc2 =-.22252093395631440426,	 /* cos(2u)	*/
		      ss2 = .97492791218182360702,	 /* sin(2u)	*/
		      cc3 =-.90096886790241912622,	 /* cos(3u)	*/
		      ss3 = .43388373911755812050;	 /* sin(3u)	*/
	double re,im,rt,it
	,t1,t2,t3,t4,t5,t6,t7,t8,t9,t10,t11,t12,t13,t14
	,aj1p00r,aj1p01r,aj1p02r,aj1p03r,aj1p04r,aj1p05r,aj1p06r,aj1p07r,aj1p08r,aj1p09r,aj1p10r,aj1p11r,aj1p12r,aj1p13r,aj1p14r,aj1p15r,aj1p16r,aj1p17r,aj1p18r,aj1p19r,aj1p20r,aj1p21r,aj1p22r,aj1p23r,aj1p24r,aj1p25r,aj1p26r,aj1p27r
	,aj1p00i,aj1p01i,aj1p02i,aj1p03i,aj1p04i,aj1p05i,aj1p06i,aj1p07i,aj1p08i,aj1p09i,aj1p10i,aj1p11i,aj1p12i,aj1p13i,aj1p14i,aj1p15i,aj1p16i,aj1p17i,aj1p18i,aj1p19i,aj1p20i,aj1p21i,aj1p22i,aj1p23i,aj1p24i,aj1p25i,aj1p26i,aj1p27i
	,cy00,cy01,cy02,cy03,cy04,cy05,cy06,cy07,cy08,cy09,cy10,cy11,cy12,cy13,cy14,cy15,cy16,cy17,cy18,cy19,cy20,cy21,cy22,cy23,cy24,cy25,cy26,cy27,temp,scale;
#if PFETCH
	double *add0, *add;
#endif
	int err;
	FILE *f1,*f2;
	static int first_entry=TRUE;

/*...stuff for the reduced-length DWT weights array is here:	*/
	int n_div_nwt;
	int col,co2,co3,m,m2,n_minus_sil,n_minus_silp1,sinwt,sinwtm1;
	double wt,wtinv,wtl,wtlp1,wtn,wtnm1,wtA,wtB,wtC;

/*...change n28 and n_div_wt to non-static to work around a gcc compiler bug. */
	n28   = n/28;
	n_div_nwt = n28 >> nwt_bits;

	if((n_div_nwt << nwt_bits) != n28)
	{
	  if(INTERACT)printf("FATAL: iter = %10d; NWT_BITS does not divide N/28 in radix28_ditN_cy_dif1.\n",iter);
	  f1 = fopen(   OFILE,"a");
	  f2 = fopen(STATFILE,"a");
	  fprintf(f1,"FATAL: iter = %10d; NWT_BITS does not divide N/28 in radix28_ditN_cy_dif1.\n",iter);
	  fprintf(f2,"FATAL: iter = %10d; NWT_BITS does not divide N/28 in radix28_ditN_cy_dif1.\n",iter);
	  fclose(f1);
	  fclose(f2);
	  err=1;
	  return(err);
	}

	if(p != psave)
	{
	  first_entry=TRUE;
	}

/*...initialize things upon first entry	*/

	if(first_entry)
	{
	  psave = p;
	  first_entry=FALSE;
	  radix_inv = (double)((long double)1.0L/(long double)28.0L);
	  n2inv     = (double)((long double)1.0L/(long double)(n/2));

	  bw    = p%n;		/* Number of bigwords in the Crandall/Fagin mixed-radix representation = (Mersenne exponent) mod (vector length).	*/
	  sw    = n - bw;	/* Number of smallwords.	*/

/*   constant index offsets for array load/stores are here.	*/

	  p01 = n28;
	  p02 = p01 + p01;
	  p03 = p02 + p01;
	  p04 = p03 + p01;
	  p05 = p04 + p01;
	  p06 = p05 + p01;
	  p07 = p06 + p01;
	  p08 = p07 + p01;
	  p09 = p08 + p01;
	  p10 = p09 + p01;
	  p11 = p10 + p01;
	  p12 = p11 + p01;
	  p13 = p12 + p01;
	  p14 = p13 + p01;
	  p15 = p14 + p01;
	  p16 = p15 + p01;
	  p17 = p16 + p01;
	  p18 = p17 + p01;
	  p19 = p18 + p01;
	  p20 = p19 + p01;
	  p21 = p20 + p01;
	  p22 = p21 + p01;
	  p23 = p22 + p01;
	  p24 = p23 + p01;
	  p25 = p24 + p01;
	  p26 = p25 + p01;
	  p27 = p26 + p01;

	  p01 = p01 + ( (p01 >> DAT_BITS) << PAD_BITS );
	  p02 = p02 + ( (p02 >> DAT_BITS) << PAD_BITS );
	  p03 = p03 + ( (p03 >> DAT_BITS) << PAD_BITS );
	  p04 = p04 + ( (p04 >> DAT_BITS) << PAD_BITS );
	  p05 = p05 + ( (p05 >> DAT_BITS) << PAD_BITS );
	  p06 = p06 + ( (p06 >> DAT_BITS) << PAD_BITS );
	  p07 = p07 + ( (p07 >> DAT_BITS) << PAD_BITS );
	  p08 = p08 + ( (p08 >> DAT_BITS) << PAD_BITS );
	  p09 = p09 + ( (p09 >> DAT_BITS) << PAD_BITS );
	  p10 = p10 + ( (p10 >> DAT_BITS) << PAD_BITS );
	  p11 = p11 + ( (p11 >> DAT_BITS) << PAD_BITS );
	  p12 = p12 + ( (p12 >> DAT_BITS) << PAD_BITS );
	  p13 = p13 + ( (p13 >> DAT_BITS) << PAD_BITS );
	  p14 = p14 + ( (p14 >> DAT_BITS) << PAD_BITS );
	  p15 = p15 + ( (p15 >> DAT_BITS) << PAD_BITS );
	  p16 = p16 + ( (p16 >> DAT_BITS) << PAD_BITS );
	  p17 = p17 + ( (p17 >> DAT_BITS) << PAD_BITS );
	  p18 = p18 + ( (p18 >> DAT_BITS) << PAD_BITS );
	  p19 = p19 + ( (p19 >> DAT_BITS) << PAD_BITS );
	  p20 = p20 + ( (p20 >> DAT_BITS) << PAD_BITS );
	  p21 = p21 + ( (p21 >> DAT_BITS) << PAD_BITS );
	  p22 = p22 + ( (p22 >> DAT_BITS) << PAD_BITS );
	  p23 = p23 + ( (p23 >> DAT_BITS) << PAD_BITS );
	  p24 = p24 + ( (p24 >> DAT_BITS) << PAD_BITS );
	  p25 = p25 + ( (p25 >> DAT_BITS) << PAD_BITS );
	  p26 = p26 + ( (p26 >> DAT_BITS) << PAD_BITS );
	  p27 = p27 + ( (p27 >> DAT_BITS) << PAD_BITS );

	  bjmodnini=0;
	  for(j=0; j < n28; j++)
	  {
	    bjmodnini -= sw; bjmodnini = bjmodnini + ( (-(int)((uint32)bjmodnini >> 31)) & n);
	  }
	}

/*...The radix-28 final DIT pass is here.	*/

	cy00=-2;		/* init carries	*/
	cy01= 0;
	cy02= 0;
	cy03= 0;
	cy04= 0;
	cy05= 0;
	cy06= 0;
	cy07= 0;
	cy08= 0;
	cy09= 0;
	cy10= 0;
	cy11= 0;
	cy12= 0;
	cy13= 0;
	cy14= 0;
	cy15= 0;
	cy16= 0;
	cy17= 0;
	cy18= 0;
	cy19= 0;
	cy20= 0;
	cy21= 0;
	cy22= 0;
	cy23= 0;
	cy24= 0;
	cy25= 0;
	cy26= 0;
	cy27= 0;

	iroot = 0;	/* init sincos array index	*/
	root_incr = 1;	/* init sincos array index increment (set = 1 for normal carry pass, = 0 for wrapper pass)	*/
	scale = n2inv;	/* init inverse-weight scale factor  (set = 2/n for normal carry pass, = 1 for wrapper pass)	*/

	jstart = 0;
	jhi = jstart+nwt-1;
	khi = n_div_nwt;

for(outer=0; outer <= 1; outer++)
{
	i = 1;		/* Pointer to the BASE and BASEINV arrays. lowest-order digit is always a bigword (i = 1).	*/

	bjmodn00= 0;
	bjmodn01= bjmodnini;
	bjmodn02= bjmodn01+bjmodnini-n; bjmodn02= bjmodn02+ ( (-(int)((uint32)bjmodn02>> 31)) & n);
	bjmodn03= bjmodn02+bjmodnini-n; bjmodn03= bjmodn03+ ( (-(int)((uint32)bjmodn03>> 31)) & n);
	bjmodn04= bjmodn03+bjmodnini-n; bjmodn04= bjmodn04+ ( (-(int)((uint32)bjmodn04>> 31)) & n);
	bjmodn05= bjmodn04+bjmodnini-n; bjmodn05= bjmodn05+ ( (-(int)((uint32)bjmodn05>> 31)) & n);
	bjmodn06= bjmodn05+bjmodnini-n; bjmodn06= bjmodn06+ ( (-(int)((uint32)bjmodn06>> 31)) & n);
	bjmodn07= bjmodn06+bjmodnini-n; bjmodn07= bjmodn07+ ( (-(int)((uint32)bjmodn07>> 31)) & n);
	bjmodn08= bjmodn07+bjmodnini-n; bjmodn08= bjmodn08+ ( (-(int)((uint32)bjmodn08>> 31)) & n);
	bjmodn09= bjmodn08+bjmodnini-n; bjmodn09= bjmodn09+ ( (-(int)((uint32)bjmodn09>> 31)) & n);
	bjmodn10= bjmodn09+bjmodnini-n; bjmodn10= bjmodn10+ ( (-(int)((uint32)bjmodn10>> 31)) & n);
	bjmodn11= bjmodn10+bjmodnini-n; bjmodn11= bjmodn11+ ( (-(int)((uint32)bjmodn11>> 31)) & n);
	bjmodn12= bjmodn11+bjmodnini-n; bjmodn12= bjmodn12+ ( (-(int)((uint32)bjmodn12>> 31)) & n);
	bjmodn13= bjmodn12+bjmodnini-n; bjmodn13= bjmodn13+ ( (-(int)((uint32)bjmodn13>> 31)) & n);
	bjmodn14= bjmodn13+bjmodnini-n; bjmodn14= bjmodn14+ ( (-(int)((uint32)bjmodn14>> 31)) & n);
	bjmodn15= bjmodn14+bjmodnini-n; bjmodn15= bjmodn15+ ( (-(int)((uint32)bjmodn15>> 31)) & n);
	bjmodn16= bjmodn15+bjmodnini-n; bjmodn16= bjmodn16+ ( (-(int)((uint32)bjmodn16>> 31)) & n);
	bjmodn17= bjmodn16+bjmodnini-n; bjmodn17= bjmodn17+ ( (-(int)((uint32)bjmodn17>> 31)) & n);
	bjmodn18= bjmodn17+bjmodnini-n; bjmodn18= bjmodn18+ ( (-(int)((uint32)bjmodn18>> 31)) & n);
	bjmodn19= bjmodn18+bjmodnini-n; bjmodn19= bjmodn19+ ( (-(int)((uint32)bjmodn19>> 31)) & n);
	bjmodn20= bjmodn19+bjmodnini-n; bjmodn20= bjmodn20+ ( (-(int)((uint32)bjmodn20>> 31)) & n);
	bjmodn21= bjmodn20+bjmodnini-n; bjmodn21= bjmodn21+ ( (-(int)((uint32)bjmodn21>> 31)) & n);
	bjmodn22= bjmodn21+bjmodnini-n; bjmodn22= bjmodn22+ ( (-(int)((uint32)bjmodn22>> 31)) & n);
	bjmodn23= bjmodn22+bjmodnini-n; bjmodn23= bjmodn23+ ( (-(int)((uint32)bjmodn23>> 31)) & n);
	bjmodn24= bjmodn23+bjmodnini-n; bjmodn24= bjmodn24+ ( (-(int)((uint32)bjmodn24>> 31)) & n);
	bjmodn25= bjmodn24+bjmodnini-n; bjmodn25= bjmodn25+ ( (-(int)((uint32)bjmodn25>> 31)) & n);
	bjmodn26= bjmodn25+bjmodnini-n; bjmodn26= bjmodn26+ ( (-(int)((uint32)bjmodn26>> 31)) & n);
	bjmodn27= bjmodn26+bjmodnini-n; bjmodn27= bjmodn27+ ( (-(int)((uint32)bjmodn27>> 31)) & n);

	col=0;
	co2=(n >> nwt_bits)-1+28;
	co3=co2-28;		/* At the start of each new j-loop, co3=co2-radix(1)	*/

	for(k=1; k <= khi; k++)	/* Do n/(radix(1)*nwt) outer loop executions...	*/
	{
	  for(j=jstart; j<jhi; j += 2)	/* Each inner loop execution processes (radix(1)*nwt) array data.	*/
	  {
	    j1 = j + ( (j >> DAT_BITS) << PAD_BITS );	/* padded-array fetch index is here	*/

/*...The radix-28 DIT pass is here:	*/

/* EWM: 10/18/04: We swap the odd-index outputs of each of the radix-4 DIT transforms (1<=>3, 5<=>7, etc.) so that the indexing
                  of all the radix-7 transforms (really just the 2nd and 4th of these) winds up being in-place. This allows us
                  to properly re-use the ajp1 variables in the carry-pass version of this routine.
*/

/*...gather the needed data (28 64-bit complex, i.e. 56 64-bit reals) and do 7 radix-4 transforms...*/
                     /*                                      outputs                                      */ /*                          inputs                           */
	RADIX_04_DIT(a[j1    ],a[j1+1    ],a[j1+p01],a[j1+1+p01],a[j1+p03],a[j1+1+p03],a[j1+p02],a[j1+1+p02],aj1p00r,aj1p00i,aj1p03r,aj1p03i,aj1p02r,aj1p02i,aj1p01r,aj1p01i,rt,it);
	RADIX_04_DIT(a[j1+p15],a[j1+1+p15],a[j1+p14],a[j1+1+p14],a[j1+p13],a[j1+1+p13],a[j1+p12],a[j1+1+p12],aj1p04r,aj1p04i,aj1p07r,aj1p07i,aj1p06r,aj1p06i,aj1p05r,aj1p05i,rt,it);
	RADIX_04_DIT(a[j1+p25],a[j1+1+p25],a[j1+p24],a[j1+1+p24],a[j1+p26],a[j1+1+p26],a[j1+p27],a[j1+1+p27],aj1p08r,aj1p08i,aj1p11r,aj1p11i,aj1p10r,aj1p10i,aj1p09r,aj1p09i,rt,it);
	RADIX_04_DIT(a[j1+p09],a[j1+1+p09],a[j1+p08],a[j1+1+p08],a[j1+p10],a[j1+1+p10],a[j1+p11],a[j1+1+p11],aj1p12r,aj1p12i,aj1p15r,aj1p15i,aj1p14r,aj1p14i,aj1p13r,aj1p13i,rt,it);
	RADIX_04_DIT(a[j1+p22],a[j1+1+p22],a[j1+p23],a[j1+1+p23],a[j1+p20],a[j1+1+p20],a[j1+p21],a[j1+1+p21],aj1p16r,aj1p16i,aj1p19r,aj1p19i,aj1p18r,aj1p18i,aj1p17r,aj1p17i,rt,it);
	RADIX_04_DIT(a[j1+p06],a[j1+1+p06],a[j1+p07],a[j1+1+p07],a[j1+p04],a[j1+1+p04],a[j1+p05],a[j1+1+p05],aj1p20r,aj1p20i,aj1p23r,aj1p23i,aj1p22r,aj1p22i,aj1p21r,aj1p21i,rt,it);
	RADIX_04_DIT(a[j1+p16],a[j1+1+p16],a[j1+p17],a[j1+1+p17],a[j1+p19],a[j1+1+p19],a[j1+p18],a[j1+1+p18],aj1p24r,aj1p24i,aj1p27r,aj1p27i,aj1p26r,aj1p26i,aj1p25r,aj1p25i,rt,it);

/*...and now do 4 radix-4 transforms...*/
                     /*                                                   inputs                                                  */ /*               intermediates              */ /*                                                                     outputs                                                                         */
	RADIX_07_DFT(aj1p00r,aj1p00i,aj1p04r,aj1p04i,aj1p08r,aj1p08i,aj1p12r,aj1p12i,aj1p16r,aj1p16i,aj1p20r,aj1p20i,aj1p24r,aj1p24i,t1,t2,t3,t4,t5,t6,t7,t8,t9,t10,t11,t12,t13,t14,aj1p00r,aj1p00i,aj1p08r,aj1p08i,aj1p16r,aj1p16i,aj1p24r,aj1p24i,aj1p04r,aj1p04i,aj1p12r,aj1p12i,aj1p20r,aj1p20i,rt,it,re,im);
	RADIX_07_DFT(aj1p03r,aj1p03i,aj1p07r,aj1p07i,aj1p11r,aj1p11i,aj1p15r,aj1p15i,aj1p19r,aj1p19i,aj1p23r,aj1p23i,aj1p27r,aj1p27i,t1,t2,t3,t4,t5,t6,t7,t8,t9,t10,t11,t12,t13,t14,aj1p07r,aj1p07i,aj1p15r,aj1p15i,aj1p23r,aj1p23i,aj1p03r,aj1p03i,aj1p11r,aj1p11i,aj1p19r,aj1p19i,aj1p27r,aj1p27i,rt,it,re,im);
	RADIX_07_DFT(aj1p02r,aj1p02i,aj1p06r,aj1p06i,aj1p10r,aj1p10i,aj1p14r,aj1p14i,aj1p18r,aj1p18i,aj1p22r,aj1p22i,aj1p26r,aj1p26i,t1,t2,t3,t4,t5,t6,t7,t8,t9,t10,t11,t12,t13,t14,aj1p14r,aj1p14i,aj1p22r,aj1p22i,aj1p02r,aj1p02i,aj1p10r,aj1p10i,aj1p18r,aj1p18i,aj1p26r,aj1p26i,aj1p06r,aj1p06i,rt,it,re,im);
	RADIX_07_DFT(aj1p01r,aj1p01i,aj1p05r,aj1p05i,aj1p09r,aj1p09i,aj1p13r,aj1p13i,aj1p17r,aj1p17i,aj1p21r,aj1p21i,aj1p25r,aj1p25i,t1,t2,t3,t4,t5,t6,t7,t8,t9,t10,t11,t12,t13,t14,aj1p21r,aj1p21i,aj1p01r,aj1p01i,aj1p09r,aj1p09i,aj1p17r,aj1p17i,aj1p25r,aj1p25i,aj1p05r,aj1p05i,aj1p13r,aj1p13i,rt,it,re,im);

/*...Now do the carries. Since the outputs would
    normally be getting dispatched to 28 separate blocks of the A-array, we need 28 separate carries.	*/

	    l= j & (nwt-1);
	    n_minus_sil   = n-si[l  ];
	    n_minus_silp1 = n-si[l+1];
	    sinwt   = si[nwt-l  ];
	    sinwtm1 = si[nwt-l-1];

	    wtl     =wt0[    l  ];
	    wtn     =wt0[nwt-l  ]*scale;	/* Include 1/(n/2) scale factor of inverse transform here...	*/
	    wtlp1   =wt0[    l+1];
	    wtnm1   =wt0[nwt-l-1]*scale;	/* ...and here.	*/

/*...set0 is slightly different from others:	*/

	   cmplx_carry_norm_nocheck0(aj1p00r,aj1p00i,cy00,bjmodn00   );
	    cmplx_carry_norm_nocheck(aj1p01r,aj1p01i,cy01,bjmodn01,1 );
	    cmplx_carry_norm_nocheck(aj1p02r,aj1p02i,cy02,bjmodn02,2 );
	    cmplx_carry_norm_nocheck(aj1p03r,aj1p03i,cy03,bjmodn03,3 );
	    cmplx_carry_norm_nocheck(aj1p04r,aj1p04i,cy04,bjmodn04,4 );
	    cmplx_carry_norm_nocheck(aj1p05r,aj1p05i,cy05,bjmodn05,5 );
	    cmplx_carry_norm_nocheck(aj1p06r,aj1p06i,cy06,bjmodn06,6 );
	    cmplx_carry_norm_nocheck(aj1p07r,aj1p07i,cy07,bjmodn07,7 );
	    cmplx_carry_norm_nocheck(aj1p08r,aj1p08i,cy08,bjmodn08,8 );
	    cmplx_carry_norm_nocheck(aj1p09r,aj1p09i,cy09,bjmodn09,9 );
	    cmplx_carry_norm_nocheck(aj1p10r,aj1p10i,cy10,bjmodn10,10);
	    cmplx_carry_norm_nocheck(aj1p11r,aj1p11i,cy11,bjmodn11,11);
	    cmplx_carry_norm_nocheck(aj1p12r,aj1p12i,cy12,bjmodn12,12);
	    cmplx_carry_norm_nocheck(aj1p13r,aj1p13i,cy13,bjmodn13,13);
	    cmplx_carry_norm_nocheck(aj1p14r,aj1p14i,cy14,bjmodn14,14);
	    cmplx_carry_norm_nocheck(aj1p15r,aj1p15i,cy15,bjmodn15,15);
	    cmplx_carry_norm_nocheck(aj1p16r,aj1p16i,cy16,bjmodn16,16);
	    cmplx_carry_norm_nocheck(aj1p17r,aj1p17i,cy17,bjmodn17,17);
	    cmplx_carry_norm_nocheck(aj1p18r,aj1p18i,cy18,bjmodn18,18);
	    cmplx_carry_norm_nocheck(aj1p19r,aj1p19i,cy19,bjmodn19,19);
	    cmplx_carry_norm_nocheck(aj1p20r,aj1p20i,cy20,bjmodn20,20);
	    cmplx_carry_norm_nocheck(aj1p21r,aj1p21i,cy21,bjmodn21,21);
	    cmplx_carry_norm_nocheck(aj1p22r,aj1p22i,cy22,bjmodn22,22);
	    cmplx_carry_norm_nocheck(aj1p23r,aj1p23i,cy23,bjmodn23,23);
	    cmplx_carry_norm_nocheck(aj1p24r,aj1p24i,cy24,bjmodn24,24);
	    cmplx_carry_norm_nocheck(aj1p25r,aj1p25i,cy25,bjmodn25,25);
	    cmplx_carry_norm_nocheck(aj1p26r,aj1p26i,cy26,bjmodn26,26);
	    cmplx_carry_norm_nocheck(aj1p27r,aj1p27i,cy27,bjmodn27,27);

	    i =((uint32)(sw - bjmodn00) >> 31);	/* get ready for the next set...	*/
	    co2=co3;	/* For all data but the first set in each j-block, co2=co3. Thus, after the first block of data is done
			   and only then: for all subsequent blocks it's superfluous), this assignment decrements co2 by radix(1).	*/

/*...The radix-28 DIF pass is here:	*/
#if PFETCH
add0 = &a[j1];
prefetch_p_doubles(add0);
#endif

/*...gather the needed data (28 64-bit complex, i.e. 56 64-bit reals) and do 4 radix-7 transforms...*/
                     /*                                                                      inputs                                                                         */ /*               intermediates              */ /*                                                  outputs                                                  */
#if PFETCH
	RADIX_07_DFT_PFETCH(aj1p00r,aj1p00i,aj1p24r,aj1p24i,aj1p20r,aj1p20i,aj1p16r,aj1p16i,aj1p12r,aj1p12i,aj1p08r,aj1p08i,aj1p04r,aj1p04i,t1,t2,t3,t4,t5,t6,t7,t8,t9,t10,t11,t12,t13,t14,aj1p00r,aj1p00i,aj1p04r,aj1p04i,aj1p08r,aj1p08i,aj1p12r,aj1p12i,aj1p16r,aj1p16i,aj1p20r,aj1p20i,aj1p24r,aj1p24i,rt,it,re,im,p01,p02,p03);
	RADIX_07_DFT_PFETCH(aj1p21r,aj1p21i,aj1p17r,aj1p17i,aj1p13r,aj1p13i,aj1p09r,aj1p09i,aj1p05r,aj1p05i,aj1p01r,aj1p01i,aj1p25r,aj1p25i,t1,t2,t3,t4,t5,t6,t7,t8,t9,t10,t11,t12,t13,t14,aj1p01r,aj1p01i,aj1p05r,aj1p05i,aj1p09r,aj1p09i,aj1p13r,aj1p13i,aj1p17r,aj1p17i,aj1p21r,aj1p21i,aj1p25r,aj1p25i,rt,it,re,im,p04,p05,p06);
	RADIX_07_DFT_PFETCH(aj1p14r,aj1p14i,aj1p10r,aj1p10i,aj1p06r,aj1p06i,aj1p02r,aj1p02i,aj1p26r,aj1p26i,aj1p22r,aj1p22i,aj1p18r,aj1p18i,t1,t2,t3,t4,t5,t6,t7,t8,t9,t10,t11,t12,t13,t14,aj1p02r,aj1p02i,aj1p06r,aj1p06i,aj1p10r,aj1p10i,aj1p14r,aj1p14i,aj1p18r,aj1p18i,aj1p22r,aj1p22i,aj1p26r,aj1p26i,rt,it,re,im,p07,p08,p09);
	RADIX_07_DFT_PFETCH(aj1p07r,aj1p07i,aj1p03r,aj1p03i,aj1p27r,aj1p27i,aj1p23r,aj1p23i,aj1p19r,aj1p19i,aj1p15r,aj1p15i,aj1p11r,aj1p11i,t1,t2,t3,t4,t5,t6,t7,t8,t9,t10,t11,t12,t13,t14,aj1p03r,aj1p03i,aj1p07r,aj1p07i,aj1p11r,aj1p11i,aj1p15r,aj1p15i,aj1p19r,aj1p19i,aj1p23r,aj1p23i,aj1p27r,aj1p27i,rt,it,re,im,p10,p11,p12);
#else
	RADIX_07_DFT       (aj1p00r,aj1p00i,aj1p24r,aj1p24i,aj1p20r,aj1p20i,aj1p16r,aj1p16i,aj1p12r,aj1p12i,aj1p08r,aj1p08i,aj1p04r,aj1p04i,t1,t2,t3,t4,t5,t6,t7,t8,t9,t10,t11,t12,t13,t14,aj1p00r,aj1p00i,aj1p04r,aj1p04i,aj1p08r,aj1p08i,aj1p12r,aj1p12i,aj1p16r,aj1p16i,aj1p20r,aj1p20i,aj1p24r,aj1p24i,rt,it,re,im);
	RADIX_07_DFT       (aj1p21r,aj1p21i,aj1p17r,aj1p17i,aj1p13r,aj1p13i,aj1p09r,aj1p09i,aj1p05r,aj1p05i,aj1p01r,aj1p01i,aj1p25r,aj1p25i,t1,t2,t3,t4,t5,t6,t7,t8,t9,t10,t11,t12,t13,t14,aj1p01r,aj1p01i,aj1p05r,aj1p05i,aj1p09r,aj1p09i,aj1p13r,aj1p13i,aj1p17r,aj1p17i,aj1p21r,aj1p21i,aj1p25r,aj1p25i,rt,it,re,im);
	RADIX_07_DFT       (aj1p14r,aj1p14i,aj1p10r,aj1p10i,aj1p06r,aj1p06i,aj1p02r,aj1p02i,aj1p26r,aj1p26i,aj1p22r,aj1p22i,aj1p18r,aj1p18i,t1,t2,t3,t4,t5,t6,t7,t8,t9,t10,t11,t12,t13,t14,aj1p02r,aj1p02i,aj1p06r,aj1p06i,aj1p10r,aj1p10i,aj1p14r,aj1p14i,aj1p18r,aj1p18i,aj1p22r,aj1p22i,aj1p26r,aj1p26i,rt,it,re,im);
	RADIX_07_DFT       (aj1p07r,aj1p07i,aj1p03r,aj1p03i,aj1p27r,aj1p27i,aj1p23r,aj1p23i,aj1p19r,aj1p19i,aj1p15r,aj1p15i,aj1p11r,aj1p11i,t1,t2,t3,t4,t5,t6,t7,t8,t9,t10,t11,t12,t13,t14,aj1p03r,aj1p03i,aj1p07r,aj1p07i,aj1p11r,aj1p11i,aj1p15r,aj1p15i,aj1p19r,aj1p19i,aj1p23r,aj1p23i,aj1p27r,aj1p27i,rt,it,re,im);
#endif

/*...and now do 7 radix-4 transforms...*/
                     /*                          inputs                           */ /*                                      outputs                                      */
#if PFETCH
	add = add0+p13;
	prefetch_p_doubles(add);

	RADIX_04_DIF_PFETCH(aj1p00r,aj1p00i,aj1p01r,aj1p01i,aj1p02r,aj1p02i,aj1p03r,aj1p03i,a[j1    ],a[j1+1    ],a[j1+p01],a[j1+1+p01],a[j1+p02],a[j1+1+p02],a[j1+p03],a[j1+1+p03],rt,it,p14,p15);
	RADIX_04_DIF_PFETCH(aj1p04r,aj1p04i,aj1p05r,aj1p05i,aj1p06r,aj1p06i,aj1p07r,aj1p07i,a[j1+p25],a[j1+1+p25],a[j1+p24],a[j1+1+p24],a[j1+p27],a[j1+1+p27],a[j1+p26],a[j1+1+p26],rt,it,p16,p17);
	RADIX_04_DIF_PFETCH(aj1p08r,aj1p08i,aj1p09r,aj1p09i,aj1p10r,aj1p10i,aj1p11r,aj1p11i,a[j1+p22],a[j1+1+p22],a[j1+p23],a[j1+1+p23],a[j1+p21],a[j1+1+p21],a[j1+p20],a[j1+1+p20],rt,it,p18,p19);
	RADIX_04_DIF_PFETCH(aj1p12r,aj1p12i,aj1p13r,aj1p13i,aj1p14r,aj1p14i,aj1p15r,aj1p15i,a[j1+p16],a[j1+1+p16],a[j1+p17],a[j1+1+p17],a[j1+p18],a[j1+1+p18],a[j1+p19],a[j1+1+p19],rt,it,p20,p21);
	RADIX_04_DIF_PFETCH(aj1p16r,aj1p16i,aj1p17r,aj1p17i,aj1p18r,aj1p18i,aj1p19r,aj1p19i,a[j1+p15],a[j1+1+p15],a[j1+p14],a[j1+1+p14],a[j1+p12],a[j1+1+p12],a[j1+p13],a[j1+1+p13],rt,it,p22,p23);
	RADIX_04_DIF_PFETCH(aj1p20r,aj1p20i,aj1p21r,aj1p21i,aj1p22r,aj1p22i,aj1p23r,aj1p23i,a[j1+p09],a[j1+1+p09],a[j1+p08],a[j1+1+p08],a[j1+p11],a[j1+1+p11],a[j1+p10],a[j1+1+p10],rt,it,p24,p25);
	RADIX_04_DIF_PFETCH(aj1p24r,aj1p24i,aj1p25r,aj1p25i,aj1p26r,aj1p26i,aj1p27r,aj1p27i,a[j1+p06],a[j1+1+p06],a[j1+p07],a[j1+1+p07],a[j1+p05],a[j1+1+p05],a[j1+p04],a[j1+1+p04],rt,it,p26,p27);
#else
	RADIX_04_DIF       (aj1p00r,aj1p00i,aj1p01r,aj1p01i,aj1p02r,aj1p02i,aj1p03r,aj1p03i,a[j1    ],a[j1+1    ],a[j1+p01],a[j1+1+p01],a[j1+p02],a[j1+1+p02],a[j1+p03],a[j1+1+p03],rt,it);
	RADIX_04_DIF       (aj1p04r,aj1p04i,aj1p05r,aj1p05i,aj1p06r,aj1p06i,aj1p07r,aj1p07i,a[j1+p25],a[j1+1+p25],a[j1+p24],a[j1+1+p24],a[j1+p27],a[j1+1+p27],a[j1+p26],a[j1+1+p26],rt,it);
	RADIX_04_DIF       (aj1p08r,aj1p08i,aj1p09r,aj1p09i,aj1p10r,aj1p10i,aj1p11r,aj1p11i,a[j1+p22],a[j1+1+p22],a[j1+p23],a[j1+1+p23],a[j1+p21],a[j1+1+p21],a[j1+p20],a[j1+1+p20],rt,it);
	RADIX_04_DIF       (aj1p12r,aj1p12i,aj1p13r,aj1p13i,aj1p14r,aj1p14i,aj1p15r,aj1p15i,a[j1+p16],a[j1+1+p16],a[j1+p17],a[j1+1+p17],a[j1+p18],a[j1+1+p18],a[j1+p19],a[j1+1+p19],rt,it);
	RADIX_04_DIF       (aj1p16r,aj1p16i,aj1p17r,aj1p17i,aj1p18r,aj1p18i,aj1p19r,aj1p19i,a[j1+p15],a[j1+1+p15],a[j1+p14],a[j1+1+p14],a[j1+p12],a[j1+1+p12],a[j1+p13],a[j1+1+p13],rt,it);
	RADIX_04_DIF       (aj1p20r,aj1p20i,aj1p21r,aj1p21i,aj1p22r,aj1p22i,aj1p23r,aj1p23i,a[j1+p09],a[j1+1+p09],a[j1+p08],a[j1+1+p08],a[j1+p11],a[j1+1+p11],a[j1+p10],a[j1+1+p10],rt,it);
	RADIX_04_DIF       (aj1p24r,aj1p24i,aj1p25r,aj1p25i,aj1p26r,aj1p26i,aj1p27r,aj1p27i,a[j1+p06],a[j1+1+p06],a[j1+p07],a[j1+1+p07],a[j1+p05],a[j1+1+p05],a[j1+p04],a[j1+1+p04],rt,it);
#endif

	  iroot += root_incr;		/* increment sincos index.	*/

	  }

	  jstart += nwt;
	  jhi    += nwt;
	  col += 28;
	  co3 -= 28;

	}

	if(root_incr==0)break;

/*   Wraparound carry cleanup loop is here: ***
!
!   (1) Invert the radix-28 forward DIF FFT of the first block of 28 complex elements in A and unweight;
!   (2) Propagate cleanup carries among the real and imaginary parts of the 28 outputs of (1);
!   (3) Reweight and perform a radix-28 forward DIF FFT on the result of (2);
!   (4) If any of the exit carries from (2) are nonzero, advance to the next 28 elements and repeat (1-4).
*/
	t1  = cy27;
	cy27= cy26;
	cy26= cy25;
	cy25= cy24;
	cy24= cy23;
	cy23= cy22;
	cy22= cy21;
	cy21= cy20;
	cy20= cy19;
	cy19= cy18;
	cy18= cy17;
	cy17= cy16;
	cy16= cy15;
	cy15= cy14;
	cy14= cy13;
	cy13= cy12;
	cy12= cy11;
	cy11= cy10;
	cy10= cy09;
	cy09= cy08;
	cy08= cy07;
	cy07= cy06;
	cy06= cy05;
	cy05= cy04;
	cy04= cy03;
	cy03= cy02;
	cy02= cy01;
	cy01= cy00;
	cy00= t1  ;

	iroot = 0;
	root_incr = 0;
	scale = 1;

	jstart = 0;
	jhi = 3;
	khi = 1;

	for(j=0; j<=jhi; j++)
	{
	  a[j    ] *= radix_inv;
	  a[j+p01] *= radix_inv;
	  a[j+p02] *= radix_inv;
	  a[j+p03] *= radix_inv;
	  a[j+p04] *= radix_inv;
	  a[j+p05] *= radix_inv;
	  a[j+p06] *= radix_inv;
	  a[j+p07] *= radix_inv;
	  a[j+p08] *= radix_inv;
	  a[j+p09] *= radix_inv;
	  a[j+p10] *= radix_inv;
	  a[j+p11] *= radix_inv;
	  a[j+p12] *= radix_inv;
	  a[j+p13] *= radix_inv;
	  a[j+p14] *= radix_inv;
	  a[j+p15] *= radix_inv;
	  a[j+p16] *= radix_inv;
	  a[j+p17] *= radix_inv;
	  a[j+p18] *= radix_inv;
	  a[j+p19] *= radix_inv;
	  a[j+p20] *= radix_inv;
	  a[j+p21] *= radix_inv;
	  a[j+p22] *= radix_inv;
	  a[j+p23] *= radix_inv;
	  a[j+p24] *= radix_inv;
	  a[j+p25] *= radix_inv;
	  a[j+p26] *= radix_inv;
	  a[j+p27] *= radix_inv;
	}
}

	if(fabs(cy00)+fabs(cy01)+fabs(cy02)+fabs(cy03)+fabs(cy04)+fabs(cy05)+fabs(cy06)+fabs(cy07)+fabs(cy08)+fabs(cy09)+fabs(cy10)+fabs(cy11)+fabs(cy12)+fabs(cy13)+fabs(cy14)+fabs(cy15)+fabs(cy16)+fabs(cy17)+fabs(cy18)+fabs(cy19)+fabs(cy20)+fabs(cy21)+fabs(cy22)+fabs(cy23)+fabs(cy24)+fabs(cy25)+fabs(cy26)+fabs(cy27) != 0.0)
	{
	    if(INTERACT)printf("FATAL: iter = %10d; nonzero exit carry in radix28_ditN_cy_dif1.\n",iter);
	    f1 = fopen(   OFILE,"a");
	    f2 = fopen(STATFILE,"a");
	    fprintf(f1,"FATAL: iter = %10d; nonzero exit carry in radix28_ditN_cy_dif1.\n",iter);
	    fprintf(f2,"FATAL: iter = %10d; nonzero exit carry in radix28_ditN_cy_dif1.\n",iter);
	    fclose(f1);
	    fclose(f2);
	    err=1;
	    return(err);
	}

	return(0);
}

/***************/

void radix28_dit_pass1(double a[], int n)
{
/*
!...Acronym: DIF = Decimation In Frequency
!
!...Subroutine to perform an initial radix-28 complex DIT FFT pass on the data in the length-N real vector A.
!
!   See the documentation in radix16_dif_pass for further details on storage and indexing.
!
!   See the documentation in radix7_dif_pass for details on the radix-15 subtransforms.
*/
	int j,j1;
	static int n28,p01,p02,p03,p04,p05,p06,p07,p08,p09,p10,p11,p12,p13,p14,p15,p16,p17,p18,p19,p20,p21,p22,p23,p24,p25,p26,p27, first_entry=TRUE;
	static double cc1 = .62348980185873353053,	 /* cos(u) = Real part of exp(i*2*pi/7), the radix-7 fundamental sincos datum	*/
		      ss1 = .78183148246802980870,	 /* sin(u) = Imag part of exp(i*2*pi/7).	*/
		      cc2 =-.22252093395631440426,	 /* cos(2u)	*/
		      ss2 = .97492791218182360702,	 /* sin(2u)	*/
		      cc3 =-.90096886790241912622,	 /* cos(3u)	*/
		      ss3 = .43388373911755812050;	 /* sin(3u)	*/
	double re,im,rt,it
	,t1,t2,t3,t4,t5,t6,t7,t8,t9,t10,t11,t12,t13,t14
	,aj1p00r,aj1p01r,aj1p02r,aj1p03r,aj1p04r,aj1p05r,aj1p06r,aj1p07r,aj1p08r,aj1p09r,aj1p10r,aj1p11r,aj1p12r,aj1p13r,aj1p14r,aj1p15r,aj1p16r,aj1p17r,aj1p18r,aj1p19r,aj1p20r,aj1p21r,aj1p22r,aj1p23r,aj1p24r,aj1p25r,aj1p26r,aj1p27r
	,aj1p00i,aj1p01i,aj1p02i,aj1p03i,aj1p04i,aj1p05i,aj1p06i,aj1p07i,aj1p08i,aj1p09i,aj1p10i,aj1p11i,aj1p12i,aj1p13i,aj1p14i,aj1p15i,aj1p16i,aj1p17i,aj1p18i,aj1p19i,aj1p20i,aj1p21i,aj1p22i,aj1p23i,aj1p24i,aj1p25i,aj1p26i,aj1p27i;

	if(!first_entry && (n/28) != n28)	/* New runlength?	*/
	{
	  first_entry=TRUE;
	}

/*...initialize things upon first entry	*/

	if(first_entry)
	{
	  first_entry=FALSE;
	  n28=n/28;

/*   constant index offsets for array load/stores are here.	*/

	  p01 = n28;
	  p02 = p01 + p01;
	  p03 = p02 + p01;
	  p04 = p03 + p01;
	  p05 = p04 + p01;
	  p06 = p05 + p01;
	  p07 = p06 + p01;
	  p08 = p07 + p01;
	  p09 = p08 + p01;
	  p10 = p09 + p01;
	  p11 = p10 + p01;
	  p12 = p11 + p01;
	  p13 = p12 + p01;
	  p14 = p13 + p01;
	  p15 = p14 + p01;
	  p16 = p15 + p01;
	  p17 = p16 + p01;
	  p18 = p17 + p01;
	  p19 = p18 + p01;
	  p20 = p19 + p01;
	  p21 = p20 + p01;
	  p22 = p21 + p01;
	  p23 = p22 + p01;
	  p24 = p23 + p01;
	  p25 = p24 + p01;
	  p26 = p25 + p01;
	  p27 = p26 + p01;

	  p01 = p01 + ( (p01 >> DAT_BITS) << PAD_BITS );
	  p02 = p02 + ( (p02 >> DAT_BITS) << PAD_BITS );
	  p03 = p03 + ( (p03 >> DAT_BITS) << PAD_BITS );
	  p04 = p04 + ( (p04 >> DAT_BITS) << PAD_BITS );
	  p05 = p05 + ( (p05 >> DAT_BITS) << PAD_BITS );
	  p06 = p06 + ( (p06 >> DAT_BITS) << PAD_BITS );
	  p07 = p07 + ( (p07 >> DAT_BITS) << PAD_BITS );
	  p08 = p08 + ( (p08 >> DAT_BITS) << PAD_BITS );
	  p09 = p09 + ( (p09 >> DAT_BITS) << PAD_BITS );
	  p10 = p10 + ( (p10 >> DAT_BITS) << PAD_BITS );
	  p11 = p11 + ( (p11 >> DAT_BITS) << PAD_BITS );
	  p12 = p12 + ( (p12 >> DAT_BITS) << PAD_BITS );
	  p13 = p13 + ( (p13 >> DAT_BITS) << PAD_BITS );
	  p14 = p14 + ( (p14 >> DAT_BITS) << PAD_BITS );
	  p15 = p15 + ( (p15 >> DAT_BITS) << PAD_BITS );
	  p16 = p16 + ( (p16 >> DAT_BITS) << PAD_BITS );
	  p17 = p17 + ( (p17 >> DAT_BITS) << PAD_BITS );
	  p18 = p18 + ( (p18 >> DAT_BITS) << PAD_BITS );
	  p19 = p19 + ( (p19 >> DAT_BITS) << PAD_BITS );
	  p20 = p20 + ( (p20 >> DAT_BITS) << PAD_BITS );
	  p21 = p21 + ( (p21 >> DAT_BITS) << PAD_BITS );
	  p22 = p22 + ( (p22 >> DAT_BITS) << PAD_BITS );
	  p23 = p23 + ( (p23 >> DAT_BITS) << PAD_BITS );
	  p24 = p24 + ( (p24 >> DAT_BITS) << PAD_BITS );
	  p25 = p25 + ( (p25 >> DAT_BITS) << PAD_BITS );
	  p26 = p26 + ( (p26 >> DAT_BITS) << PAD_BITS );
	  p27 = p27 + ( (p27 >> DAT_BITS) << PAD_BITS );
	}

/*...The radix-28 pass is here.	*/

	for(j=0; j < n28; j += 2)
	{
	  j1 = j  + ( (j  >> DAT_BITS) << PAD_BITS );

/*
Twiddleless version requires us to swap inputs as follows:
indices  0, 1, 2, 3, 4, 5, 6, 7, 8, 9,10,11,12,13,14,15,16,17,18,19,20,21,22,23,24,25,26,27
      -> 0,24,20,16,12, 8, 4,21,17,13, 9, 5, 1,25,14,10, 6, 2,26,22,18, 7, 3,27,23,19,15,11

I.e. start out with first quartet of indices {0,7,14,21}, permute those according to
  {0,7,14,21}*27%28 = {0,21,14,7}, then each is head of a length-7 list of indices with decrement 4

Remember, inputs to DIT are bit-reversed, so
a[0, 1, 2, 3, 4, 5, 6, 7, 8, 9,10,11,12,13,14,15,16,17,18,19,20,21,22,23,24,25,26,27] contain
x[0,14, 7,21, 1,15, 8,22, 2,16, 9,23, 3,17,10,24, 4,18,11,25, 5,19,12,26, 6,20,13,27], which get swapped to
x[0,14,21, 7,24,10,17, 3,20, 6,13,27,16, 2, 9,23,12,26, 5,19, 8,22, 1,15, 4,18,25,11], which means the a-indices get swapped as
a[0, 1, 3, 2,15,14,13,12,25,24,26,27, 9, 8,10,11,22,23,20,21, 6, 7, 4, 5,16,17,19,18].
*/
/* EWM: 10/18/04: We swap the odd-index outputs of each of the radix-4 DIT transforms (1<=>3, 5<=>7, etc.) so that the indexing
                  of all the radix-7 transforms (really just the 2nd and 4th of these) winds up being in-place. This allows us
                  to properly re-use the ajp1 variables in the carry-pass version of this routine.
                  (NOTE: Didn't bother to do the swapping in the macro-less version of this routine - would only need to
                   do the swapping there if for some reason we wanted the macro-less DIT code in the carry routines.)
*/
#if 1
/*...gather the needed data (28 64-bit complex, i.e. 56 64-bit reals) and do 7 radix-4 transforms...*/
                     /*                                      outputs                                      */ /*                          inputs                           */
	RADIX_04_DIT(a[j1    ],a[j1+1    ],a[j1+p01],a[j1+1+p01],a[j1+p03],a[j1+1+p03],a[j1+p02],a[j1+1+p02],aj1p00r,aj1p00i,aj1p03r,aj1p03i,aj1p02r,aj1p02i,aj1p01r,aj1p01i,rt,it);
	RADIX_04_DIT(a[j1+p15],a[j1+1+p15],a[j1+p14],a[j1+1+p14],a[j1+p13],a[j1+1+p13],a[j1+p12],a[j1+1+p12],aj1p04r,aj1p04i,aj1p07r,aj1p07i,aj1p06r,aj1p06i,aj1p05r,aj1p05i,rt,it);
	RADIX_04_DIT(a[j1+p25],a[j1+1+p25],a[j1+p24],a[j1+1+p24],a[j1+p26],a[j1+1+p26],a[j1+p27],a[j1+1+p27],aj1p08r,aj1p08i,aj1p11r,aj1p11i,aj1p10r,aj1p10i,aj1p09r,aj1p09i,rt,it);
	RADIX_04_DIT(a[j1+p09],a[j1+1+p09],a[j1+p08],a[j1+1+p08],a[j1+p10],a[j1+1+p10],a[j1+p11],a[j1+1+p11],aj1p12r,aj1p12i,aj1p15r,aj1p15i,aj1p14r,aj1p14i,aj1p13r,aj1p13i,rt,it);
	RADIX_04_DIT(a[j1+p22],a[j1+1+p22],a[j1+p23],a[j1+1+p23],a[j1+p20],a[j1+1+p20],a[j1+p21],a[j1+1+p21],aj1p16r,aj1p16i,aj1p19r,aj1p19i,aj1p18r,aj1p18i,aj1p17r,aj1p17i,rt,it);
	RADIX_04_DIT(a[j1+p06],a[j1+1+p06],a[j1+p07],a[j1+1+p07],a[j1+p04],a[j1+1+p04],a[j1+p05],a[j1+1+p05],aj1p20r,aj1p20i,aj1p23r,aj1p23i,aj1p22r,aj1p22i,aj1p21r,aj1p21i,rt,it);
	RADIX_04_DIT(a[j1+p16],a[j1+1+p16],a[j1+p17],a[j1+1+p17],a[j1+p19],a[j1+1+p19],a[j1+p18],a[j1+1+p18],aj1p24r,aj1p24i,aj1p27r,aj1p27i,aj1p26r,aj1p26i,aj1p25r,aj1p25i,rt,it);

/*...and now do 4 radix-4 transforms...*/
                     /*                                                   inputs                                                  */ /*               intermediates              */ /*                                                                     outputs                                                                         */
	RADIX_07_DFT(aj1p00r,aj1p00i,aj1p04r,aj1p04i,aj1p08r,aj1p08i,aj1p12r,aj1p12i,aj1p16r,aj1p16i,aj1p20r,aj1p20i,aj1p24r,aj1p24i,t1,t2,t3,t4,t5,t6,t7,t8,t9,t10,t11,t12,t13,t14,a[j1    ],a[j1+1    ],a[j1+p08],a[j1+1+p08],a[j1+p16],a[j1+1+p16],a[j1+p24],a[j1+1+p24],a[j1+p04],a[j1+1+p04],a[j1+p12],a[j1+1+p12],a[j1+p20],a[j1+1+p20],rt,it,re,im);
	RADIX_07_DFT(aj1p03r,aj1p03i,aj1p07r,aj1p07i,aj1p11r,aj1p11i,aj1p15r,aj1p15i,aj1p19r,aj1p19i,aj1p23r,aj1p23i,aj1p27r,aj1p27i,t1,t2,t3,t4,t5,t6,t7,t8,t9,t10,t11,t12,t13,t14,a[j1+p07],a[j1+1+p07],a[j1+p15],a[j1+1+p15],a[j1+p23],a[j1+1+p23],a[j1+p03],a[j1+1+p03],a[j1+p11],a[j1+1+p11],a[j1+p19],a[j1+1+p19],a[j1+p27],a[j1+1+p27],rt,it,re,im);
	RADIX_07_DFT(aj1p02r,aj1p02i,aj1p06r,aj1p06i,aj1p10r,aj1p10i,aj1p14r,aj1p14i,aj1p18r,aj1p18i,aj1p22r,aj1p22i,aj1p26r,aj1p26i,t1,t2,t3,t4,t5,t6,t7,t8,t9,t10,t11,t12,t13,t14,a[j1+p14],a[j1+1+p14],a[j1+p22],a[j1+1+p22],a[j1+p02],a[j1+1+p02],a[j1+p10],a[j1+1+p10],a[j1+p18],a[j1+1+p18],a[j1+p26],a[j1+1+p26],a[j1+p06],a[j1+1+p06],rt,it,re,im);
	RADIX_07_DFT(aj1p01r,aj1p01i,aj1p05r,aj1p05i,aj1p09r,aj1p09i,aj1p13r,aj1p13i,aj1p17r,aj1p17i,aj1p21r,aj1p21i,aj1p25r,aj1p25i,t1,t2,t3,t4,t5,t6,t7,t8,t9,t10,t11,t12,t13,t14,a[j1+p21],a[j1+1+p21],a[j1+p01],a[j1+1+p01],a[j1+p09],a[j1+1+p09],a[j1+p17],a[j1+1+p17],a[j1+p25],a[j1+1+p25],a[j1+p05],a[j1+1+p05],a[j1+p13],a[j1+1+p13],rt,it,re,im);
#else

/*...gather the needed data (28 64-bit complex, i.e. 56 64-bit reals) and do 7 radix-4 transforms,	*/
/*...Block 1:	*/
	aj1p00r=a[j1    ];	aj1p00i=a[j1+1    ];
	rt     =a[j1+p01];	it     =a[j1+1+p01];
	aj1p01r=aj1p00r-rt;  	aj1p00r=aj1p00r+rt;
	aj1p01i=aj1p00i-it;	aj1p00i=aj1p00i+it;

	aj1p02r=a[j1+p03];	aj1p02i=a[j1+1+p03];
	rt     =a[j1+p02];	it     =a[j1+1+p02];
	aj1p03r=aj1p02r-rt;  	aj1p02r=aj1p02r+rt;
	aj1p03i=aj1p02i-it;  	aj1p02i=aj1p02i+it;

	rt =aj1p02r;	aj1p02r=aj1p00r-rt;	aj1p00r=aj1p00r+rt;
	it =aj1p02i;	aj1p02i=aj1p00i-it;	aj1p00i=aj1p00i+it;

	rt =aj1p03r;	aj1p03r=aj1p01r-aj1p03i;aj1p01r=aj1p01r+aj1p03i;
			aj1p03i=aj1p01i+rt;	aj1p01i=aj1p01i-rt;

/*...Block 2:	*/
	aj1p04r=a[j1+p15];	aj1p04i=a[j1+1+p15];
	rt     =a[j1+p14];	it     =a[j1+1+p14];
	aj1p05r=aj1p04r-rt;  	aj1p04r=aj1p04r+rt;
	aj1p05i=aj1p04i-it;	aj1p04i=aj1p04i+it;

	aj1p06r=a[j1+p13];	aj1p06i=a[j1+1+p13];
	rt     =a[j1+p12];	it     =a[j1+1+p12];
	aj1p07r=aj1p06r-rt;  	aj1p06r=aj1p06r+rt;
	aj1p07i=aj1p06i-it;  	aj1p06i=aj1p06i+it;

	rt =aj1p06r;	aj1p06r=aj1p04r-rt;	aj1p04r=aj1p04r+rt;
	it =aj1p06i;	aj1p06i=aj1p04i-it;	aj1p04i=aj1p04i+it;

	rt =aj1p07r;	aj1p07r=aj1p05r-aj1p07i;aj1p05r=aj1p05r+aj1p07i;
			aj1p07i=aj1p05i+rt;	aj1p05i=aj1p05i-rt;

/*...Block 3:	*/
	aj1p08r=a[j1+p25];	aj1p08i=a[j1+1+p25];
	rt     =a[j1+p24];	it     =a[j1+1+p24];
	aj1p09r=aj1p08r-rt;  	aj1p08r=aj1p08r+rt;
	aj1p09i=aj1p08i-it;	aj1p08i=aj1p08i+it;

	aj1p10r=a[j1+p26];	aj1p10i=a[j1+1+p26];
	rt     =a[j1+p27];	it     =a[j1+1+p27];
	aj1p11r=aj1p10r-rt;  	aj1p10r=aj1p10r+rt;
	aj1p11i=aj1p10i-it;  	aj1p10i=aj1p10i+it;

	rt =aj1p10r;	aj1p10r=aj1p08r-rt;	aj1p08r=aj1p08r+rt;
	it =aj1p10i;	aj1p10i=aj1p08i-it;	aj1p08i=aj1p08i+it;

	rt =aj1p11r;	aj1p11r=aj1p09r-aj1p11i;aj1p09r=aj1p09r+aj1p11i;
			aj1p11i=aj1p09i+rt;	aj1p09i=aj1p09i-rt;

/*...Block 4:	*/
	aj1p12r=a[j1+p09];	aj1p12i=a[j1+1+p09];
	rt     =a[j1+p08];	it     =a[j1+1+p08];
	aj1p13r=aj1p12r-rt;  	aj1p12r=aj1p12r+rt;
	aj1p13i=aj1p12i-it;	aj1p12i=aj1p12i+it;

	aj1p14r=a[j1+p10];	aj1p14i=a[j1+1+p10];
	rt     =a[j1+p11];	it     =a[j1+1+p11];
	aj1p15r=aj1p14r-rt;  	aj1p14r=aj1p14r+rt;
	aj1p15i=aj1p14i-it;  	aj1p14i=aj1p14i+it;

	rt =aj1p14r;	aj1p14r=aj1p12r-rt;	aj1p12r=aj1p12r+rt;
	it =aj1p14i;	aj1p14i=aj1p12i-it;	aj1p12i=aj1p12i+it;

	rt =aj1p15r;	aj1p15r=aj1p13r-aj1p15i;aj1p13r=aj1p13r+aj1p15i;
			aj1p15i=aj1p13i+rt;	aj1p13i=aj1p13i-rt;

/*...Block 5:	*/
	aj1p16r=a[j1+p22];	aj1p16i=a[j1+1+p22];
	rt     =a[j1+p23];	it     =a[j1+1+p23];
	aj1p17r=aj1p16r-rt;  	aj1p16r=aj1p16r+rt;
	aj1p17i=aj1p16i-it;	aj1p16i=aj1p16i+it;

	aj1p18r=a[j1+p20];	aj1p18i=a[j1+1+p20];
	rt     =a[j1+p21];	it     =a[j1+1+p21];
	aj1p19r=aj1p18r-rt;  	aj1p18r=aj1p18r+rt;
	aj1p19i=aj1p18i-it;  	aj1p18i=aj1p18i+it;

	rt =aj1p18r;	aj1p18r=aj1p16r-rt;	aj1p16r=aj1p16r+rt;
	it =aj1p18i;	aj1p18i=aj1p16i-it;	aj1p16i=aj1p16i+it;

	rt =aj1p19r;	aj1p19r=aj1p17r-aj1p19i;aj1p17r=aj1p17r+aj1p19i;
			aj1p19i=aj1p17i+rt;	aj1p17i=aj1p17i-rt;

/*...Block 6:	*/
	aj1p20r=a[j1+p06];	aj1p20i=a[j1+1+p06];
	rt     =a[j1+p07];	it     =a[j1+1+p07];
	aj1p21r=aj1p20r-rt;  	aj1p20r=aj1p20r+rt;
	aj1p21i=aj1p20i-it;	aj1p20i=aj1p20i+it;

	aj1p22r=a[j1+p04];	aj1p22i=a[j1+1+p04];
	rt     =a[j1+p05];	it     =a[j1+1+p05];
	aj1p23r=aj1p22r-rt;  	aj1p22r=aj1p22r+rt;
	aj1p23i=aj1p22i-it;  	aj1p22i=aj1p22i+it;

	rt =aj1p22r;	aj1p22r=aj1p20r-rt;	aj1p20r=aj1p20r+rt;
	it =aj1p22i;	aj1p22i=aj1p20i-it;	aj1p20i=aj1p20i+it;

	rt =aj1p23r;	aj1p23r=aj1p21r-aj1p23i;aj1p21r=aj1p21r+aj1p23i;
			aj1p23i=aj1p21i+rt;	aj1p21i=aj1p21i-rt;

/*...Block 7:	*/
	aj1p24r=a[j1+p16];	aj1p24i=a[j1+1+p16];
	rt     =a[j1+p17];	it     =a[j1+1+p17];
	aj1p25r=aj1p24r-rt;  	aj1p24r=aj1p24r+rt;
	aj1p25i=aj1p24i-it;	aj1p24i=aj1p24i+it;

	aj1p26r=a[j1+p19];	aj1p26i=a[j1+1+p19];
	rt     =a[j1+p18];	it     =a[j1+1+p18];
	aj1p27r=aj1p26r-rt;  	aj1p26r=aj1p26r+rt;
	aj1p27i=aj1p26i-it;  	aj1p26i=aj1p26i+it;

	rt =aj1p26r;	aj1p26r=aj1p24r-rt;	aj1p24r=aj1p24r+rt;
	it =aj1p26i;	aj1p26i=aj1p24i-it;	aj1p24i=aj1p24i+it;

	rt =aj1p27r;	aj1p27r=aj1p25r-aj1p27i;aj1p25r=aj1p25r+aj1p27i;
			aj1p27i=aj1p25i+rt;	aj1p25i=aj1p25i-rt;

/*...and now do 4 radix-7 transforms.	*/
/*...Block 1:	*/

	  t1 =aj1p00r;			t2 =aj1p00i;
	  t3 =aj1p04r+aj1p24r;		t4 =aj1p04i+aj1p24i;	/* x1+x6 */
	  t5 =aj1p08r+aj1p20r;		t6 =aj1p08i+aj1p20i;	/* x2+x5 */
	  t7 =aj1p12r+aj1p16r;		t8 =aj1p12i+aj1p16i;	/* x3+x4 */
	  t9 =aj1p12r-aj1p16r;		t10=aj1p12i-aj1p16i;	/* x3-x4 */
	  t11=aj1p08r-aj1p20r;		t12=aj1p08i-aj1p20i;	/* x2-x5 */
	  t13=aj1p04r-aj1p24r;		t14=aj1p04i-aj1p24i;	/* x1-x6 */

	  a[j1    ] = t1+t3+t5+t7;	a[j1+1    ] = t2+t4+t6+t8;
	  rt = t1+cc1*t3+cc2*t5+cc3*t7;	it = t2+cc1*t4+cc2*t6+cc3*t8;
	  re = t1+cc2*t3+cc3*t5+cc1*t7;	im = t2+cc2*t4+cc3*t6+cc1*t8;
	  t1 = t1+cc3*t3+cc1*t5+cc2*t7;	t2 = t2+cc3*t4+cc1*t6+cc2*t8;

	  t3 = ss1*t13+ss2*t11+ss3*t9;	t4 = ss1*t14+ss2*t12+ss3*t10;
	  t5 = ss2*t13-ss3*t11-ss1*t9;	t6 = ss2*t14-ss3*t12-ss1*t10;
	  t7 = ss3*t13-ss1*t11+ss2*t9;	t8 = ss3*t14-ss1*t12+ss2*t10;
	/* Output permutation causes signs to get flipped here: */
	  a[j1+p08]=rt-t4;		a[j1+1+p08]=it+t3;
	  a[j1+p16]=re-t6;		a[j1+1+p16]=im+t5;
	  a[j1+p24]=t1-t8;		a[j1+1+p24]=t2+t7;
	  a[j1+p04]=t1+t8;		a[j1+1+p04]=t2-t7;
	  a[j1+p12]=re+t6;		a[j1+1+p12]=im-t5;
	  a[j1+p20]=rt+t4;		a[j1+1+p20]=it-t3;

/*...Block 2:	*/

	  t1 =aj1p01r;			t2 =aj1p01i;
	  t3 =aj1p05r+aj1p25r;		t4 =aj1p05i+aj1p25i;	/* x1+x6 */
	  t5 =aj1p09r+aj1p21r;		t6 =aj1p09i+aj1p21i;	/* x2+x5 */
	  t7 =aj1p13r+aj1p17r;		t8 =aj1p13i+aj1p17i;	/* x3+x4 */
	  t9 =aj1p13r-aj1p17r;		t10=aj1p13i-aj1p17i;	/* x3-x4 */
	  t11=aj1p09r-aj1p21r;		t12=aj1p09i-aj1p21i;	/* x2-x5 */
	  t13=aj1p05r-aj1p25r;		t14=aj1p05i-aj1p25i;	/* x1-x6 */

	  a[j1+p07] = t1+t3+t5+t7;	a[j1+1+p07] = t2+t4+t6+t8;
	  rt = t1+cc1*t3+cc2*t5+cc3*t7;	it = t2+cc1*t4+cc2*t6+cc3*t8;
	  re = t1+cc2*t3+cc3*t5+cc1*t7;	im = t2+cc2*t4+cc3*t6+cc1*t8;
	  t1 = t1+cc3*t3+cc1*t5+cc2*t7;	t2 = t2+cc3*t4+cc1*t6+cc2*t8;

	  t3 = ss1*t13+ss2*t11+ss3*t9;	t4 = ss1*t14+ss2*t12+ss3*t10;
	  t5 = ss2*t13-ss3*t11-ss1*t9;	t6 = ss2*t14-ss3*t12-ss1*t10;
	  t7 = ss3*t13-ss1*t11+ss2*t9;	t8 = ss3*t14-ss1*t12+ss2*t10;
	/* Output permutation causes signs to get flipped here: */
	  a[j1+p15]=rt-t4;		a[j1+1+p15]=it+t3;
	  a[j1+p23]=re-t6;		a[j1+1+p23]=im+t5;
	  a[j1+p03]=t1-t8;		a[j1+1+p03]=t2+t7;
	  a[j1+p11]=t1+t8;		a[j1+1+p11]=t2-t7;
	  a[j1+p19]=re+t6;		a[j1+1+p19]=im-t5;
	  a[j1+p27]=rt+t4;		a[j1+1+p27]=it-t3;

/*...Block 3:	*/

	  t1 =aj1p02r;			t2 =aj1p02i;
	  t3 =aj1p06r+aj1p26r;		t4 =aj1p06i+aj1p26i;	/* x1+x6 */
	  t5 =aj1p10r+aj1p22r;		t6 =aj1p10i+aj1p22i;	/* x2+x5 */
	  t7 =aj1p14r+aj1p18r;		t8 =aj1p14i+aj1p18i;	/* x3+x4 */
	  t9 =aj1p14r-aj1p18r;		t10=aj1p14i-aj1p18i;	/* x3-x4 */
	  t11=aj1p10r-aj1p22r;		t12=aj1p10i-aj1p22i;	/* x2-x5 */
	  t13=aj1p06r-aj1p26r;		t14=aj1p06i-aj1p26i;	/* x1-x6 */

	  a[j1+p14] = t1+t3+t5+t7;	a[j1+1+p14] = t2+t4+t6+t8;
	  rt = t1+cc1*t3+cc2*t5+cc3*t7;	it = t2+cc1*t4+cc2*t6+cc3*t8;
	  re = t1+cc2*t3+cc3*t5+cc1*t7;	im = t2+cc2*t4+cc3*t6+cc1*t8;
	  t1 = t1+cc3*t3+cc1*t5+cc2*t7;	t2 = t2+cc3*t4+cc1*t6+cc2*t8;

	  t3 = ss1*t13+ss2*t11+ss3*t9;	t4 = ss1*t14+ss2*t12+ss3*t10;
	  t5 = ss2*t13-ss3*t11-ss1*t9;	t6 = ss2*t14-ss3*t12-ss1*t10;
	  t7 = ss3*t13-ss1*t11+ss2*t9;	t8 = ss3*t14-ss1*t12+ss2*t10;
	/* Output permutation causes signs to get flipped here: */
	  a[j1+p22]=rt-t4;		a[j1+1+p22]=it+t3;
	  a[j1+p02]=re-t6;		a[j1+1+p02]=im+t5;
	  a[j1+p10]=t1-t8;		a[j1+1+p10]=t2+t7;
	  a[j1+p18]=t1+t8;		a[j1+1+p18]=t2-t7;
	  a[j1+p26]=re+t6;		a[j1+1+p26]=im-t5;
	  a[j1+p06]=rt+t4;		a[j1+1+p06]=it-t3;

/*...Block 4:	*/

	  t1 =aj1p03r;			t2 =aj1p03i;
	  t3 =aj1p07r+aj1p27r;		t4 =aj1p07i+aj1p27i;	/* x1+x6 */
	  t5 =aj1p11r+aj1p23r;		t6 =aj1p11i+aj1p23i;	/* x2+x5 */
	  t7 =aj1p15r+aj1p19r;		t8 =aj1p15i+aj1p19i;	/* x3+x4 */
	  t9 =aj1p15r-aj1p19r;		t10=aj1p15i-aj1p19i;	/* x3-x4 */
	  t11=aj1p11r-aj1p23r;		t12=aj1p11i-aj1p23i;	/* x2-x5 */
	  t13=aj1p07r-aj1p27r;		t14=aj1p07i-aj1p27i;	/* x1-x6 */

	  a[j1+p21] = t1+t3+t5+t7;	a[j1+1+p21] = t2+t4+t6+t8;
	  rt = t1+cc1*t3+cc2*t5+cc3*t7;	it = t2+cc1*t4+cc2*t6+cc3*t8;
	  re = t1+cc2*t3+cc3*t5+cc1*t7;	im = t2+cc2*t4+cc3*t6+cc1*t8;
	  t1 = t1+cc3*t3+cc1*t5+cc2*t7;	t2 = t2+cc3*t4+cc1*t6+cc2*t8;

	  t3 = ss1*t13+ss2*t11+ss3*t9;	t4 = ss1*t14+ss2*t12+ss3*t10;
	  t5 = ss2*t13-ss3*t11-ss1*t9;	t6 = ss2*t14-ss3*t12-ss1*t10;
	  t7 = ss3*t13-ss1*t11+ss2*t9;	t8 = ss3*t14-ss1*t12+ss2*t10;
	/* Output permutation causes signs to get flipped here: */
	  a[j1+p01]=rt-t4;		a[j1+1+p01]=it+t3;
	  a[j1+p09]=re-t6;		a[j1+1+p09]=im+t5;
	  a[j1+p17]=t1-t8;		a[j1+1+p17]=t2+t7;
	  a[j1+p25]=t1+t8;		a[j1+1+p25]=t2-t7;
	  a[j1+p05]=re+t6;		a[j1+1+p05]=im-t5;
	  a[j1+p13]=rt+t4;		a[j1+1+p13]=it-t3;
#endif
	}
}

