/*******************************************************************************
*                                                                              *
*   (C) 1997-2004 by Ernst W. Mayer.                                           *
*                                                                              *
*  This program is free software; you can redistribute it and/or modify it     *
*  under the terms of the GNU General Public License as published by the       *
*  Free Software Foundation; either version 2 of the License, or (at your      *
*  option) any later version.                                                  *
*                                                                              *
*  This program is distributed in the hope that it will be useful, but WITHOUT *
*  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or       *
*  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for   *
*  more details.                                                               *
*                                                                              *
*  You should have received a copy of the GNU General Public License along     *
*  with this program; see the file GPL.txt.  If not, you may view one at       *
*  http://www.fsf.org/licenses/licenses.html, or obtain one by writing to the  *
*  Free Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA     *
*  02111-1307, USA.                                                            *
*                                                                              *
*******************************************************************************/

#include "Mlucas.h"

/***************/

void radix8_dif_pass1(double a[], int n)
{
/*
!...Acronym: DIF = Decimation In Frequency
!
!...Subroutine to perform an initial radix-8 complex DIF FFT pass on the data in the length-N real vector A.
!
!   See the documentation in radix16_dif_pass for further details.
*/
	int j,j1,j2,j3,j4,j5,j6,j7,j8;
	static int n8, first_entry=TRUE;
	double rt,it
		,t1,t2,t3,t4,t5,t6,t7,t8,t9,t10,t11,t12,t13,t14,t15,t16;

	if(!first_entry && (n >> 3) != n8)	/* New runlength?	*/
	{
	  first_entry=TRUE;
	}

/*...initialize things upon first entry	*/

	if(first_entry)
	{
	  first_entry=FALSE;
	  n8 = n/8;
	}

/*...The radix-8 pass is here.	*/

      for(j=0; j < n8; j += 2)
      {
	  j2 = j  +n8;
	  j3 = j2 +n8;
	  j4 = j3 +n8;
	  j5 = j4 +n8;
	  j6 = j5 +n8;
	  j7 = j6 +n8;
	  j8 = j7 +n8;

	  j1 = j  + ( (j  >> DAT_BITS) << PAD_BITS );
	  j2 = j2 + ( (j2 >> DAT_BITS) << PAD_BITS );
	  j3 = j3 + ( (j3 >> DAT_BITS) << PAD_BITS );
	  j4 = j4 + ( (j4 >> DAT_BITS) << PAD_BITS );
	  j5 = j5 + ( (j5 >> DAT_BITS) << PAD_BITS );
	  j6 = j6 + ( (j6 >> DAT_BITS) << PAD_BITS );
	  j7 = j7 + ( (j7 >> DAT_BITS) << PAD_BITS );
	  j8 = j8 + ( (j8 >> DAT_BITS) << PAD_BITS );

/*       gather the needed data (8 64-bit complex, i.e. 16 64-bit reals) and combine to get the 4 length-2 transforms...	*/
	  t1 =a[j1];		t2 =a[j1+1];
	  rt =a[j5];		it =a[j5+1];
	  t3 =t1 -rt;		t4 =t2 -it;
	  t1 =t1 +rt;		t2 =t2 +it;

	  t5 =a[j3];		t6 =a[j3+1];
	  rt =a[j7];		it =a[j7+1];
	  t7 =t5 -rt;		t8 =t6 -it;
	  t5 =t5 +rt;		t6 =t6 +it;

	  t9 =a[j2];		t10=a[j2+1];
	  rt =a[j6];		it =a[j6+1];
	  t11=t9 -rt;		t12=t10-it;
	  t9 =t9 +rt;		t10=t10+it;

	  t13=a[j4];		t14=a[j4+1];
	  rt =a[j8];		it =a[j8+1];
	  t15=t13-rt;		t16=t14-it;
	  t13=t13+rt;		t14=t14+it;
/*       combine to get the 2 length-4 transforms...	*/
	  rt =t5;		it =t6;
	  t5 =t1 -rt;		t6 =t2 -it;
	  t1 =t1 +rt;		t2 =t2 +it;

	  rt =t7;		it =t8;
	  t7 =t3 +it;		t8 =t4 -rt;
	  t3 =t3 -it;		t4 =t4 +rt;

	  rt =t13;		it =t14;
	  t13=t9 -rt;		t14=t10-it;
	  t9 =t9 +rt;		t10=t10+it;

	  rt =t15;		it =t16;
	  t15=t11+it;		t16=t12-rt;
	  t11=t11-it;		t12=t12+rt;
/*       now combine the two half-transforms	*/
	  a[j1]=t1+t9;		a[j1+1]=t2+t10;
	  a[j2]=t1-t9;		a[j2+1]=t2-t10;

	  a[j3]=t5-t14;		a[j3+1]=t6+t13;
	  a[j4]=t5+t14;		a[j4+1]=t6-t13;

	  rt =(t11-t12)*ISRT2;	it =(t11+t12)*ISRT2;
	  a[j5]=t3+rt;		a[j5+1]=t4+it;
	  a[j6]=t3-rt;		a[j6+1]=t4-it;

	  rt =(t15+t16)*ISRT2;	it =(t16-t15)*ISRT2;
	  a[j7]=t7-rt;		a[j7+1]=t8-it;
	  a[j8]=t7+rt;		a[j8+1]=t8+it;
	}
}

/***************/

int radix8_ditN_cy_dif1(double a[], int n, int nwt, int nwt_bits, double wt0[], double wt1[], int si[], int nrt, int nrt_bits, struct complex rt0[],
                         struct complex rt1[], double base[], double baseinv[], int iter, double *fracmax, int p)
{
/*
!...Acronym: DWT = Discrete Weighted Transform, DIT = Decimation In Time, DIF = Decimation In Frequency
!
!...Performs a final radix-8 complex DIT pass, an inverse DWT weighting, a carry propagation,
!   a forward DWT weighting, and an initial radix-8 complex DIF pass on the data in the length-N real vector A.
!
!   Data enter and are returned in the A-array.
!
!   See the documentation in mers_mod_square and radix16_dif_pass for further details on the array
!   storage scheme, and radix8_ditN_cy_dif1 for details on the reduced-length weights array scheme.
*/
	int n8,bjmodn0,bjmodn1,bjmodn2,bjmodn3,bjmodn4,bjmodn5,bjmodn6,bjmodn7,i,j,j1,jstart,jhi,iroot,root_incr,k1,k2,k,khi,l,outer;
	static int psave=0,bw,sw,bjmodnini,nm1,nrtm1,p1,p2,p3,p4,p5,p6,p7;
	const double one_half[3] = {1.0, 0.5, 0.25};	/* Needed for small-weights-tables scheme */
	static double radix_inv,n2inv;
	double rt,it
#if !USE_NEW_DIT
		,c1,c2,c3,c4,c5,c6,c7
		,s1,s2,s3,s4,s5,s6,s7
#endif
		,t1,t2,t3,t4,t5,t6,t7,t8,t9,t10,t11,t12,t13,t14,t15,t16
		,aj1p0r,aj1p1r,aj1p2r,aj1p3r,aj1p4r,aj1p5r,aj1p6r,aj1p7r
		,aj1p0i,aj1p1i,aj1p2i,aj1p3i,aj1p4i,aj1p5i,aj1p6i,aj1p7i
		,cy0,cy1,cy2,cy3,cy4,cy5,cy6,cy7,temp,frac,scale;
	double maxerr = 0.0;
#if PFETCH
	double *add0, *add;
#endif
	int err;
	FILE *f1,*f2;
	static int first_entry=TRUE;

/*...stuff for the reduced-length DWT weights array is here:	*/
	int n_div_nwt;
	int col,co2,co3,m,m2,n_minus_sil,n_minus_silp1,sinwt,sinwtm1;
	double wt,wtinv,wtl,wtlp1,wtn,wtnm1,wtA,wtB,wtC;

/*...change n8 and n_div_wt to non-static to work around a gcc compiler bug. */
	n8   = n/8;
	n_div_nwt = n8 >> nwt_bits;

	if((n_div_nwt << nwt_bits) != n8)
	{
	  if(INTERACT)printf("FATAL: iter = %10d; NWT_BITS does not divide N/8 in radix8_ditN_cy_dif1.\n",iter);
	  f1 = fopen(   OFILE,"a");
	  f2 = fopen(STATFILE,"a");
	  fprintf(f1,"FATAL: iter = %10d; NWT_BITS does not divide N/8 in radix8_ditN_cy_dif1.\n",iter);
	  fprintf(f2,"FATAL: iter = %10d; NWT_BITS does not divide N/8 in radix8_ditN_cy_dif1.\n",iter);
	  fclose(f1);
	  fclose(f2);
	  err=1;
	  return(err);
	}

	if(p != psave)
	{
	  first_entry=TRUE;
	}

/*...initialize things upon first entry	*/

	if(first_entry)
	{
	  psave = p;
	  first_entry=FALSE;
	  radix_inv = 0.125;
	  n2inv = 1.0/(n/2);

	  bw    = p%n;		/* Number of bigwords in the Crandall/Fagin mixed-radix representation = (Mersenne exponent) mod (vector length).	*/
	  sw    = n - bw;	/* Number of smallwords.	*/

	  nm1   = n-1;
	  nrtm1 = nrt-1;

/*   constant index offsets for load/stores are here.	*/

	  p1 = n8 + ( (n8 >> DAT_BITS) << PAD_BITS );
	  p2 = p1 +p1;
	  p3 = p2 +p1;
	  p4 = p3 +p1;
	  p5 = p4 +p1;
	  p6 = p5 +p1;
	  p7 = p6 +p1;

	  bjmodnini=0;
	  for(j=0; j < n8; j++)
	  {
	    bjmodnini -= sw; bjmodnini = bjmodnini + ( (-(int)((uint32)bjmodnini >> 31)) & n);
	  }
	}

/*...The radix-8 final DIT pass is here.	*/

	cy0 =-2;	/* init carries	*/
	cy1 = 0;
	cy2 = 0;
	cy3 = 0;
	cy4 = 0;
	cy5 = 0;
	cy6 = 0;
	cy7 = 0;

	*fracmax=0;	/* init max. fractional error	*/

	iroot = 0;	/* init sincos array index	*/
	root_incr = 1;	/* init sincos array index increment (set = 1 for normal carry pass, = 0 for wrapper pass)	*/

	scale = n2inv;	/* init inverse-weight scale factor  (set = 2/n for normal carry pass, = 1 for wrapper pass)	*/

	jstart = 0;
	jhi = jstart+nwt-1;
	khi = n_div_nwt;

for(outer=0; outer <= 1; outer++)
{
	i = 1;		/* Pointer to the BASE and BASEINV arrays. lowest-order digit is always a bigword (i = 1).	*/

	bjmodn0 = 0;
	bjmodn1 = bjmodnini;
	bjmodn2 = bjmodn1 +bjmodnini-n; bjmodn2 = bjmodn2 + ( (-(int)((uint32)bjmodn2 >> 31)) & n);
	bjmodn3 = bjmodn2 +bjmodnini-n; bjmodn3 = bjmodn3 + ( (-(int)((uint32)bjmodn3 >> 31)) & n);
	bjmodn4 = bjmodn3 +bjmodnini-n; bjmodn4 = bjmodn4 + ( (-(int)((uint32)bjmodn4 >> 31)) & n);
	bjmodn5 = bjmodn4 +bjmodnini-n; bjmodn5 = bjmodn5 + ( (-(int)((uint32)bjmodn5 >> 31)) & n);
	bjmodn6 = bjmodn5 +bjmodnini-n; bjmodn6 = bjmodn6 + ( (-(int)((uint32)bjmodn6 >> 31)) & n);
	bjmodn7 = bjmodn6 +bjmodnini-n; bjmodn7 = bjmodn7 + ( (-(int)((uint32)bjmodn7 >> 31)) & n);

	col=0;
	co2=(n >> nwt_bits)-1+8;
	co3=co2-8;		/* At the start of each new j-loop, co3=co2-radix(1)	*/

	for(k=1; k <= khi; k++)	/* Do n/(radix(1)*nwt) outer loop executions...	*/
	{
	  for(j=jstart; j<jhi; j += 2)	/* Each inner loop execution processes (radix(1)*nwt) array data.	*/
	  {
	    j1 = j + ( (j >> DAT_BITS) << PAD_BITS );	/* padded-array fetch index is here	*/

#if USE_NEW_DIT
	      t1 =a[j1   ];	t2 =a[j1+1   ];
	      rt =a[j1+p1];	it =a[j1+1+p1];
	      t3 =t1 -rt;	t4 =t2 -it;
	      t1 =t1 +rt;	t2 =t2 +it;

	      t5 =a[j1+p2];	t6 =a[j1+1+p2];
	      rt =a[j1+p3];	it =a[j1+1+p3];
	      t7 =t5 -rt;	t8 =t6 -it;
	      t5 =t5 +rt;	t6 =t6 +it;

	      t9 =a[j1+p4];	t10=a[j1+1+p4];
	      rt =a[j1+p5];	it =a[j1+1+p5];
	      t11=t9 -rt;	t12=t10-it;
	      t9 =t9 +rt;	t10=t10+it;

	      t13=a[j1+p6];	t14=a[j1+1+p6];
	      rt =a[j1+p7];	it =a[j1+1+p7];
	      t15=t13-rt;	t16=t14-it;
	      t13=t13+rt;	t14=t14+it;
#else
/*	here are the needed sincos data - these will be processed below in bit-reversed order.	*/

	    l = iroot;

	      k1=(l & nrtm1);
	      k2=(l >> nrt_bits);
	      l += iroot;
	      t1=rt0[k1].re;	t2=rt0[k1].im;
	      t3=rt1[k2].re;	t4=rt1[k2].im;
	      c1=t1*t3-t2*t4;	s1=t1*t4+t2*t3;

	      k1=(l & nrtm1);
	      k2=(l >> nrt_bits);
	      l += l + iroot;
	      t1=rt0[k1].re;	t2=rt0[k1].im;
	      t3=rt1[k2].re;	t4=rt1[k2].im;
	      c2=t1*t3-t2*t4;	s2=t1*t4+t2*t3;

	      k1=(l & nrtm1);
	      k2=(l >> nrt_bits);
	      t1=rt0[k1].re;	t2=rt0[k1].im;
	      t3=rt1[k2].re;	t4=rt1[k2].im;
	      c5=t1*t3-t2*t4;	s5=t1*t4+t2*t3;

	      t1=c1*c5; t2=c1*s5; t3=s1*c5; t4=s1*s5;
	      c4=t1+t4; s4=t2-t3; c6=t1-t4; s6=t2+t3;

	      t1=c2*c5; t2=c2*s5; t3=s2*c5; t4=s2*s5;
	      c3=t1+t4; s3=t2-t3; c7=t1-t4; s7=t2+t3;

/*       gather the needed data (8 64-bit complex, i.e. 16 64-bit reals) and get the 4 length-2 transforms...	*/
	      t1 =a[j1   ];			t2 =a[j1+1   ];
	      rt =a[j1+p1]*c4 +a[j1+1+p1]*s4;	it =a[j1+1+p1]*c4 -a[j1+p1]*s4;
	      t3 =t1 -rt;			t4 =t2 -it;
	      t1 =t1 +rt;			t2 =t2 +it;

	      t5 =a[j1+p2]*c2 +a[j1+1+p2]*s2;	t6 =a[j1+1+p2]*c2 -a[j1+p2]*s2;
	      rt =a[j1+p3]*c6 +a[j1+1+p3]*s6;	it =a[j1+1+p3]*c6 -a[j1+p3]*s6;
	      t7 =t5 -rt;			t8 =t6 -it;
	      t5 =t5 +rt;			t6 =t6 +it;

	      t9 =a[j1+p4]*c1 +a[j1+1+p4]*s1;	t10=a[j1+1+p4]*c1 -a[j1+p4]*s1;
	      rt =a[j1+p5]*c5 +a[j1+1+p5]*s5;	it =a[j1+1+p5]*c5 -a[j1+p5]*s5;
	      t11=t9 -rt;			t12=t10-it;
	      t9 =t9 +rt;			t10=t10+it;

	      t13=a[j1+p6]*c3 +a[j1+1+p6]*s3;	t14=a[j1+1+p6]*c3 -a[j1+p6]*s3;
	      rt =a[j1+p7]*c7 +a[j1+1+p7]*s7;	it =a[j1+1+p7]*c7 -a[j1+p7]*s7;
	      t15=t13-rt;			t16=t14-it;
	      t13=t13+rt;			t14=t14+it;
#endif
/*       combine to get the 2 length-4 transforms...	*/
	      rt =t5;	t5 =t1 -rt;	t1 =t1 +rt;
	      it =t6;	t6 =t2 -it;	t2 =t2 +it;

	      rt =t7;	t7 =t3 -t8;	t3 =t3 +t8;
			t8 =t4 +rt;	t4 =t4 -rt;

	      rt =t13;	t13=t9 -rt;	t9 =t9 +rt;
	      it =t14;	t14=t10-it;	t10=t10+it;

	      rt =t15;	t15=t11-t16;	t11=t11+t16;
			t16=t12+rt;	t12=t12-rt;
/*       now combine the two half-transforms	*/
	      aj1p0r=t1+t9;		aj1p0i=t2+t10;
	      aj1p4r=t1-t9;		aj1p4i=t2-t10;

	      rt=(t11+t12)*ISRT2;	it=(t11-t12)*ISRT2;
	      aj1p1r=t3+rt;		aj1p1i=t4-it;
	      aj1p5r=t3-rt;		aj1p5i=t4+it;

	      aj1p2r=t5+t14;		aj1p2i=t6-t13;
	      aj1p6r=t5-t14;		aj1p6i=t6+t13;

	      rt=(t15-t16)*ISRT2;	it=(t15+t16)*ISRT2;
	      aj1p3r=t7-rt;		aj1p3i=t8-it;
	      aj1p7r=t7+rt;		aj1p7i=t8+it;

/*...and combine those to complete the radix-8 transform and do the carries. Since the outputs would
     normally be getting dispatched to 8 separate blocks of the A-array, we need 8 separate carries.	*/

	/* We can use the identity WT(J)*WT(N-J) = 2 here to obviate the
	   need for an array of inverse weights - if we need the inverse
	   of WT(J), just use WT(N-J)/2. Together with a reduced-array weights scheme
	   this needs L* = (N - J) & (NWT - 1) = NWT - [J & (NWT - 1)] = NWT - L;
	   also needs 	*/

	    l= j & (nwt-1);			/* We want (S*J mod N) - SI(L) for all 16 carries, so precompute	*/
	    n_minus_sil   = n-si[l  ];		/* N - SI(L) and for each J, find N - (B*J mod N) - SI(L)		*/
	    n_minus_silp1 = n-si[l+1];		/* For the inverse weight, want (S*(N - J) mod N) - SI(NWT - L) =	*/
	    sinwt   = si[nwt-l  ];		/*	= N - (S*J mod N) - SI(NWT - L) = (B*J mod N) - SI(NWT - L).	*/
	    sinwtm1 = si[nwt-l-1];

	    wtl     =wt0[    l  ];
	    wtn     =wt0[nwt-l  ]*scale;	/* Include 1/(n/2) scale factor of inverse transform here...	*/
	    wtlp1   =wt0[    l+1];
	    wtnm1   =wt0[nwt-l-1]*scale;	/* ...and here.	*/

/*...set0 is slightly different from others:	*/
	   cmplx_carry_norm_pow2_errcheck0(aj1p0r,aj1p0i,cy0,bjmodn0);
	    cmplx_carry_norm_pow2_errcheck(aj1p1r,aj1p1i,cy1,bjmodn1,0x1);
	    cmplx_carry_norm_pow2_errcheck(aj1p2r,aj1p2i,cy2,bjmodn2,0x2);
	    cmplx_carry_norm_pow2_errcheck(aj1p3r,aj1p3i,cy3,bjmodn3,0x3);
	    cmplx_carry_norm_pow2_errcheck(aj1p4r,aj1p4i,cy4,bjmodn4,0x4);
	    cmplx_carry_norm_pow2_errcheck(aj1p5r,aj1p5i,cy5,bjmodn5,0x5);
	    cmplx_carry_norm_pow2_errcheck(aj1p6r,aj1p6i,cy6,bjmodn6,0x6);
	    cmplx_carry_norm_pow2_errcheck(aj1p7r,aj1p7i,cy7,bjmodn7,0x7);

	    i =((uint32)(sw - bjmodn0) >> 31);	/* get ready for the next set...	*/
	    co2=co3;	/* For all data but the first set in each j-block, co2=co3. Thus, after the first block of data is done
			   (and only then: for all subsequent blocks it's superfluous), this assignment decrements co2 by radix(1).	*/

/*...The radix-8 DIF pass is here:	*/
#if PFETCH
add0 = &a[j1];
prefetch_p_doubles(add);
#endif
/*       gather the needed data (8 64-bit complex, i.e. 16 64-bit reals) and combine to get the 4 length-2 transforms...	*/
	  t3 =aj1p0r -aj1p4r;	t4 =aj1p0i -aj1p4i;
	  t1 =aj1p0r +aj1p4r;	t2 =aj1p0i +aj1p4i;

	  t7 =aj1p2r -aj1p6r;	t8 =aj1p2i -aj1p6i;
	  t5 =aj1p2r +aj1p6r;	t6 =aj1p2i +aj1p6i;
#if PFETCH
add = add0+p1;
prefetch_p_doubles(add);
#endif
	  t11=aj1p1r -aj1p5r;	t12=aj1p1i -aj1p5i;
	  t9 =aj1p1r +aj1p5r;	t10=aj1p1i +aj1p5i;

	  t15=aj1p3r -aj1p7r;	t16=aj1p3i -aj1p7i;
	  t13=aj1p3r +aj1p7r;	t14=aj1p3i +aj1p7i;
#if PFETCH
add = add0+p2;
prefetch_p_doubles(add);
#endif
/*       combine to get the 2 length-4 transforms...	*/
	  rt =t5;	t5 =t1 -rt;	t1 =t1 +rt;
	  it =t6;	t6 =t2 -it;	t2 =t2 +it;

	  rt =t7;	t7 =t3 +t8;	t3 =t3 -t8;
			t8 =t4 -rt;	t4 =t4 +rt;
#if PFETCH
add = add0+p3;
prefetch_p_doubles(add);
#endif
	  rt =t13;	t13=t9 -rt;	t9 =t9 +rt;
	  it =t14;	t14=t10-it;	t10=t10+it;

	  rt =t15;	t15=t11+t16;	t11=t11-t16;
			t16=t12-rt;	t12=t12+rt;
/*       now combine the two half-transforms	*/
#if PFETCH
add = add0+p4;
prefetch_p_doubles(add);
#endif
	  a[j1   ]=t1+t9;	a[j1+1   ]=t2+t10;
	  a[j1+p1]=t1-t9;	a[j1+1+p1]=t2-t10;
#if PFETCH
add = add0+p5;
prefetch_p_doubles(add);
#endif
	  a[j1+p2]=t5-t14;	a[j1+1+p2]=t6+t13;
	  a[j1+p3]=t5+t14;	a[j1+1+p3]=t6-t13;
#if PFETCH
add = add0+p6;
prefetch_p_doubles(add);
#endif
	  rt =(t11-t12)*ISRT2;	it =(t11+t12)*ISRT2;
	  a[j1+p4]=t3+rt;	a[j1+1+p4]=t4+it;
	  a[j1+p5]=t3-rt;	a[j1+1+p5]=t4-it;
#if PFETCH
add = add0+p7;
prefetch_p_doubles(add);
#endif
	  rt =(t15+t16)*ISRT2;	it =(t16-t15)*ISRT2;
	  a[j1+p6]=t7-rt;	a[j1+1+p6]=t8-it;
	  a[j1+p7]=t7+rt;	a[j1+1+p7]=t8+it;

	  iroot += root_incr;		/* increment sincos index.	*/
	  }

	  jstart += nwt;
	  jhi    += nwt;
	  col += 8;
	  co3 -= 8;

	}

	if(root_incr==0)break;

/*
!   Wraparound carry cleanup loop is here:
!
!   The cleanup carries from the end of each length-N/8 block into the begining of the next
!   can all be neatly processed as follows:
!
!   (1) Invert the radix-8 forward DIF FFT of the first block of 8 complex elements in A and unweight;
!   (2) Propagate cleanup carries among the real and imaginary parts of the 8 outputs of (1);
!   (3) Reweight and perform a radix-8 forward DIF FFT on the result of (2);
!   (4) If any of the exit carries from (2) are nonzero, advance to the next 8 elements and repeat (1-4).
*/
	t1  = cy7;
	cy7 = cy6;
	cy6 = cy5;
	cy5 = cy4;
	cy4 = cy3;
	cy3 = cy2;
	cy2 = cy1;
	cy1 = cy0;
	cy0 = t1;

	iroot = 0;
	root_incr = 0;
	scale = 1;

	jstart = 0;
	jhi = 3;
	khi = 1;

	for(j=0; j<=jhi; j++)
	{
	  a[j   ] *= radix_inv;
	  a[j+p1] *= radix_inv;
	  a[j+p2] *= radix_inv;
	  a[j+p3] *= radix_inv;
	  a[j+p4] *= radix_inv;
	  a[j+p5] *= radix_inv;
	  a[j+p6] *= radix_inv;
	  a[j+p7] *= radix_inv;
	}
}

	if(fabs(cy0)+fabs(cy1)+fabs(cy2)+fabs(cy3)+fabs(cy4)+fabs(cy5)+fabs(cy6)+fabs(cy7) != 0.0)
	{
	    if(INTERACT)printf("FATAL: iter = %10d; nonzero exit carry in radix8_ditN_cy_dif1.\n",iter);
	    f1 = fopen(   OFILE,"a");
	    f2 = fopen(STATFILE,"a");
	    fprintf(f1,"FATAL: iter = %10d; nonzero exit carry in radix8_ditN_cy_dif1.\n",iter);
	    fprintf(f2,"FATAL: iter = %10d; nonzero exit carry in radix8_ditN_cy_dif1.\n",iter);
	    fclose(f1);
	    fclose(f2);
	    err=1;
	    return(err);
	}
	*fracmax = maxerr;
	return(0);
}

/***************/

int radix8_ditN_cy_dif1_nochk(double a[], int n, int nwt, int nwt_bits, double wt0[], double wt1[], int si[], int nrt, int nrt_bits, struct complex rt0[],
                         struct complex rt1[], double base[], double baseinv[], int iter                 , int p)
{
/*
!...Acronym: DWT = Discrete Weighted Transform, DIT = Decimation In Time, DIF = Decimation In Frequency
!
!...Performs a final radix-8 complex DIT pass, an inverse DWT weighting, a carry propagation,
!   a forward DWT weighting, and an initial radix-8 complex DIF pass on the data in the length-N real vector A.
!
!   Data enter and are returned in the A-array.
!
!   See the documentation in mers_mod_square and radix16_dif_pass for further details on the array
!   storage scheme, and radix8_ditN_cy_dif1 for details on the reduced-length weights array scheme.
*/
	int n8,bjmodn0,bjmodn1,bjmodn2,bjmodn3,bjmodn4,bjmodn5,bjmodn6,bjmodn7,i,j,j1,jstart,jhi,iroot,root_incr,k1,k2,k,khi,l,outer;
	static int psave=0,bw,sw,bjmodnini,nm1,nrtm1,p1,p2,p3,p4,p5,p6,p7;
	const double one_half[3] = {1.0, 0.5, 0.25};	/* Needed for small-weights-tables scheme */
	static double radix_inv,n2inv;
	double rt,it
#if !USE_NEW_DIT
		,c1,c2,c3,c4,c5,c6,c7
		,s1,s2,s3,s4,s5,s6,s7
#endif
		,t1,t2,t3,t4,t5,t6,t7,t8,t9,t10,t11,t12,t13,t14,t15,t16
		,aj1p0r,aj1p1r,aj1p2r,aj1p3r,aj1p4r,aj1p5r,aj1p6r,aj1p7r
		,aj1p0i,aj1p1i,aj1p2i,aj1p3i,aj1p4i,aj1p5i,aj1p6i,aj1p7i
		,cy0,cy1,cy2,cy3,cy4,cy5,cy6,cy7,temp,scale;
#if PFETCH
	double *add0, *add;
#endif
	int err;
	FILE *f1,*f2;
	static int first_entry=TRUE;

/*...stuff for the reduced-length DWT weights array is here:	*/
	int n_div_nwt;
	int col,co2,co3,m,m2,n_minus_sil,n_minus_silp1,sinwt,sinwtm1;
	double wt,wtinv,wtl,wtlp1,wtn,wtnm1,wtA,wtB,wtC;

/*...change n8 and n_div_wt to non-static to work around a gcc compiler bug. */
	n8   = n/8;
	n_div_nwt = n8 >> nwt_bits;

	if((n_div_nwt << nwt_bits) != n8)
	{
	  if(INTERACT)printf("FATAL: iter = %10d; NWT_BITS does not divide N/8 in radix8_ditN_cy_dif1.\n",iter);
	  f1 = fopen(   OFILE,"a");
	  f2 = fopen(STATFILE,"a");
	  fprintf(f1,"FATAL: iter = %10d; NWT_BITS does not divide N/8 in radix8_ditN_cy_dif1.\n",iter);
	  fprintf(f2,"FATAL: iter = %10d; NWT_BITS does not divide N/8 in radix8_ditN_cy_dif1.\n",iter);
	  fclose(f1);
	  fclose(f2);
	  err=1;
	  return(err);
	}

	if(p != psave)
	{
	  first_entry=TRUE;
	}

/*...initialize things upon first entry	*/

	if(first_entry)
	{
	  psave = p;
	  first_entry=FALSE;
	  radix_inv = 0.125;
	  n2inv = 1.0/(n/2);

	  bw    = p%n;		/* Number of bigwords in the Crandall/Fagin mixed-radix representation = (Mersenne exponent) mod (vector length).	*/
	  sw    = n - bw;	/* Number of smallwords.	*/

	  nm1   = n-1;
	  nrtm1 = nrt-1;

/*   constant index offsets for load/stores are here.	*/

	  p1 = n8 + ( (n8 >> DAT_BITS) << PAD_BITS );
	  p2 = p1 +p1;
	  p3 = p2 +p1;
	  p4 = p3 +p1;
	  p5 = p4 +p1;
	  p6 = p5 +p1;
	  p7 = p6 +p1;

	  bjmodnini=0;
	  for(j=0; j < n8; j++)
	  {
	    bjmodnini -= sw; bjmodnini = bjmodnini + ( (-(int)((uint32)bjmodnini >> 31)) & n);
	  }
	}

/*...The radix-8 final DIT pass is here.	*/

	cy0 =-2;	/* init carries	*/
	cy1 = 0;
	cy2 = 0;
	cy3 = 0;
	cy4 = 0;
	cy5 = 0;
	cy6 = 0;
	cy7 = 0;

	iroot = 0;	/* init sincos array index	*/
	root_incr = 1;	/* init sincos array index increment (set = 1 for normal carry pass, = 0 for wrapper pass)	*/
	scale = n2inv;	/* init inverse-weight scale factor  (set = 2/n for normal carry pass, = 1 for wrapper pass)	*/

	jstart = 0;
	jhi = jstart+nwt-1;
	khi = n_div_nwt;

for(outer=0; outer <= 1; outer++)
{
	i = 1;		/* Pointer to the BASE and BASEINV arrays. lowest-order digit is always a bigword (i = 1).	*/

	bjmodn0 = 0;
	bjmodn1 = bjmodnini;
	bjmodn2 = bjmodn1 +bjmodnini-n; bjmodn2 = bjmodn2 + ( (-(int)((uint32)bjmodn2 >> 31)) & n);
	bjmodn3 = bjmodn2 +bjmodnini-n; bjmodn3 = bjmodn3 + ( (-(int)((uint32)bjmodn3 >> 31)) & n);
	bjmodn4 = bjmodn3 +bjmodnini-n; bjmodn4 = bjmodn4 + ( (-(int)((uint32)bjmodn4 >> 31)) & n);
	bjmodn5 = bjmodn4 +bjmodnini-n; bjmodn5 = bjmodn5 + ( (-(int)((uint32)bjmodn5 >> 31)) & n);
	bjmodn6 = bjmodn5 +bjmodnini-n; bjmodn6 = bjmodn6 + ( (-(int)((uint32)bjmodn6 >> 31)) & n);
	bjmodn7 = bjmodn6 +bjmodnini-n; bjmodn7 = bjmodn7 + ( (-(int)((uint32)bjmodn7 >> 31)) & n);

	col=0;
	co2=(n >> nwt_bits)-1+8;
	co3=co2-8;		/* At the start of each new j-loop, co3=co2-radix(1)	*/

	for(k=1; k <= khi; k++)	/* Do n/(radix(1)*nwt) outer loop executions...	*/
	{
	  for(j=jstart; j<jhi; j += 2)	/* Each inner loop execution processes (radix(1)*nwt) array data.	*/
	  {
	    j1 = j + ( (j >> DAT_BITS) << PAD_BITS );	/* padded-array fetch index is here	*/

#if USE_NEW_DIT
	      t1 =a[j1   ];	t2 =a[j1+1   ];
	      rt =a[j1+p1];	it =a[j1+1+p1];
	      t3 =t1 -rt;	t4 =t2 -it;
	      t1 =t1 +rt;	t2 =t2 +it;

	      t5 =a[j1+p2];	t6 =a[j1+1+p2];
	      rt =a[j1+p3];	it =a[j1+1+p3];
	      t7 =t5 -rt;	t8 =t6 -it;
	      t5 =t5 +rt;	t6 =t6 +it;

	      t9 =a[j1+p4];	t10=a[j1+1+p4];
	      rt =a[j1+p5];	it =a[j1+1+p5];
	      t11=t9 -rt;	t12=t10-it;
	      t9 =t9 +rt;	t10=t10+it;

	      t13=a[j1+p6];	t14=a[j1+1+p6];
	      rt =a[j1+p7];	it =a[j1+1+p7];
	      t15=t13-rt;	t16=t14-it;
	      t13=t13+rt;	t14=t14+it;
#else
/*	here are the needed sincos data - these will be processed below in bit-reversed order.	*/

	    l = iroot;

	      k1=(l & nrtm1);
	      k2=(l >> nrt_bits);
	      l += iroot;
	      t1=rt0[k1].re;	t2=rt0[k1].im;
	      t3=rt1[k2].re;	t4=rt1[k2].im;
	      c1=t1*t3-t2*t4;	s1=t1*t4+t2*t3;

	      k1=(l & nrtm1);
	      k2=(l >> nrt_bits);
	      l += l + iroot;
	      t1=rt0[k1].re;	t2=rt0[k1].im;
	      t3=rt1[k2].re;	t4=rt1[k2].im;
	      c2=t1*t3-t2*t4;	s2=t1*t4+t2*t3;

	      k1=(l & nrtm1);
	      k2=(l >> nrt_bits);
	      t1=rt0[k1].re;	t2=rt0[k1].im;
	      t3=rt1[k2].re;	t4=rt1[k2].im;
	      c5=t1*t3-t2*t4;	s5=t1*t4+t2*t3;

	      t1=c1*c5; t2=c1*s5; t3=s1*c5; t4=s1*s5;
	      c4=t1+t4; s4=t2-t3; c6=t1-t4; s6=t2+t3;

	      t1=c2*c5; t2=c2*s5; t3=s2*c5; t4=s2*s5;
	      c3=t1+t4; s3=t2-t3; c7=t1-t4; s7=t2+t3;

/*       gather the needed data (8 64-bit complex, i.e. 16 64-bit reals) and get the 4 length-2 transforms...	*/
	      t1 =a[j1   ];			t2 =a[j1+1   ];
	      rt =a[j1+p1]*c4 +a[j1+1+p1]*s4;	it =a[j1+1+p1]*c4 -a[j1+p1]*s4;
	      t3 =t1 -rt;			t4 =t2 -it;
	      t1 =t1 +rt;			t2 =t2 +it;

	      t5 =a[j1+p2]*c2 +a[j1+1+p2]*s2;	t6 =a[j1+1+p2]*c2 -a[j1+p2]*s2;
	      rt =a[j1+p3]*c6 +a[j1+1+p3]*s6;	it =a[j1+1+p3]*c6 -a[j1+p3]*s6;
	      t7 =t5 -rt;			t8 =t6 -it;
	      t5 =t5 +rt;			t6 =t6 +it;

	      t9 =a[j1+p4]*c1 +a[j1+1+p4]*s1;	t10=a[j1+1+p4]*c1 -a[j1+p4]*s1;
	      rt =a[j1+p5]*c5 +a[j1+1+p5]*s5;	it =a[j1+1+p5]*c5 -a[j1+p5]*s5;
	      t11=t9 -rt;			t12=t10-it;
	      t9 =t9 +rt;			t10=t10+it;

	      t13=a[j1+p6]*c3 +a[j1+1+p6]*s3;	t14=a[j1+1+p6]*c3 -a[j1+p6]*s3;
	      rt =a[j1+p7]*c7 +a[j1+1+p7]*s7;	it =a[j1+1+p7]*c7 -a[j1+p7]*s7;
	      t15=t13-rt;			t16=t14-it;
	      t13=t13+rt;			t14=t14+it;
#endif
/*       combine to get the 2 length-4 transforms...	*/
	      rt =t5;	t5 =t1 -rt;	t1 =t1 +rt;
	      it =t6;	t6 =t2 -it;	t2 =t2 +it;

	      rt =t7;	t7 =t3 -t8;	t3 =t3 +t8;
			t8 =t4 +rt;	t4 =t4 -rt;

	      rt =t13;	t13=t9 -rt;	t9 =t9 +rt;
	      it =t14;	t14=t10-it;	t10=t10+it;

	      rt =t15;	t15=t11-t16;	t11=t11+t16;
			t16=t12+rt;	t12=t12-rt;
/*       now combine the two half-transforms	*/
	      aj1p0r=t1+t9;		aj1p0i=t2+t10;
	      aj1p4r=t1-t9;		aj1p4i=t2-t10;

	      rt=(t11+t12)*ISRT2;	it=(t11-t12)*ISRT2;
	      aj1p1r=t3+rt;		aj1p1i=t4-it;
	      aj1p5r=t3-rt;		aj1p5i=t4+it;

	      aj1p2r=t5+t14;		aj1p2i=t6-t13;
	      aj1p6r=t5-t14;		aj1p6i=t6+t13;

	      rt=(t15-t16)*ISRT2;	it=(t15+t16)*ISRT2;
	      aj1p3r=t7-rt;		aj1p3i=t8-it;
	      aj1p7r=t7+rt;		aj1p7i=t8+it;

/*...and combine those to complete the radix-8 transform and do the carries. Since the outputs would
     normally be getting dispatched to 8 separate blocks of the A-array, we need 8 separate carries.	*/

	/* We can use the identity WT(J)*WT(N-J) = 2 here to obviate the
	   need for an array of inverse weights - if we need the inverse
	   of WT(J), just use WT(N-J)/2. Together with a reduced-array weights scheme
	   this needs L* = (N - J) & (NWT - 1) = NWT - [J & (NWT - 1)] = NWT - L;
	   also needs 	*/

	    l= j & (nwt-1);			/* We want (S*J mod N) - SI(L) for all 16 carries, so precompute	*/
	    n_minus_sil   = n-si[l  ];		/* N - SI(L) and for each J, find N - (B*J mod N) - SI(L)		*/
	    n_minus_silp1 = n-si[l+1];		/* For the inverse weight, want (S*(N - J) mod N) - SI(NWT - L) =	*/
	    sinwt   = si[nwt-l  ];		/*	= N - (S*J mod N) - SI(NWT - L) = (B*J mod N) - SI(NWT - L).	*/
	    sinwtm1 = si[nwt-l-1];

	    wtl     =wt0[    l  ];
	    wtn     =wt0[nwt-l  ]*scale;	/* Include 1/(n/2) scale factor of inverse transform here...	*/
	    wtlp1   =wt0[    l+1];
	    wtnm1   =wt0[nwt-l-1]*scale;	/* ...and here.	*/

/*...set0 is slightly different from others:	*/
	   cmplx_carry_norm_pow2_nocheck0(aj1p0r,aj1p0i,cy0,bjmodn0);
	    cmplx_carry_norm_pow2_nocheck(aj1p1r,aj1p1i,cy1,bjmodn1,0x1);
	    cmplx_carry_norm_pow2_nocheck(aj1p2r,aj1p2i,cy2,bjmodn2,0x2);
	    cmplx_carry_norm_pow2_nocheck(aj1p3r,aj1p3i,cy3,bjmodn3,0x3);
	    cmplx_carry_norm_pow2_nocheck(aj1p4r,aj1p4i,cy4,bjmodn4,0x4);
	    cmplx_carry_norm_pow2_nocheck(aj1p5r,aj1p5i,cy5,bjmodn5,0x5);
	    cmplx_carry_norm_pow2_nocheck(aj1p6r,aj1p6i,cy6,bjmodn6,0x6);
	    cmplx_carry_norm_pow2_nocheck(aj1p7r,aj1p7i,cy7,bjmodn7,0x7);

	    i =((uint32)(sw - bjmodn0) >> 31);	/* get ready for the next set...	*/
	    co2=co3;	/* For all data but the first set in each j-block, co2=co3. Thus, after the first block of data is done
			   (and only then: for all subsequent blocks it's superfluous), this assignment decrements co2 by radix(1).	*/

/*...The radix-8 DIF pass is here:	*/
#if PFETCH
add0 = &a[j1];
prefetch_p_doubles(add);
#endif
/*       gather the needed data (8 64-bit complex, i.e. 16 64-bit reals) and combine to get the 4 length-2 transforms...	*/
	  t3 =aj1p0r -aj1p4r;	t4 =aj1p0i -aj1p4i;
	  t1 =aj1p0r +aj1p4r;	t2 =aj1p0i +aj1p4i;

	  t7 =aj1p2r -aj1p6r;	t8 =aj1p2i -aj1p6i;
	  t5 =aj1p2r +aj1p6r;	t6 =aj1p2i +aj1p6i;
#if PFETCH
add = add0+p1;
prefetch_p_doubles(add);
#endif
	  t11=aj1p1r -aj1p5r;	t12=aj1p1i -aj1p5i;
	  t9 =aj1p1r +aj1p5r;	t10=aj1p1i +aj1p5i;

	  t15=aj1p3r -aj1p7r;	t16=aj1p3i -aj1p7i;
	  t13=aj1p3r +aj1p7r;	t14=aj1p3i +aj1p7i;
#if PFETCH
add = add0+p2;
prefetch_p_doubles(add);
#endif
/*       combine to get the 2 length-4 transforms...	*/
	  rt =t5;	t5 =t1 -rt;	t1 =t1 +rt;
	  it =t6;	t6 =t2 -it;	t2 =t2 +it;

	  rt =t7;	t7 =t3 +t8;	t3 =t3 -t8;
			t8 =t4 -rt;	t4 =t4 +rt;
#if PFETCH
add = add0+p3;
prefetch_p_doubles(add);
#endif
	  rt =t13;	t13=t9 -rt;	t9 =t9 +rt;
	  it =t14;	t14=t10-it;	t10=t10+it;

	  rt =t15;	t15=t11+t16;	t11=t11-t16;
			t16=t12-rt;	t12=t12+rt;
/*       now combine the two half-transforms	*/
#if PFETCH
add = add0+p4;
prefetch_p_doubles(add);
#endif
	  a[j1   ]=t1+t9;	a[j1+1   ]=t2+t10;
	  a[j1+p1]=t1-t9;	a[j1+1+p1]=t2-t10;
#if PFETCH
add = add0+p5;
prefetch_p_doubles(add);
#endif
	  a[j1+p2]=t5-t14;	a[j1+1+p2]=t6+t13;
	  a[j1+p3]=t5+t14;	a[j1+1+p3]=t6-t13;
#if PFETCH
add = add0+p6;
prefetch_p_doubles(add);
#endif
	  rt =(t11-t12)*ISRT2;	it =(t11+t12)*ISRT2;
	  a[j1+p4]=t3+rt;	a[j1+1+p4]=t4+it;
	  a[j1+p5]=t3-rt;	a[j1+1+p5]=t4-it;
#if PFETCH
add = add0+p7;
prefetch_p_doubles(add);
#endif
	  rt =(t15+t16)*ISRT2;	it =(t16-t15)*ISRT2;
	  a[j1+p6]=t7-rt;	a[j1+1+p6]=t8-it;
	  a[j1+p7]=t7+rt;	a[j1+1+p7]=t8+it;

	  iroot += root_incr;		/* increment sincos index.	*/

	  }

	  jstart += nwt;
	  jhi    += nwt;
	  col += 8;
	  co3 -= 8;

	}

	if(root_incr==0)break;

/*
!   Wraparound carry cleanup loop is here:
!
!   The cleanup carries from the end of each length-N/8 block into the begining of the next
!   can all be neatly processed as follows:
!
!   (1) Invert the radix-8 forward DIF FFT of the first block of 8 complex elements in A and unweight;
!   (2) Propagate cleanup carries among the real and imaginary parts of the 8 outputs of (1);
!   (3) Reweight and perform a radix-8 forward DIF FFT on the result of (2);
!   (4) If any of the exit carries from (2) are nonzero, advance to the next 8 elements and repeat (1-4).
*/
	t1  = cy7;
	cy7 = cy6;
	cy6 = cy5;
	cy5 = cy4;
	cy4 = cy3;
	cy3 = cy2;
	cy2 = cy1;
	cy1 = cy0;
	cy0 = t1;

	iroot = 0;
	root_incr = 0;
	scale = 1;

	jstart = 0;
	jhi = 3;
	khi = 1;

	for(j=0; j<=jhi; j++)
	{
	  a[j   ] *= radix_inv;
	  a[j+p1] *= radix_inv;
	  a[j+p2] *= radix_inv;
	  a[j+p3] *= radix_inv;
	  a[j+p4] *= radix_inv;
	  a[j+p5] *= radix_inv;
	  a[j+p6] *= radix_inv;
	  a[j+p7] *= radix_inv;
	}
}

	if(fabs(cy0)+fabs(cy1)+fabs(cy2)+fabs(cy3)+fabs(cy4)+fabs(cy5)+fabs(cy6)+fabs(cy7) != 0.0)
	{
	    if(INTERACT)printf("FATAL: iter = %10d; nonzero exit carry in radix8_ditN_cy_dif1_nochk.\n",iter);
	    f1 = fopen(   OFILE,"a");
	    f2 = fopen(STATFILE,"a");
	    fprintf(f1,"FATAL: iter = %10d; nonzero exit carry in radix8_ditN_cy_dif1_nochk.\n",iter);
	    fprintf(f2,"FATAL: iter = %10d; nonzero exit carry in radix8_ditN_cy_dif1_nochk.\n",iter);
	    fclose(f1);
	    fclose(f2);
	    err=1;
	    return(err);
	}

	return(0);
}

/***************/

void radix8_dit_pass1(double a[], int n)
{
/*
!...Acronym: DIT = Decimation In Time
!
!...Subroutine to perform an initial radix-8 complex inverse DIT FFT pass on the data in the length-N real vector A.
!
!   This routine is designed exclusively to undo the effects of radix8_dif_pass1,
!   i.e. to reobtain the raw all-integer residue vector at the end of an iteration cycle.
!
!   See the documentation in radix16_dif_pass for further details on the array indexing,
!   and radix8_dif_pass1 for details on the algorithm.
*/
	int j,j1,j2,j3,j4,j5,j6,j7,j8;
	static int n8, first_entry=TRUE;
	double rt,it
		,t1,t2,t3,t4,t5,t6,t7,t8,t9,t10,t11,t12,t13,t14,t15,t16;

	if(!first_entry && (n >> 3) != n8)	/* New runlength?	*/
	{
	  first_entry=TRUE;
	}

/*...initialize things upon first entry	*/

	if(first_entry)
	{
	  first_entry=FALSE;
	  n8 = n/8;
	}

/*...The radix-8 pass is here.	*/

      for(j=0; j < n8; j += 2)
      {
	  j2 = j  +n8;
	  j3 = j2 +n8;
	  j4 = j3 +n8;
	  j5 = j4 +n8;
	  j6 = j5 +n8;
	  j7 = j6 +n8;
	  j8 = j7 +n8;

	  j1 = j  + ( (j  >> DAT_BITS) << PAD_BITS );
	  j2 = j2 + ( (j2 >> DAT_BITS) << PAD_BITS );
	  j3 = j3 + ( (j3 >> DAT_BITS) << PAD_BITS );
	  j4 = j4 + ( (j4 >> DAT_BITS) << PAD_BITS );
	  j5 = j5 + ( (j5 >> DAT_BITS) << PAD_BITS );
	  j6 = j6 + ( (j6 >> DAT_BITS) << PAD_BITS );
	  j7 = j7 + ( (j7 >> DAT_BITS) << PAD_BITS );
	  j8 = j8 + ( (j8 >> DAT_BITS) << PAD_BITS );

/*       gather the needed data (8 64-bit complex, i.e. 16 64-bit reals) and combine to get the 4 length-2 transforms...	*/
	  t1 =a[j1];			t2 =a[j1+1];
	    rt =a[j2];			it =a[j2+1];
	    t3 =t1 -rt;			t4 =t2 -it;
	    t1 =t1 +rt;			t2 =t2 +it;

	    t5 =a[j3];			t6 =a[j3+1];
	    rt =a[j4];			it =a[j4+1];
	    t7 =t5 -rt;			t8 =t6 -it;
	    t5 =t5 +rt;			t6 =t6 +it;

	    t9 =a[j5];			t10=a[j5+1];
	    rt =a[j6];			it =a[j6+1];
	    t11=t9 -rt;			t12=t10-it;
	    t9 =t9 +rt;			t10=t10+it;

	    t13=a[j7];			t14=a[j7+1];
	    rt =a[j8];			it =a[j8+1];
	    t15=t13-rt;			t16=t14-it;
	    t13=t13+rt;			t14=t14+it;
/*       combine to get the 2 length-4 transform...	*/
	    rt =t5;			it =t6;
	    t5 =t1 -rt;			t6 =t2 -it;
	    t1 =t1 +rt;			t2 =t2 +it;

	    rt =t7;			it =t8;
	    t7 =t3 -it;			t8 =t4 +rt;
	    t3 =t3 +it;			t4 =t4 -rt;

	    rt =t13;			it =t14;
	    t13=t9 -rt;			t14=t10-it;
	    t9 =t9 +rt;			t10=t10+it;

	    rt =t15;			it =t16;
	    t15=t11-it;			t16=t12+rt;
	    t11=t11+it;			t12=t12-rt;
/*       now combine the two half-transforms	*/
	    a[j1]=t1+t9;		a[j1+1]=t2+t10;
	    a[j5]=t1-t9;		a[j5+1]=t2-t10;

	    rt=(t11+t12)*ISRT2;		it=(t11-t12)*ISRT2;
	    a[j2]=t3+rt;		a[j2+1]=t4-it;
	    a[j6]=t3-rt;		a[j6+1]=t4+it;

	    a[j3]=t5+t14;		a[j3+1]=t6-t13;
	    a[j7]=t5-t14;		a[j7+1]=t6+t13;

	    rt=(t15-t16)*ISRT2;		it=(t15+t16)*ISRT2;
	    a[j4]=t7-rt;		a[j4+1]=t8-it;
	    a[j8]=t7+rt;		a[j8+1]=t8+it;
	}
}
