/*******************************************************************************
*                                                                              *
*   (C) 1997-2002 by Ernst W. Mayer.                                           *
*                                                                              *
*  This program is free software; you can redistribute it and/or modify it     *
*  under the terms of the GNU General Public License as published by the       *
*  Free Software Foundation; either version 2 of the License, or (at your      *
*  option) any later version.                                                  *
*                                                                              *
*  This program is distributed in the hope that it will be useful, but WITHOUT *
*  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or       *
*  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for   *
*  more details.                                                               *
*                                                                              *
*  You should have received a copy of the GNU General Public License along     *
*  with this program; see the file GPL.txt.  If not, you may view one at       *
*  http://www.fsf.org/licenses/licenses.html, or obtain one by writing to the  *
*  Free Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA     *
*  02111-1307, USA.                                                            *
*                                                                              *
*******************************************************************************/

/****************************************************************************
 * We now include this header file if it was not included before.
 ****************************************************************************/
#ifndef types_h_included
#define types_h_included

/* Typedefs */

/*...useful utility parameters */

#undef TRUE
#define TRUE	1

#undef FALSE
#define FALSE	0

/* Basic integer types - yes, I'm assuming 'int' means 32 bits, but this assumption gets checked
at the start of program execution, so we're not flying blind: */
typedef          int		 int32;
typedef unsigned int		uint32;
/* MSVC doesn't like 'long long', and of course MS has their own completely non-portable substitute:	*/
#ifdef WINDOWS
	typedef   signed __int64	 int64;
	typedef unsigned __int64	uint64;
#else
	typedef          long long	 int64;
	typedef unsigned long long	uint64;
#endif

/* Some useful simple macros: */
#undef  nint
#define nint(x) floor(x + 0.5)			/* Slow version of nearest-int; for doubles we use a much faster trick,
						   but this is useful for reference purposes and error checking */
#undef  max
#undef  min
#define max(a,b) ((a) > (b) ? (a) : (b))	/* NOTE: when calling these macros, do NOT allow either argument to be a    */
#define min(a,b) ((a) < (b) ? (a) : (b))	/* function call, since that may result in the function being called twice. */

#undef  labs					/* 64-bit absval is not portable, so alias it. */
#define labs(a) ((a) >= 0 ? (a) : (-a))

#endif	/* types_h_included */
