/*******************************************************************************
*                                                                              *
*   (C) 1997-2002 by Ernst W. Mayer.                                           *
*                                                                              *
*  This program is free software; you can redistribute it and/or modify it     *
*  under the terms of the GNU General Public License as published by the       *
*  Free Software Foundation; either version 2 of the License, or (at your      *
*  option) any later version.                                                  *
*                                                                              *
*  This program is distributed in the hope that it will be useful, but WITHOUT *
*  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or       *
*  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for   *
*  more details.                                                               *
*                                                                              *
*  You should have received a copy of the GNU General Public License along     *
*  with this program; see the file GPL.txt.  If not, you may view one at       *
*  http://www.fsf.org/licenses/licenses.html, or obtain one by writing to the  *
*  Free Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA     *
*  02111-1307, USA.                                                            *
*                                                                              *
*******************************************************************************/

#include "Mlucas.h"

/***************/

void check_nbits_in_types()
{
/* The check-nint code below didn't work properly with (double)twopi and twopi a
   #defined long double, so store the test floats into local doubles first. */

double tpi;

/* On x86, make sure we're using the full 64 mantissa bits of the floating-point registers: */
#if(defined(__i386__))
unsigned short FPUCTRL;
	/* Copy the FPU control word set by the compiler to a local variable... */
	#if(defined(__MWERKS__))
		__asm	fstcw	FPUCTRL
	#elif(defined(__GNUC__) || defined(__GNUG__))
		__asm__ volatile ("fstcw %0" : "=m" (FPUCTRL) );
	#endif
	/* ...and inform the user if the compiler isn't specifying 64-bit mantissa mode. */
	switch (FPUCTRL & 0x0300) {
	case 0x0300:
		break;
	case 0x0200:
		printf("INFO: compiler sets x86 FPU to 53-bit mantissa mode.\n");
		break;
	case 0x0000:
		printf("INFO: compiler sets x86 FPU to 24-bit mantissa mode.\n");
		break;
	default:
		printf("INFO: compiler sets x86 FPU to unknown precision. FPU Control Word: %4X\n", FPUCTRL);
	}
	/* Irrespective of what values the compiler set for bits 8 and 9 of FPUCTRL,
	set them both to 1 to get the full 64-mantissa-bit precision available on the x86. */
	printf("Setting x86 FPU to 64-bit mantissa mode.\n");
	FPUCTRL |= 0x0300;
	/* ...and then reload the FPU control word for the changes to take effect. */
	#if(defined(__MWERKS__))
		__asm	fldcw	FPUCTRL
	#elif(defined(__GNUC__) || defined(__GNUG__))
		__asm__ volatile ("fldcw %0" :: "m" (FPUCTRL) );
	#endif
#endif

#if DEBUG
	printf("INFO: Program compiled with debugging diagnostics ON.\n");
#endif

#ifdef PPC
	printf("PowerPC...accept no substitutes.\n");
#endif

#if   defined(ia64_gcc)
	printf("Itanium, compiled with Gnu C compiler.\n");
#elif defined(ia64_icc)
	printf("Itanium, compiled with Intel C compiler Version %u.\n", __INTEL_COMPILER);
#elif defined(ia64_hpc)
	printf("Itanium, compiled with HP C compiler.\n");
	/*fesetprec(FE_EXTPREC);	<== This doesn't have the hoped-for effect; need to discuss with HP compiler group*/
#elif defined(hppa_hpc)
	printf("PA-RISC, compiled with HP C compiler.\n");
#endif

#if PFETCH
	printf("INFO: Using prefetch.\n");
#endif

/* Check lengths of basic data types: */
	if(!(sizeof( int32) == 4)) { printf("ERROR 10 in util.c\n"); exit(EXIT_FAILURE); }
	if(!(sizeof(uint32) == 4)) { printf("ERROR 12 in util.c\n"); exit(EXIT_FAILURE); }
	if(!(sizeof( int64) == 8)) { printf("ERROR 14 in util.c\n"); exit(EXIT_FAILURE); }
	if(!(sizeof(uint64) == 8)) { printf("ERROR 16 in util.c\n"); exit(EXIT_FAILURE); }
	if(!(sizeof(  long) == sizeof(void *)))   { printf("ERROR 18 in util.c\n"); exit(EXIT_FAILURE); }	/* ALIGN_DOUBLES assumes this. */
	if(!(sizeof(long long) >= 2*sizeof(int))) { printf("ERROR 20 in util.c\n"); exit(EXIT_FAILURE); }	/* Make sure long long large enough to hold product of 2 ints	*/

/* Make sure the Sparc V9+ prefetch is being implemented */
#if(defined(__SUN_PREFETCH))
   printf("Sparc CPU supports prefetch\n");
#endif

/* Use qfloat routines to set various global floating-point constants and the local var. tpi: */
	ISRT2 = qfdbl(qfmul_pow2(QSQRT2, -1));		/* 1/sqrt2	*/
	tpi   = qfdbl(Q2PI);

/* Set the value of the round constant used for fast NINT emulation: */
/* X86 64-mantissa-bit register doubles: */
//#if(defined(__i386__) || defined(__ia64__) || (defined(__ia64) && defined(hppa_hpc) && defined(_FPEVAL_EXTENDED)))	/* Last of these is for HP C or C++ compiler for HPUX */
#if(defined(__i386__) || (defined(__ia64) && defined(ia64_hpc) && defined(_FPEVAL_EXTENDED)))	/* Last of these is for HP C or C++ compiler for HPUX */
	RND_A = 3.0*0x4000000*0x2000000*0x800;
	RND_B =12.0*0x2000000*0x1000000*0x800;
	printf("INFO: using 80-bit-double form of rounding constant\n");
/* These assume IEEE64-compliant double-precision hardware arithmetic. */
#else
	RND_A = 3.0*0x4000000*0x2000000;
	RND_B =12.0*0x2000000*0x1000000;
	printf("INFO: using 64-bit-double form of rounding constant\n");
#endif

	/*if( ((tpi+RND_A)-RND_B) != 6.0 || ((ISRT2+RND_A)-RND_B) != 1.0 )
	{
		printf("INFO: 2*pi  = %20.15f,  rnd(2*pi ) = %20.15f\n", tpi  , (tpi  +RND_A)-RND_B );
		printf("INFO: ISRT2 = %20.15f,  rnd(ISRT2) = %20.15f\n", ISRT2, (ISRT2+RND_A)-RND_B );
		{ printf("ERROR 30 in util.c\n"); exit(EXIT_FAILURE); }
	}*/
}

/***************/

int leadz(int x)
{
/*   Returns number of leading (leftmost) zeros of the integer x.	*/

	uint32 i;
	for(i = 1; i < 8*sizeof(int); i++)
	{
		if(x < 0)break;
		x <<= 1;
	}
	return(i-1);
}

/***************/

int trailz(int x)
{
/*   Returns number of trailing (rightmost) zeros of the 8-byte integer x.	*/

	uint32 i;
	for(i = 1; i < 8*sizeof(int); i++)
	{
		if(x & 1)break;
		x >>= 1;
	}
	return(i-1);
}

/***************/

int pprimeF(uint32 p, uint32 z)
{
/* returns 1 if p is a base-z Fermat pseudoprimes, 0 otherwise. */

	unsigned long long pt = p, zt = z, y = 1, n = p-1, flag;

	while(n)
	{
		flag = n & 1;
		n >>= 1;
		if(flag) y = (y*zt)%pt;
		zt = (zt*zt)%pt;
		if(!zt) return 0;
	}
	return((int)(y==1));
}

/********************/

void errprint_sincos(double *x, double *y, double theta)
{
/*...This is a workaround for the the DEC Unix V4.0 real*16 sincos bug, which causes problems when the argument
      = m*pi/4 +- pi/512, m integer. If an error in the real*16 result is detected, we use the real*8 value. Since this
     tends to happen for a small fraction of the sincos data, the overall accuracy of the calculation is essentially unchanged.
     To see whether this occurs on your system and for your runlength, uncomment the print statements below.	*/

	double tmp;

	tmp = cos(theta);
	if(fabs(*x - tmp) > 1e-10)
	{
	  printf("WARNING: real*16 cosine error ; using double-precision values instead.\n");
	  printf("theta = %20.15f, lond double = %20.15f, double = %20.15f\n", theta,*x,tmp);
	  *x = tmp;
	}

	tmp = sin(theta);
	if(fabs(*y - tmp) > 1e-10)
	{
	  printf("WARNING: real*16 sine error ; using double-precision values instead.\n");
	  printf("theta = %20.15f, lond double = %20.15f, double = %20.15f\n", theta,*y,tmp);
	  *y = tmp;
	}
}

/********************/

void hex_res64(double a[], int n, int p, int *nbits, char *hex_res)
{
/*
!   Character function to take an n-digit LL residue, convert it to positive-digit form,
!   and pretty-print the lowest-order 16 hex digits.
*/
	int bimodn,bits,cy,i,ii,j;
	int bw,sw,base_small,bits_small;
	int64  itmp;
	uint64 res64 = 0;
	static int hex_chars[16] = {'0','1','2','3','4','5','6','7','8','9','A','B','C','D','E','F'};

	hex_res[0]='\0';
	hex_res[16]='\0';
	bits_small = p/n;
	base_small = 1 << bits_small;
	bw = p%n;	/* cf. MERS_MOD_SQUARE	*/
	sw = n - bw;
/*
!...if most-significant digit in the balanced-representation form is < 0, add M(p) to the residue.
!   Since we assume M(p) has >> 64 bits and we seek only the lowest-order 64, can do this by just
!   subtracting 1 from a(1) and then normalizing the lowest-order few digits that give a total >= 64 bits.
*/
	cy=0;		/* init carry.	*/

	for(i=n-1; i >= 0; i--)
	{
	  j = i + ( (i >> DAT_BITS) << PAD_BITS );
	  if(a[j] != 0.0)
	  {
	    if(a[j] < 0.0) cy=-1;	/* If leading digit < 0, add M(p) to the residue, which amounts to subtracting 1 from the least-significant word.	*/
	    break;
	  }
	}

/*...Now form the hexadecimal residue, converting digits to positive-digit form along the way...	*/

	bimodn = 0;
	ii     = 1;	/* lowest-order digit always a bigword.	*/
	*nbits = 0;	/* Total bits accumulated so far in the hex residue	*/

	for(i=0; i < n; i++)
	{
	  j = i + ( (i >> DAT_BITS) << PAD_BITS );

	  bits = bits_small + ii;	/* number of bits in the current word.	*/

	  itmp = (int64)(a[j] + cy);	/* current digit in int*4 form, subtracting any borrow from the previous digit.	*/
	  if(itmp < 0)			/* If current digit < 0, add the current base and set carry into next-higher digit = -1	*/
	  {
	    itmp += (base_small << ii);
	    cy = -1;
	  }
	  else
	  {
	    cy = 0;
	  }

	  res64 += (itmp << *nbits);
	  *nbits += bits;
	  if(*nbits >= 64){ *nbits = 64; break; }

	  bimodn += bw;
	  if(bimodn >= n) bimodn -= n;
	  ii = (uint32)(sw - bimodn) >> 31;
	}

	for(i=15; i >= 16-(*nbits>>2); i--)
	{
	  hex_res[i] = hex_chars[res64 & 15];
	  res64 >>= 4;
	}

	for(i=15-(*nbits>>2); i >= 0; i--)
	{
	  hex_res[i] = 'x';
	}
	return;
}

/********************/

/* Complex multiplication */
struct complex cmul(struct complex *a, struct complex *b)
{
	struct complex cout;
	cout.re = (*a).re*(*b).re - (*a).im*(*b).im;
	cout.im = (*a).re*(*b).im + (*a).im*(*b).re;
	return cout;
}

