/*******************************************************************************
*                                                                              *
*   (C) 1997-2004 by Ernst W. Mayer.                                           *
*                                                                              *
*  This program is free software; you can redistribute it and/or modify it     *
*  under the terms of the GNU General Public License as published by the       *
*  Free Software Foundation; either version 2 of the License, or (at your      *
*  option) any later version.                                                  *
*                                                                              *
*  This program is distributed in the hope that it will be useful, but WITHOUT *
*  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or       *
*  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for   *
*  more details.                                                               *
*                                                                              *
*  You should have received a copy of the GNU General Public License along     *
*  with this program; see the file GPL.txt.  If not, you may view one at       *
*  http://www.fsf.org/licenses/licenses.html, or obtain one by writing to the  *
*  Free Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA     *
*  02111-1307, USA.                                                            *
*                                                                              *
*******************************************************************************/

#include "Mlucas.h"

/***************/

void radix36_dif_pass1(double a[], int n)
{
/*
!...Acronym: DIF = Decimation In Frequency
!
!...Subroutine to perform an initial radix-36 complex DIF FFT pass on the data in the length-N real vector A.
*/
	int j,jt,j1;
	static int n36, first_entry=TRUE
		    ,p01,p02,p03,p04,p05,p06,p07,p08,p09,p10,p11,p12,p13,p14,p15,p16,p17
		,p18,p19,p20,p21,p22,p23,p24,p25,p26,p27,p28,p29,p30,p31,p32,p33,p34,p35;
	static double c   =  0.76604444311897803520,	/* cos(2*pi/9) */
		      s   =  0.64278760968653932631,	/* sin(2*pi/9) */
		      c2  =  0.17364817766693034887,	/* cos(2*u) */
		      s2  =  0.98480775301220805936,	/* sin(2*u) */
		      c3m1= -1.50000000000000000000,	/* cos(3*u)-1 */
		      s3  =  0.86602540378443864677,	/* sin(3*u) */
		      c4  = -0.93969262078590838404,	/* cos(4*u) */
		      s4  =  0.34202014332566873307;	/* sin(4*u) */
	double rt,it,re
		,t00,t01,t02,t03,t04,t05,t06,t07,t08,t09,t0a,t0b,t0c,t0d,t0e,t0f,t0g,t0h
		,t10,t11,t12,t13,t14,t15,t16,t17,t18,t19,t1a,t1b,t1c,t1d,t1e,t1f,t1g,t1h
		,t20,t21,t22,t23,t24,t25,t26,t27,t28,t29,t2a,t2b,t2c,t2d,t2e,t2f,t2g,t2h
		,t30,t31,t32,t33,t34,t35,t36,t37,t38,t39,t3a,t3b,t3c,t3d,t3e,t3f,t3g,t3h;

	if(!first_entry && (n/36) != n36)	/* New runlength?	*/
	{
	  first_entry=TRUE;
	}

/*...initialize things upon first entry	*/

	if(first_entry)
	{
	  first_entry=FALSE;
	  n36=n/36;

/*   constant index offsets for array load/stores are here.	*/

	  p01 = n36;
	  p02 = p01 + p01;
	  p03 = p02 + p01;
	  p04 = p03 + p01;
	  p05 = p04 + p01;
	  p06 = p05 + p01;
	  p07 = p06 + p01;
	  p08 = p07 + p01;
	  p09 = p08 + p01;
	  p10 = p09 + p01;
	  p11 = p10 + p01;
	  p12 = p11 + p01;
	  p13 = p12 + p01;
	  p14 = p13 + p01;
	  p15 = p14 + p01;
	  p16 = p15 + p01;
	  p17 = p16 + p01;
	  p18 = p17 + p01;
	  p19 = p18 + p01;
	  p20 = p19 + p01;
	  p21 = p20 + p01;
	  p22 = p21 + p01;
	  p23 = p22 + p01;
	  p24 = p23 + p01;
	  p25 = p24 + p01;
	  p26 = p25 + p01;
	  p27 = p26 + p01;
	  p28 = p27 + p01;
	  p29 = p28 + p01;
	  p30 = p29 + p01;
	  p31 = p30 + p01;
	  p32 = p31 + p01;
	  p33 = p32 + p01;
	  p34 = p33 + p01;
	  p35 = p34 + p01;

	  p01 = p01 + ( (p01 >> DAT_BITS) << PAD_BITS );
	  p02 = p02 + ( (p02 >> DAT_BITS) << PAD_BITS );
	  p03 = p03 + ( (p03 >> DAT_BITS) << PAD_BITS );
	  p04 = p04 + ( (p04 >> DAT_BITS) << PAD_BITS );
	  p05 = p05 + ( (p05 >> DAT_BITS) << PAD_BITS );
	  p06 = p06 + ( (p06 >> DAT_BITS) << PAD_BITS );
	  p07 = p07 + ( (p07 >> DAT_BITS) << PAD_BITS );
	  p08 = p08 + ( (p08 >> DAT_BITS) << PAD_BITS );
	  p09 = p09 + ( (p09 >> DAT_BITS) << PAD_BITS );
	  p10 = p10 + ( (p10 >> DAT_BITS) << PAD_BITS );
	  p11 = p11 + ( (p11 >> DAT_BITS) << PAD_BITS );
	  p12 = p12 + ( (p12 >> DAT_BITS) << PAD_BITS );
	  p13 = p13 + ( (p13 >> DAT_BITS) << PAD_BITS );
	  p14 = p14 + ( (p14 >> DAT_BITS) << PAD_BITS );
	  p15 = p15 + ( (p15 >> DAT_BITS) << PAD_BITS );
	  p16 = p16 + ( (p16 >> DAT_BITS) << PAD_BITS );
	  p17 = p17 + ( (p17 >> DAT_BITS) << PAD_BITS );
	  p18 = p18 + ( (p18 >> DAT_BITS) << PAD_BITS );
	  p19 = p19 + ( (p19 >> DAT_BITS) << PAD_BITS );
	  p20 = p20 + ( (p20 >> DAT_BITS) << PAD_BITS );
	  p21 = p21 + ( (p21 >> DAT_BITS) << PAD_BITS );
	  p22 = p22 + ( (p22 >> DAT_BITS) << PAD_BITS );
	  p23 = p23 + ( (p23 >> DAT_BITS) << PAD_BITS );
	  p24 = p24 + ( (p24 >> DAT_BITS) << PAD_BITS );
	  p25 = p25 + ( (p25 >> DAT_BITS) << PAD_BITS );
	  p26 = p26 + ( (p26 >> DAT_BITS) << PAD_BITS );
	  p27 = p27 + ( (p27 >> DAT_BITS) << PAD_BITS );
	  p28 = p28 + ( (p28 >> DAT_BITS) << PAD_BITS );
	  p29 = p29 + ( (p29 >> DAT_BITS) << PAD_BITS );
	  p30 = p30 + ( (p30 >> DAT_BITS) << PAD_BITS );
	  p31 = p31 + ( (p31 >> DAT_BITS) << PAD_BITS );
	  p32 = p32 + ( (p32 >> DAT_BITS) << PAD_BITS );
	  p33 = p33 + ( (p33 >> DAT_BITS) << PAD_BITS );
	  p34 = p34 + ( (p34 >> DAT_BITS) << PAD_BITS );
	  p35 = p35 + ( (p35 >> DAT_BITS) << PAD_BITS );
	}

/*...The radix-36 pass is here.	*/

      for(j=0; j < n36; j += 2)
      {
	  jt = j  + ( (j  >> DAT_BITS) << PAD_BITS );
/*
Twiddleless version requires us to swap inputs as follows:
indices  0, 1, 2, 3, 4, 5, 6, 7, 8, 9,10,11,12,13,14,15,16,17,18,19,20,21,22,23,24,25,26,27,28,29,30,31,32,33,34,35
      -> 0,27,18, 9,32,23,14, 5,28,19,10, 1,24,15, 6,33,20,11, 2,29,16, 7,34,25,12, 3,30,21, 8,35,26,17, 4,31,22,13.
I.e. start out with first nonet of indices {0,4,8,12,16,20,24,28,32}, permute those according to
{0,4,8,12,16,20,24,28,32}*35%36 = {0,32,28,24,20,16,12,8,4}, then each is head of a length-4 list of indices with decrement 9.
*/

/*...First radix-9 transform:	*/
	/* permuted version of jt + p00,p12,p24 */
	  t00 =a[jt    ];		t01 =a[jt    +1];
	  t02 =a[jt+p24]+a[jt+p12];	t03 =a[jt+p24+1]+a[jt+p12+1];
	  t04 =a[jt+p24]-a[jt+p12];	t05 =a[jt+p24+1]-a[jt+p12+1];
	  t00 =t00+t02;			t01 =t01+t03;
	  t02 =t00+c3m1*t02;		t03 =t01+c3m1*t03;
	  rt  =s3*t04;			it  =s3*t05;
	  t04 =t02+it;			t05 =t03-rt;
	  t02 =t02-it;			t03 =t03+rt;
	/* permuted version of jt + p04,p16,p28 */
	  t06 =a[jt+p32];		t07 =a[jt+p32+1];
	  t08 =a[jt+p20]+a[jt+p08];	t09 =a[jt+p20+1]+a[jt+p08+1];
	  t0a =a[jt+p20]-a[jt+p08];	t0b =a[jt+p20+1]-a[jt+p08+1];
	  t06 =t06+t08;			t07 =t07+t09;
	  t08 =t06+c3m1*t08;		t09 =t07+c3m1*t09;
	  rt  =s3*t0a;			it  =s3*t0b;
	  t0a =t08+it;			t0b =t09-rt;
	  t08 =t08-it;			t09 =t09+rt;
	/* permuted version of jt + p08,p20,p32 */
	  t0c =a[jt+p28];		t0d =a[jt+p28+1];
	  t0e =a[jt+p16]+a[jt+p04];	t0f =a[jt+p16+1]+a[jt+p04+1];
	  t0g =a[jt+p16]-a[jt+p04];	t0h =a[jt+p16+1]-a[jt+p04+1];
	  t0c =t0c+t0e;			t0d =t0d+t0f;
	  t0e =t0c+c3m1*t0e;		t0f =t0d+c3m1*t0f;
	  rt  =s3*t0g;			it  =s3*t0h;
	  t0g =t0e+it;			t0h =t0f-rt;
	  t0e =t0e-it;			t0f =t0f+rt;
	/* Twiddles: */
	  rt  =t06;			it  =t07;
	  t06 =rt+t0c;			t07 =it+t0d;
	  t0c =rt-t0c;			t0d =it-t0d;
	  t00 =t00+t06;			t01 =t01+t07;
	  t06 =t00+c3m1*t06;		t07 =t01+c3m1*t07;
	  rt  =s3*t0c;			it  =s3*t0d;
	  t0c =t06+it;			t0d =t07-rt;
	  t06 =t06-it;			t07 =t07+rt;

	  rt  =t08*c -t09*s;		it  =t08*s +t09*c;
	  re  =t0e*c2-t0f*s2;		t0f =t0e*s2+t0f*c2;	t0e=re;
	  t08 =rt+t0e;			t09 =it+t0f;
	  t0e =rt-t0e;			t0f =it-t0f;
	  t02 =t02+t08;			t03 =t03+t09;
	  t08 =t02+c3m1*t08;		t09 =t03+c3m1*t09;
	  rt  =s3*t0e;			it  =s3*t0f;
	  t0e =t08+it;			t0f =t09-rt;
	  t08 =t08-it;			t09 =t09+rt;

	  rt  =t0a*c2-t0b*s2;		it  =t0a*s2+t0b*c2;
	  re  =t0g*c4-t0h*s4;		t0h =t0g*s4+t0h*c4;	t0g=re;
	  t0a =rt+t0g;			t0b =it+t0h;
	  t0g =rt-t0g;			t0h =it-t0h;
	  t04 =t04+t0a;			t05 =t05+t0b;
	  t0a =t04+c3m1*t0a;		t0b =t05+c3m1*t0b;
	  rt  =s3*t0g;			it  =s3*t0h;
	  t0g =t0a+it;			t0h =t0b-rt;
	  t0a =t0a-it;			t0b =t0b+rt;

/*...Second radix-9 transform:	*/
	/* permuted version of jt + p01,p13,p25 */
	  t10 =a[jt+p27];		t11 =a[jt+p27+1];
	  t12 =a[jt+p15]+a[jt+p03];	t13 =a[jt+p15+1]+a[jt+p03+1];
	  t14 =a[jt+p15]-a[jt+p03];	t15 =a[jt+p15+1]-a[jt+p03+1];
	  t10 =t10+t12;			t11 =t11+t13;
	  t12 =t10+c3m1*t12;		t13 =t11+c3m1*t13;
	  rt  =s3*t14;			it  =s3*t15;
	  t14 =t12+it;			t15 =t13-rt;
	  t12 =t12-it;			t13 =t13+rt;
	/* permuted version of jt + p05,p17,p29 */
	  t16 =a[jt+p23];		t17 =a[jt+p23+1];
	  t18 =a[jt+p11]+a[jt+p35];	t19 =a[jt+p11+1]+a[jt+p35+1];
	  t1a =a[jt+p11]-a[jt+p35];	t1b =a[jt+p11+1]-a[jt+p35+1];
	  t16 =t16+t18;			t17 =t17+t19;
	  t18 =t16+c3m1*t18;		t19 =t17+c3m1*t19;
	  rt  =s3*t1a;			it  =s3*t1b;
	  t1a =t18+it;			t1b =t19-rt;
	  t18 =t18-it;			t19 =t19+rt;
	/* permuted version of jt + p09,p21,p33 */
	  t1c =a[jt+p19];		t1d =a[jt+p19+1];
	  t1e =a[jt+p07]+a[jt+p31];	t1f =a[jt+p07+1]+a[jt+p31+1];
	  t1g =a[jt+p07]-a[jt+p31];	t1h =a[jt+p07+1]-a[jt+p31+1];
	  t1c =t1c+t1e;			t1d =t1d+t1f;
	  t1e =t1c+c3m1*t1e;		t1f =t1d+c3m1*t1f;
	  rt  =s3*t1g;			it  =s3*t1h;
	  t1g =t1e+it;			t1h =t1f-rt;
	  t1e =t1e-it;			t1f =t1f+rt;
	/* Twiddles: */
	  rt  =t16;			it  =t17;
	  t16 =rt+t1c;			t17 =it+t1d;
	  t1c =rt-t1c;			t1d =it-t1d;
	  t10 =t10+t16;			t11 =t11+t17;
	  t16 =t10+c3m1*t16;		t17 =t11+c3m1*t17;
	  rt  =s3*t1c;			it  =s3*t1d;
	  t1c =t16+it;			t1d =t17-rt;
	  t16 =t16-it;			t17 =t17+rt;

	  rt  =t18*c -t19*s;		it  =t18*s +t19*c;
	  re  =t1e*c2-t1f*s2;		t1f =t1e*s2+t1f*c2;	t1e=re;
	  t18 =rt+t1e;			t19 =it+t1f;
	  t1e =rt-t1e;			t1f =it-t1f;
	  t12 =t12+t18;			t13 =t13+t19;
	  t18 =t12+c3m1*t18;		t19 =t13+c3m1*t19;
	  rt  =s3*t1e;			it  =s3*t1f;
	  t1e =t18+it;			t1f =t19-rt;
	  t18 =t18-it;			t19 =t19+rt;

	  rt  =t1a*c2-t1b*s2;		it  =t1a*s2+t1b*c2;
	  re  =t1g*c4-t1h*s4;		t1h =t1g*s4+t1h*c4;	t1g=re;
	  t1a =rt+t1g;			t1b =it+t1h;
	  t1g =rt-t1g;			t1h =it-t1h;
	  t14 =t14+t1a;			t15 =t15+t1b;
	  t1a =t14+c3m1*t1a;		t1b =t15+c3m1*t1b;
	  rt  =s3*t1g;			it  =s3*t1h;
	  t1g =t1a+it;			t1h =t1b-rt;
	  t1a =t1a-it;			t1b =t1b+rt;

/*...Third radix-9 transform:	*/
	/* permuted version of jt + p02,p14,p26 */
	  t20 =a[jt+p18];		t21 =a[jt+p18+1];
	  t22 =a[jt+p06]+a[jt+p30];	t23 =a[jt+p06+1]+a[jt+p30+1];
	  t24 =a[jt+p06]-a[jt+p30];	t25 =a[jt+p06+1]-a[jt+p30+1];
	  t20 =t20+t22;			t21 =t21+t23;
	  t22 =t20+c3m1*t22;		t23 =t21+c3m1*t23;
	  rt  =s3*t24;			it  =s3*t25;
	  t24 =t22+it;			t25 =t23-rt;
	  t22 =t22-it;			t23 =t23+rt;
	/* permuted version of jt + p06,p18,p30 */
	  t26 =a[jt+p14];		t27 =a[jt+p14+1];
	  t28 =a[jt+p02]+a[jt+p26];	t29 =a[jt+p02+1]+a[jt+p26+1];
	  t2a =a[jt+p02]-a[jt+p26];	t2b =a[jt+p02+1]-a[jt+p26+1];
	  t26 =t26+t28;			t27 =t27+t29;
	  t28 =t26+c3m1*t28;		t29 =t27+c3m1*t29;
	  rt  =s3*t2a;			it  =s3*t2b;
	  t2a =t28+it;			t2b =t29-rt;
	  t28 =t28-it;			t29 =t29+rt;
	/* permuted version of jt + p10,p22,p34 */
	  t2c =a[jt+p10];		t2d =a[jt+p10+1];
	  t2e =a[jt+p34]+a[jt+p22];	t2f =a[jt+p34+1]+a[jt+p22+1];
	  t2g =a[jt+p34]-a[jt+p22];	t2h =a[jt+p34+1]-a[jt+p22+1];
	  t2c =t2c+t2e;			t2d =t2d+t2f;
	  t2e =t2c+c3m1*t2e;		t2f =t2d+c3m1*t2f;
	  rt  =s3*t2g;			it  =s3*t2h;
	  t2g =t2e+it;			t2h =t2f-rt;
	  t2e =t2e-it;			t2f =t2f+rt;
	/* Twiddles: */
	  rt  =t26;			it  =t27;
	  t26 =rt+t2c;			t27 =it+t2d;
	  t2c =rt-t2c;			t2d =it-t2d;
	  t20 =t20+t26;			t21 =t21+t27;
	  t26 =t20+c3m1*t26;		t27 =t21+c3m1*t27;
	  rt  =s3*t2c;			it  =s3*t2d;
	  t2c =t26+it;			t2d =t27-rt;
	  t26 =t26-it;			t27 =t27+rt;

	  rt  =t28*c -t29*s;		it  =t28*s +t29*c;
	  re  =t2e*c2-t2f*s2;		t2f =t2e*s2+t2f*c2;	t2e=re;
	  t28 =rt+t2e;			t29 =it+t2f;
	  t2e =rt-t2e;			t2f =it-t2f;
	  t22 =t22+t28;			t23 =t23+t29;
	  t28 =t22+c3m1*t28;		t29 =t23+c3m1*t29;
	  rt  =s3*t2e;			it  =s3*t2f;
	  t2e =t28+it;			t2f =t29-rt;
	  t28 =t28-it;			t29 =t29+rt;

	  rt  =t2a*c2-t2b*s2;		it  =t2a*s2+t2b*c2;
	  re  =t2g*c4-t2h*s4;		t2h =t2g*s4+t2h*c4;	t2g=re;
	  t2a =rt+t2g;			t2b =it+t2h;
	  t2g =rt-t2g;			t2h =it-t2h;
	  t24 =t24+t2a;			t25 =t25+t2b;
	  t2a =t24+c3m1*t2a;		t2b =t25+c3m1*t2b;
	  rt  =s3*t2g;			it  =s3*t2h;
	  t2g =t2a+it;			t2h =t2b-rt;
	  t2a =t2a-it;			t2b =t2b+rt;

/*...Fourth radix-9 transform:	*/
	/* permuted version of jt + p03,p15,p27 */
	  t30 =a[jt+p09];		t31 =a[jt+p09+1];
	  t32 =a[jt+p33]+a[jt+p21];	t33 =a[jt+p33+1]+a[jt+p21+1];
	  t34 =a[jt+p33]-a[jt+p21];	t35 =a[jt+p33+1]-a[jt+p21+1];
	  t30 =t30+t32;			t31 =t31+t33;
	  t32 =t30+c3m1*t32;		t33 =t31+c3m1*t33;
	  rt  =s3*t34;			it  =s3*t35;
	  t34 =t32+it;			t35 =t33-rt;
	  t32 =t32-it;			t33 =t33+rt;
	/* permuted version of jt + p07,p19,p31 */
	  t36 =a[jt+p05];		t37 =a[jt+p05+1];
	  t38 =a[jt+p29]+a[jt+p17];	t39 =a[jt+p29+1]+a[jt+p17+1];
	  t3a =a[jt+p29]-a[jt+p17];	t3b =a[jt+p29+1]-a[jt+p17+1];
	  t36 =t36+t38;			t37 =t37+t39;
	  t38 =t36+c3m1*t38;		t39 =t37+c3m1*t39;
	  rt  =s3*t3a;			it  =s3*t3b;
	  t3a =t38+it;			t3b =t39-rt;
	  t38 =t38-it;			t39 =t39+rt;
	/* permuted version of jt + p11,p23,p35 */
	  t3c =a[jt+p01];		t3d =a[jt+p01+1];
	  t3e =a[jt+p25]+a[jt+p13];	t3f =a[jt+p25+1]+a[jt+p13+1];
	  t3g =a[jt+p25]-a[jt+p13];	t3h =a[jt+p25+1]-a[jt+p13+1];
	  t3c =t3c+t3e;			t3d =t3d+t3f;
	  t3e =t3c+c3m1*t3e;		t3f =t3d+c3m1*t3f;
	  rt  =s3*t3g;			it  =s3*t3h;
	  t3g =t3e+it;			t3h =t3f-rt;
	  t3e =t3e-it;			t3f =t3f+rt;
	/* Twiddles: */
	  rt  =t36;			it  =t37;
	  t36 =rt+t3c;			t37 =it+t3d;
	  t3c =rt-t3c;			t3d =it-t3d;
	  t30 =t30+t36;			t31 =t31+t37;
	  t36 =t30+c3m1*t36;		t37 =t31+c3m1*t37;
	  rt  =s3*t3c;			it  =s3*t3d;
	  t3c =t36+it;			t3d =t37-rt;
	  t36 =t36-it;			t37 =t37+rt;

	  rt  =t38*c -t39*s;		it  =t38*s +t39*c;
	  re  =t3e*c2-t3f*s2;		t3f =t3e*s2+t3f*c2;	t3e=re;
	  t38 =rt+t3e;			t39 =it+t3f;
	  t3e =rt-t3e;			t3f =it-t3f;
	  t32 =t32+t38;			t33 =t33+t39;
	  t38 =t32+c3m1*t38;		t39 =t33+c3m1*t39;
	  rt  =s3*t3e;			it  =s3*t3f;
	  t3e =t38+it;			t3f =t39-rt;
	  t38 =t38-it;			t39 =t39+rt;

	  rt  =t3a*c2-t3b*s2;		it  =t3a*s2+t3b*c2;
	  re  =t3g*c4-t3h*s4;		t3h =t3g*s4+t3h*c4;	t3g=re;
	  t3a =rt+t3g;			t3b =it+t3h;
	  t3g =rt-t3g;			t3h =it-t3h;
	  t34 =t34+t3a;			t35 =t35+t3b;
	  t3a =t34+c3m1*t3a;		t3b =t35+c3m1*t3b;
	  rt  =s3*t3g;			it  =s3*t3h;
	  t3g =t3a+it;			t3h =t3b-rt;
	  t3a =t3a-it;			t3b =t3b+rt;

/*...and now do nine radix-4 transforms:	*/

/*...Block 1: */
	j1=jt;
	    rt =t20;	t20=t00-rt;	t00=t00+rt;
	    it =t21;	t21=t01-it;	t01=t01+it;

	    rt =t30;	t30=t10-rt;	t10=t10+rt;
	    it =t31;	t31=t11-it;	t11=t11+it;

	    a[j1    ]=t00+t10;	a[j1+1    ]=t01+t11;
	    a[j1+p01]=t00-t10;	a[j1+1+p01]=t01-t11;

	    a[j1+p03]=t20-t31;	a[j1+1+p03]=t21+t30;	/* mpy by I is inlined here...	*/
	    a[j1+p02]=t20+t31;	a[j1+1+p02]=t21-t30;

/*...Block 2: */
	j1=jt+p32;
	    rt =t22;	t22=t02-rt;	t02=t02+rt;
	    it =t23;	t23=t03-it;	t03=t03+it;

	    rt =t32;	t32=t12-rt;	t12=t12+rt;
	    it =t33;	t33=t13-it;	t13=t13+it;

	    a[j1    ]=t02+t12;	a[j1+1    ]=t03+t13;
	    a[j1+p01]=t02-t12;	a[j1+1+p01]=t03-t13;

	    a[j1+p03]=t22-t33;	a[j1+1+p03]=t23+t32;	/* mpy by I is inlined here...	*/
	    a[j1+p02]=t22+t33;	a[j1+1+p02]=t23-t32;

/*...Block 3: */
	j1=jt+p20;
	    rt =t24;	t24=t04-rt;	t04=t04+rt;
	    it =t25;	t25=t05-it;	t05=t05+it;

	    rt =t34;	t34=t14-rt;	t14=t14+rt;
	    it =t35;	t35=t15-it;	t15=t15+it;

	    a[j1+p02]=t04+t14;	a[j1+1+p02]=t05+t15;
	    a[j1+p03]=t04-t14;	a[j1+1+p03]=t05-t15;

	    a[j1    ]=t24-t35;	a[j1+1    ]=t25+t34;	/* mpy by I is inlined here...	*/
	    a[j1+p01]=t24+t35;	a[j1+1+p01]=t25-t34;

/*...Block 4: */
	j1=jt+p08;
	    rt =t26;	t26=t06-rt;	t06=t06+rt;
	    it =t27;	t27=t07-it;	t07=t07+it;

	    rt =t36;	t36=t16-rt;	t16=t16+rt;
	    it =t37;	t37=t17-it;	t17=t17+it;

	    a[j1+p01]=t06+t16;	a[j1+1+p01]=t07+t17;
	    a[j1    ]=t06-t16;	a[j1+1    ]=t07-t17;

	    a[j1+p02]=t26-t37;	a[j1+1+p02]=t27+t36;	/* mpy by I is inlined here...	*/
	    a[j1+p03]=t26+t37;	a[j1+1+p03]=t27-t36;

/*...Block 5: */
	j1=jt+p28;
	    rt =t28;	t28=t08-rt;	t08=t08+rt;
	    it =t29;	t29=t09-it;	t09=t09+it;

	    rt =t38;	t38=t18-rt;	t18=t18+rt;
	    it =t39;	t39=t19-it;	t19=t19+it;

	    a[j1+p03]=t08+t18;	a[j1+1+p03]=t09+t19;
	    a[j1+p02]=t08-t18;	a[j1+1+p02]=t09-t19;

	    a[j1+p01]=t28-t39;	a[j1+1+p01]=t29+t38;	/* mpy by I is inlined here...	*/
	    a[j1    ]=t28+t39;	a[j1+1    ]=t29-t38;

/*...Block 6: */
	j1=jt+p16;
	    rt =t2a;	t2a=t0a-rt;	t0a=t0a+rt;
	    it =t2b;	t2b=t0b-it;	t0b=t0b+it;

	    rt =t3a;	t3a=t1a-rt;	t1a=t1a+rt;
	    it =t3b;	t3b=t1b-it;	t1b=t1b+it;

	    a[j1    ]=t0a+t1a;	a[j1+1    ]=t0b+t1b;
	    a[j1+p01]=t0a-t1a;	a[j1+1+p01]=t0b-t1b;

	    a[j1+p03]=t2a-t3b;	a[j1+1+p03]=t2b+t3a;	/* mpy by I is inlined here...	*/
	    a[j1+p02]=t2a+t3b;	a[j1+1+p02]=t2b-t3a;

/*...Block 7: */
	j1=jt+p04;
	    rt =t2c;	t2c=t0c-rt;	t0c=t0c+rt;
	    it =t2d;	t2d=t0d-it;	t0d=t0d+it;

	    rt =t3c;	t3c=t1c-rt;	t1c=t1c+rt;
	    it =t3d;	t3d=t1d-it;	t1d=t1d+it;

	    a[j1+p02]=t0c+t1c;	a[j1+1+p02]=t0d+t1d;
	    a[j1+p03]=t0c-t1c;	a[j1+1+p03]=t0d-t1d;

	    a[j1    ]=t2c-t3d;	a[j1+1    ]=t2d+t3c;	/* mpy by I is inlined here...	*/
	    a[j1+p01]=t2c+t3d;	a[j1+1+p01]=t2d-t3c;

/*...Block 8: */
	j1=jt+p24;
	    rt =t2e;	t2e=t0e-rt;	t0e=t0e+rt;
	    it =t2f;	t2f=t0f-it;	t0f=t0f+it;

	    rt =t3e;	t3e=t1e-rt;	t1e=t1e+rt;
	    it =t3f;	t3f=t1f-it;	t1f=t1f+it;

	    a[j1+p01]=t0e+t1e;	a[j1+1+p01]=t0f+t1f;
	    a[j1    ]=t0e-t1e;	a[j1+1    ]=t0f-t1f;

	    a[j1+p02]=t2e-t3f;	a[j1+1+p02]=t2f+t3e;	/* mpy by I is inlined here...	*/
	    a[j1+p03]=t2e+t3f;	a[j1+1+p03]=t2f-t3e;

/*...Block 9: */
	j1=jt+p12;
	    rt =t2g;	t2g=t0g-rt;	t0g=t0g+rt;
	    it =t2h;	t2h=t0h-it;	t0h=t0h+it;

	    rt =t3g;	t3g=t1g-rt;	t1g=t1g+rt;
	    it =t3h;	t3h=t1h-it;	t1h=t1h+it;

	    a[j1+p03]=t0g+t1g;	a[j1+1+p03]=t0h+t1h;
	    a[j1+p02]=t0g-t1g;	a[j1+1+p02]=t0h-t1h;

	    a[j1+p01]=t2g-t3h;	a[j1+1+p01]=t2h+t3g;	/* mpy by I is inlined here...	*/
	    a[j1    ]=t2g+t3h;	a[j1+1    ]=t2h-t3g;
								/* Totals: 534 FADD, 160 FMUL	*/
	}
}

/**************/

int radix36_ditN_cy_dif1(double a[], int n, int nwt, int nwt_bits, double wt0[], double wt1[], int si[], double base[], double baseinv[], int iter, double *fracmax, int p)
{
/*
!...Acronym: DWT = Discrete Weighted Transform, DIT = Decimation In Time, DIF = Decimation In Frequency
!
!...Performs a final radix-36 complex DIT pass, an inverse DWT weighting, a carry propagation,
!   a forward DWT weighting, and an initial radix-36 complex DIF pass on the data in the length-N real vector A.
!
!   Data enter and are returned in the A-array.
!
!   See the documentation in mers_mod_square and radix16_dif_pass for further details on the array
!   storage scheme, and radix8_ditN_cy_dif1 for details on the reduced-length weights array scheme.
*/
	int n36, bjmodn00,bjmodn01,bjmodn02,bjmodn03,bjmodn04,bjmodn05,bjmodn06,bjmodn07,bjmodn08,bjmodn09,bjmodn10,bjmodn11,bjmodn12,bjmodn13,bjmodn14,bjmodn15,bjmodn16,bjmodn17
		,bjmodn18,bjmodn19,bjmodn20,bjmodn21,bjmodn22,bjmodn23,bjmodn24,bjmodn25,bjmodn26,bjmodn27,bjmodn28,bjmodn29,bjmodn30,bjmodn31,bjmodn32,bjmodn33,bjmodn34,bjmodn35
		,i,j,jt,j1,jstart,jhi,iroot,root_incr,k,khi,l,outer;
	static int psave=0,bw,sw,bjmodnini,nrtm1,p01,p02,p03,p04,p05,p06,p07,p08,p09,p10,p11,p12,p13,p14,p15,p16,p17
		,p18,p19,p20,p21,p22,p23,p24,p25,p26,p27,p28,p29,p30,p31,p32,p33,p34,p35;
	const  double one_half[3] = {1.0, 0.5, 0.25};	/* Needed for small-weights-tables scheme */
	static double c   =  0.76604444311897803520,	/* cos(2*pi/9) */
		      s   =  0.64278760968653932631,	/* sin(2*pi/9) */
		      c2  =  0.17364817766693034887,	/* cos(2*u) */
		      s2  =  0.98480775301220805936,	/* sin(2*u) */
		      c3m1= -1.50000000000000000000,	/* cos(3*u)-1 */
		      s3  =  0.86602540378443864677,	/* sin(3*u) */
		      c4  = -0.93969262078590838404,	/* cos(4*u) */
		      s4  =  0.34202014332566873307;	/* sin(4*u) */
	static double radix_inv, n2inv;
	double rt,it,re
		,t00,t01,t02,t03,t04,t05,t06,t07,t08,t09,t0a,t0b,t0c,t0d,t0e,t0f,t0g,t0h
		,t10,t11,t12,t13,t14,t15,t16,t17,t18,t19,t1a,t1b,t1c,t1d,t1e,t1f,t1g,t1h
		,t20,t21,t22,t23,t24,t25,t26,t27,t28,t29,t2a,t2b,t2c,t2d,t2e,t2f,t2g,t2h
		,t30,t31,t32,t33,t34,t35,t36,t37,t38,t39,t3a,t3b,t3c,t3d,t3e,t3f,t3g,t3h
	,aj1p00r,aj1p01r,aj1p02r,aj1p03r,aj1p04r,aj1p05r,aj1p06r,aj1p07r,aj1p08r,aj1p09r,aj1p10r,aj1p11r,aj1p12r,aj1p13r,aj1p14r,aj1p15r,aj1p16r,aj1p17r,aj1p18r,aj1p19r,aj1p20r,aj1p21r,aj1p22r,aj1p23r,aj1p24r,aj1p25r,aj1p26r,aj1p27r,aj1p28r,aj1p29r,aj1p30r,aj1p31r,aj1p32r,aj1p33r,aj1p34r,aj1p35r
	,aj1p00i,aj1p01i,aj1p02i,aj1p03i,aj1p04i,aj1p05i,aj1p06i,aj1p07i,aj1p08i,aj1p09i,aj1p10i,aj1p11i,aj1p12i,aj1p13i,aj1p14i,aj1p15i,aj1p16i,aj1p17i,aj1p18i,aj1p19i,aj1p20i,aj1p21i,aj1p22i,aj1p23i,aj1p24i,aj1p25i,aj1p26i,aj1p27i,aj1p28i,aj1p29i,aj1p30i,aj1p31i,aj1p32i,aj1p33i,aj1p34i,aj1p35i
		,cy00,cy01,cy02,cy03,cy04,cy05,cy06,cy07,cy08,cy09,cy10,cy11,cy12,cy13,cy14,cy15,cy16,cy17,cy18,cy19,cy20,cy21,cy22,cy23,cy24,cy25,cy26,cy27,cy28,cy29,cy30,cy31,cy32,cy33,cy34,cy35,temp,frac,scale;
	double maxerr = 0.0;
#if PFETCH
	double *add0, *add;
#endif
	int err;
	FILE *f1,*f2;
	static int first_entry=TRUE;

/*...stuff for the reduced-length DWT weights array is here:	*/
	int n_div_nwt;
	int col,co2,co3,m,m2,n_minus_sil,n_minus_silp1,sinwt,sinwtm1;
	double wt,wtinv,wtl,wtlp1,wtn,wtnm1,wtA,wtB,wtC;

/*...change n36 and n_div_wt to non-static to work around a gcc compiler bug. */
	n36   = n/36;
	n_div_nwt = n36 >> nwt_bits;

	if((n_div_nwt << nwt_bits) != n36)
	{
	  if(INTERACT)printf("FATAL: iter = %10d; NWT_BITS does not divide N/36 in radix36_ditN_cy_dif1.\n",iter);
	  f1 = fopen(   OFILE,"a");
	  f2 = fopen(STATFILE,"a");
	  fprintf(f1,"FATAL: iter = %10d; NWT_BITS does not divide N/36 in radix36_ditN_cy_dif1.\n",iter);
	  fprintf(f2,"FATAL: iter = %10d; NWT_BITS does not divide N/36 in radix36_ditN_cy_dif1.\n",iter);
	  fclose(f1);
	  fclose(f2);
	  err=1;
	  return(err);
	}

	if(p != psave)
	{
	  first_entry=TRUE;
	}

/*...initialize things upon first entry	*/

	if(first_entry)
	{
	  psave = p;
	  first_entry=FALSE;
	  radix_inv = (double)((long double)1.0L/(long double)36.0L);
	  n2inv     = (double)((long double)1.0L/(long double)(n/2));

	  bw    = p%n;		/* Number of bigwords in the Crandall/Fagin mixed-radix representation = (Mersenne exponent) mod (vector length).	*/
	  sw    = n - bw;	/* Number of smallwords.	*/

/*   constant index offsets for array load/stores are here.	*/

	  p01 = n36;
	  p02 = p01 + p01;
	  p03 = p02 + p01;
	  p04 = p03 + p01;
	  p05 = p04 + p01;
	  p06 = p05 + p01;
	  p07 = p06 + p01;
	  p08 = p07 + p01;
	  p09 = p08 + p01;
	  p10 = p09 + p01;
	  p11 = p10 + p01;
	  p12 = p11 + p01;
	  p13 = p12 + p01;
	  p14 = p13 + p01;
	  p15 = p14 + p01;
	  p16 = p15 + p01;
	  p17 = p16 + p01;
	  p18 = p17 + p01;
	  p19 = p18 + p01;
	  p20 = p19 + p01;
	  p21 = p20 + p01;
	  p22 = p21 + p01;
	  p23 = p22 + p01;
	  p24 = p23 + p01;
	  p25 = p24 + p01;
	  p26 = p25 + p01;
	  p27 = p26 + p01;
	  p28 = p27 + p01;
	  p29 = p28 + p01;
	  p30 = p29 + p01;
	  p31 = p30 + p01;
	  p32 = p31 + p01;
	  p33 = p32 + p01;
	  p34 = p33 + p01;
	  p35 = p34 + p01;

	  p01 = p01 + ( (p01 >> DAT_BITS) << PAD_BITS );
	  p02 = p02 + ( (p02 >> DAT_BITS) << PAD_BITS );
	  p03 = p03 + ( (p03 >> DAT_BITS) << PAD_BITS );
	  p04 = p04 + ( (p04 >> DAT_BITS) << PAD_BITS );
	  p05 = p05 + ( (p05 >> DAT_BITS) << PAD_BITS );
	  p06 = p06 + ( (p06 >> DAT_BITS) << PAD_BITS );
	  p07 = p07 + ( (p07 >> DAT_BITS) << PAD_BITS );
	  p08 = p08 + ( (p08 >> DAT_BITS) << PAD_BITS );
	  p09 = p09 + ( (p09 >> DAT_BITS) << PAD_BITS );
	  p10 = p10 + ( (p10 >> DAT_BITS) << PAD_BITS );
	  p11 = p11 + ( (p11 >> DAT_BITS) << PAD_BITS );
	  p12 = p12 + ( (p12 >> DAT_BITS) << PAD_BITS );
	  p13 = p13 + ( (p13 >> DAT_BITS) << PAD_BITS );
	  p14 = p14 + ( (p14 >> DAT_BITS) << PAD_BITS );
	  p15 = p15 + ( (p15 >> DAT_BITS) << PAD_BITS );
	  p16 = p16 + ( (p16 >> DAT_BITS) << PAD_BITS );
	  p17 = p17 + ( (p17 >> DAT_BITS) << PAD_BITS );
	  p18 = p18 + ( (p18 >> DAT_BITS) << PAD_BITS );
	  p19 = p19 + ( (p19 >> DAT_BITS) << PAD_BITS );
	  p20 = p20 + ( (p20 >> DAT_BITS) << PAD_BITS );
	  p21 = p21 + ( (p21 >> DAT_BITS) << PAD_BITS );
	  p22 = p22 + ( (p22 >> DAT_BITS) << PAD_BITS );
	  p23 = p23 + ( (p23 >> DAT_BITS) << PAD_BITS );
	  p24 = p24 + ( (p24 >> DAT_BITS) << PAD_BITS );
	  p25 = p25 + ( (p25 >> DAT_BITS) << PAD_BITS );
	  p26 = p26 + ( (p26 >> DAT_BITS) << PAD_BITS );
	  p27 = p27 + ( (p27 >> DAT_BITS) << PAD_BITS );
	  p28 = p28 + ( (p28 >> DAT_BITS) << PAD_BITS );
	  p29 = p29 + ( (p29 >> DAT_BITS) << PAD_BITS );
	  p30 = p30 + ( (p30 >> DAT_BITS) << PAD_BITS );
	  p31 = p31 + ( (p31 >> DAT_BITS) << PAD_BITS );
	  p32 = p32 + ( (p32 >> DAT_BITS) << PAD_BITS );
	  p33 = p33 + ( (p33 >> DAT_BITS) << PAD_BITS );
	  p34 = p34 + ( (p34 >> DAT_BITS) << PAD_BITS );
	  p35 = p35 + ( (p35 >> DAT_BITS) << PAD_BITS );

	  bjmodnini=0;
	  for(j=0; j < n36; j++)
	  {
	    bjmodnini -= sw; bjmodnini = bjmodnini + ( (-(int)((uint32)bjmodnini >> 31)) & n);
	  }
	}

/*...The radix-36 final DIT pass is here.	*/

	cy00=-2;		/* init carries	*/
	cy01= 0;
	cy02= 0;
	cy03= 0;
	cy04= 0;
	cy05= 0;
	cy06= 0;
	cy07= 0;
	cy08= 0;
	cy09= 0;
	cy10= 0;
	cy11= 0;
	cy12= 0;
	cy13= 0;
	cy14= 0;
	cy15= 0;
	cy16= 0;
	cy17= 0;
	cy18= 0;
	cy19= 0;
	cy20= 0;
	cy21= 0;
	cy22= 0;
	cy23= 0;
	cy24= 0;
	cy25= 0;
	cy26= 0;
	cy27= 0;
	cy28= 0;
	cy29= 0;
	cy30= 0;
	cy31= 0;
	cy32= 0;
	cy33= 0;
	cy34= 0;
	cy35= 0;

	*fracmax=0;	/* init max. fractional error	*/

	iroot = 0;	/* init sincos array index	*/
	root_incr = 1;	/* init sincos array index increment (set = 1 for normal carry pass, = 0 for wrapper pass)	*/
	scale = n2inv;	/* init inverse-weight scale factor  (set = 2/n for normal carry pass, = 1 for wrapper pass)	*/

	jstart = 0;
	jhi = jstart+nwt-1;
	khi = n_div_nwt;

for(outer=0; outer <= 1; outer++)
{
	i = 1;		/* Pointer to the BASE and BASEINV arrays. lowest-order digit is always a bigword (i = 1).	*/

	bjmodn00= 0;
	bjmodn01= bjmodnini;
	bjmodn02= bjmodn01+bjmodnini-n; bjmodn02= bjmodn02+ ( (-(int)((uint32)bjmodn02>> 31)) & n);
	bjmodn03= bjmodn02+bjmodnini-n; bjmodn03= bjmodn03+ ( (-(int)((uint32)bjmodn03>> 31)) & n);
	bjmodn04= bjmodn03+bjmodnini-n; bjmodn04= bjmodn04+ ( (-(int)((uint32)bjmodn04>> 31)) & n);
	bjmodn05= bjmodn04+bjmodnini-n; bjmodn05= bjmodn05+ ( (-(int)((uint32)bjmodn05>> 31)) & n);
	bjmodn06= bjmodn05+bjmodnini-n; bjmodn06= bjmodn06+ ( (-(int)((uint32)bjmodn06>> 31)) & n);
	bjmodn07= bjmodn06+bjmodnini-n; bjmodn07= bjmodn07+ ( (-(int)((uint32)bjmodn07>> 31)) & n);
	bjmodn08= bjmodn07+bjmodnini-n; bjmodn08= bjmodn08+ ( (-(int)((uint32)bjmodn08>> 31)) & n);
	bjmodn09= bjmodn08+bjmodnini-n; bjmodn09= bjmodn09+ ( (-(int)((uint32)bjmodn09>> 31)) & n);
	bjmodn10= bjmodn09+bjmodnini-n; bjmodn10= bjmodn10+ ( (-(int)((uint32)bjmodn10>> 31)) & n);
	bjmodn11= bjmodn10+bjmodnini-n; bjmodn11= bjmodn11+ ( (-(int)((uint32)bjmodn11>> 31)) & n);
	bjmodn12= bjmodn11+bjmodnini-n; bjmodn12= bjmodn12+ ( (-(int)((uint32)bjmodn12>> 31)) & n);
	bjmodn13= bjmodn12+bjmodnini-n; bjmodn13= bjmodn13+ ( (-(int)((uint32)bjmodn13>> 31)) & n);
	bjmodn14= bjmodn13+bjmodnini-n; bjmodn14= bjmodn14+ ( (-(int)((uint32)bjmodn14>> 31)) & n);
	bjmodn15= bjmodn14+bjmodnini-n; bjmodn15= bjmodn15+ ( (-(int)((uint32)bjmodn15>> 31)) & n);
	bjmodn16= bjmodn15+bjmodnini-n; bjmodn16= bjmodn16+ ( (-(int)((uint32)bjmodn16>> 31)) & n);
	bjmodn17= bjmodn16+bjmodnini-n; bjmodn17= bjmodn17+ ( (-(int)((uint32)bjmodn17>> 31)) & n);
	bjmodn18= bjmodn17+bjmodnini-n; bjmodn18= bjmodn18+ ( (-(int)((uint32)bjmodn18>> 31)) & n);
	bjmodn19= bjmodn18+bjmodnini-n; bjmodn19= bjmodn19+ ( (-(int)((uint32)bjmodn19>> 31)) & n);
	bjmodn20= bjmodn19+bjmodnini-n; bjmodn20= bjmodn20+ ( (-(int)((uint32)bjmodn20>> 31)) & n);
	bjmodn21= bjmodn20+bjmodnini-n; bjmodn21= bjmodn21+ ( (-(int)((uint32)bjmodn21>> 31)) & n);
	bjmodn22= bjmodn21+bjmodnini-n; bjmodn22= bjmodn22+ ( (-(int)((uint32)bjmodn22>> 31)) & n);
	bjmodn23= bjmodn22+bjmodnini-n; bjmodn23= bjmodn23+ ( (-(int)((uint32)bjmodn23>> 31)) & n);
	bjmodn24= bjmodn23+bjmodnini-n; bjmodn24= bjmodn24+ ( (-(int)((uint32)bjmodn24>> 31)) & n);
	bjmodn25= bjmodn24+bjmodnini-n; bjmodn25= bjmodn25+ ( (-(int)((uint32)bjmodn25>> 31)) & n);
	bjmodn26= bjmodn25+bjmodnini-n; bjmodn26= bjmodn26+ ( (-(int)((uint32)bjmodn26>> 31)) & n);
	bjmodn27= bjmodn26+bjmodnini-n; bjmodn27= bjmodn27+ ( (-(int)((uint32)bjmodn27>> 31)) & n);
	bjmodn28= bjmodn27+bjmodnini-n; bjmodn28= bjmodn28+ ( (-(int)((uint32)bjmodn28>> 31)) & n);
	bjmodn29= bjmodn28+bjmodnini-n; bjmodn29= bjmodn29+ ( (-(int)((uint32)bjmodn29>> 31)) & n);
	bjmodn30= bjmodn29+bjmodnini-n; bjmodn30= bjmodn30+ ( (-(int)((uint32)bjmodn30>> 31)) & n);
	bjmodn31= bjmodn30+bjmodnini-n; bjmodn31= bjmodn31+ ( (-(int)((uint32)bjmodn31>> 31)) & n);
	bjmodn32= bjmodn31+bjmodnini-n; bjmodn32= bjmodn32+ ( (-(int)((uint32)bjmodn32>> 31)) & n);
	bjmodn33= bjmodn32+bjmodnini-n; bjmodn33= bjmodn33+ ( (-(int)((uint32)bjmodn33>> 31)) & n);
	bjmodn34= bjmodn33+bjmodnini-n; bjmodn34= bjmodn34+ ( (-(int)((uint32)bjmodn34>> 31)) & n);
	bjmodn35= bjmodn34+bjmodnini-n; bjmodn35= bjmodn35+ ( (-(int)((uint32)bjmodn35>> 31)) & n);

	col=0;
	co2=(n >> nwt_bits)-1+36;
	co3=co2-36;		/* At the start of each new j-loop, co3=co2-radix(1)	*/

	for(k=1; k <= khi; k++)	/* Do n/(radix(1)*nwt) outer loop executions...	*/
	{
	  for(j=jstart; j<jhi; j += 2)	/* Each inner loop execution processes (radix(1)*nwt) array data.	*/
	  {
	    jt = j + ( (j >> DAT_BITS) << PAD_BITS );	/* padded-array fetch index is here	*/
/*
!...gather the needed data (36 64-bit complex, i.e. 72 64-bit reals) and do a radix-36 DIT transform...
*/
/*...Block 1:	*/
	j1=jt;
	t00=a[j1    ];	t01=a[j1+1    ];
	rt =a[j1+p01];	it =a[j1+1+p01];
	t10=t00-rt;  	t00=t00+rt;
	t11=t01-it;	t01=t01+it;

	t20=a[j1+p03];	t21=a[j1+1+p03];
	rt =a[j1+p02];	it =a[j1+1+p02];
	t30=t20-rt;  	t20=t20+rt;
	t31=t21-it;  	t21=t21+it;

	rt =t20;	t20=t00-rt;	t00=t00+rt;
	it =t21;	t21=t01-it;	t01=t01+it;

	rt =t30;	t30=t10-t31;	t10=t10+t31;
			t31=t11+rt;	t11=t11-rt;

/*...Block 2:	*/
	j1=jt+p08;
	t02=a[j1+p01];	t03=a[j1+1+p01];
	rt =a[j1    ];	it =a[j1+1    ];
	t12=t02-rt;  	t02=t02+rt;
	t13=t03-it;	t03=t03+it;

	t22=a[j1+p02];	t23=a[j1+1+p02];
	rt =a[j1+p03];	it =a[j1+1+p03];
	t32=t22-rt;  	t22=t22+rt;
	t33=t23-it;  	t23=t23+it;

	rt =t22;	t22=t02-rt;	t02=t02+rt;
	it =t23;	t23=t03-it;	t03=t03+it;

	rt =t32;	t32=t12-t33;	t12=t12+t33;
			t33=t13+rt;	t13=t13-rt;

/*...Block 3:	*/
	j1=jt+p04;
	t04=a[j1+p02];	t05=a[j1+1+p02];
	rt =a[j1+p03];	it =a[j1+1+p03];
	t14=t04-rt;  	t04=t04+rt;
	t15=t05-it;	t05=t05+it;

	t24=a[j1    ];	t25=a[j1+1    ];
	rt =a[j1+p01];	it =a[j1+1+p01];
	t34=t24-rt;  	t24=t24+rt;
	t35=t25-it;  	t25=t25+it;

	rt =t24;	t24=t04-rt;	t04=t04+rt;
	it =t25;	t25=t05-it;	t05=t05+it;

	rt =t34;	t34=t14-t35;	t14=t14+t35;
			t35=t15+rt;	t15=t15-rt;

/*...Block 4:	*/
	j1=jt+p28;
	t06=a[j1+p03];	t07=a[j1+1+p03];
	rt =a[j1+p02];	it =a[j1+1+p02];
	t16=t06-rt;  	t06=t06+rt;
	t17=t07-it;	t07=t07+it;

	t26=a[j1+p01];	t27=a[j1+1+p01];
	rt =a[j1    ];	it =a[j1+1    ];
	t36=t26-rt;  	t26=t26+rt;
	t37=t27-it;  	t27=t27+it;

	rt =t26;	t26=t06-rt;	t06=t06+rt;
	it =t27;	t27=t07-it;	t07=t07+it;

	rt =t36;	t36=t16-t37;	t16=t16+t37;
			t37=t17+rt;	t17=t17-rt;

/*...Block 5:	*/
	j1=jt+p24;
	t08=a[j1+p01];	t09=a[j1+1+p01];
	rt =a[j1    ];	it =a[j1+1    ];
	t18=t08-rt;  	t08=t08+rt;
	t19=t09-it;	t09=t09+it;

	t28=a[j1+p02];	t29=a[j1+1+p02];
	rt =a[j1+p03];	it =a[j1+1+p03];
	t38=t28-rt;  	t28=t28+rt;
	t39=t29-it;  	t29=t29+it;

	rt =t28;	t28=t08-rt;	t08=t08+rt;
	it =t29;	t29=t09-it;	t09=t09+it;

	rt =t38;	t38=t18-t39;	t18=t18+t39;
			t39=t19+rt;	t19=t19-rt;

/*...Block 6:	*/
	j1=jt+p32;
	t0a=a[j1    ];	t0b=a[j1+1    ];
	rt =a[j1+p01];	it =a[j1+1+p01];
	t1a=t0a-rt;  	t0a=t0a+rt;
	t1b=t0b-it;	t0b=t0b+it;

	t2a=a[j1+p03];	t2b=a[j1+1+p03];
	rt =a[j1+p02];	it =a[j1+1+p02];
	t3a=t2a-rt;  	t2a=t2a+rt;
	t3b=t2b-it;  	t2b=t2b+it;

	rt =t2a;	t2a=t0a-rt;	t0a=t0a+rt;
	it =t2b;	t2b=t0b-it;	t0b=t0b+it;

	rt =t3a;	t3a=t1a-t3b;	t1a=t1a+t3b;
			t3b=t1b+rt;	t1b=t1b-rt;

/*...Block 7:	*/
	j1=jt+p12;
	t0c=a[j1+p03];	t0d=a[j1+1+p03];
	rt =a[j1+p02];	it =a[j1+1+p02];
	t1c=t0c-rt;  	t0c=t0c+rt;
	t1d=t0d-it;	t0d=t0d+it;

	t2c=a[j1+p01];	t2d=a[j1+1+p01];
	rt =a[j1    ];	it =a[j1+1    ];
	t3c=t2c-rt;  	t2c=t2c+rt;
	t3d=t2d-it;  	t2d=t2d+it;

	rt =t2c;	t2c=t0c-rt;	t0c=t0c+rt;
	it =t2d;	t2d=t0d-it;	t0d=t0d+it;

	rt =t3c;	t3c=t1c-t3d;	t1c=t1c+t3d;
			t3d=t1d+rt;	t1d=t1d-rt;

/*...Block 8:	*/
	j1=jt+p20;
	t0e=a[j1+p02];	t0f=a[j1+1+p02];
	rt =a[j1+p03];	it =a[j1+1+p03];
	t1e=t0e-rt;  	t0e=t0e+rt;
	t1f=t0f-it;	t0f=t0f+it;

	t2e=a[j1    ];	t2f=a[j1+1    ];
	rt =a[j1+p01];	it =a[j1+1+p01];
	t3e=t2e-rt;  	t2e=t2e+rt;
	t3f=t2f-it;  	t2f=t2f+it;

	rt =t2e;	t2e=t0e-rt;	t0e=t0e+rt;
	it =t2f;	t2f=t0f-it;	t0f=t0f+it;

	rt =t3e;	t3e=t1e-t3f;	t1e=t1e+t3f;
			t3f=t1f+rt;	t1f=t1f-rt;

/*...Block 9:	*/
	j1=jt+p16;
	t0g=a[j1    ];	t0h=a[j1+1    ];
	rt =a[j1+p01];	it =a[j1+1+p01];
	t1g=t0g-rt;  	t0g=t0g+rt;
	t1h=t0h-it;	t0h=t0h+it;

	t2g=a[j1+p03];	t2h=a[j1+1+p03];
	rt =a[j1+p02];	it =a[j1+1+p02];
	t3g=t2g-rt;  	t2g=t2g+rt;
	t3h=t2h-it;  	t2h=t2h+it;

	rt =t2g;	t2g=t0g-rt;	t0g=t0g+rt;
	it =t2h;	t2h=t0h-it;	t0h=t0h+it;

	rt =t3g;	t3g=t1g-t3h;	t1g=t1g+t3h;
			t3h=t1h+rt;	t1h=t1h-rt;

/*       ...and now do four radix-9 transforms.	*/

/*...First radix-9 transform:	*/
	  rt  =t02;			it  =t03;
	  t02 =rt+t04;			t03 =it+t05;
	  t04 =rt-t04;			t05 =it-t05;
	  t00 =t00+t02;			t01 =t01+t03;
	  t02 =t00+c3m1*t02;		t03 =t01+c3m1*t03;
	  rt  =s3*t04;			it  =s3*t05;
	  t04 =t02-it;			t05 =t03+rt;
	  t02 =t02+it;			t03 =t03-rt;

	  rt  =t08;			it  =t09;
	  t08 =rt+t0a;			t09 =it+t0b;
	  t0a =rt-t0a;			t0b =it-t0b;
	  t06 =t06+t08;			t07 =t07+t09;
	  t08 =t06+c3m1*t08;		t09 =t07+c3m1*t09;
	  rt  =s3*t0a;			it  =s3*t0b;
	  t0a =t08-it;			t0b =t09+rt;
	  t08 =t08+it;			t09 =t09-rt;

	  rt  =t0e;			it  =t0f;
	  t0e =rt+t0g;			t0f =it+t0h;
	  t0g =rt-t0g;			t0h =it-t0h;
	  t0c =t0c+t0e;			t0d =t0d+t0f;
	  t0e =t0c+c3m1*t0e;		t0f =t0d+c3m1*t0f;
	  rt  =s3*t0g;			it  =s3*t0h;
	  t0g =t0e-it;			t0h =t0f+rt;
	  t0e =t0e+it;			t0f =t0f-rt;
	/* Twiddles: */
	  rt  =t06;			it  =t07;
	  t06 =rt+t0c;			t07 =it+t0d;
	  t0c =rt-t0c;			t0d =it-t0d;
	  t00 =t00+t06;			t01 =t01+t07;
	  aj1p00r=t00;			aj1p00i=t01;
	  t06 =t00+c3m1*t06;		t07 =t01+c3m1*t07;
	  rt  =s3*t0c;			it  =s3*t0d;
	  aj1p24r=t06+it;		aj1p24i=t07-rt;
	  aj1p12r=t06-it;		aj1p12i=t07+rt;

	  rt  =t08*c +t09*s;		it  =t09*c -t08*s;
	  re  =t0e*c2+t0f*s2;		t0f =t0f*c2-t0e*s2;	t0e=re;
	  t08 =rt+t0e;			t09 =it+t0f;
	  t0e =rt-t0e;			t0f =it-t0f;
	  t02 =t02+t08;			t03 =t03+t09;
	  aj1p08r=t02;			aj1p08i=t03;
	  t08 =t02+c3m1*t08;		t09=t03+c3m1*t09;
	  rt  =s3*t0e;			it  =s3*t0f;
	  aj1p32r=t08+it;		aj1p32i=t09-rt;
	  aj1p20r=t08-it;		aj1p20i=t09+rt;

	  rt  =t0a*c2+t0b*s2;		it  =t0b*c2-t0a*s2;
	  re  =t0g*c4+t0h*s4;		t0h =t0h*c4-t0g*s4;	t0g=re;
	  t0a =rt+t0g;			t0b =it+t0h;
	  t0g =rt-t0g;			t0h =it-t0h;
	  t04 =t04+t0a;			t05 =t05+t0b;
	  aj1p16r=t04;			aj1p16i=t05;
	  t0a =t04+c3m1*t0a;		t0b =t05+c3m1*t0b;
	  rt  =s3*t0g;			it  =s3*t0h;
	  aj1p04r=t0a+it;		aj1p04i=t0b-rt;
	  aj1p28r=t0a-it;		aj1p28i=t0b+rt;

/*...Second radix-9 transform:	*/
	  rt  =t12;			it  =t13;
	  t12 =rt+t14;			t13 =it+t15;
	  t14 =rt-t14;			t15 =it-t15;
	  t10 =t10+t12;			t11 =t11+t13;
	  t12 =t10+c3m1*t12;		t13 =t11+c3m1*t13;
	  rt  =s3*t14;			it  =s3*t15;
	  t14 =t12-it;			t15 =t13+rt;
	  t12 =t12+it;			t13 =t13-rt;

	  rt  =t18;			it  =t19;
	  t18 =rt+t1a;			t19 =it+t1b;
	  t1a =rt-t1a;			t1b =it-t1b;
	  t16 =t16+t18;			t17 =t17+t19;
	  t18 =t16+c3m1*t18;		t19 =t17+c3m1*t19;
	  rt  =s3*t1a;			it  =s3*t1b;
	  t1a =t18-it;			t1b =t19+rt;
	  t18 =t18+it;			t19 =t19-rt;

	  rt  =t1e;			it  =t1f;
	  t1e =rt+t1g;			t1f =it+t1h;
	  t1g =rt-t1g;			t1h =it-t1h;
	  t1c =t1c+t1e;			t1d =t1d+t1f;
	  t1e =t1c+c3m1*t1e;		t1f =t1d+c3m1*t1f;
	  rt  =s3*t1g;			it  =s3*t1h;
	  t1g =t1e-it;			t1h =t1f+rt;
	  t1e =t1e+it;			t1f =t1f-rt;
	/* Twiddles: */
	  rt  =t16;			it  =t17;
	  t16 =rt+t1c;			t17 =it+t1d;
	  t1c =rt-t1c;			t1d =it-t1d;
	  t10 =t10+t16;			t11 =t11+t17;
	  aj1p27r=t10;			aj1p27i=t11;
	  t16 =t10+c3m1*t16;		t17 =t11+c3m1*t17;
	  rt  =s3*t1c;			it  =s3*t1d;
	  aj1p15r=t16+it;		aj1p15i=t17-rt;
	  aj1p03r=t16-it;		aj1p03i=t17+rt;

	  rt  =t18*c +t19*s;		it  =t19*c -t18*s;
	  re  =t1e*c2+t1f*s2;		t1f =t1f*c2-t1e*s2;	t1e=re;
	  t18 =rt+t1e;			t19 =it+t1f;
	  t1e =rt-t1e;			t1f =it-t1f;
	  t12 =t12+t18;			t13 =t13+t19;
	  aj1p35r=t12;			aj1p35i=t13;
	  t18 =t12+c3m1*t18;		t19=t13+c3m1*t19;
	  rt  =s3*t1e;			it  =s3*t1f;
	  aj1p23r=t18+it;		aj1p23i=t19-rt;
	  aj1p11r=t18-it;		aj1p11i=t19+rt;

	  rt  =t1a*c2+t1b*s2;		it  =t1b*c2-t1a*s2;
	  re  =t1g*c4+t1h*s4;		t1h =t1h*c4-t1g*s4;	t1g=re;
	  t1a =rt+t1g;			t1b =it+t1h;
	  t1g =rt-t1g;			t1h =it-t1h;
	  t14 =t14+t1a;			t15 =t15+t1b;
	  aj1p07r=t14;			aj1p07i=t15;
	  t1a =t14+c3m1*t1a;		t1b =t15+c3m1*t1b;
	  rt  =s3*t1g;			it  =s3*t1h;
	  aj1p31r=t1a+it;		aj1p31i=t1b-rt;
	  aj1p19r=t1a-it;		aj1p19i=t1b+rt;

/*...Third radix-9 transform:	*/
	  rt  =t22;			it  =t23;
	  t22 =rt+t24;			t23 =it+t25;
	  t24 =rt-t24;			t25 =it-t25;
	  t20 =t20+t22;			t21 =t21+t23;
	  t22 =t20+c3m1*t22;		t23 =t21+c3m1*t23;
	  rt  =s3*t24;			it  =s3*t25;
	  t24 =t22-it;			t25 =t23+rt;
	  t22 =t22+it;			t23 =t23-rt;

	  rt  =t28;			it  =t29;
	  t28 =rt+t2a;			t29 =it+t2b;
	  t2a =rt-t2a;			t2b =it-t2b;
	  t26 =t26+t28;			t27 =t27+t29;
	  t28 =t26+c3m1*t28;		t29 =t27+c3m1*t29;
	  rt  =s3*t2a;			it  =s3*t2b;
	  t2a =t28-it;			t2b =t29+rt;
	  t28 =t28+it;			t29 =t29-rt;

	  rt  =t2e;			it  =t2f;
	  t2e =rt+t2g;			t2f =it+t2h;
	  t2g =rt-t2g;			t2h =it-t2h;
	  t2c =t2c+t2e;			t2d =t2d+t2f;
	  t2e =t2c+c3m1*t2e;		t2f =t2d+c3m1*t2f;
	  rt  =s3*t2g;			it  =s3*t2h;
	  t2g =t2e-it;			t2h =t2f+rt;
	  t2e =t2e+it;			t2f =t2f-rt;
	/* Twiddles: */
	  rt  =t26;			it  =t27;
	  t26 =rt+t2c;			t27 =it+t2d;
	  t2c =rt-t2c;			t2d =it-t2d;
	  t20 =t20+t26;			t21 =t21+t27;
	  aj1p18r=t20;			aj1p18i=t21;
	  t26 =t20+c3m1*t26;		t27 =t21+c3m1*t27;
	  rt  =s3*t2c;			it  =s3*t2d;
	  aj1p06r=t26+it;		aj1p06i=t27-rt;
	  aj1p30r=t26-it;		aj1p30i=t27+rt;

	  rt  =t28*c +t29*s;		it  =t29*c -t28*s;
	  re  =t2e*c2+t2f*s2;		t2f =t2f*c2-t2e*s2;	t2e=re;
	  t28 =rt+t2e;			t29 =it+t2f;
	  t2e =rt-t2e;			t2f =it-t2f;
	  t22 =t22+t28;			t23 =t23+t29;
	  aj1p26r=t22;			aj1p26i=t23;
	  t28 =t22+c3m1*t28;		t29=t23+c3m1*t29;
	  rt  =s3*t2e;			it  =s3*t2f;
	  aj1p14r=t28+it;		aj1p14i=t29-rt;
	  aj1p02r=t28-it;		aj1p02i=t29+rt;

	  rt  =t2a*c2+t2b*s2;		it  =t2b*c2-t2a*s2;
	  re  =t2g*c4+t2h*s4;		t2h =t2h*c4-t2g*s4;	t2g=re;
	  t2a =rt+t2g;			t2b =it+t2h;
	  t2g =rt-t2g;			t2h =it-t2h;
	  t24 =t24+t2a;			t25 =t25+t2b;
	  aj1p34r=t24;			aj1p34i=t25;
	  t2a =t24+c3m1*t2a;		t2b =t25+c3m1*t2b;
	  rt  =s3*t2g;			it  =s3*t2h;
	  aj1p22r=t2a+it;		aj1p22i=t2b-rt;
	  aj1p10r=t2a-it;		aj1p10i=t2b+rt;

/*...Fourth radix-9 transform:	*/
	  rt  =t32;			it  =t33;
	  t32 =rt+t34;			t33 =it+t35;
	  t34 =rt-t34;			t35 =it-t35;
	  t30 =t30+t32;			t31 =t31+t33;
	  t32 =t30+c3m1*t32;		t33 =t31+c3m1*t33;
	  rt  =s3*t34;			it  =s3*t35;
	  t34 =t32-it;			t35 =t33+rt;
	  t32 =t32+it;			t33 =t33-rt;

	  rt  =t38;			it  =t39;
	  t38 =rt+t3a;			t39 =it+t3b;
	  t3a =rt-t3a;			t3b =it-t3b;
	  t36 =t36+t38;			t37 =t37+t39;
	  t38 =t36+c3m1*t38;		t39 =t37+c3m1*t39;
	  rt  =s3*t3a;			it  =s3*t3b;
	  t3a =t38-it;			t3b =t39+rt;
	  t38 =t38+it;			t39 =t39-rt;

	  rt  =t3e;			it  =t3f;
	  t3e =rt+t3g;			t3f =it+t3h;
	  t3g =rt-t3g;			t3h =it-t3h;
	  t3c =t3c+t3e;			t3d =t3d+t3f;
	  t3e =t3c+c3m1*t3e;		t3f =t3d+c3m1*t3f;
	  rt  =s3*t3g;			it  =s3*t3h;
	  t3g =t3e-it;			t3h =t3f+rt;
	  t3e =t3e+it;			t3f =t3f-rt;
	/* Twiddles: */
	  rt  =t36;			it  =t37;
	  t36 =rt+t3c;			t37 =it+t3d;
	  t3c =rt-t3c;			t3d =it-t3d;
	  t30 =t30+t36;			t31 =t31+t37;
	  aj1p09r=t30;			aj1p09i=t31;
	  t36 =t30+c3m1*t36;		t37 =t31+c3m1*t37;
	  rt  =s3*t3c;			it  =s3*t3d;
	  aj1p33r=t36+it;		aj1p33i=t37-rt;
	  aj1p21r=t36-it;		aj1p21i=t37+rt;

	  rt  =t38*c +t39*s;		it  =t39*c -t38*s;
	  re  =t3e*c2+t3f*s2;		t3f =t3f*c2-t3e*s2;	t3e=re;
	  t38 =rt+t3e;			t39 =it+t3f;
	  t3e =rt-t3e;			t3f =it-t3f;
	  t32 =t32+t38;			t33 =t33+t39;
	  aj1p17r=t32;			aj1p17i=t33;
	  t38 =t32+c3m1*t38;		t39=t33+c3m1*t39;
	  rt  =s3*t3e;			it  =s3*t3f;
	  aj1p05r=t38+it;		aj1p05i=t39-rt;
	  aj1p29r=t38-it;		aj1p29i=t39+rt;

	  rt  =t3a*c2+t3b*s2;		it  =t3b*c2-t3a*s2;
	  re  =t3g*c4+t3h*s4;		t3h =t3h*c4-t3g*s4;	t3g=re;
	  t3a =rt+t3g;			t3b =it+t3h;
	  t3g =rt-t3g;			t3h =it-t3h;
	  t34 =t34+t3a;			t35 =t35+t3b;
	  aj1p25r=t34;			aj1p25i=t35;
	  t3a =t34+c3m1*t3a;		t3b =t35+c3m1*t3b;
	  rt  =s3*t3g;			it  =s3*t3h;
	  aj1p13r=t3a+it;		aj1p13i=t3b-rt;
	  aj1p01r=t3a-it;		aj1p01i=t3b+rt;

/*...Now do the carries. Since the outputs would
    normally be getting dispatched to 36 separate blocks of the A-array, we need 36 separate carries.	*/

	    l= j & (nwt-1);
	    n_minus_sil   = n-si[l  ];
	    n_minus_silp1 = n-si[l+1];
	    sinwt   = si[nwt-l  ];
	    sinwtm1 = si[nwt-l-1];

	    wtl     =wt0[    l  ];
	    wtn     =wt0[nwt-l  ]*scale;	/* Include 1/(n/2) scale factor of inverse transform here...	*/
	    wtlp1   =wt0[    l+1];
	    wtnm1   =wt0[nwt-l-1]*scale;	/* ...and here.	*/

/*...set0 is slightly different from others:	*/
	   cmplx_carry_norm_errcheck0(aj1p00r,aj1p00i,cy00,bjmodn00   );
	    cmplx_carry_norm_errcheck(aj1p01r,aj1p01i,cy01,bjmodn01,1 );
	    cmplx_carry_norm_errcheck(aj1p02r,aj1p02i,cy02,bjmodn02,2 );
	    cmplx_carry_norm_errcheck(aj1p03r,aj1p03i,cy03,bjmodn03,3 );
	    cmplx_carry_norm_errcheck(aj1p04r,aj1p04i,cy04,bjmodn04,4 );
	    cmplx_carry_norm_errcheck(aj1p05r,aj1p05i,cy05,bjmodn05,5 );
	    cmplx_carry_norm_errcheck(aj1p06r,aj1p06i,cy06,bjmodn06,6 );
	    cmplx_carry_norm_errcheck(aj1p07r,aj1p07i,cy07,bjmodn07,7 );
	    cmplx_carry_norm_errcheck(aj1p08r,aj1p08i,cy08,bjmodn08,8 );
	    cmplx_carry_norm_errcheck(aj1p09r,aj1p09i,cy09,bjmodn09,9 );
	    cmplx_carry_norm_errcheck(aj1p10r,aj1p10i,cy10,bjmodn10,10);
	    cmplx_carry_norm_errcheck(aj1p11r,aj1p11i,cy11,bjmodn11,11);
	    cmplx_carry_norm_errcheck(aj1p12r,aj1p12i,cy12,bjmodn12,12);
	    cmplx_carry_norm_errcheck(aj1p13r,aj1p13i,cy13,bjmodn13,13);
	    cmplx_carry_norm_errcheck(aj1p14r,aj1p14i,cy14,bjmodn14,14);
	    cmplx_carry_norm_errcheck(aj1p15r,aj1p15i,cy15,bjmodn15,15);
	    cmplx_carry_norm_errcheck(aj1p16r,aj1p16i,cy16,bjmodn16,16);
	    cmplx_carry_norm_errcheck(aj1p17r,aj1p17i,cy17,bjmodn17,17);
	    cmplx_carry_norm_errcheck(aj1p18r,aj1p18i,cy18,bjmodn18,18);
	    cmplx_carry_norm_errcheck(aj1p19r,aj1p19i,cy19,bjmodn19,19);
	    cmplx_carry_norm_errcheck(aj1p20r,aj1p20i,cy20,bjmodn20,20);
	    cmplx_carry_norm_errcheck(aj1p21r,aj1p21i,cy21,bjmodn21,21);
	    cmplx_carry_norm_errcheck(aj1p22r,aj1p22i,cy22,bjmodn22,22);
	    cmplx_carry_norm_errcheck(aj1p23r,aj1p23i,cy23,bjmodn23,23);
	    cmplx_carry_norm_errcheck(aj1p24r,aj1p24i,cy24,bjmodn24,24);
	    cmplx_carry_norm_errcheck(aj1p25r,aj1p25i,cy25,bjmodn25,25);
	    cmplx_carry_norm_errcheck(aj1p26r,aj1p26i,cy26,bjmodn26,26);
	    cmplx_carry_norm_errcheck(aj1p27r,aj1p27i,cy27,bjmodn27,27);
	    cmplx_carry_norm_errcheck(aj1p28r,aj1p28i,cy28,bjmodn28,28);
	    cmplx_carry_norm_errcheck(aj1p29r,aj1p29i,cy29,bjmodn29,29);
	    cmplx_carry_norm_errcheck(aj1p30r,aj1p30i,cy30,bjmodn30,30);
	    cmplx_carry_norm_errcheck(aj1p31r,aj1p31i,cy31,bjmodn31,31);
	    cmplx_carry_norm_errcheck(aj1p32r,aj1p32i,cy32,bjmodn32,32);
	    cmplx_carry_norm_errcheck(aj1p33r,aj1p33i,cy33,bjmodn33,33);
	    cmplx_carry_norm_errcheck(aj1p34r,aj1p34i,cy34,bjmodn34,34);
	    cmplx_carry_norm_errcheck(aj1p35r,aj1p35i,cy35,bjmodn35,35);

	    i =((uint32)(sw - bjmodn00) >> 31);	/* get ready for the next set...	*/
	    co2=co3;	/* For all data but the first set in each j-block, co2=co3. Thus, after the first block of data is done
			   and only then: for all subsequent blocks it's superfluous), this assignment decrements co2 by radix(1).	*/

/*...The radix-36 DIF pass is here:	*/
#if PFETCH
add0 = &a[jt];
prefetch_p_doubles(add0);
#endif
/*...First radix-9 transform:	*/
	/* permuted version of jt + p00,p12,p24 */
	  t00 =aj1p00r;			t01 =aj1p00i;
	  t02 =aj1p24r+aj1p12r;		t03 =aj1p24i+aj1p12i;
	  t04 =aj1p24r-aj1p12r;		t05 =aj1p24i-aj1p12i;
	  t00 =t00+t02;			t01 =t01+t03;
	  t02 =t00+c3m1*t02;		t03 =t01+c3m1*t03;
	  rt  =s3*t04;			it  =s3*t05;
	  t04 =t02+it;			t05 =t03-rt;
	  t02 =t02-it;			t03 =t03+rt;
#if PFETCH
add = add0+p01;
prefetch_p_doubles(add);
#endif
	/* permuted version of jt + p04,p16,p28 */
	  t06 =aj1p32r;			t07 =aj1p32i;
	  t08 =aj1p20r+aj1p08r;		t09 =aj1p20i+aj1p08i;
	  t0a =aj1p20r-aj1p08r;		t0b =aj1p20i-aj1p08i;
	  t06 =t06+t08;			t07 =t07+t09;
	  t08 =t06+c3m1*t08;		t09 =t07+c3m1*t09;
	  rt  =s3*t0a;			it  =s3*t0b;
	  t0a =t08+it;			t0b =t09-rt;
	  t08 =t08-it;			t09 =t09+rt;
#if PFETCH
add = add0+p02;
prefetch_p_doubles(add);
#endif
	/* permuted version of jt + p08,p20,p32 */
	  t0c =aj1p28r;			t0d =aj1p28i;
	  t0e =aj1p16r+aj1p04r;		t0f =aj1p16i+aj1p04i;
	  t0g =aj1p16r-aj1p04r;		t0h =aj1p16i-aj1p04i;
	  t0c =t0c+t0e;			t0d =t0d+t0f;
	  t0e =t0c+c3m1*t0e;		t0f =t0d+c3m1*t0f;
	  rt  =s3*t0g;			it  =s3*t0h;
	  t0g =t0e+it;			t0h =t0f-rt;
	  t0e =t0e-it;			t0f =t0f+rt;
#if PFETCH
add = add0+p03;
prefetch_p_doubles(add);
#endif
	/* Twiddles: */
	  rt  =t06;			it  =t07;
	  t06 =rt+t0c;			t07 =it+t0d;
	  t0c =rt-t0c;			t0d =it-t0d;
	  t00 =t00+t06;			t01 =t01+t07;
	  t06 =t00+c3m1*t06;		t07 =t01+c3m1*t07;
	  rt  =s3*t0c;			it  =s3*t0d;
	  t0c =t06+it;			t0d =t07-rt;
	  t06 =t06-it;			t07 =t07+rt;
#if PFETCH
add = add0+p04;
prefetch_p_doubles(add);
#endif
	  rt  =t08*c -t09*s;		it  =t08*s +t09*c;
	  re  =t0e*c2-t0f*s2;		t0f =t0e*s2+t0f*c2;	t0e=re;
	  t08 =rt+t0e;			t09 =it+t0f;
	  t0e =rt-t0e;			t0f =it-t0f;
	  t02 =t02+t08;			t03 =t03+t09;
	  t08 =t02+c3m1*t08;		t09 =t03+c3m1*t09;
	  rt  =s3*t0e;			it  =s3*t0f;
	  t0e =t08+it;			t0f =t09-rt;
	  t08 =t08-it;			t09 =t09+rt;
#if PFETCH
add = add0+p05;
prefetch_p_doubles(add);
#endif
	  rt  =t0a*c2-t0b*s2;		it  =t0a*s2+t0b*c2;
	  re  =t0g*c4-t0h*s4;		t0h =t0g*s4+t0h*c4;	t0g=re;
	  t0a =rt+t0g;			t0b =it+t0h;
	  t0g =rt-t0g;			t0h =it-t0h;
	  t04 =t04+t0a;			t05 =t05+t0b;
	  t0a =t04+c3m1*t0a;		t0b =t05+c3m1*t0b;
	  rt  =s3*t0g;			it  =s3*t0h;
	  t0g =t0a+it;			t0h =t0b-rt;
	  t0a =t0a-it;			t0b =t0b+rt;
#if PFETCH
add = add0+p06;
prefetch_p_doubles(add);
#endif

/*...Second radix-9 transform:	*/
	/* permuted version of jt + p01,p13,p25 */
	  t10 =aj1p27r;			t11 =aj1p27i;
	  t12 =aj1p15r+aj1p03r;		t13 =aj1p15i+aj1p03i;
	  t14 =aj1p15r-aj1p03r;		t15 =aj1p15i-aj1p03i;
	  t10 =t10+t12;			t11 =t11+t13;
	  t12 =t10+c3m1*t12;		t13 =t11+c3m1*t13;
	  rt  =s3*t14;			it  =s3*t15;
	  t14 =t12+it;			t15 =t13-rt;
	  t12 =t12-it;			t13 =t13+rt;
#if PFETCH
add = add0+p07;
prefetch_p_doubles(add);
#endif
	/* permuted version of jt + p05,p17,p29 */
	  t16 =aj1p23r;			t17 =aj1p23i;
	  t18 =aj1p11r+aj1p35r;		t19 =aj1p11i+aj1p35i;
	  t1a =aj1p11r-aj1p35r;		t1b =aj1p11i-aj1p35i;
	  t16 =t16+t18;			t17 =t17+t19;
	  t18 =t16+c3m1*t18;		t19 =t17+c3m1*t19;
	  rt  =s3*t1a;			it  =s3*t1b;
	  t1a =t18+it;			t1b =t19-rt;
	  t18 =t18-it;			t19 =t19+rt;
#if PFETCH
add = add0+p08;
prefetch_p_doubles(add);
#endif
	/* permuted version of jt + p09,p21,p33 */
	  t1c =aj1p19r;			t1d =aj1p19i;
	  t1e =aj1p07r+aj1p31r;		t1f =aj1p07i+aj1p31i;
	  t1g =aj1p07r-aj1p31r;		t1h =aj1p07i-aj1p31i;
	  t1c =t1c+t1e;			t1d =t1d+t1f;
	  t1e =t1c+c3m1*t1e;		t1f =t1d+c3m1*t1f;
	  rt  =s3*t1g;			it  =s3*t1h;
	  t1g =t1e+it;			t1h =t1f-rt;
	  t1e =t1e-it;			t1f =t1f+rt;
#if PFETCH
add = add0+p09;
prefetch_p_doubles(add);
#endif
	/* Twiddles: */
	  rt  =t16;			it  =t17;
	  t16 =rt+t1c;			t17 =it+t1d;
	  t1c =rt-t1c;			t1d =it-t1d;
	  t10 =t10+t16;			t11 =t11+t17;
	  t16 =t10+c3m1*t16;		t17 =t11+c3m1*t17;
	  rt  =s3*t1c;			it  =s3*t1d;
	  t1c =t16+it;			t1d =t17-rt;
	  t16 =t16-it;			t17 =t17+rt;
#if PFETCH
add = add0+p10;
prefetch_p_doubles(add);
#endif
	  rt  =t18*c -t19*s;		it  =t18*s +t19*c;
	  re  =t1e*c2-t1f*s2;		t1f =t1e*s2+t1f*c2;	t1e=re;
	  t18 =rt+t1e;			t19 =it+t1f;
	  t1e =rt-t1e;			t1f =it-t1f;
	  t12 =t12+t18;			t13 =t13+t19;
	  t18 =t12+c3m1*t18;		t19 =t13+c3m1*t19;
	  rt  =s3*t1e;			it  =s3*t1f;
	  t1e =t18+it;			t1f =t19-rt;
	  t18 =t18-it;			t19 =t19+rt;
#if PFETCH
add = add0+p11;
prefetch_p_doubles(add);
#endif
	  rt  =t1a*c2-t1b*s2;		it  =t1a*s2+t1b*c2;
	  re  =t1g*c4-t1h*s4;		t1h =t1g*s4+t1h*c4;	t1g=re;
	  t1a =rt+t1g;			t1b =it+t1h;
	  t1g =rt-t1g;			t1h =it-t1h;
	  t14 =t14+t1a;			t15 =t15+t1b;
	  t1a =t14+c3m1*t1a;		t1b =t15+c3m1*t1b;
	  rt  =s3*t1g;			it  =s3*t1h;
	  t1g =t1a+it;			t1h =t1b-rt;
	  t1a =t1a-it;			t1b =t1b+rt;
#if PFETCH
add = add0+p12;
prefetch_p_doubles(add);
#endif

/*...Third radix-9 transform:	*/
	/* permuted version of jt + p02,p14,p26 */
	  t20 =aj1p18r;			t21 =aj1p18i;
	  t22 =aj1p06r+aj1p30r;		t23 =aj1p06i+aj1p30i;
	  t24 =aj1p06r-aj1p30r;		t25 =aj1p06i-aj1p30i;
	  t20 =t20+t22;			t21 =t21+t23;
	  t22 =t20+c3m1*t22;		t23 =t21+c3m1*t23;
	  rt  =s3*t24;			it  =s3*t25;
	  t24 =t22+it;			t25 =t23-rt;
	  t22 =t22-it;			t23 =t23+rt;
#if PFETCH
add = add0+p13;
prefetch_p_doubles(add);
#endif
	/* permuted version of jt + p06,p18,p30 */
	  t26 =aj1p14r;			t27 =aj1p14i;
	  t28 =aj1p02r+aj1p26r;		t29 =aj1p02i+aj1p26i;
	  t2a =aj1p02r-aj1p26r;		t2b =aj1p02i-aj1p26i;
	  t26 =t26+t28;			t27 =t27+t29;
	  t28 =t26+c3m1*t28;		t29 =t27+c3m1*t29;
	  rt  =s3*t2a;			it  =s3*t2b;
	  t2a =t28+it;			t2b =t29-rt;
	  t28 =t28-it;			t29 =t29+rt;
#if PFETCH
add = add0+p14;
prefetch_p_doubles(add);
#endif
	/* permuted version of jt + p10,p22,p34 */
	  t2c =aj1p10r;			t2d =aj1p10i;
	  t2e =aj1p34r+aj1p22r;		t2f =aj1p34i+aj1p22i;
	  t2g =aj1p34r-aj1p22r;		t2h =aj1p34i-aj1p22i;
	  t2c =t2c+t2e;			t2d =t2d+t2f;
	  t2e =t2c+c3m1*t2e;		t2f =t2d+c3m1*t2f;
	  rt  =s3*t2g;			it  =s3*t2h;
	  t2g =t2e+it;			t2h =t2f-rt;
	  t2e =t2e-it;			t2f =t2f+rt;
#if PFETCH
add = add0+p15;
prefetch_p_doubles(add);
#endif
	/* Twiddles: */
	  rt  =t26;			it  =t27;
	  t26 =rt+t2c;			t27 =it+t2d;
	  t2c =rt-t2c;			t2d =it-t2d;
	  t20 =t20+t26;			t21 =t21+t27;
	  t26 =t20+c3m1*t26;		t27 =t21+c3m1*t27;
	  rt  =s3*t2c;			it  =s3*t2d;
	  t2c =t26+it;			t2d =t27-rt;
	  t26 =t26-it;			t27 =t27+rt;
#if PFETCH
add = add0+p16;
prefetch_p_doubles(add);
#endif
	  rt  =t28*c -t29*s;		it  =t28*s +t29*c;
	  re  =t2e*c2-t2f*s2;		t2f =t2e*s2+t2f*c2;	t2e=re;
	  t28 =rt+t2e;			t29 =it+t2f;
	  t2e =rt-t2e;			t2f =it-t2f;
	  t22 =t22+t28;			t23 =t23+t29;
	  t28 =t22+c3m1*t28;		t29 =t23+c3m1*t29;
	  rt  =s3*t2e;			it  =s3*t2f;
	  t2e =t28+it;			t2f =t29-rt;
	  t28 =t28-it;			t29 =t29+rt;
#if PFETCH
add = add0+p17;
prefetch_p_doubles(add);
#endif
	  rt  =t2a*c2-t2b*s2;		it  =t2a*s2+t2b*c2;
	  re  =t2g*c4-t2h*s4;		t2h =t2g*s4+t2h*c4;	t2g=re;
	  t2a =rt+t2g;			t2b =it+t2h;
	  t2g =rt-t2g;			t2h =it-t2h;
	  t24 =t24+t2a;			t25 =t25+t2b;
	  t2a =t24+c3m1*t2a;		t2b =t25+c3m1*t2b;
	  rt  =s3*t2g;			it  =s3*t2h;
	  t2g =t2a+it;			t2h =t2b-rt;
	  t2a =t2a-it;			t2b =t2b+rt;
#if PFETCH
add = add0+p18;
prefetch_p_doubles(add);
#endif

/*...Fourth radix-9 transform:	*/
	/* permuted version of jt + p03,p15,p27 */
	  t30 =aj1p09r;			t31 =aj1p09i;
	  t32 =aj1p33r+aj1p21r;		t33 =aj1p33i+aj1p21i;
	  t34 =aj1p33r-aj1p21r;		t35 =aj1p33i-aj1p21i;
	  t30 =t30+t32;			t31 =t31+t33;
	  t32 =t30+c3m1*t32;		t33 =t31+c3m1*t33;
	  rt  =s3*t34;			it  =s3*t35;
	  t34 =t32+it;			t35 =t33-rt;
	  t32 =t32-it;			t33 =t33+rt;
#if PFETCH
add = add0+p19;
prefetch_p_doubles(add);
#endif
	/* permuted version of jt + p07,p19,p31 */
	  t36 =aj1p05r;			t37 =aj1p05i;
	  t38 =aj1p29r+aj1p17r;		t39 =aj1p29i+aj1p17i;
	  t3a =aj1p29r-aj1p17r;		t3b =aj1p29i-aj1p17i;
	  t36 =t36+t38;			t37 =t37+t39;
	  t38 =t36+c3m1*t38;		t39 =t37+c3m1*t39;
	  rt  =s3*t3a;			it  =s3*t3b;
	  t3a =t38+it;			t3b =t39-rt;
	  t38 =t38-it;			t39 =t39+rt;
#if PFETCH
add = add0+p20;
prefetch_p_doubles(add);
#endif
	/* permuted version of jt + p11,p23,p35 */
	  t3c =aj1p01r;			t3d =aj1p01i;
	  t3e =aj1p25r+aj1p13r;		t3f =aj1p25i+aj1p13i;
	  t3g =aj1p25r-aj1p13r;		t3h =aj1p25i-aj1p13i;
	  t3c =t3c+t3e;			t3d =t3d+t3f;
	  t3e =t3c+c3m1*t3e;		t3f =t3d+c3m1*t3f;
	  rt  =s3*t3g;			it  =s3*t3h;
	  t3g =t3e+it;			t3h =t3f-rt;
	  t3e =t3e-it;			t3f =t3f+rt;
#if PFETCH
add = add0+p21;
prefetch_p_doubles(add);
#endif
	/* Twiddles: */
	  rt  =t36;			it  =t37;
	  t36 =rt+t3c;			t37 =it+t3d;
	  t3c =rt-t3c;			t3d =it-t3d;
	  t30 =t30+t36;			t31 =t31+t37;
	  t36 =t30+c3m1*t36;		t37 =t31+c3m1*t37;
	  rt  =s3*t3c;			it  =s3*t3d;
	  t3c =t36+it;			t3d =t37-rt;
	  t36 =t36-it;			t37 =t37+rt;
#if PFETCH
add = add0+p22;
prefetch_p_doubles(add);
#endif
	  rt  =t38*c -t39*s;		it  =t38*s +t39*c;
	  re  =t3e*c2-t3f*s2;		t3f =t3e*s2+t3f*c2;	t3e=re;
	  t38 =rt+t3e;			t39 =it+t3f;
	  t3e =rt-t3e;			t3f =it-t3f;
	  t32 =t32+t38;			t33 =t33+t39;
	  t38 =t32+c3m1*t38;		t39 =t33+c3m1*t39;
	  rt  =s3*t3e;			it  =s3*t3f;
	  t3e =t38+it;			t3f =t39-rt;
	  t38 =t38-it;			t39 =t39+rt;
#if PFETCH
add = add0+p23;
prefetch_p_doubles(add);
#endif
	  rt  =t3a*c2-t3b*s2;		it  =t3a*s2+t3b*c2;
	  re  =t3g*c4-t3h*s4;		t3h =t3g*s4+t3h*c4;	t3g=re;
	  t3a =rt+t3g;			t3b =it+t3h;
	  t3g =rt-t3g;			t3h =it-t3h;
	  t34 =t34+t3a;			t35 =t35+t3b;
	  t3a =t34+c3m1*t3a;		t3b =t35+c3m1*t3b;
	  rt  =s3*t3g;			it  =s3*t3h;
	  t3g =t3a+it;			t3h =t3b-rt;
	  t3a =t3a-it;			t3b =t3b+rt;
#if PFETCH
add = add0+p24;
prefetch_p_doubles(add);
#endif

/*...and now do nine radix-4 transforms:	*/

/*...Block 1: */
	j1=jt;
	    rt =t20;	t20=t00-rt;	t00=t00+rt;
	    it =t21;	t21=t01-it;	t01=t01+it;

	    rt =t30;	t30=t10-rt;	t10=t10+rt;
	    it =t31;	t31=t11-it;	t11=t11+it;

	    a[j1    ]=t00+t10;	a[j1+1    ]=t01+t11;
	    a[j1+p01]=t00-t10;	a[j1+1+p01]=t01-t11;

	    a[j1+p03]=t20-t31;	a[j1+1+p03]=t21+t30;	/* mpy by I is inlined here...	*/
	    a[j1+p02]=t20+t31;	a[j1+1+p02]=t21-t30;
#if PFETCH
add = add0+p25;
prefetch_p_doubles(add);
#endif
/*...Block 2: */
	j1=jt+p32;
	    rt =t22;	t22=t02-rt;	t02=t02+rt;
	    it =t23;	t23=t03-it;	t03=t03+it;

	    rt =t32;	t32=t12-rt;	t12=t12+rt;
	    it =t33;	t33=t13-it;	t13=t13+it;

	    a[j1    ]=t02+t12;	a[j1+1    ]=t03+t13;
	    a[j1+p01]=t02-t12;	a[j1+1+p01]=t03-t13;

	    a[j1+p03]=t22-t33;	a[j1+1+p03]=t23+t32;	/* mpy by I is inlined here...	*/
	    a[j1+p02]=t22+t33;	a[j1+1+p02]=t23-t32;
#if PFETCH
add = add0+p26;
prefetch_p_doubles(add);
#endif
/*...Block 3: */
	j1=jt+p20;
	    rt =t24;	t24=t04-rt;	t04=t04+rt;
	    it =t25;	t25=t05-it;	t05=t05+it;

	    rt =t34;	t34=t14-rt;	t14=t14+rt;
	    it =t35;	t35=t15-it;	t15=t15+it;

	    a[j1+p02]=t04+t14;	a[j1+1+p02]=t05+t15;
	    a[j1+p03]=t04-t14;	a[j1+1+p03]=t05-t15;

	    a[j1    ]=t24-t35;	a[j1+1    ]=t25+t34;	/* mpy by I is inlined here...	*/
	    a[j1+p01]=t24+t35;	a[j1+1+p01]=t25-t34;
#if PFETCH
add = add0+p27;
prefetch_p_doubles(add);
#endif
/*...Block 4: */
	j1=jt+p08;
	    rt =t26;	t26=t06-rt;	t06=t06+rt;
	    it =t27;	t27=t07-it;	t07=t07+it;

	    rt =t36;	t36=t16-rt;	t16=t16+rt;
	    it =t37;	t37=t17-it;	t17=t17+it;

	    a[j1+p01]=t06+t16;	a[j1+1+p01]=t07+t17;
	    a[j1    ]=t06-t16;	a[j1+1    ]=t07-t17;

	    a[j1+p02]=t26-t37;	a[j1+1+p02]=t27+t36;	/* mpy by I is inlined here...	*/
	    a[j1+p03]=t26+t37;	a[j1+1+p03]=t27-t36;
#if PFETCH
add = add0+p28;
prefetch_p_doubles(add);
#endif
/*...Block 5: */
	j1=jt+p28;
	    rt =t28;	t28=t08-rt;	t08=t08+rt;
	    it =t29;	t29=t09-it;	t09=t09+it;

	    rt =t38;	t38=t18-rt;	t18=t18+rt;
	    it =t39;	t39=t19-it;	t19=t19+it;

	    a[j1+p03]=t08+t18;	a[j1+1+p03]=t09+t19;
	    a[j1+p02]=t08-t18;	a[j1+1+p02]=t09-t19;

	    a[j1+p01]=t28-t39;	a[j1+1+p01]=t29+t38;	/* mpy by I is inlined here...	*/
	    a[j1    ]=t28+t39;	a[j1+1    ]=t29-t38;
#if PFETCH
add = add0+p29;
prefetch_p_doubles(add);
#endif
/*...Block 6: */
	j1=jt+p16;
	    rt =t2a;	t2a=t0a-rt;	t0a=t0a+rt;
	    it =t2b;	t2b=t0b-it;	t0b=t0b+it;

	    rt =t3a;	t3a=t1a-rt;	t1a=t1a+rt;
	    it =t3b;	t3b=t1b-it;	t1b=t1b+it;

	    a[j1    ]=t0a+t1a;	a[j1+1    ]=t0b+t1b;
	    a[j1+p01]=t0a-t1a;	a[j1+1+p01]=t0b-t1b;

	    a[j1+p03]=t2a-t3b;	a[j1+1+p03]=t2b+t3a;	/* mpy by I is inlined here...	*/
	    a[j1+p02]=t2a+t3b;	a[j1+1+p02]=t2b-t3a;
#if PFETCH
add = add0+p30;
prefetch_p_doubles(add);
#endif
/*...Block 7: */
	j1=jt+p04;
	    rt =t2c;	t2c=t0c-rt;	t0c=t0c+rt;
	    it =t2d;	t2d=t0d-it;	t0d=t0d+it;

	    rt =t3c;	t3c=t1c-rt;	t1c=t1c+rt;
	    it =t3d;	t3d=t1d-it;	t1d=t1d+it;

	    a[j1+p02]=t0c+t1c;	a[j1+1+p02]=t0d+t1d;
	    a[j1+p03]=t0c-t1c;	a[j1+1+p03]=t0d-t1d;

	    a[j1    ]=t2c-t3d;	a[j1+1    ]=t2d+t3c;	/* mpy by I is inlined here...	*/
	    a[j1+p01]=t2c+t3d;	a[j1+1+p01]=t2d-t3c;
#if PFETCH
add = add0+p31;
prefetch_p_doubles(add);
#endif
/*...Block 8: */
	j1=jt+p24;
	    rt =t2e;	t2e=t0e-rt;	t0e=t0e+rt;
	    it =t2f;	t2f=t0f-it;	t0f=t0f+it;

	    rt =t3e;	t3e=t1e-rt;	t1e=t1e+rt;
	    it =t3f;	t3f=t1f-it;	t1f=t1f+it;

	    a[j1+p01]=t0e+t1e;	a[j1+1+p01]=t0f+t1f;
	    a[j1    ]=t0e-t1e;	a[j1+1    ]=t0f-t1f;

	    a[j1+p02]=t2e-t3f;	a[j1+1+p02]=t2f+t3e;	/* mpy by I is inlined here...	*/
	    a[j1+p03]=t2e+t3f;	a[j1+1+p03]=t2f-t3e;
#if PFETCH
add = add0+p32;
prefetch_p_doubles(add);
#endif
/*...Block 9: */
	j1=jt+p12;
	    rt =t2g;	t2g=t0g-rt;	t0g=t0g+rt;
	    it =t2h;	t2h=t0h-it;	t0h=t0h+it;

	    rt =t3g;	t3g=t1g-rt;	t1g=t1g+rt;
	    it =t3h;	t3h=t1h-it;	t1h=t1h+it;

	    a[j1+p03]=t0g+t1g;	a[j1+1+p03]=t0h+t1h;
	    a[j1+p02]=t0g-t1g;	a[j1+1+p02]=t0h-t1h;

	    a[j1+p01]=t2g-t3h;	a[j1+1+p01]=t2h+t3g;	/* mpy by I is inlined here...	*/
	    a[j1    ]=t2g+t3h;	a[j1+1    ]=t2h-t3g;
#if PFETCH
add = add0+p33;
prefetch_p_doubles(add);
#endif
#if PFETCH
add = add0+p34;
prefetch_p_doubles(add);
#endif
#if PFETCH
add = add0+p35;
prefetch_p_doubles(add);
#endif

	  iroot += root_incr;		/* increment sincos index.	*/

	  }

	  jstart += nwt;
	  jhi    += nwt;
	  col += 36;
	  co3 -= 36;

	}

	if(root_incr==0)break;

/*   Wraparound carry cleanup loop is here: ***
!
!   (1) Invert the radix-36 forward DIF FFT of the first block of 36 complex elements in A and unweight;
!   (2) Propagate cleanup carries among the real and imaginary parts of the 36 outputs of (1);
!   (3) Reweight and perform a radix-36 forward DIF FFT on the result of (2);
!   (4) If any of the exit carries from (2) are nonzero, advance to the next 36 elements and repeat (1-4).
*/
	t00 = cy35;
	cy35= cy34;
	cy34= cy33;
	cy33= cy32;
	cy32= cy31;
	cy31= cy30;
	cy30= cy29;
	cy29= cy28;
	cy28= cy27;
	cy27= cy26;
	cy26= cy25;
	cy25= cy24;
	cy24= cy23;
	cy23= cy22;
	cy22= cy21;
	cy21= cy20;
	cy20= cy19;
	cy19= cy18;
	cy18= cy17;
	cy17= cy16;
	cy16= cy15;
	cy15= cy14;
	cy14= cy13;
	cy13= cy12;
	cy12= cy11;
	cy11= cy10;
	cy10= cy09;
	cy09= cy08;
	cy08= cy07;
	cy07= cy06;
	cy06= cy05;
	cy05= cy04;
	cy04= cy03;
	cy03= cy02;
	cy02= cy01;
	cy01= cy00;
	cy00= t00 ;

	iroot = 0;
	root_incr = 0;
	scale = 1;

	jstart = 0;
	jhi = 3;
	khi = 1;

	for(j=0; j<=jhi; j++)
	{
	  a[j    ] *= radix_inv;
	  a[j+p01] *= radix_inv;
	  a[j+p02] *= radix_inv;
	  a[j+p03] *= radix_inv;
	  a[j+p04] *= radix_inv;
	  a[j+p05] *= radix_inv;
	  a[j+p06] *= radix_inv;
	  a[j+p07] *= radix_inv;
	  a[j+p08] *= radix_inv;
	  a[j+p09] *= radix_inv;
	  a[j+p10] *= radix_inv;
	  a[j+p11] *= radix_inv;
	  a[j+p12] *= radix_inv;
	  a[j+p13] *= radix_inv;
	  a[j+p14] *= radix_inv;
	  a[j+p15] *= radix_inv;
	  a[j+p16] *= radix_inv;
	  a[j+p17] *= radix_inv;
	  a[j+p18] *= radix_inv;
	  a[j+p19] *= radix_inv;
	  a[j+p20] *= radix_inv;
	  a[j+p21] *= radix_inv;
	  a[j+p22] *= radix_inv;
	  a[j+p23] *= radix_inv;
	  a[j+p24] *= radix_inv;
	  a[j+p25] *= radix_inv;
	  a[j+p26] *= radix_inv;
	  a[j+p27] *= radix_inv;
	  a[j+p28] *= radix_inv;
	  a[j+p29] *= radix_inv;
	  a[j+p30] *= radix_inv;
	  a[j+p31] *= radix_inv;
	  a[j+p32] *= radix_inv;
	  a[j+p33] *= radix_inv;
	  a[j+p34] *= radix_inv;
	  a[j+p35] *= radix_inv;
	}
}

	if(fabs(cy00)+fabs(cy01)+fabs(cy02)+fabs(cy03)+fabs(cy04)+fabs(cy05)+fabs(cy06)+fabs(cy07)+fabs(cy08)+fabs(cy09)+fabs(cy10)+fabs(cy11)+fabs(cy12)+fabs(cy13)+fabs(cy14)+fabs(cy15)+fabs(cy16)+fabs(cy17)+fabs(cy18)+fabs(cy19)+fabs(cy20)+fabs(cy21)+fabs(cy22)+fabs(cy23)+fabs(cy24)+fabs(cy25)+fabs(cy26)+fabs(cy27)+fabs(cy28)+fabs(cy29)+fabs(cy30)+fabs(cy31)+fabs(cy32)+fabs(cy33)+fabs(cy34)+fabs(cy35) != 0.0)
	{
	    if(INTERACT)printf("FATAL: iter = %10d; nonzero exit carry in radix36_ditN_cy_dif1.\n",iter);
	    f1 = fopen(   OFILE,"a");
	    f2 = fopen(STATFILE,"a");
	    fprintf(f1,"FATAL: iter = %10d; nonzero exit carry in radix36_ditN_cy_dif1.\n",iter);
	    fprintf(f2,"FATAL: iter = %10d; nonzero exit carry in radix36_ditN_cy_dif1.\n",iter);
	    fclose(f1);
	    fclose(f2);
	    err=1;
	    return(err);
	}
	*fracmax = maxerr;
	return(0);
}

/***************/

int radix36_ditN_cy_dif1_nochk(double a[], int n, int nwt, int nwt_bits, double wt0[], double wt1[], int si[], double base[], double baseinv[], int iter                 , int p)
{
/*
!...Acronym: DWT = Discrete Weighted Transform, DIT = Decimation In Time, DIF = Decimation In Frequency
!
!...Performs a final radix-36 complex DIT pass, an inverse DWT weighting, a carry propagation,
!   a forward DWT weighting, and an initial radix-36 complex DIF pass on the data in the length-N real vector A.
!
!   Data enter and are returned in the A-array.
!
!   See the documentation in mers_mod_square and radix16_dif_pass for further details on the array
!   storage scheme, and radix8_ditN_cy_dif1 for details on the reduced-length weights array scheme.
*/
	int n36, bjmodn00,bjmodn01,bjmodn02,bjmodn03,bjmodn04,bjmodn05,bjmodn06,bjmodn07,bjmodn08,bjmodn09,bjmodn10,bjmodn11,bjmodn12,bjmodn13,bjmodn14,bjmodn15,bjmodn16,bjmodn17
		,bjmodn18,bjmodn19,bjmodn20,bjmodn21,bjmodn22,bjmodn23,bjmodn24,bjmodn25,bjmodn26,bjmodn27,bjmodn28,bjmodn29,bjmodn30,bjmodn31,bjmodn32,bjmodn33,bjmodn34,bjmodn35
		,i,j,jt,j1,jstart,jhi,iroot,root_incr,k,khi,l,outer;
	static int psave=0,bw,sw,bjmodnini,nrtm1,p01,p02,p03,p04,p05,p06,p07,p08,p09,p10,p11,p12,p13,p14,p15,p16,p17
		,p18,p19,p20,p21,p22,p23,p24,p25,p26,p27,p28,p29,p30,p31,p32,p33,p34,p35;
	const  double one_half[3] = {1.0, 0.5, 0.25};	/* Needed for small-weights-tables scheme */
	static double c   =  0.76604444311897803520,	/* cos(2*pi/9) */
		      s   =  0.64278760968653932631,	/* sin(2*pi/9) */
		      c2  =  0.17364817766693034887,	/* cos(2*u) */
		      s2  =  0.98480775301220805936,	/* sin(2*u) */
		      c3m1= -1.50000000000000000000,	/* cos(3*u)-1 */
		      s3  =  0.86602540378443864677,	/* sin(3*u) */
		      c4  = -0.93969262078590838404,	/* cos(4*u) */
		      s4  =  0.34202014332566873307;	/* sin(4*u) */
	static double radix_inv, n2inv;
	double rt,it,re
		,t00,t01,t02,t03,t04,t05,t06,t07,t08,t09,t0a,t0b,t0c,t0d,t0e,t0f,t0g,t0h
		,t10,t11,t12,t13,t14,t15,t16,t17,t18,t19,t1a,t1b,t1c,t1d,t1e,t1f,t1g,t1h
		,t20,t21,t22,t23,t24,t25,t26,t27,t28,t29,t2a,t2b,t2c,t2d,t2e,t2f,t2g,t2h
		,t30,t31,t32,t33,t34,t35,t36,t37,t38,t39,t3a,t3b,t3c,t3d,t3e,t3f,t3g,t3h
	,aj1p00r,aj1p01r,aj1p02r,aj1p03r,aj1p04r,aj1p05r,aj1p06r,aj1p07r,aj1p08r,aj1p09r,aj1p10r,aj1p11r,aj1p12r,aj1p13r,aj1p14r,aj1p15r,aj1p16r,aj1p17r,aj1p18r,aj1p19r,aj1p20r,aj1p21r,aj1p22r,aj1p23r,aj1p24r,aj1p25r,aj1p26r,aj1p27r,aj1p28r,aj1p29r,aj1p30r,aj1p31r,aj1p32r,aj1p33r,aj1p34r,aj1p35r
	,aj1p00i,aj1p01i,aj1p02i,aj1p03i,aj1p04i,aj1p05i,aj1p06i,aj1p07i,aj1p08i,aj1p09i,aj1p10i,aj1p11i,aj1p12i,aj1p13i,aj1p14i,aj1p15i,aj1p16i,aj1p17i,aj1p18i,aj1p19i,aj1p20i,aj1p21i,aj1p22i,aj1p23i,aj1p24i,aj1p25i,aj1p26i,aj1p27i,aj1p28i,aj1p29i,aj1p30i,aj1p31i,aj1p32i,aj1p33i,aj1p34i,aj1p35i
		,cy00,cy01,cy02,cy03,cy04,cy05,cy06,cy07,cy08,cy09,cy10,cy11,cy12,cy13,cy14,cy15,cy16,cy17,cy18,cy19,cy20,cy21,cy22,cy23,cy24,cy25,cy26,cy27,cy28,cy29,cy30,cy31,cy32,cy33,cy34,cy35,temp,scale;
#if PFETCH
	double *add0, *add;
#endif
	int err;
	FILE *f1,*f2;
	static int first_entry=TRUE;

/*...stuff for the reduced-length DWT weights array is here:	*/
	int n_div_nwt;
	int col,co2,co3,m,m2,n_minus_sil,n_minus_silp1,sinwt,sinwtm1;
	double wt,wtinv,wtl,wtlp1,wtn,wtnm1,wtA,wtB,wtC;

/*...change n36 and n_div_wt to non-static to work around a gcc compiler bug. */
	n36   = n/36;
	n_div_nwt = n36 >> nwt_bits;

	if((n_div_nwt << nwt_bits) != n36)
	{
	  if(INTERACT)printf("FATAL: iter = %10d; NWT_BITS does not divide N/36 in radix36_ditN_cy_dif1.\n",iter);
	  f1 = fopen(   OFILE,"a");
	  f2 = fopen(STATFILE,"a");
	  fprintf(f1,"FATAL: iter = %10d; NWT_BITS does not divide N/36 in radix36_ditN_cy_dif1.\n",iter);
	  fprintf(f2,"FATAL: iter = %10d; NWT_BITS does not divide N/36 in radix36_ditN_cy_dif1.\n",iter);
	  fclose(f1);
	  fclose(f2);
	  err=1;
	  return(err);
	}

	if(p != psave)
	{
	  first_entry=TRUE;
	}

/*...initialize things upon first entry	*/

	if(first_entry)
	{
	  psave = p;
	  first_entry=FALSE;
	  radix_inv = (double)((long double)1.0L/(long double)36.0L);
	  n2inv     = (double)((long double)1.0L/(long double)(n/2));

	  bw    = p%n;		/* Number of bigwords in the Crandall/Fagin mixed-radix representation = (Mersenne exponent) mod (vector length).	*/
	  sw    = n - bw;	/* Number of smallwords.	*/

/*   constant index offsets for array load/stores are here.	*/

	  p01 = n36;
	  p02 = p01 + p01;
	  p03 = p02 + p01;
	  p04 = p03 + p01;
	  p05 = p04 + p01;
	  p06 = p05 + p01;
	  p07 = p06 + p01;
	  p08 = p07 + p01;
	  p09 = p08 + p01;
	  p10 = p09 + p01;
	  p11 = p10 + p01;
	  p12 = p11 + p01;
	  p13 = p12 + p01;
	  p14 = p13 + p01;
	  p15 = p14 + p01;
	  p16 = p15 + p01;
	  p17 = p16 + p01;
	  p18 = p17 + p01;
	  p19 = p18 + p01;
	  p20 = p19 + p01;
	  p21 = p20 + p01;
	  p22 = p21 + p01;
	  p23 = p22 + p01;
	  p24 = p23 + p01;
	  p25 = p24 + p01;
	  p26 = p25 + p01;
	  p27 = p26 + p01;
	  p28 = p27 + p01;
	  p29 = p28 + p01;
	  p30 = p29 + p01;
	  p31 = p30 + p01;
	  p32 = p31 + p01;
	  p33 = p32 + p01;
	  p34 = p33 + p01;
	  p35 = p34 + p01;

	  p01 = p01 + ( (p01 >> DAT_BITS) << PAD_BITS );
	  p02 = p02 + ( (p02 >> DAT_BITS) << PAD_BITS );
	  p03 = p03 + ( (p03 >> DAT_BITS) << PAD_BITS );
	  p04 = p04 + ( (p04 >> DAT_BITS) << PAD_BITS );
	  p05 = p05 + ( (p05 >> DAT_BITS) << PAD_BITS );
	  p06 = p06 + ( (p06 >> DAT_BITS) << PAD_BITS );
	  p07 = p07 + ( (p07 >> DAT_BITS) << PAD_BITS );
	  p08 = p08 + ( (p08 >> DAT_BITS) << PAD_BITS );
	  p09 = p09 + ( (p09 >> DAT_BITS) << PAD_BITS );
	  p10 = p10 + ( (p10 >> DAT_BITS) << PAD_BITS );
	  p11 = p11 + ( (p11 >> DAT_BITS) << PAD_BITS );
	  p12 = p12 + ( (p12 >> DAT_BITS) << PAD_BITS );
	  p13 = p13 + ( (p13 >> DAT_BITS) << PAD_BITS );
	  p14 = p14 + ( (p14 >> DAT_BITS) << PAD_BITS );
	  p15 = p15 + ( (p15 >> DAT_BITS) << PAD_BITS );
	  p16 = p16 + ( (p16 >> DAT_BITS) << PAD_BITS );
	  p17 = p17 + ( (p17 >> DAT_BITS) << PAD_BITS );
	  p18 = p18 + ( (p18 >> DAT_BITS) << PAD_BITS );
	  p19 = p19 + ( (p19 >> DAT_BITS) << PAD_BITS );
	  p20 = p20 + ( (p20 >> DAT_BITS) << PAD_BITS );
	  p21 = p21 + ( (p21 >> DAT_BITS) << PAD_BITS );
	  p22 = p22 + ( (p22 >> DAT_BITS) << PAD_BITS );
	  p23 = p23 + ( (p23 >> DAT_BITS) << PAD_BITS );
	  p24 = p24 + ( (p24 >> DAT_BITS) << PAD_BITS );
	  p25 = p25 + ( (p25 >> DAT_BITS) << PAD_BITS );
	  p26 = p26 + ( (p26 >> DAT_BITS) << PAD_BITS );
	  p27 = p27 + ( (p27 >> DAT_BITS) << PAD_BITS );
	  p28 = p28 + ( (p28 >> DAT_BITS) << PAD_BITS );
	  p29 = p29 + ( (p29 >> DAT_BITS) << PAD_BITS );
	  p30 = p30 + ( (p30 >> DAT_BITS) << PAD_BITS );
	  p31 = p31 + ( (p31 >> DAT_BITS) << PAD_BITS );
	  p32 = p32 + ( (p32 >> DAT_BITS) << PAD_BITS );
	  p33 = p33 + ( (p33 >> DAT_BITS) << PAD_BITS );
	  p34 = p34 + ( (p34 >> DAT_BITS) << PAD_BITS );
	  p35 = p35 + ( (p35 >> DAT_BITS) << PAD_BITS );

	  bjmodnini=0;
	  for(j=0; j < n36; j++)
	  {
	    bjmodnini -= sw; bjmodnini = bjmodnini + ( (-(int)((uint32)bjmodnini >> 31)) & n);
	  }
	}

/*...The radix-36 final DIT pass is here.	*/

	cy00=-2;		/* init carries	*/
	cy01= 0;
	cy02= 0;
	cy03= 0;
	cy04= 0;
	cy05= 0;
	cy06= 0;
	cy07= 0;
	cy08= 0;
	cy09= 0;
	cy10= 0;
	cy11= 0;
	cy12= 0;
	cy13= 0;
	cy14= 0;
	cy15= 0;
	cy16= 0;
	cy17= 0;
	cy18= 0;
	cy19= 0;
	cy20= 0;
	cy21= 0;
	cy22= 0;
	cy23= 0;
	cy24= 0;
	cy25= 0;
	cy26= 0;
	cy27= 0;
	cy28= 0;
	cy29= 0;
	cy30= 0;
	cy31= 0;
	cy32= 0;
	cy33= 0;
	cy34= 0;
	cy35= 0;

	iroot = 0;	/* init sincos array index	*/
	root_incr = 1;	/* init sincos array index increment (set = 1 for normal carry pass, = 0 for wrapper pass)	*/
	scale = n2inv;	/* init inverse-weight scale factor  (set = 2/n for normal carry pass, = 1 for wrapper pass)	*/

	jstart = 0;
	jhi = jstart+nwt-1;
	khi = n_div_nwt;

for(outer=0; outer <= 1; outer++)
{
	i = 1;		/* Pointer to the BASE and BASEINV arrays. lowest-order digit is always a bigword (i = 1).	*/

	bjmodn00= 0;
	bjmodn01= bjmodnini;
	bjmodn02= bjmodn01+bjmodnini-n; bjmodn02= bjmodn02+ ( (-(int)((uint32)bjmodn02>> 31)) & n);
	bjmodn03= bjmodn02+bjmodnini-n; bjmodn03= bjmodn03+ ( (-(int)((uint32)bjmodn03>> 31)) & n);
	bjmodn04= bjmodn03+bjmodnini-n; bjmodn04= bjmodn04+ ( (-(int)((uint32)bjmodn04>> 31)) & n);
	bjmodn05= bjmodn04+bjmodnini-n; bjmodn05= bjmodn05+ ( (-(int)((uint32)bjmodn05>> 31)) & n);
	bjmodn06= bjmodn05+bjmodnini-n; bjmodn06= bjmodn06+ ( (-(int)((uint32)bjmodn06>> 31)) & n);
	bjmodn07= bjmodn06+bjmodnini-n; bjmodn07= bjmodn07+ ( (-(int)((uint32)bjmodn07>> 31)) & n);
	bjmodn08= bjmodn07+bjmodnini-n; bjmodn08= bjmodn08+ ( (-(int)((uint32)bjmodn08>> 31)) & n);
	bjmodn09= bjmodn08+bjmodnini-n; bjmodn09= bjmodn09+ ( (-(int)((uint32)bjmodn09>> 31)) & n);
	bjmodn10= bjmodn09+bjmodnini-n; bjmodn10= bjmodn10+ ( (-(int)((uint32)bjmodn10>> 31)) & n);
	bjmodn11= bjmodn10+bjmodnini-n; bjmodn11= bjmodn11+ ( (-(int)((uint32)bjmodn11>> 31)) & n);
	bjmodn12= bjmodn11+bjmodnini-n; bjmodn12= bjmodn12+ ( (-(int)((uint32)bjmodn12>> 31)) & n);
	bjmodn13= bjmodn12+bjmodnini-n; bjmodn13= bjmodn13+ ( (-(int)((uint32)bjmodn13>> 31)) & n);
	bjmodn14= bjmodn13+bjmodnini-n; bjmodn14= bjmodn14+ ( (-(int)((uint32)bjmodn14>> 31)) & n);
	bjmodn15= bjmodn14+bjmodnini-n; bjmodn15= bjmodn15+ ( (-(int)((uint32)bjmodn15>> 31)) & n);
	bjmodn16= bjmodn15+bjmodnini-n; bjmodn16= bjmodn16+ ( (-(int)((uint32)bjmodn16>> 31)) & n);
	bjmodn17= bjmodn16+bjmodnini-n; bjmodn17= bjmodn17+ ( (-(int)((uint32)bjmodn17>> 31)) & n);
	bjmodn18= bjmodn17+bjmodnini-n; bjmodn18= bjmodn18+ ( (-(int)((uint32)bjmodn18>> 31)) & n);
	bjmodn19= bjmodn18+bjmodnini-n; bjmodn19= bjmodn19+ ( (-(int)((uint32)bjmodn19>> 31)) & n);
	bjmodn20= bjmodn19+bjmodnini-n; bjmodn20= bjmodn20+ ( (-(int)((uint32)bjmodn20>> 31)) & n);
	bjmodn21= bjmodn20+bjmodnini-n; bjmodn21= bjmodn21+ ( (-(int)((uint32)bjmodn21>> 31)) & n);
	bjmodn22= bjmodn21+bjmodnini-n; bjmodn22= bjmodn22+ ( (-(int)((uint32)bjmodn22>> 31)) & n);
	bjmodn23= bjmodn22+bjmodnini-n; bjmodn23= bjmodn23+ ( (-(int)((uint32)bjmodn23>> 31)) & n);
	bjmodn24= bjmodn23+bjmodnini-n; bjmodn24= bjmodn24+ ( (-(int)((uint32)bjmodn24>> 31)) & n);
	bjmodn25= bjmodn24+bjmodnini-n; bjmodn25= bjmodn25+ ( (-(int)((uint32)bjmodn25>> 31)) & n);
	bjmodn26= bjmodn25+bjmodnini-n; bjmodn26= bjmodn26+ ( (-(int)((uint32)bjmodn26>> 31)) & n);
	bjmodn27= bjmodn26+bjmodnini-n; bjmodn27= bjmodn27+ ( (-(int)((uint32)bjmodn27>> 31)) & n);
	bjmodn28= bjmodn27+bjmodnini-n; bjmodn28= bjmodn28+ ( (-(int)((uint32)bjmodn28>> 31)) & n);
	bjmodn29= bjmodn28+bjmodnini-n; bjmodn29= bjmodn29+ ( (-(int)((uint32)bjmodn29>> 31)) & n);
	bjmodn30= bjmodn29+bjmodnini-n; bjmodn30= bjmodn30+ ( (-(int)((uint32)bjmodn30>> 31)) & n);
	bjmodn31= bjmodn30+bjmodnini-n; bjmodn31= bjmodn31+ ( (-(int)((uint32)bjmodn31>> 31)) & n);
	bjmodn32= bjmodn31+bjmodnini-n; bjmodn32= bjmodn32+ ( (-(int)((uint32)bjmodn32>> 31)) & n);
	bjmodn33= bjmodn32+bjmodnini-n; bjmodn33= bjmodn33+ ( (-(int)((uint32)bjmodn33>> 31)) & n);
	bjmodn34= bjmodn33+bjmodnini-n; bjmodn34= bjmodn34+ ( (-(int)((uint32)bjmodn34>> 31)) & n);
	bjmodn35= bjmodn34+bjmodnini-n; bjmodn35= bjmodn35+ ( (-(int)((uint32)bjmodn35>> 31)) & n);

	col=0;
	co2=(n >> nwt_bits)-1+36;
	co3=co2-36;		/* At the start of each new j-loop, co3=co2-radix(1)	*/

	for(k=1; k <= khi; k++)	/* Do n/(radix(1)*nwt) outer loop executions...	*/
	{
	  for(j=jstart; j<jhi; j += 2)	/* Each inner loop execution processes (radix(1)*nwt) array data.	*/
	  {
	    jt = j + ( (j >> DAT_BITS) << PAD_BITS );	/* padded-array fetch index is here	*/
/*
!...gather the needed data (36 64-bit complex, i.e. 72 64-bit reals) and do a radix-36 DIT transform...
*/
/*...Block 1:	*/
	j1=jt;
	t00=a[j1    ];	t01=a[j1+1    ];
	rt =a[j1+p01];	it =a[j1+1+p01];
	t10=t00-rt;  	t00=t00+rt;
	t11=t01-it;	t01=t01+it;

	t20=a[j1+p03];	t21=a[j1+1+p03];
	rt =a[j1+p02];	it =a[j1+1+p02];
	t30=t20-rt;  	t20=t20+rt;
	t31=t21-it;  	t21=t21+it;

	rt =t20;	t20=t00-rt;	t00=t00+rt;
	it =t21;	t21=t01-it;	t01=t01+it;

	rt =t30;	t30=t10-t31;	t10=t10+t31;
			t31=t11+rt;	t11=t11-rt;

/*...Block 2:	*/
	j1=jt+p08;
	t02=a[j1+p01];	t03=a[j1+1+p01];
	rt =a[j1    ];	it =a[j1+1    ];
	t12=t02-rt;  	t02=t02+rt;
	t13=t03-it;	t03=t03+it;

	t22=a[j1+p02];	t23=a[j1+1+p02];
	rt =a[j1+p03];	it =a[j1+1+p03];
	t32=t22-rt;  	t22=t22+rt;
	t33=t23-it;  	t23=t23+it;

	rt =t22;	t22=t02-rt;	t02=t02+rt;
	it =t23;	t23=t03-it;	t03=t03+it;

	rt =t32;	t32=t12-t33;	t12=t12+t33;
			t33=t13+rt;	t13=t13-rt;

/*...Block 3:	*/
	j1=jt+p04;
	t04=a[j1+p02];	t05=a[j1+1+p02];
	rt =a[j1+p03];	it =a[j1+1+p03];
	t14=t04-rt;  	t04=t04+rt;
	t15=t05-it;	t05=t05+it;

	t24=a[j1    ];	t25=a[j1+1    ];
	rt =a[j1+p01];	it =a[j1+1+p01];
	t34=t24-rt;  	t24=t24+rt;
	t35=t25-it;  	t25=t25+it;

	rt =t24;	t24=t04-rt;	t04=t04+rt;
	it =t25;	t25=t05-it;	t05=t05+it;

	rt =t34;	t34=t14-t35;	t14=t14+t35;
			t35=t15+rt;	t15=t15-rt;

/*...Block 4:	*/
	j1=jt+p28;
	t06=a[j1+p03];	t07=a[j1+1+p03];
	rt =a[j1+p02];	it =a[j1+1+p02];
	t16=t06-rt;  	t06=t06+rt;
	t17=t07-it;	t07=t07+it;

	t26=a[j1+p01];	t27=a[j1+1+p01];
	rt =a[j1    ];	it =a[j1+1    ];
	t36=t26-rt;  	t26=t26+rt;
	t37=t27-it;  	t27=t27+it;

	rt =t26;	t26=t06-rt;	t06=t06+rt;
	it =t27;	t27=t07-it;	t07=t07+it;

	rt =t36;	t36=t16-t37;	t16=t16+t37;
			t37=t17+rt;	t17=t17-rt;

/*...Block 5:	*/
	j1=jt+p24;
	t08=a[j1+p01];	t09=a[j1+1+p01];
	rt =a[j1    ];	it =a[j1+1    ];
	t18=t08-rt;  	t08=t08+rt;
	t19=t09-it;	t09=t09+it;

	t28=a[j1+p02];	t29=a[j1+1+p02];
	rt =a[j1+p03];	it =a[j1+1+p03];
	t38=t28-rt;  	t28=t28+rt;
	t39=t29-it;  	t29=t29+it;

	rt =t28;	t28=t08-rt;	t08=t08+rt;
	it =t29;	t29=t09-it;	t09=t09+it;

	rt =t38;	t38=t18-t39;	t18=t18+t39;
			t39=t19+rt;	t19=t19-rt;

/*...Block 6:	*/
	j1=jt+p32;
	t0a=a[j1    ];	t0b=a[j1+1    ];
	rt =a[j1+p01];	it =a[j1+1+p01];
	t1a=t0a-rt;  	t0a=t0a+rt;
	t1b=t0b-it;	t0b=t0b+it;

	t2a=a[j1+p03];	t2b=a[j1+1+p03];
	rt =a[j1+p02];	it =a[j1+1+p02];
	t3a=t2a-rt;  	t2a=t2a+rt;
	t3b=t2b-it;  	t2b=t2b+it;

	rt =t2a;	t2a=t0a-rt;	t0a=t0a+rt;
	it =t2b;	t2b=t0b-it;	t0b=t0b+it;

	rt =t3a;	t3a=t1a-t3b;	t1a=t1a+t3b;
			t3b=t1b+rt;	t1b=t1b-rt;

/*...Block 7:	*/
	j1=jt+p12;
	t0c=a[j1+p03];	t0d=a[j1+1+p03];
	rt =a[j1+p02];	it =a[j1+1+p02];
	t1c=t0c-rt;  	t0c=t0c+rt;
	t1d=t0d-it;	t0d=t0d+it;

	t2c=a[j1+p01];	t2d=a[j1+1+p01];
	rt =a[j1    ];	it =a[j1+1    ];
	t3c=t2c-rt;  	t2c=t2c+rt;
	t3d=t2d-it;  	t2d=t2d+it;

	rt =t2c;	t2c=t0c-rt;	t0c=t0c+rt;
	it =t2d;	t2d=t0d-it;	t0d=t0d+it;

	rt =t3c;	t3c=t1c-t3d;	t1c=t1c+t3d;
			t3d=t1d+rt;	t1d=t1d-rt;

/*...Block 8:	*/
	j1=jt+p20;
	t0e=a[j1+p02];	t0f=a[j1+1+p02];
	rt =a[j1+p03];	it =a[j1+1+p03];
	t1e=t0e-rt;  	t0e=t0e+rt;
	t1f=t0f-it;	t0f=t0f+it;

	t2e=a[j1    ];	t2f=a[j1+1    ];
	rt =a[j1+p01];	it =a[j1+1+p01];
	t3e=t2e-rt;  	t2e=t2e+rt;
	t3f=t2f-it;  	t2f=t2f+it;

	rt =t2e;	t2e=t0e-rt;	t0e=t0e+rt;
	it =t2f;	t2f=t0f-it;	t0f=t0f+it;

	rt =t3e;	t3e=t1e-t3f;	t1e=t1e+t3f;
			t3f=t1f+rt;	t1f=t1f-rt;

/*...Block 9:	*/
	j1=jt+p16;
	t0g=a[j1    ];	t0h=a[j1+1    ];
	rt =a[j1+p01];	it =a[j1+1+p01];
	t1g=t0g-rt;  	t0g=t0g+rt;
	t1h=t0h-it;	t0h=t0h+it;

	t2g=a[j1+p03];	t2h=a[j1+1+p03];
	rt =a[j1+p02];	it =a[j1+1+p02];
	t3g=t2g-rt;  	t2g=t2g+rt;
	t3h=t2h-it;  	t2h=t2h+it;

	rt =t2g;	t2g=t0g-rt;	t0g=t0g+rt;
	it =t2h;	t2h=t0h-it;	t0h=t0h+it;

	rt =t3g;	t3g=t1g-t3h;	t1g=t1g+t3h;
			t3h=t1h+rt;	t1h=t1h-rt;

/*       ...and now do four radix-9 transforms.	*/

/*...First radix-9 transform:	*/
	  rt  =t02;			it  =t03;
	  t02 =rt+t04;			t03 =it+t05;
	  t04 =rt-t04;			t05 =it-t05;
	  t00 =t00+t02;			t01 =t01+t03;
	  t02 =t00+c3m1*t02;		t03 =t01+c3m1*t03;
	  rt  =s3*t04;			it  =s3*t05;
	  t04 =t02-it;			t05 =t03+rt;
	  t02 =t02+it;			t03 =t03-rt;

	  rt  =t08;			it  =t09;
	  t08 =rt+t0a;			t09 =it+t0b;
	  t0a =rt-t0a;			t0b =it-t0b;
	  t06 =t06+t08;			t07 =t07+t09;
	  t08 =t06+c3m1*t08;		t09 =t07+c3m1*t09;
	  rt  =s3*t0a;			it  =s3*t0b;
	  t0a =t08-it;			t0b =t09+rt;
	  t08 =t08+it;			t09 =t09-rt;

	  rt  =t0e;			it  =t0f;
	  t0e =rt+t0g;			t0f =it+t0h;
	  t0g =rt-t0g;			t0h =it-t0h;
	  t0c =t0c+t0e;			t0d =t0d+t0f;
	  t0e =t0c+c3m1*t0e;		t0f =t0d+c3m1*t0f;
	  rt  =s3*t0g;			it  =s3*t0h;
	  t0g =t0e-it;			t0h =t0f+rt;
	  t0e =t0e+it;			t0f =t0f-rt;
	/* Twiddles: */
	  rt  =t06;			it  =t07;
	  t06 =rt+t0c;			t07 =it+t0d;
	  t0c =rt-t0c;			t0d =it-t0d;
	  t00 =t00+t06;			t01 =t01+t07;
	  aj1p00r=t00;			aj1p00i=t01;
	  t06 =t00+c3m1*t06;		t07 =t01+c3m1*t07;
	  rt  =s3*t0c;			it  =s3*t0d;
	  aj1p24r=t06+it;		aj1p24i=t07-rt;
	  aj1p12r=t06-it;		aj1p12i=t07+rt;

	  rt  =t08*c +t09*s;		it  =t09*c -t08*s;
	  re  =t0e*c2+t0f*s2;		t0f =t0f*c2-t0e*s2;	t0e=re;
	  t08 =rt+t0e;			t09 =it+t0f;
	  t0e =rt-t0e;			t0f =it-t0f;
	  t02 =t02+t08;			t03 =t03+t09;
	  aj1p08r=t02;			aj1p08i=t03;
	  t08 =t02+c3m1*t08;		t09=t03+c3m1*t09;
	  rt  =s3*t0e;			it  =s3*t0f;
	  aj1p32r=t08+it;		aj1p32i=t09-rt;
	  aj1p20r=t08-it;		aj1p20i=t09+rt;

	  rt  =t0a*c2+t0b*s2;		it  =t0b*c2-t0a*s2;
	  re  =t0g*c4+t0h*s4;		t0h =t0h*c4-t0g*s4;	t0g=re;
	  t0a =rt+t0g;			t0b =it+t0h;
	  t0g =rt-t0g;			t0h =it-t0h;
	  t04 =t04+t0a;			t05 =t05+t0b;
	  aj1p16r=t04;			aj1p16i=t05;
	  t0a =t04+c3m1*t0a;		t0b =t05+c3m1*t0b;
	  rt  =s3*t0g;			it  =s3*t0h;
	  aj1p04r=t0a+it;		aj1p04i=t0b-rt;
	  aj1p28r=t0a-it;		aj1p28i=t0b+rt;

/*...Second radix-9 transform:	*/
	  rt  =t12;			it  =t13;
	  t12 =rt+t14;			t13 =it+t15;
	  t14 =rt-t14;			t15 =it-t15;
	  t10 =t10+t12;			t11 =t11+t13;
	  t12 =t10+c3m1*t12;		t13 =t11+c3m1*t13;
	  rt  =s3*t14;			it  =s3*t15;
	  t14 =t12-it;			t15 =t13+rt;
	  t12 =t12+it;			t13 =t13-rt;

	  rt  =t18;			it  =t19;
	  t18 =rt+t1a;			t19 =it+t1b;
	  t1a =rt-t1a;			t1b =it-t1b;
	  t16 =t16+t18;			t17 =t17+t19;
	  t18 =t16+c3m1*t18;		t19 =t17+c3m1*t19;
	  rt  =s3*t1a;			it  =s3*t1b;
	  t1a =t18-it;			t1b =t19+rt;
	  t18 =t18+it;			t19 =t19-rt;

	  rt  =t1e;			it  =t1f;
	  t1e =rt+t1g;			t1f =it+t1h;
	  t1g =rt-t1g;			t1h =it-t1h;
	  t1c =t1c+t1e;			t1d =t1d+t1f;
	  t1e =t1c+c3m1*t1e;		t1f =t1d+c3m1*t1f;
	  rt  =s3*t1g;			it  =s3*t1h;
	  t1g =t1e-it;			t1h =t1f+rt;
	  t1e =t1e+it;			t1f =t1f-rt;
	/* Twiddles: */
	  rt  =t16;			it  =t17;
	  t16 =rt+t1c;			t17 =it+t1d;
	  t1c =rt-t1c;			t1d =it-t1d;
	  t10 =t10+t16;			t11 =t11+t17;
	  aj1p27r=t10;			aj1p27i=t11;
	  t16 =t10+c3m1*t16;		t17 =t11+c3m1*t17;
	  rt  =s3*t1c;			it  =s3*t1d;
	  aj1p15r=t16+it;		aj1p15i=t17-rt;
	  aj1p03r=t16-it;		aj1p03i=t17+rt;

	  rt  =t18*c +t19*s;		it  =t19*c -t18*s;
	  re  =t1e*c2+t1f*s2;		t1f =t1f*c2-t1e*s2;	t1e=re;
	  t18 =rt+t1e;			t19 =it+t1f;
	  t1e =rt-t1e;			t1f =it-t1f;
	  t12 =t12+t18;			t13 =t13+t19;
	  aj1p35r=t12;			aj1p35i=t13;
	  t18 =t12+c3m1*t18;		t19=t13+c3m1*t19;
	  rt  =s3*t1e;			it  =s3*t1f;
	  aj1p23r=t18+it;		aj1p23i=t19-rt;
	  aj1p11r=t18-it;		aj1p11i=t19+rt;

	  rt  =t1a*c2+t1b*s2;		it  =t1b*c2-t1a*s2;
	  re  =t1g*c4+t1h*s4;		t1h =t1h*c4-t1g*s4;	t1g=re;
	  t1a =rt+t1g;			t1b =it+t1h;
	  t1g =rt-t1g;			t1h =it-t1h;
	  t14 =t14+t1a;			t15 =t15+t1b;
	  aj1p07r=t14;			aj1p07i=t15;
	  t1a =t14+c3m1*t1a;		t1b =t15+c3m1*t1b;
	  rt  =s3*t1g;			it  =s3*t1h;
	  aj1p31r=t1a+it;		aj1p31i=t1b-rt;
	  aj1p19r=t1a-it;		aj1p19i=t1b+rt;

/*...Third radix-9 transform:	*/
	  rt  =t22;			it  =t23;
	  t22 =rt+t24;			t23 =it+t25;
	  t24 =rt-t24;			t25 =it-t25;
	  t20 =t20+t22;			t21 =t21+t23;
	  t22 =t20+c3m1*t22;		t23 =t21+c3m1*t23;
	  rt  =s3*t24;			it  =s3*t25;
	  t24 =t22-it;			t25 =t23+rt;
	  t22 =t22+it;			t23 =t23-rt;

	  rt  =t28;			it  =t29;
	  t28 =rt+t2a;			t29 =it+t2b;
	  t2a =rt-t2a;			t2b =it-t2b;
	  t26 =t26+t28;			t27 =t27+t29;
	  t28 =t26+c3m1*t28;		t29 =t27+c3m1*t29;
	  rt  =s3*t2a;			it  =s3*t2b;
	  t2a =t28-it;			t2b =t29+rt;
	  t28 =t28+it;			t29 =t29-rt;

	  rt  =t2e;			it  =t2f;
	  t2e =rt+t2g;			t2f =it+t2h;
	  t2g =rt-t2g;			t2h =it-t2h;
	  t2c =t2c+t2e;			t2d =t2d+t2f;
	  t2e =t2c+c3m1*t2e;		t2f =t2d+c3m1*t2f;
	  rt  =s3*t2g;			it  =s3*t2h;
	  t2g =t2e-it;			t2h =t2f+rt;
	  t2e =t2e+it;			t2f =t2f-rt;
	/* Twiddles: */
	  rt  =t26;			it  =t27;
	  t26 =rt+t2c;			t27 =it+t2d;
	  t2c =rt-t2c;			t2d =it-t2d;
	  t20 =t20+t26;			t21 =t21+t27;
	  aj1p18r=t20;			aj1p18i=t21;
	  t26 =t20+c3m1*t26;		t27 =t21+c3m1*t27;
	  rt  =s3*t2c;			it  =s3*t2d;
	  aj1p06r=t26+it;		aj1p06i=t27-rt;
	  aj1p30r=t26-it;		aj1p30i=t27+rt;

	  rt  =t28*c +t29*s;		it  =t29*c -t28*s;
	  re  =t2e*c2+t2f*s2;		t2f =t2f*c2-t2e*s2;	t2e=re;
	  t28 =rt+t2e;			t29 =it+t2f;
	  t2e =rt-t2e;			t2f =it-t2f;
	  t22 =t22+t28;			t23 =t23+t29;
	  aj1p26r=t22;			aj1p26i=t23;
	  t28 =t22+c3m1*t28;		t29=t23+c3m1*t29;
	  rt  =s3*t2e;			it  =s3*t2f;
	  aj1p14r=t28+it;		aj1p14i=t29-rt;
	  aj1p02r=t28-it;		aj1p02i=t29+rt;

	  rt  =t2a*c2+t2b*s2;		it  =t2b*c2-t2a*s2;
	  re  =t2g*c4+t2h*s4;		t2h =t2h*c4-t2g*s4;	t2g=re;
	  t2a =rt+t2g;			t2b =it+t2h;
	  t2g =rt-t2g;			t2h =it-t2h;
	  t24 =t24+t2a;			t25 =t25+t2b;
	  aj1p34r=t24;			aj1p34i=t25;
	  t2a =t24+c3m1*t2a;		t2b =t25+c3m1*t2b;
	  rt  =s3*t2g;			it  =s3*t2h;
	  aj1p22r=t2a+it;		aj1p22i=t2b-rt;
	  aj1p10r=t2a-it;		aj1p10i=t2b+rt;

/*...Fourth radix-9 transform:	*/
	  rt  =t32;			it  =t33;
	  t32 =rt+t34;			t33 =it+t35;
	  t34 =rt-t34;			t35 =it-t35;
	  t30 =t30+t32;			t31 =t31+t33;
	  t32 =t30+c3m1*t32;		t33 =t31+c3m1*t33;
	  rt  =s3*t34;			it  =s3*t35;
	  t34 =t32-it;			t35 =t33+rt;
	  t32 =t32+it;			t33 =t33-rt;

	  rt  =t38;			it  =t39;
	  t38 =rt+t3a;			t39 =it+t3b;
	  t3a =rt-t3a;			t3b =it-t3b;
	  t36 =t36+t38;			t37 =t37+t39;
	  t38 =t36+c3m1*t38;		t39 =t37+c3m1*t39;
	  rt  =s3*t3a;			it  =s3*t3b;
	  t3a =t38-it;			t3b =t39+rt;
	  t38 =t38+it;			t39 =t39-rt;

	  rt  =t3e;			it  =t3f;
	  t3e =rt+t3g;			t3f =it+t3h;
	  t3g =rt-t3g;			t3h =it-t3h;
	  t3c =t3c+t3e;			t3d =t3d+t3f;
	  t3e =t3c+c3m1*t3e;		t3f =t3d+c3m1*t3f;
	  rt  =s3*t3g;			it  =s3*t3h;
	  t3g =t3e-it;			t3h =t3f+rt;
	  t3e =t3e+it;			t3f =t3f-rt;
	/* Twiddles: */
	  rt  =t36;			it  =t37;
	  t36 =rt+t3c;			t37 =it+t3d;
	  t3c =rt-t3c;			t3d =it-t3d;
	  t30 =t30+t36;			t31 =t31+t37;
	  aj1p09r=t30;			aj1p09i=t31;
	  t36 =t30+c3m1*t36;		t37 =t31+c3m1*t37;
	  rt  =s3*t3c;			it  =s3*t3d;
	  aj1p33r=t36+it;		aj1p33i=t37-rt;
	  aj1p21r=t36-it;		aj1p21i=t37+rt;

	  rt  =t38*c +t39*s;		it  =t39*c -t38*s;
	  re  =t3e*c2+t3f*s2;		t3f =t3f*c2-t3e*s2;	t3e=re;
	  t38 =rt+t3e;			t39 =it+t3f;
	  t3e =rt-t3e;			t3f =it-t3f;
	  t32 =t32+t38;			t33 =t33+t39;
	  aj1p17r=t32;			aj1p17i=t33;
	  t38 =t32+c3m1*t38;		t39=t33+c3m1*t39;
	  rt  =s3*t3e;			it  =s3*t3f;
	  aj1p05r=t38+it;		aj1p05i=t39-rt;
	  aj1p29r=t38-it;		aj1p29i=t39+rt;

	  rt  =t3a*c2+t3b*s2;		it  =t3b*c2-t3a*s2;
	  re  =t3g*c4+t3h*s4;		t3h =t3h*c4-t3g*s4;	t3g=re;
	  t3a =rt+t3g;			t3b =it+t3h;
	  t3g =rt-t3g;			t3h =it-t3h;
	  t34 =t34+t3a;			t35 =t35+t3b;
	  aj1p25r=t34;			aj1p25i=t35;
	  t3a =t34+c3m1*t3a;		t3b =t35+c3m1*t3b;
	  rt  =s3*t3g;			it  =s3*t3h;
	  aj1p13r=t3a+it;		aj1p13i=t3b-rt;
	  aj1p01r=t3a-it;		aj1p01i=t3b+rt;

/*...Now do the carries. Since the outputs would
    normally be getting dispatched to 36 separate blocks of the A-array, we need 36 separate carries.	*/

	    l= j & (nwt-1);
	    n_minus_sil   = n-si[l  ];
	    n_minus_silp1 = n-si[l+1];
	    sinwt   = si[nwt-l  ];
	    sinwtm1 = si[nwt-l-1];

	    wtl     =wt0[    l  ];
	    wtn     =wt0[nwt-l  ]*scale;	/* Include 1/(n/2) scale factor of inverse transform here...	*/
	    wtlp1   =wt0[    l+1];
	    wtnm1   =wt0[nwt-l-1]*scale;	/* ...and here.	*/

/*...set0 is slightly different from others:	*/
	   cmplx_carry_norm_nocheck0(aj1p00r,aj1p00i,cy00,bjmodn00   );
	    cmplx_carry_norm_nocheck(aj1p01r,aj1p01i,cy01,bjmodn01,1 );
	    cmplx_carry_norm_nocheck(aj1p02r,aj1p02i,cy02,bjmodn02,2 );
	    cmplx_carry_norm_nocheck(aj1p03r,aj1p03i,cy03,bjmodn03,3 );
	    cmplx_carry_norm_nocheck(aj1p04r,aj1p04i,cy04,bjmodn04,4 );
	    cmplx_carry_norm_nocheck(aj1p05r,aj1p05i,cy05,bjmodn05,5 );
	    cmplx_carry_norm_nocheck(aj1p06r,aj1p06i,cy06,bjmodn06,6 );
	    cmplx_carry_norm_nocheck(aj1p07r,aj1p07i,cy07,bjmodn07,7 );
	    cmplx_carry_norm_nocheck(aj1p08r,aj1p08i,cy08,bjmodn08,8 );
	    cmplx_carry_norm_nocheck(aj1p09r,aj1p09i,cy09,bjmodn09,9 );
	    cmplx_carry_norm_nocheck(aj1p10r,aj1p10i,cy10,bjmodn10,10);
	    cmplx_carry_norm_nocheck(aj1p11r,aj1p11i,cy11,bjmodn11,11);
	    cmplx_carry_norm_nocheck(aj1p12r,aj1p12i,cy12,bjmodn12,12);
	    cmplx_carry_norm_nocheck(aj1p13r,aj1p13i,cy13,bjmodn13,13);
	    cmplx_carry_norm_nocheck(aj1p14r,aj1p14i,cy14,bjmodn14,14);
	    cmplx_carry_norm_nocheck(aj1p15r,aj1p15i,cy15,bjmodn15,15);
	    cmplx_carry_norm_nocheck(aj1p16r,aj1p16i,cy16,bjmodn16,16);
	    cmplx_carry_norm_nocheck(aj1p17r,aj1p17i,cy17,bjmodn17,17);
	    cmplx_carry_norm_nocheck(aj1p18r,aj1p18i,cy18,bjmodn18,18);
	    cmplx_carry_norm_nocheck(aj1p19r,aj1p19i,cy19,bjmodn19,19);
	    cmplx_carry_norm_nocheck(aj1p20r,aj1p20i,cy20,bjmodn20,20);
	    cmplx_carry_norm_nocheck(aj1p21r,aj1p21i,cy21,bjmodn21,21);
	    cmplx_carry_norm_nocheck(aj1p22r,aj1p22i,cy22,bjmodn22,22);
	    cmplx_carry_norm_nocheck(aj1p23r,aj1p23i,cy23,bjmodn23,23);
	    cmplx_carry_norm_nocheck(aj1p24r,aj1p24i,cy24,bjmodn24,24);
	    cmplx_carry_norm_nocheck(aj1p25r,aj1p25i,cy25,bjmodn25,25);
	    cmplx_carry_norm_nocheck(aj1p26r,aj1p26i,cy26,bjmodn26,26);
	    cmplx_carry_norm_nocheck(aj1p27r,aj1p27i,cy27,bjmodn27,27);
	    cmplx_carry_norm_nocheck(aj1p28r,aj1p28i,cy28,bjmodn28,28);
	    cmplx_carry_norm_nocheck(aj1p29r,aj1p29i,cy29,bjmodn29,29);
	    cmplx_carry_norm_nocheck(aj1p30r,aj1p30i,cy30,bjmodn30,30);
	    cmplx_carry_norm_nocheck(aj1p31r,aj1p31i,cy31,bjmodn31,31);
	    cmplx_carry_norm_nocheck(aj1p32r,aj1p32i,cy32,bjmodn32,32);
	    cmplx_carry_norm_nocheck(aj1p33r,aj1p33i,cy33,bjmodn33,33);
	    cmplx_carry_norm_nocheck(aj1p34r,aj1p34i,cy34,bjmodn34,34);
	    cmplx_carry_norm_nocheck(aj1p35r,aj1p35i,cy35,bjmodn35,35);

	    i =((uint32)(sw - bjmodn00) >> 31);	/* get ready for the next set...	*/
	    co2=co3;	/* For all data but the first set in each j-block, co2=co3. Thus, after the first block of data is done
			   and only then: for all subsequent blocks it's superfluous), this assignment decrements co2 by radix(1).	*/

/*...The radix-36 DIF pass is here:	*/
#if PFETCH
add0 = &a[jt];
prefetch_p_doubles(add0);
#endif
/*...First radix-9 transform:	*/
	/* permuted version of jt + p00,p12,p24 */
	  t00 =aj1p00r;			t01 =aj1p00i;
	  t02 =aj1p24r+aj1p12r;		t03 =aj1p24i+aj1p12i;
	  t04 =aj1p24r-aj1p12r;		t05 =aj1p24i-aj1p12i;
	  t00 =t00+t02;			t01 =t01+t03;
	  t02 =t00+c3m1*t02;		t03 =t01+c3m1*t03;
	  rt  =s3*t04;			it  =s3*t05;
	  t04 =t02+it;			t05 =t03-rt;
	  t02 =t02-it;			t03 =t03+rt;
#if PFETCH
add = add0+p01;
prefetch_p_doubles(add);
#endif
	/* permuted version of jt + p04,p16,p28 */
	  t06 =aj1p32r;			t07 =aj1p32i;
	  t08 =aj1p20r+aj1p08r;		t09 =aj1p20i+aj1p08i;
	  t0a =aj1p20r-aj1p08r;		t0b =aj1p20i-aj1p08i;
	  t06 =t06+t08;			t07 =t07+t09;
	  t08 =t06+c3m1*t08;		t09 =t07+c3m1*t09;
	  rt  =s3*t0a;			it  =s3*t0b;
	  t0a =t08+it;			t0b =t09-rt;
	  t08 =t08-it;			t09 =t09+rt;
#if PFETCH
add = add0+p02;
prefetch_p_doubles(add);
#endif
	/* permuted version of jt + p08,p20,p32 */
	  t0c =aj1p28r;			t0d =aj1p28i;
	  t0e =aj1p16r+aj1p04r;		t0f =aj1p16i+aj1p04i;
	  t0g =aj1p16r-aj1p04r;		t0h =aj1p16i-aj1p04i;
	  t0c =t0c+t0e;			t0d =t0d+t0f;
	  t0e =t0c+c3m1*t0e;		t0f =t0d+c3m1*t0f;
	  rt  =s3*t0g;			it  =s3*t0h;
	  t0g =t0e+it;			t0h =t0f-rt;
	  t0e =t0e-it;			t0f =t0f+rt;
#if PFETCH
add = add0+p03;
prefetch_p_doubles(add);
#endif
	/* Twiddles: */
	  rt  =t06;			it  =t07;
	  t06 =rt+t0c;			t07 =it+t0d;
	  t0c =rt-t0c;			t0d =it-t0d;
	  t00 =t00+t06;			t01 =t01+t07;
	  t06 =t00+c3m1*t06;		t07 =t01+c3m1*t07;
	  rt  =s3*t0c;			it  =s3*t0d;
	  t0c =t06+it;			t0d =t07-rt;
	  t06 =t06-it;			t07 =t07+rt;
#if PFETCH
add = add0+p04;
prefetch_p_doubles(add);
#endif
	  rt  =t08*c -t09*s;		it  =t08*s +t09*c;
	  re  =t0e*c2-t0f*s2;		t0f =t0e*s2+t0f*c2;	t0e=re;
	  t08 =rt+t0e;			t09 =it+t0f;
	  t0e =rt-t0e;			t0f =it-t0f;
	  t02 =t02+t08;			t03 =t03+t09;
	  t08 =t02+c3m1*t08;		t09 =t03+c3m1*t09;
	  rt  =s3*t0e;			it  =s3*t0f;
	  t0e =t08+it;			t0f =t09-rt;
	  t08 =t08-it;			t09 =t09+rt;
#if PFETCH
add = add0+p05;
prefetch_p_doubles(add);
#endif
	  rt  =t0a*c2-t0b*s2;		it  =t0a*s2+t0b*c2;
	  re  =t0g*c4-t0h*s4;		t0h =t0g*s4+t0h*c4;	t0g=re;
	  t0a =rt+t0g;			t0b =it+t0h;
	  t0g =rt-t0g;			t0h =it-t0h;
	  t04 =t04+t0a;			t05 =t05+t0b;
	  t0a =t04+c3m1*t0a;		t0b =t05+c3m1*t0b;
	  rt  =s3*t0g;			it  =s3*t0h;
	  t0g =t0a+it;			t0h =t0b-rt;
	  t0a =t0a-it;			t0b =t0b+rt;
#if PFETCH
add = add0+p06;
prefetch_p_doubles(add);
#endif

/*...Second radix-9 transform:	*/
	/* permuted version of jt + p01,p13,p25 */
	  t10 =aj1p27r;			t11 =aj1p27i;
	  t12 =aj1p15r+aj1p03r;		t13 =aj1p15i+aj1p03i;
	  t14 =aj1p15r-aj1p03r;		t15 =aj1p15i-aj1p03i;
	  t10 =t10+t12;			t11 =t11+t13;
	  t12 =t10+c3m1*t12;		t13 =t11+c3m1*t13;
	  rt  =s3*t14;			it  =s3*t15;
	  t14 =t12+it;			t15 =t13-rt;
	  t12 =t12-it;			t13 =t13+rt;
#if PFETCH
add = add0+p07;
prefetch_p_doubles(add);
#endif
	/* permuted version of jt + p05,p17,p29 */
	  t16 =aj1p23r;			t17 =aj1p23i;
	  t18 =aj1p11r+aj1p35r;		t19 =aj1p11i+aj1p35i;
	  t1a =aj1p11r-aj1p35r;		t1b =aj1p11i-aj1p35i;
	  t16 =t16+t18;			t17 =t17+t19;
	  t18 =t16+c3m1*t18;		t19 =t17+c3m1*t19;
	  rt  =s3*t1a;			it  =s3*t1b;
	  t1a =t18+it;			t1b =t19-rt;
	  t18 =t18-it;			t19 =t19+rt;
#if PFETCH
add = add0+p08;
prefetch_p_doubles(add);
#endif
	/* permuted version of jt + p09,p21,p33 */
	  t1c =aj1p19r;			t1d =aj1p19i;
	  t1e =aj1p07r+aj1p31r;		t1f =aj1p07i+aj1p31i;
	  t1g =aj1p07r-aj1p31r;		t1h =aj1p07i-aj1p31i;
	  t1c =t1c+t1e;			t1d =t1d+t1f;
	  t1e =t1c+c3m1*t1e;		t1f =t1d+c3m1*t1f;
	  rt  =s3*t1g;			it  =s3*t1h;
	  t1g =t1e+it;			t1h =t1f-rt;
	  t1e =t1e-it;			t1f =t1f+rt;
#if PFETCH
add = add0+p09;
prefetch_p_doubles(add);
#endif
	/* Twiddles: */
	  rt  =t16;			it  =t17;
	  t16 =rt+t1c;			t17 =it+t1d;
	  t1c =rt-t1c;			t1d =it-t1d;
	  t10 =t10+t16;			t11 =t11+t17;
	  t16 =t10+c3m1*t16;		t17 =t11+c3m1*t17;
	  rt  =s3*t1c;			it  =s3*t1d;
	  t1c =t16+it;			t1d =t17-rt;
	  t16 =t16-it;			t17 =t17+rt;
#if PFETCH
add = add0+p10;
prefetch_p_doubles(add);
#endif
	  rt  =t18*c -t19*s;		it  =t18*s +t19*c;
	  re  =t1e*c2-t1f*s2;		t1f =t1e*s2+t1f*c2;	t1e=re;
	  t18 =rt+t1e;			t19 =it+t1f;
	  t1e =rt-t1e;			t1f =it-t1f;
	  t12 =t12+t18;			t13 =t13+t19;
	  t18 =t12+c3m1*t18;		t19 =t13+c3m1*t19;
	  rt  =s3*t1e;			it  =s3*t1f;
	  t1e =t18+it;			t1f =t19-rt;
	  t18 =t18-it;			t19 =t19+rt;
#if PFETCH
add = add0+p11;
prefetch_p_doubles(add);
#endif
	  rt  =t1a*c2-t1b*s2;		it  =t1a*s2+t1b*c2;
	  re  =t1g*c4-t1h*s4;		t1h =t1g*s4+t1h*c4;	t1g=re;
	  t1a =rt+t1g;			t1b =it+t1h;
	  t1g =rt-t1g;			t1h =it-t1h;
	  t14 =t14+t1a;			t15 =t15+t1b;
	  t1a =t14+c3m1*t1a;		t1b =t15+c3m1*t1b;
	  rt  =s3*t1g;			it  =s3*t1h;
	  t1g =t1a+it;			t1h =t1b-rt;
	  t1a =t1a-it;			t1b =t1b+rt;
#if PFETCH
add = add0+p12;
prefetch_p_doubles(add);
#endif

/*...Third radix-9 transform:	*/
	/* permuted version of jt + p02,p14,p26 */
	  t20 =aj1p18r;			t21 =aj1p18i;
	  t22 =aj1p06r+aj1p30r;		t23 =aj1p06i+aj1p30i;
	  t24 =aj1p06r-aj1p30r;		t25 =aj1p06i-aj1p30i;
	  t20 =t20+t22;			t21 =t21+t23;
	  t22 =t20+c3m1*t22;		t23 =t21+c3m1*t23;
	  rt  =s3*t24;			it  =s3*t25;
	  t24 =t22+it;			t25 =t23-rt;
	  t22 =t22-it;			t23 =t23+rt;
#if PFETCH
add = add0+p13;
prefetch_p_doubles(add);
#endif
	/* permuted version of jt + p06,p18,p30 */
	  t26 =aj1p14r;			t27 =aj1p14i;
	  t28 =aj1p02r+aj1p26r;		t29 =aj1p02i+aj1p26i;
	  t2a =aj1p02r-aj1p26r;		t2b =aj1p02i-aj1p26i;
	  t26 =t26+t28;			t27 =t27+t29;
	  t28 =t26+c3m1*t28;		t29 =t27+c3m1*t29;
	  rt  =s3*t2a;			it  =s3*t2b;
	  t2a =t28+it;			t2b =t29-rt;
	  t28 =t28-it;			t29 =t29+rt;
#if PFETCH
add = add0+p14;
prefetch_p_doubles(add);
#endif
	/* permuted version of jt + p10,p22,p34 */
	  t2c =aj1p10r;			t2d =aj1p10i;
	  t2e =aj1p34r+aj1p22r;		t2f =aj1p34i+aj1p22i;
	  t2g =aj1p34r-aj1p22r;		t2h =aj1p34i-aj1p22i;
	  t2c =t2c+t2e;			t2d =t2d+t2f;
	  t2e =t2c+c3m1*t2e;		t2f =t2d+c3m1*t2f;
	  rt  =s3*t2g;			it  =s3*t2h;
	  t2g =t2e+it;			t2h =t2f-rt;
	  t2e =t2e-it;			t2f =t2f+rt;
#if PFETCH
add = add0+p15;
prefetch_p_doubles(add);
#endif
	/* Twiddles: */
	  rt  =t26;			it  =t27;
	  t26 =rt+t2c;			t27 =it+t2d;
	  t2c =rt-t2c;			t2d =it-t2d;
	  t20 =t20+t26;			t21 =t21+t27;
	  t26 =t20+c3m1*t26;		t27 =t21+c3m1*t27;
	  rt  =s3*t2c;			it  =s3*t2d;
	  t2c =t26+it;			t2d =t27-rt;
	  t26 =t26-it;			t27 =t27+rt;
#if PFETCH
add = add0+p16;
prefetch_p_doubles(add);
#endif
	  rt  =t28*c -t29*s;		it  =t28*s +t29*c;
	  re  =t2e*c2-t2f*s2;		t2f =t2e*s2+t2f*c2;	t2e=re;
	  t28 =rt+t2e;			t29 =it+t2f;
	  t2e =rt-t2e;			t2f =it-t2f;
	  t22 =t22+t28;			t23 =t23+t29;
	  t28 =t22+c3m1*t28;		t29 =t23+c3m1*t29;
	  rt  =s3*t2e;			it  =s3*t2f;
	  t2e =t28+it;			t2f =t29-rt;
	  t28 =t28-it;			t29 =t29+rt;
#if PFETCH
add = add0+p17;
prefetch_p_doubles(add);
#endif
	  rt  =t2a*c2-t2b*s2;		it  =t2a*s2+t2b*c2;
	  re  =t2g*c4-t2h*s4;		t2h =t2g*s4+t2h*c4;	t2g=re;
	  t2a =rt+t2g;			t2b =it+t2h;
	  t2g =rt-t2g;			t2h =it-t2h;
	  t24 =t24+t2a;			t25 =t25+t2b;
	  t2a =t24+c3m1*t2a;		t2b =t25+c3m1*t2b;
	  rt  =s3*t2g;			it  =s3*t2h;
	  t2g =t2a+it;			t2h =t2b-rt;
	  t2a =t2a-it;			t2b =t2b+rt;
#if PFETCH
add = add0+p18;
prefetch_p_doubles(add);
#endif

/*...Fourth radix-9 transform:	*/
	/* permuted version of jt + p03,p15,p27 */
	  t30 =aj1p09r;			t31 =aj1p09i;
	  t32 =aj1p33r+aj1p21r;		t33 =aj1p33i+aj1p21i;
	  t34 =aj1p33r-aj1p21r;		t35 =aj1p33i-aj1p21i;
	  t30 =t30+t32;			t31 =t31+t33;
	  t32 =t30+c3m1*t32;		t33 =t31+c3m1*t33;
	  rt  =s3*t34;			it  =s3*t35;
	  t34 =t32+it;			t35 =t33-rt;
	  t32 =t32-it;			t33 =t33+rt;
#if PFETCH
add = add0+p19;
prefetch_p_doubles(add);
#endif
	/* permuted version of jt + p07,p19,p31 */
	  t36 =aj1p05r;			t37 =aj1p05i;
	  t38 =aj1p29r+aj1p17r;		t39 =aj1p29i+aj1p17i;
	  t3a =aj1p29r-aj1p17r;		t3b =aj1p29i-aj1p17i;
	  t36 =t36+t38;			t37 =t37+t39;
	  t38 =t36+c3m1*t38;		t39 =t37+c3m1*t39;
	  rt  =s3*t3a;			it  =s3*t3b;
	  t3a =t38+it;			t3b =t39-rt;
	  t38 =t38-it;			t39 =t39+rt;
#if PFETCH
add = add0+p20;
prefetch_p_doubles(add);
#endif
	/* permuted version of jt + p11,p23,p35 */
	  t3c =aj1p01r;			t3d =aj1p01i;
	  t3e =aj1p25r+aj1p13r;		t3f =aj1p25i+aj1p13i;
	  t3g =aj1p25r-aj1p13r;		t3h =aj1p25i-aj1p13i;
	  t3c =t3c+t3e;			t3d =t3d+t3f;
	  t3e =t3c+c3m1*t3e;		t3f =t3d+c3m1*t3f;
	  rt  =s3*t3g;			it  =s3*t3h;
	  t3g =t3e+it;			t3h =t3f-rt;
	  t3e =t3e-it;			t3f =t3f+rt;
#if PFETCH
add = add0+p21;
prefetch_p_doubles(add);
#endif
	/* Twiddles: */
	  rt  =t36;			it  =t37;
	  t36 =rt+t3c;			t37 =it+t3d;
	  t3c =rt-t3c;			t3d =it-t3d;
	  t30 =t30+t36;			t31 =t31+t37;
	  t36 =t30+c3m1*t36;		t37 =t31+c3m1*t37;
	  rt  =s3*t3c;			it  =s3*t3d;
	  t3c =t36+it;			t3d =t37-rt;
	  t36 =t36-it;			t37 =t37+rt;
#if PFETCH
add = add0+p22;
prefetch_p_doubles(add);
#endif
	  rt  =t38*c -t39*s;		it  =t38*s +t39*c;
	  re  =t3e*c2-t3f*s2;		t3f =t3e*s2+t3f*c2;	t3e=re;
	  t38 =rt+t3e;			t39 =it+t3f;
	  t3e =rt-t3e;			t3f =it-t3f;
	  t32 =t32+t38;			t33 =t33+t39;
	  t38 =t32+c3m1*t38;		t39 =t33+c3m1*t39;
	  rt  =s3*t3e;			it  =s3*t3f;
	  t3e =t38+it;			t3f =t39-rt;
	  t38 =t38-it;			t39 =t39+rt;
#if PFETCH
add = add0+p23;
prefetch_p_doubles(add);
#endif
	  rt  =t3a*c2-t3b*s2;		it  =t3a*s2+t3b*c2;
	  re  =t3g*c4-t3h*s4;		t3h =t3g*s4+t3h*c4;	t3g=re;
	  t3a =rt+t3g;			t3b =it+t3h;
	  t3g =rt-t3g;			t3h =it-t3h;
	  t34 =t34+t3a;			t35 =t35+t3b;
	  t3a =t34+c3m1*t3a;		t3b =t35+c3m1*t3b;
	  rt  =s3*t3g;			it  =s3*t3h;
	  t3g =t3a+it;			t3h =t3b-rt;
	  t3a =t3a-it;			t3b =t3b+rt;
#if PFETCH
add = add0+p24;
prefetch_p_doubles(add);
#endif

/*...and now do nine radix-4 transforms:	*/

/*...Block 1: */
	j1=jt;
	    rt =t20;	t20=t00-rt;	t00=t00+rt;
	    it =t21;	t21=t01-it;	t01=t01+it;

	    rt =t30;	t30=t10-rt;	t10=t10+rt;
	    it =t31;	t31=t11-it;	t11=t11+it;

	    a[j1    ]=t00+t10;	a[j1+1    ]=t01+t11;
	    a[j1+p01]=t00-t10;	a[j1+1+p01]=t01-t11;

	    a[j1+p03]=t20-t31;	a[j1+1+p03]=t21+t30;	/* mpy by I is inlined here...	*/
	    a[j1+p02]=t20+t31;	a[j1+1+p02]=t21-t30;
#if PFETCH
add = add0+p25;
prefetch_p_doubles(add);
#endif
/*...Block 2: */
	j1=jt+p32;
	    rt =t22;	t22=t02-rt;	t02=t02+rt;
	    it =t23;	t23=t03-it;	t03=t03+it;

	    rt =t32;	t32=t12-rt;	t12=t12+rt;
	    it =t33;	t33=t13-it;	t13=t13+it;

	    a[j1    ]=t02+t12;	a[j1+1    ]=t03+t13;
	    a[j1+p01]=t02-t12;	a[j1+1+p01]=t03-t13;

	    a[j1+p03]=t22-t33;	a[j1+1+p03]=t23+t32;	/* mpy by I is inlined here...	*/
	    a[j1+p02]=t22+t33;	a[j1+1+p02]=t23-t32;
#if PFETCH
add = add0+p26;
prefetch_p_doubles(add);
#endif
/*...Block 3: */
	j1=jt+p20;
	    rt =t24;	t24=t04-rt;	t04=t04+rt;
	    it =t25;	t25=t05-it;	t05=t05+it;

	    rt =t34;	t34=t14-rt;	t14=t14+rt;
	    it =t35;	t35=t15-it;	t15=t15+it;

	    a[j1+p02]=t04+t14;	a[j1+1+p02]=t05+t15;
	    a[j1+p03]=t04-t14;	a[j1+1+p03]=t05-t15;

	    a[j1    ]=t24-t35;	a[j1+1    ]=t25+t34;	/* mpy by I is inlined here...	*/
	    a[j1+p01]=t24+t35;	a[j1+1+p01]=t25-t34;
#if PFETCH
add = add0+p27;
prefetch_p_doubles(add);
#endif
/*...Block 4: */
	j1=jt+p08;
	    rt =t26;	t26=t06-rt;	t06=t06+rt;
	    it =t27;	t27=t07-it;	t07=t07+it;

	    rt =t36;	t36=t16-rt;	t16=t16+rt;
	    it =t37;	t37=t17-it;	t17=t17+it;

	    a[j1+p01]=t06+t16;	a[j1+1+p01]=t07+t17;
	    a[j1    ]=t06-t16;	a[j1+1    ]=t07-t17;

	    a[j1+p02]=t26-t37;	a[j1+1+p02]=t27+t36;	/* mpy by I is inlined here...	*/
	    a[j1+p03]=t26+t37;	a[j1+1+p03]=t27-t36;
#if PFETCH
add = add0+p28;
prefetch_p_doubles(add);
#endif
/*...Block 5: */
	j1=jt+p28;
	    rt =t28;	t28=t08-rt;	t08=t08+rt;
	    it =t29;	t29=t09-it;	t09=t09+it;

	    rt =t38;	t38=t18-rt;	t18=t18+rt;
	    it =t39;	t39=t19-it;	t19=t19+it;

	    a[j1+p03]=t08+t18;	a[j1+1+p03]=t09+t19;
	    a[j1+p02]=t08-t18;	a[j1+1+p02]=t09-t19;

	    a[j1+p01]=t28-t39;	a[j1+1+p01]=t29+t38;	/* mpy by I is inlined here...	*/
	    a[j1    ]=t28+t39;	a[j1+1    ]=t29-t38;
#if PFETCH
add = add0+p29;
prefetch_p_doubles(add);
#endif
/*...Block 6: */
	j1=jt+p16;
	    rt =t2a;	t2a=t0a-rt;	t0a=t0a+rt;
	    it =t2b;	t2b=t0b-it;	t0b=t0b+it;

	    rt =t3a;	t3a=t1a-rt;	t1a=t1a+rt;
	    it =t3b;	t3b=t1b-it;	t1b=t1b+it;

	    a[j1    ]=t0a+t1a;	a[j1+1    ]=t0b+t1b;
	    a[j1+p01]=t0a-t1a;	a[j1+1+p01]=t0b-t1b;

	    a[j1+p03]=t2a-t3b;	a[j1+1+p03]=t2b+t3a;	/* mpy by I is inlined here...	*/
	    a[j1+p02]=t2a+t3b;	a[j1+1+p02]=t2b-t3a;
#if PFETCH
add = add0+p30;
prefetch_p_doubles(add);
#endif
/*...Block 7: */
	j1=jt+p04;
	    rt =t2c;	t2c=t0c-rt;	t0c=t0c+rt;
	    it =t2d;	t2d=t0d-it;	t0d=t0d+it;

	    rt =t3c;	t3c=t1c-rt;	t1c=t1c+rt;
	    it =t3d;	t3d=t1d-it;	t1d=t1d+it;

	    a[j1+p02]=t0c+t1c;	a[j1+1+p02]=t0d+t1d;
	    a[j1+p03]=t0c-t1c;	a[j1+1+p03]=t0d-t1d;

	    a[j1    ]=t2c-t3d;	a[j1+1    ]=t2d+t3c;	/* mpy by I is inlined here...	*/
	    a[j1+p01]=t2c+t3d;	a[j1+1+p01]=t2d-t3c;
#if PFETCH
add = add0+p31;
prefetch_p_doubles(add);
#endif
/*...Block 8: */
	j1=jt+p24;
	    rt =t2e;	t2e=t0e-rt;	t0e=t0e+rt;
	    it =t2f;	t2f=t0f-it;	t0f=t0f+it;

	    rt =t3e;	t3e=t1e-rt;	t1e=t1e+rt;
	    it =t3f;	t3f=t1f-it;	t1f=t1f+it;

	    a[j1+p01]=t0e+t1e;	a[j1+1+p01]=t0f+t1f;
	    a[j1    ]=t0e-t1e;	a[j1+1    ]=t0f-t1f;

	    a[j1+p02]=t2e-t3f;	a[j1+1+p02]=t2f+t3e;	/* mpy by I is inlined here...	*/
	    a[j1+p03]=t2e+t3f;	a[j1+1+p03]=t2f-t3e;
#if PFETCH
add = add0+p32;
prefetch_p_doubles(add);
#endif
/*...Block 9: */
	j1=jt+p12;
	    rt =t2g;	t2g=t0g-rt;	t0g=t0g+rt;
	    it =t2h;	t2h=t0h-it;	t0h=t0h+it;

	    rt =t3g;	t3g=t1g-rt;	t1g=t1g+rt;
	    it =t3h;	t3h=t1h-it;	t1h=t1h+it;

	    a[j1+p03]=t0g+t1g;	a[j1+1+p03]=t0h+t1h;
	    a[j1+p02]=t0g-t1g;	a[j1+1+p02]=t0h-t1h;

	    a[j1+p01]=t2g-t3h;	a[j1+1+p01]=t2h+t3g;	/* mpy by I is inlined here...	*/
	    a[j1    ]=t2g+t3h;	a[j1+1    ]=t2h-t3g;
#if PFETCH
add = add0+p33;
prefetch_p_doubles(add);
#endif
#if PFETCH
add = add0+p34;
prefetch_p_doubles(add);
#endif
#if PFETCH
add = add0+p35;
prefetch_p_doubles(add);
#endif

	  iroot += root_incr;		/* increment sincos index.	*/

	  }

	  jstart += nwt;
	  jhi    += nwt;
	  col += 36;
	  co3 -= 36;

	}

	if(root_incr==0)break;

/*   Wraparound carry cleanup loop is here: ***
!
!   (1) Invert the radix-36 forward DIF FFT of the first block of 36 complex elements in A and unweight;
!   (2) Propagate cleanup carries among the real and imaginary parts of the 36 outputs of (1);
!   (3) Reweight and perform a radix-36 forward DIF FFT on the result of (2);
!   (4) If any of the exit carries from (2) are nonzero, advance to the next 36 elements and repeat (1-4).
*/
	t00 = cy35;
	cy35= cy34;
	cy34= cy33;
	cy33= cy32;
	cy32= cy31;
	cy31= cy30;
	cy30= cy29;
	cy29= cy28;
	cy28= cy27;
	cy27= cy26;
	cy26= cy25;
	cy25= cy24;
	cy24= cy23;
	cy23= cy22;
	cy22= cy21;
	cy21= cy20;
	cy20= cy19;
	cy19= cy18;
	cy18= cy17;
	cy17= cy16;
	cy16= cy15;
	cy15= cy14;
	cy14= cy13;
	cy13= cy12;
	cy12= cy11;
	cy11= cy10;
	cy10= cy09;
	cy09= cy08;
	cy08= cy07;
	cy07= cy06;
	cy06= cy05;
	cy05= cy04;
	cy04= cy03;
	cy03= cy02;
	cy02= cy01;
	cy01= cy00;
	cy00= t00 ;

	iroot = 0;
	root_incr = 0;
	scale = 1;

	jstart = 0;
	jhi = 3;
	khi = 1;

	for(j=0; j<=jhi; j++)
	{
	  a[j    ] *= radix_inv;
	  a[j+p01] *= radix_inv;
	  a[j+p02] *= radix_inv;
	  a[j+p03] *= radix_inv;
	  a[j+p04] *= radix_inv;
	  a[j+p05] *= radix_inv;
	  a[j+p06] *= radix_inv;
	  a[j+p07] *= radix_inv;
	  a[j+p08] *= radix_inv;
	  a[j+p09] *= radix_inv;
	  a[j+p10] *= radix_inv;
	  a[j+p11] *= radix_inv;
	  a[j+p12] *= radix_inv;
	  a[j+p13] *= radix_inv;
	  a[j+p14] *= radix_inv;
	  a[j+p15] *= radix_inv;
	  a[j+p16] *= radix_inv;
	  a[j+p17] *= radix_inv;
	  a[j+p18] *= radix_inv;
	  a[j+p19] *= radix_inv;
	  a[j+p20] *= radix_inv;
	  a[j+p21] *= radix_inv;
	  a[j+p22] *= radix_inv;
	  a[j+p23] *= radix_inv;
	  a[j+p24] *= radix_inv;
	  a[j+p25] *= radix_inv;
	  a[j+p26] *= radix_inv;
	  a[j+p27] *= radix_inv;
	  a[j+p28] *= radix_inv;
	  a[j+p29] *= radix_inv;
	  a[j+p30] *= radix_inv;
	  a[j+p31] *= radix_inv;
	  a[j+p32] *= radix_inv;
	  a[j+p33] *= radix_inv;
	  a[j+p34] *= radix_inv;
	  a[j+p35] *= radix_inv;
	}
}

	if(fabs(cy00)+fabs(cy01)+fabs(cy02)+fabs(cy03)+fabs(cy04)+fabs(cy05)+fabs(cy06)+fabs(cy07)+fabs(cy08)+fabs(cy09)+fabs(cy10)+fabs(cy11)+fabs(cy12)+fabs(cy13)+fabs(cy14)+fabs(cy15)+fabs(cy16)+fabs(cy17)+fabs(cy18)+fabs(cy19)+fabs(cy20)+fabs(cy21)+fabs(cy22)+fabs(cy23)+fabs(cy24)+fabs(cy25)+fabs(cy26)+fabs(cy27)+fabs(cy28)+fabs(cy29)+fabs(cy30)+fabs(cy31)+fabs(cy32)+fabs(cy33)+fabs(cy34)+fabs(cy35) != 0.0)
	{
	    if(INTERACT)printf("FATAL: iter = %10d; nonzero exit carry in radix36_ditN_cy_dif1.\n",iter);
	    f1 = fopen(   OFILE,"a");
	    f2 = fopen(STATFILE,"a");
	    fprintf(f1,"FATAL: iter = %10d; nonzero exit carry in radix36_ditN_cy_dif1.\n",iter);
	    fprintf(f2,"FATAL: iter = %10d; nonzero exit carry in radix36_ditN_cy_dif1.\n",iter);
	    fclose(f1);
	    fclose(f2);
	    err=1;
	    return(err);
	}

	return(0);
}

/***************/

void radix36_dit_pass1(double a[], int n)
{
/*
!...Acronym: DIT = Decimation In Time
!
!...Subroutine to perform an initial radix-36 complex DIT FFT pass on the data in the length-N real vector A.
!
!   This routine is designed exclusively to undo the effects of radix36_dif_pass1,
!   i.e. to reobtain the raw all-integer residue vector at the end of an iteration cycle.
*/
	int j,jt,j1;
	static int n36, first_entry=TRUE
		    ,p01,p02,p03,p04,p05,p06,p07,p08,p09,p10,p11,p12,p13,p14,p15,p16,p17
		,p18,p19,p20,p21,p22,p23,p24,p25,p26,p27,p28,p29,p30,p31,p32,p33,p34,p35;
	static double c   =  0.76604444311897803520,	/* cos(2*pi/9) */
		      s   =  0.64278760968653932631,	/* sin(2*pi/9) */
		      c2  =  0.17364817766693034887,	/* cos(2*u) */
		      s2  =  0.98480775301220805936,	/* sin(2*u) */
		      c3m1= -1.50000000000000000000,	/* cos(3*u)-1 */
		      s3  =  0.86602540378443864677,	/* sin(3*u) */
		      c4  = -0.93969262078590838404,	/* cos(4*u) */
		      s4  =  0.34202014332566873307;	/* sin(4*u) */
	double rt,it,re
		,t00,t01,t02,t03,t04,t05,t06,t07,t08,t09,t0a,t0b,t0c,t0d,t0e,t0f,t0g,t0h
		,t10,t11,t12,t13,t14,t15,t16,t17,t18,t19,t1a,t1b,t1c,t1d,t1e,t1f,t1g,t1h
		,t20,t21,t22,t23,t24,t25,t26,t27,t28,t29,t2a,t2b,t2c,t2d,t2e,t2f,t2g,t2h
		,t30,t31,t32,t33,t34,t35,t36,t37,t38,t39,t3a,t3b,t3c,t3d,t3e,t3f,t3g,t3h;

	if(!first_entry && (n/36) != n36)	/* New runlength?	*/
	{
	  first_entry=TRUE;
	}

/*...initialize things upon first entry	*/

	if(first_entry)
	{
	  first_entry=FALSE;
	  n36=n/36;

/*   constant index offsets for array load/stores are here.	*/

	  p01 = n36;
	  p02 = p01 + p01;
	  p03 = p02 + p01;
	  p04 = p03 + p01;
	  p05 = p04 + p01;
	  p06 = p05 + p01;
	  p07 = p06 + p01;
	  p08 = p07 + p01;
	  p09 = p08 + p01;
	  p10 = p09 + p01;
	  p11 = p10 + p01;
	  p12 = p11 + p01;
	  p13 = p12 + p01;
	  p14 = p13 + p01;
	  p15 = p14 + p01;
	  p16 = p15 + p01;
	  p17 = p16 + p01;
	  p18 = p17 + p01;
	  p19 = p18 + p01;
	  p20 = p19 + p01;
	  p21 = p20 + p01;
	  p22 = p21 + p01;
	  p23 = p22 + p01;
	  p24 = p23 + p01;
	  p25 = p24 + p01;
	  p26 = p25 + p01;
	  p27 = p26 + p01;
	  p28 = p27 + p01;
	  p29 = p28 + p01;
	  p30 = p29 + p01;
	  p31 = p30 + p01;
	  p32 = p31 + p01;
	  p33 = p32 + p01;
	  p34 = p33 + p01;
	  p35 = p34 + p01;

	  p01 = p01 + ( (p01 >> DAT_BITS) << PAD_BITS );
	  p02 = p02 + ( (p02 >> DAT_BITS) << PAD_BITS );
	  p03 = p03 + ( (p03 >> DAT_BITS) << PAD_BITS );
	  p04 = p04 + ( (p04 >> DAT_BITS) << PAD_BITS );
	  p05 = p05 + ( (p05 >> DAT_BITS) << PAD_BITS );
	  p06 = p06 + ( (p06 >> DAT_BITS) << PAD_BITS );
	  p07 = p07 + ( (p07 >> DAT_BITS) << PAD_BITS );
	  p08 = p08 + ( (p08 >> DAT_BITS) << PAD_BITS );
	  p09 = p09 + ( (p09 >> DAT_BITS) << PAD_BITS );
	  p10 = p10 + ( (p10 >> DAT_BITS) << PAD_BITS );
	  p11 = p11 + ( (p11 >> DAT_BITS) << PAD_BITS );
	  p12 = p12 + ( (p12 >> DAT_BITS) << PAD_BITS );
	  p13 = p13 + ( (p13 >> DAT_BITS) << PAD_BITS );
	  p14 = p14 + ( (p14 >> DAT_BITS) << PAD_BITS );
	  p15 = p15 + ( (p15 >> DAT_BITS) << PAD_BITS );
	  p16 = p16 + ( (p16 >> DAT_BITS) << PAD_BITS );
	  p17 = p17 + ( (p17 >> DAT_BITS) << PAD_BITS );
	  p18 = p18 + ( (p18 >> DAT_BITS) << PAD_BITS );
	  p19 = p19 + ( (p19 >> DAT_BITS) << PAD_BITS );
	  p20 = p20 + ( (p20 >> DAT_BITS) << PAD_BITS );
	  p21 = p21 + ( (p21 >> DAT_BITS) << PAD_BITS );
	  p22 = p22 + ( (p22 >> DAT_BITS) << PAD_BITS );
	  p23 = p23 + ( (p23 >> DAT_BITS) << PAD_BITS );
	  p24 = p24 + ( (p24 >> DAT_BITS) << PAD_BITS );
	  p25 = p25 + ( (p25 >> DAT_BITS) << PAD_BITS );
	  p26 = p26 + ( (p26 >> DAT_BITS) << PAD_BITS );
	  p27 = p27 + ( (p27 >> DAT_BITS) << PAD_BITS );
	  p28 = p28 + ( (p28 >> DAT_BITS) << PAD_BITS );
	  p29 = p29 + ( (p29 >> DAT_BITS) << PAD_BITS );
	  p30 = p30 + ( (p30 >> DAT_BITS) << PAD_BITS );
	  p31 = p31 + ( (p31 >> DAT_BITS) << PAD_BITS );
	  p32 = p32 + ( (p32 >> DAT_BITS) << PAD_BITS );
	  p33 = p33 + ( (p33 >> DAT_BITS) << PAD_BITS );
	  p34 = p34 + ( (p34 >> DAT_BITS) << PAD_BITS );
	  p35 = p35 + ( (p35 >> DAT_BITS) << PAD_BITS );
	}

/*...The radix-36 pass is here.	*/

      for(j=0; j < n36; j += 2)
      {
	  jt = j  + ( (j  >> DAT_BITS) << PAD_BITS );
/*
Twiddleless version requires us to swap inputs as follows:
indices  0, 1, 2, 3, 4, 5, 6, 7, 8, 9,10,11,12,13,14,15,16,17,18,19,20,21,22,23,24,25,26,27,28,29,30,31,32,33,34,35
      -> 0,32,28,24,20,16,12, 8, 4,27,23,19,15,11, 7, 3,35,31,18,14,10, 6, 2,34,30,26,22, 9, 5, 1,33,29,25,21,17,13

I.e. start out with first quartet of indices {0,9,18,27}, permute those according to
  {0,9,18,27}*35%36 = {0,27,18,9}, then each is head of a length-9 list of indices with decrement 4

Remember, inputs to DIT are bit-reversed, so
a[0, 1, 2, 3, 4, 5, 6, 7, 8, 9,10,11,12,13,14,15,16,17,18,19,20,21,22,23,24,25,26,27,28,29,30,31,32,33,34,35] contain
x[0,18, 9,27, 3,21,12,30, 6,24,15,33, 1,19,10,28, 4,22,13,31, 7,25,16,34, 2,20,11,29, 5,23,14,32, 8,26,17,35], which get swapped to
x[0,18,27, 9,24, 6,15,33,12,30, 3,21,32,14,23, 5,20, 2,11,29, 8,26,35,17,28,10,19, 1,16,34, 7,25, 4,22,31,13], which means the a-indices get swapped as
a[0, 1, 3, 2, 9, 8,10,11, 6, 7, 4, 5,31,30,29,28,25,24,26,27,32,33,35,34,15,14,13,12,22,23,20,21,16,17,19,18].
*/
/*...Block 1:	*/
	j1=jt;
	t00=a[j1    ];	t01=a[j1+1    ];
	rt =a[j1+p01];	it =a[j1+1+p01];
	t10=t00-rt;  	t00=t00+rt;
	t11=t01-it;	t01=t01+it;

	t20=a[j1+p03];	t21=a[j1+1+p03];
	rt =a[j1+p02];	it =a[j1+1+p02];
	t30=t20-rt;  	t20=t20+rt;
	t31=t21-it;  	t21=t21+it;

	rt =t20;	t20=t00-rt;	t00=t00+rt;
	it =t21;	t21=t01-it;	t01=t01+it;

	rt =t30;	t30=t10-t31;	t10=t10+t31;
			t31=t11+rt;	t11=t11-rt;

/*...Block 2:	*/
	j1=jt+p08;
	t02=a[j1+p01];	t03=a[j1+1+p01];
	rt =a[j1    ];	it =a[j1+1    ];
	t12=t02-rt;  	t02=t02+rt;
	t13=t03-it;	t03=t03+it;

	t22=a[j1+p02];	t23=a[j1+1+p02];
	rt =a[j1+p03];	it =a[j1+1+p03];
	t32=t22-rt;  	t22=t22+rt;
	t33=t23-it;  	t23=t23+it;

	rt =t22;	t22=t02-rt;	t02=t02+rt;
	it =t23;	t23=t03-it;	t03=t03+it;

	rt =t32;	t32=t12-t33;	t12=t12+t33;
			t33=t13+rt;	t13=t13-rt;

/*...Block 3:	*/
	j1=jt+p04;
	t04=a[j1+p02];	t05=a[j1+1+p02];
	rt =a[j1+p03];	it =a[j1+1+p03];
	t14=t04-rt;  	t04=t04+rt;
	t15=t05-it;	t05=t05+it;

	t24=a[j1    ];	t25=a[j1+1    ];
	rt =a[j1+p01];	it =a[j1+1+p01];
	t34=t24-rt;  	t24=t24+rt;
	t35=t25-it;  	t25=t25+it;

	rt =t24;	t24=t04-rt;	t04=t04+rt;
	it =t25;	t25=t05-it;	t05=t05+it;

	rt =t34;	t34=t14-t35;	t14=t14+t35;
			t35=t15+rt;	t15=t15-rt;

/*...Block 4:	*/
	j1=jt+p28;
	t06=a[j1+p03];	t07=a[j1+1+p03];
	rt =a[j1+p02];	it =a[j1+1+p02];
	t16=t06-rt;  	t06=t06+rt;
	t17=t07-it;	t07=t07+it;

	t26=a[j1+p01];	t27=a[j1+1+p01];
	rt =a[j1    ];	it =a[j1+1    ];
	t36=t26-rt;  	t26=t26+rt;
	t37=t27-it;  	t27=t27+it;

	rt =t26;	t26=t06-rt;	t06=t06+rt;
	it =t27;	t27=t07-it;	t07=t07+it;

	rt =t36;	t36=t16-t37;	t16=t16+t37;
			t37=t17+rt;	t17=t17-rt;

/*...Block 5:	*/
	j1=jt+p24;
	t08=a[j1+p01];	t09=a[j1+1+p01];
	rt =a[j1    ];	it =a[j1+1    ];
	t18=t08-rt;  	t08=t08+rt;
	t19=t09-it;	t09=t09+it;

	t28=a[j1+p02];	t29=a[j1+1+p02];
	rt =a[j1+p03];	it =a[j1+1+p03];
	t38=t28-rt;  	t28=t28+rt;
	t39=t29-it;  	t29=t29+it;

	rt =t28;	t28=t08-rt;	t08=t08+rt;
	it =t29;	t29=t09-it;	t09=t09+it;

	rt =t38;	t38=t18-t39;	t18=t18+t39;
			t39=t19+rt;	t19=t19-rt;

/*...Block 6:	*/
	j1=jt+p32;
	t0a=a[j1    ];	t0b=a[j1+1    ];
	rt =a[j1+p01];	it =a[j1+1+p01];
	t1a=t0a-rt;  	t0a=t0a+rt;
	t1b=t0b-it;	t0b=t0b+it;

	t2a=a[j1+p03];	t2b=a[j1+1+p03];
	rt =a[j1+p02];	it =a[j1+1+p02];
	t3a=t2a-rt;  	t2a=t2a+rt;
	t3b=t2b-it;  	t2b=t2b+it;

	rt =t2a;	t2a=t0a-rt;	t0a=t0a+rt;
	it =t2b;	t2b=t0b-it;	t0b=t0b+it;

	rt =t3a;	t3a=t1a-t3b;	t1a=t1a+t3b;
			t3b=t1b+rt;	t1b=t1b-rt;

/*...Block 7:	*/
	j1=jt+p12;
	t0c=a[j1+p03];	t0d=a[j1+1+p03];
	rt =a[j1+p02];	it =a[j1+1+p02];
	t1c=t0c-rt;  	t0c=t0c+rt;
	t1d=t0d-it;	t0d=t0d+it;

	t2c=a[j1+p01];	t2d=a[j1+1+p01];
	rt =a[j1    ];	it =a[j1+1    ];
	t3c=t2c-rt;  	t2c=t2c+rt;
	t3d=t2d-it;  	t2d=t2d+it;

	rt =t2c;	t2c=t0c-rt;	t0c=t0c+rt;
	it =t2d;	t2d=t0d-it;	t0d=t0d+it;

	rt =t3c;	t3c=t1c-t3d;	t1c=t1c+t3d;
			t3d=t1d+rt;	t1d=t1d-rt;

/*...Block 8:	*/
	j1=jt+p20;
	t0e=a[j1+p02];	t0f=a[j1+1+p02];
	rt =a[j1+p03];	it =a[j1+1+p03];
	t1e=t0e-rt;  	t0e=t0e+rt;
	t1f=t0f-it;	t0f=t0f+it;

	t2e=a[j1    ];	t2f=a[j1+1    ];
	rt =a[j1+p01];	it =a[j1+1+p01];
	t3e=t2e-rt;  	t2e=t2e+rt;
	t3f=t2f-it;  	t2f=t2f+it;

	rt =t2e;	t2e=t0e-rt;	t0e=t0e+rt;
	it =t2f;	t2f=t0f-it;	t0f=t0f+it;

	rt =t3e;	t3e=t1e-t3f;	t1e=t1e+t3f;
			t3f=t1f+rt;	t1f=t1f-rt;

/*...Block 9:	*/
	j1=jt+p16;
	t0g=a[j1    ];	t0h=a[j1+1    ];
	rt =a[j1+p01];	it =a[j1+1+p01];
	t1g=t0g-rt;  	t0g=t0g+rt;
	t1h=t0h-it;	t0h=t0h+it;

	t2g=a[j1+p03];	t2h=a[j1+1+p03];
	rt =a[j1+p02];	it =a[j1+1+p02];
	t3g=t2g-rt;  	t2g=t2g+rt;
	t3h=t2h-it;  	t2h=t2h+it;

	rt =t2g;	t2g=t0g-rt;	t0g=t0g+rt;
	it =t2h;	t2h=t0h-it;	t0h=t0h+it;

	rt =t3g;	t3g=t1g-t3h;	t1g=t1g+t3h;
			t3h=t1h+rt;	t1h=t1h-rt;

/*       ...and now do four radix-9 transforms.	*/

/*...First radix-9 transform:	*/
	  rt  =t02;			it  =t03;
	  t02 =rt+t04;			t03 =it+t05;
	  t04 =rt-t04;			t05 =it-t05;
	  t00 =t00+t02;			t01 =t01+t03;
	  t02 =t00+c3m1*t02;		t03 =t01+c3m1*t03;
	  rt  =s3*t04;			it  =s3*t05;
	  t04 =t02-it;			t05 =t03+rt;
	  t02 =t02+it;			t03 =t03-rt;

	  rt  =t08;			it  =t09;
	  t08 =rt+t0a;			t09 =it+t0b;
	  t0a =rt-t0a;			t0b =it-t0b;
	  t06 =t06+t08;			t07 =t07+t09;
	  t08 =t06+c3m1*t08;		t09 =t07+c3m1*t09;
	  rt  =s3*t0a;			it  =s3*t0b;
	  t0a =t08-it;			t0b =t09+rt;
	  t08 =t08+it;			t09 =t09-rt;

	  rt  =t0e;			it  =t0f;
	  t0e =rt+t0g;			t0f =it+t0h;
	  t0g =rt-t0g;			t0h =it-t0h;
	  t0c =t0c+t0e;			t0d =t0d+t0f;
	  t0e =t0c+c3m1*t0e;		t0f =t0d+c3m1*t0f;
	  rt  =s3*t0g;			it  =s3*t0h;
	  t0g =t0e-it;			t0h =t0f+rt;
	  t0e =t0e+it;			t0f =t0f-rt;
	/* Twiddles: */
	  rt  =t06;			it  =t07;
	  t06 =rt+t0c;			t07 =it+t0d;
	  t0c =rt-t0c;			t0d =it-t0d;
	  t00 =t00+t06;			t01 =t01+t07;
	  a[jt    ]=t00;		a[jt+1    ]=t01;
	  t06 =t00+c3m1*t06;		t07 =t01+c3m1*t07;
	  rt  =s3*t0c;			it  =s3*t0d;
	  a[jt+p24]=t06+it;		a[jt+1+p24]=t07-rt;
	  a[jt+p12]=t06-it;		a[jt+1+p12]=t07+rt;

	  rt  =t08*c +t09*s;		it  =t09*c -t08*s;
	  re  =t0e*c2+t0f*s2;		t0f =t0f*c2-t0e*s2;	t0e=re;
	  t08 =rt+t0e;			t09 =it+t0f;
	  t0e =rt-t0e;			t0f =it-t0f;
	  t02 =t02+t08;			t03 =t03+t09;
	  a[jt+p08]=t02;		a[jt+1+p08]=t03;
	  t08 =t02+c3m1*t08;		t09=t03+c3m1*t09;
	  rt  =s3*t0e;			it  =s3*t0f;
	  a[jt+p32]=t08+it;		a[jt+1+p32]=t09-rt;
	  a[jt+p20]=t08-it;		a[jt+1+p20]=t09+rt;

	  rt  =t0a*c2+t0b*s2;		it  =t0b*c2-t0a*s2;
	  re  =t0g*c4+t0h*s4;		t0h =t0h*c4-t0g*s4;	t0g=re;
	  t0a =rt+t0g;			t0b =it+t0h;
	  t0g =rt-t0g;			t0h =it-t0h;
	  t04 =t04+t0a;			t05 =t05+t0b;
	  a[jt+p16]=t04;		a[jt+1+p16]=t05;
	  t0a =t04+c3m1*t0a;		t0b =t05+c3m1*t0b;
	  rt  =s3*t0g;			it  =s3*t0h;
	  a[jt+p04]=t0a+it;		a[jt+1+p04]=t0b-rt;
	  a[jt+p28]=t0a-it;		a[jt+1+p28]=t0b+rt;

/*...Second radix-9 transform:	*/
	  rt  =t12;			it  =t13;
	  t12 =rt+t14;			t13 =it+t15;
	  t14 =rt-t14;			t15 =it-t15;
	  t10 =t10+t12;			t11 =t11+t13;
	  t12 =t10+c3m1*t12;		t13 =t11+c3m1*t13;
	  rt  =s3*t14;			it  =s3*t15;
	  t14 =t12-it;			t15 =t13+rt;
	  t12 =t12+it;			t13 =t13-rt;

	  rt  =t18;			it  =t19;
	  t18 =rt+t1a;			t19 =it+t1b;
	  t1a =rt-t1a;			t1b =it-t1b;
	  t16 =t16+t18;			t17 =t17+t19;
	  t18 =t16+c3m1*t18;		t19 =t17+c3m1*t19;
	  rt  =s3*t1a;			it  =s3*t1b;
	  t1a =t18-it;			t1b =t19+rt;
	  t18 =t18+it;			t19 =t19-rt;

	  rt  =t1e;			it  =t1f;
	  t1e =rt+t1g;			t1f =it+t1h;
	  t1g =rt-t1g;			t1h =it-t1h;
	  t1c =t1c+t1e;			t1d =t1d+t1f;
	  t1e =t1c+c3m1*t1e;		t1f =t1d+c3m1*t1f;
	  rt  =s3*t1g;			it  =s3*t1h;
	  t1g =t1e-it;			t1h =t1f+rt;
	  t1e =t1e+it;			t1f =t1f-rt;
	/* Twiddles: */
	  rt  =t16;			it  =t17;
	  t16 =rt+t1c;			t17 =it+t1d;
	  t1c =rt-t1c;			t1d =it-t1d;
	  t10 =t10+t16;			t11 =t11+t17;
	  a[jt+p27]=t10;		a[jt+1+p27]=t11;
	  t16 =t10+c3m1*t16;		t17 =t11+c3m1*t17;
	  rt  =s3*t1c;			it  =s3*t1d;
	  a[jt+p15]=t16+it;		a[jt+1+p15]=t17-rt;
	  a[jt+p03]=t16-it;		a[jt+1+p03]=t17+rt;

	  rt  =t18*c +t19*s;		it  =t19*c -t18*s;
	  re  =t1e*c2+t1f*s2;		t1f =t1f*c2-t1e*s2;	t1e=re;
	  t18 =rt+t1e;			t19 =it+t1f;
	  t1e =rt-t1e;			t1f =it-t1f;
	  t12 =t12+t18;			t13 =t13+t19;
	  a[jt+p35]=t12;		a[jt+1+p35]=t13;
	  t18 =t12+c3m1*t18;		t19=t13+c3m1*t19;
	  rt  =s3*t1e;			it  =s3*t1f;
	  a[jt+p23]=t18+it;		a[jt+1+p23]=t19-rt;
	  a[jt+p11]=t18-it;		a[jt+1+p11]=t19+rt;

	  rt  =t1a*c2+t1b*s2;		it  =t1b*c2-t1a*s2;
	  re  =t1g*c4+t1h*s4;		t1h =t1h*c4-t1g*s4;	t1g=re;
	  t1a =rt+t1g;			t1b =it+t1h;
	  t1g =rt-t1g;			t1h =it-t1h;
	  t14 =t14+t1a;			t15 =t15+t1b;
	  a[jt+p07]=t14;		a[jt+1+p07]=t15;
	  t1a =t14+c3m1*t1a;		t1b =t15+c3m1*t1b;
	  rt  =s3*t1g;			it  =s3*t1h;
	  a[jt+p31]=t1a+it;		a[jt+1+p31]=t1b-rt;
	  a[jt+p19]=t1a-it;		a[jt+1+p19]=t1b+rt;

/*...Third radix-9 transform:	*/
	  rt  =t22;			it  =t23;
	  t22 =rt+t24;			t23 =it+t25;
	  t24 =rt-t24;			t25 =it-t25;
	  t20 =t20+t22;			t21 =t21+t23;
	  t22 =t20+c3m1*t22;		t23 =t21+c3m1*t23;
	  rt  =s3*t24;			it  =s3*t25;
	  t24 =t22-it;			t25 =t23+rt;
	  t22 =t22+it;			t23 =t23-rt;

	  rt  =t28;			it  =t29;
	  t28 =rt+t2a;			t29 =it+t2b;
	  t2a =rt-t2a;			t2b =it-t2b;
	  t26 =t26+t28;			t27 =t27+t29;
	  t28 =t26+c3m1*t28;		t29 =t27+c3m1*t29;
	  rt  =s3*t2a;			it  =s3*t2b;
	  t2a =t28-it;			t2b =t29+rt;
	  t28 =t28+it;			t29 =t29-rt;

	  rt  =t2e;			it  =t2f;
	  t2e =rt+t2g;			t2f =it+t2h;
	  t2g =rt-t2g;			t2h =it-t2h;
	  t2c =t2c+t2e;			t2d =t2d+t2f;
	  t2e =t2c+c3m1*t2e;		t2f =t2d+c3m1*t2f;
	  rt  =s3*t2g;			it  =s3*t2h;
	  t2g =t2e-it;			t2h =t2f+rt;
	  t2e =t2e+it;			t2f =t2f-rt;
	/* Twiddles: */
	  rt  =t26;			it  =t27;
	  t26 =rt+t2c;			t27 =it+t2d;
	  t2c =rt-t2c;			t2d =it-t2d;
	  t20 =t20+t26;			t21 =t21+t27;
	  a[jt+p18]=t20;		a[jt+1+p18]=t21;
	  t26 =t20+c3m1*t26;		t27 =t21+c3m1*t27;
	  rt  =s3*t2c;			it  =s3*t2d;
	  a[jt+p06]=t26+it;		a[jt+1+p06]=t27-rt;
	  a[jt+p30]=t26-it;		a[jt+1+p30]=t27+rt;

	  rt  =t28*c +t29*s;		it  =t29*c -t28*s;
	  re  =t2e*c2+t2f*s2;		t2f =t2f*c2-t2e*s2;	t2e=re;
	  t28 =rt+t2e;			t29 =it+t2f;
	  t2e =rt-t2e;			t2f =it-t2f;
	  t22 =t22+t28;			t23 =t23+t29;
	  a[jt+p26]=t22;		a[jt+1+p26]=t23;
	  t28 =t22+c3m1*t28;		t29=t23+c3m1*t29;
	  rt  =s3*t2e;			it  =s3*t2f;
	  a[jt+p14]=t28+it;		a[jt+1+p14]=t29-rt;
	  a[jt+p02]=t28-it;		a[jt+1+p02]=t29+rt;

	  rt  =t2a*c2+t2b*s2;		it  =t2b*c2-t2a*s2;
	  re  =t2g*c4+t2h*s4;		t2h =t2h*c4-t2g*s4;	t2g=re;
	  t2a =rt+t2g;			t2b =it+t2h;
	  t2g =rt-t2g;			t2h =it-t2h;
	  t24 =t24+t2a;			t25 =t25+t2b;
	  a[jt+p34]=t24;		a[jt+1+p34]=t25;
	  t2a =t24+c3m1*t2a;		t2b =t25+c3m1*t2b;
	  rt  =s3*t2g;			it  =s3*t2h;
	  a[jt+p22]=t2a+it;		a[jt+1+p22]=t2b-rt;
	  a[jt+p10]=t2a-it;		a[jt+1+p10]=t2b+rt;

/*...Fourth radix-9 transform:	*/
	  rt  =t32;			it  =t33;
	  t32 =rt+t34;			t33 =it+t35;
	  t34 =rt-t34;			t35 =it-t35;
	  t30 =t30+t32;			t31 =t31+t33;
	  t32 =t30+c3m1*t32;		t33 =t31+c3m1*t33;
	  rt  =s3*t34;			it  =s3*t35;
	  t34 =t32-it;			t35 =t33+rt;
	  t32 =t32+it;			t33 =t33-rt;

	  rt  =t38;			it  =t39;
	  t38 =rt+t3a;			t39 =it+t3b;
	  t3a =rt-t3a;			t3b =it-t3b;
	  t36 =t36+t38;			t37 =t37+t39;
	  t38 =t36+c3m1*t38;		t39 =t37+c3m1*t39;
	  rt  =s3*t3a;			it  =s3*t3b;
	  t3a =t38-it;			t3b =t39+rt;
	  t38 =t38+it;			t39 =t39-rt;

	  rt  =t3e;			it  =t3f;
	  t3e =rt+t3g;			t3f =it+t3h;
	  t3g =rt-t3g;			t3h =it-t3h;
	  t3c =t3c+t3e;			t3d =t3d+t3f;
	  t3e =t3c+c3m1*t3e;		t3f =t3d+c3m1*t3f;
	  rt  =s3*t3g;			it  =s3*t3h;
	  t3g =t3e-it;			t3h =t3f+rt;
	  t3e =t3e+it;			t3f =t3f-rt;
	/* Twiddles: */
	  rt  =t36;			it  =t37;
	  t36 =rt+t3c;			t37 =it+t3d;
	  t3c =rt-t3c;			t3d =it-t3d;
	  t30 =t30+t36;			t31 =t31+t37;
	  a[jt+p09]=t30;		a[jt+1+p09]=t31;
	  t36 =t30+c3m1*t36;		t37 =t31+c3m1*t37;
	  rt  =s3*t3c;			it  =s3*t3d;
	  a[jt+p33]=t36+it;		a[jt+1+p33]=t37-rt;
	  a[jt+p21]=t36-it;		a[jt+1+p21]=t37+rt;

	  rt  =t38*c +t39*s;		it  =t39*c -t38*s;
	  re  =t3e*c2+t3f*s2;		t3f =t3f*c2-t3e*s2;	t3e=re;
	  t38 =rt+t3e;			t39 =it+t3f;
	  t3e =rt-t3e;			t3f =it-t3f;
	  t32 =t32+t38;			t33 =t33+t39;
	  a[jt+p17]=t32;		a[jt+1+p17]=t33;
	  t38 =t32+c3m1*t38;		t39=t33+c3m1*t39;
	  rt  =s3*t3e;			it  =s3*t3f;
	  a[jt+p05]=t38+it;		a[jt+1+p05]=t39-rt;
	  a[jt+p29]=t38-it;		a[jt+1+p29]=t39+rt;

	  rt  =t3a*c2+t3b*s2;		it  =t3b*c2-t3a*s2;
	  re  =t3g*c4+t3h*s4;		t3h =t3h*c4-t3g*s4;	t3g=re;
	  t3a =rt+t3g;			t3b =it+t3h;
	  t3g =rt-t3g;			t3h =it-t3h;
	  t34 =t34+t3a;			t35 =t35+t3b;
	  a[jt+p25]=t34;		a[jt+1+p25]=t35;
	  t3a =t34+c3m1*t3a;		t3b =t35+c3m1*t3b;
	  rt  =s3*t3g;			it  =s3*t3h;
	  a[jt+p13]=t3a+it;		a[jt+1+p13]=t3b-rt;
	  a[jt+p01]=t3a-it;		a[jt+1+p01]=t3b+rt;
	}
}

